
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: builtin.C,v 2.5 1995/06/17 20:32:58 davidsi Exp $
//
// $Log: builtin.C,v $
// Revision 2.5  1995/06/17  20:32:58  davidsi
// - Added support for the char type.
//
// Revision 2.4  1995/06/02  20:04:28  davidsi
// - In builtin() added calls to cmd_math() and cmd_window().
// - node_builtin::types_of_arguments() now correctly uses the enumerated type
//   for writing errors out.
//
// Revision 2.3  1995/05/19  18:49:43  davidsi
// - Tidying up and add GNU copyright message.
//
// Revision 2.2  1995/04/16  19:56:14  davidsi
// - Added support for builtin functions that wish to return values rather
//   than passing via arguments.  The return_value member is NULL or a
//   pointer to the value that is to returned. It can be set with the
//   SetReturn() function member.
// - Use DialogPrintf() for print() so output goes to the xsm window and not
//   stdout.
//
// Revision 2.1  1995/04/03  20:42:17  davidsi
// - builtin() now also calls cmd_window() to handle all window/drawing
//   commands and types.
//
// Revision 2.0  1995/01/19  21:58:31  davidsi
// - Initial version.
//
//

#include "stdio.h"
#include "stdlib.h"

#include "error.h"
#include "builtin.h"
#include "dialog.h"

extern errstream cerror;		// error stream
extern diagstream cinfo;		// information stream
class value_int *INT_VALUE;		// reference for integer types
class value_string *STRING_VALUE;	// reference for string types
class value_real *REAL_VALUE;		// reference for real types
class value_char *CHAR_VALUE;		// reference for char types

extern void xsmbuiltin(class context *);
extern void cmd_window(class context *);
extern void cmd_math(class context *);
extern void cmd_file(class context *);


//
// print
//
// Simple print command
//
// INPUTS :	bptr		The builtin information
// RETURNS:	None
//
int
print(class node_builtin &bptr)
{
	class value *val;
	char *str;
	int i=0;

	while ( (val=bptr[i])!=NULL) {
		if ( val->load(str)== -1 )
			cerror << not_string << "\n";
		else {
			DialogPrintf("%s",str);
		}
		i++;
	}

	return(0);
}

//
// node_builtin::getarg
//
// Reads the arguments being passed to the builtin function, caching
// them ready for access by that function.
//
// INPUTS :	None
// RETURNS:	None
//
void
node_builtin::getarg()
{
	int i;

	num_args=0;

	if (cur_arg==NULL || cur_context==NULL )
		return;

        for(i=0;i<MAX_ARGS;i++) {
		val[i]=cur_arg->evaluate(cur_context);
                if (val[i]==NULL) {
			cerror << not_evaluated << "\n";
			break;
		}
		num_args++;
		cur_arg=cur_arg->next();
		if (!cur_arg)
			break;
        }
	if (i==MAX_ARGS) 
		cerror << "more arguments than unwrapper can manage\n";
}


//
// node_builtin::freearg
//
// Disposes of all arguments passed to the builtin function
//
// INPUTS :	None
// RETURNS:	None
//
void
node_builtin::freearg()
{
	int i;
	
	for(i=0;i<num_args;i++) 
		val[i]->dispose();
}


//
// node_builtin::[] index
//
// Returns the requested argument
//
// INPUTS :	index		The argument that is to be returned
// RETURNS:	The argument
//
class value *
node_builtin::operator[] (int index)
{
	if (index>=num_args)
		return(NULL);
	return(val[index]);
}


//
// node_builtin::number_of_arguments
//
// Returns the number of arguments being passed to the builtin function.
//
// INPUTS :	None
// RETURNS:	The number of arguments being passed.
//
int 
node_builtin::number_of_arguments()
{
	return(num_args);
}


//
// node_builtin::types_of_arguments num req_type
//
// Validates that the arguments being handled by the builtin 
// function match those given by the user.
//
// INPUTS :	num		Number of arguments
//		req_type	Array specifying the type of each argument
// RETURNS:	TRUE if arguments match, else FALSE.
//
int
node_builtin::types_of_arguments(int num,class value **req_type)
{
	int i;

	//
	// Check correct number of arguments
	//
	if (num!=num_args) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}
	for(i=0;i<num;i++) {
		if ( req_type[i] && !(val[i]->type_equal(req_type[i])) ) {
			cerror << type_mismatch << " ";
			cerror << "argument "<<i<<" was incorrect\n";
			return(-1);
		}
	}
	
	return(0);
}



//
// builtint context
//
// Defines all types and functions that are made available by this
// module
//
// INPUTS :	context		The context within which the definitions
//				will be made
// RETURNS:	None
//
void
builtin(class context *cont)
{
	class node_builtin *nb;

	nb=new node_builtin(cont,"print",print);

	//
	// Add standard variable types
	//
	cont->addtype("int",new class value_int);
	INT_VALUE=new class value_int;

	cont->addtype("string",new class value_string);
	STRING_VALUE=new class value_string;

	cont->addtype("real",new class value_real);
	REAL_VALUE=new class value_real;

	cont->addtype("function",new class value_func);

	cont->addtype("char",new class value_char);
	CHAR_VALUE=new class value_char;

	xsmbuiltin(cont);
	cmd_window(cont);
	cmd_math(cont);
	cmd_file(cont);
}

//
// node_builtin::SetReturn value
//
// Sets the return value for a builtin function
//
// INPUTS :	value		The value to be returned at the end of
//				the builtin function.
// RETURNS:	None		 
//
void
node_builtin::SetReturn(class value *v)
{
	return_value=v;
}


//
// node_builtin::evaluate
//
// Evaluates a built in function using the specified arguments within
// the current context.
//
class value *
node_builtin::evaluate(class context *cont,
		class context *argcont,
		class node_arg *args
		)
{
	cur_context=argcont;
	cur_arg=args;

	getarg();
	//
	// Unless set otherwise the function returns nothing (SetReturn() )
	//
	return_value=NULL;

	funcaddr(this);

	freearg();

	if (return_value==NULL)
		return(	new class value );
	
	return(return_value);
}


//
// Constructor for node_builtin 
//
// When a builtin function declaration is built, the underlying 
// node_funcdecl() is not to important - so it gets created with
// lots of NULL values!
//
node_builtin::node_builtin(class context *cont,
		char *vname,
		int (*addr)(class node_builtin &) ) :
	node_funcdecl(	(class node_funcdecl *)NULL,
			(class node_funcdecl *)NULL,
			(class node *)NULL 
	             ) 
{
	cur_arg=NULL;
	cur_context=NULL;

	//
	// Create a symbol table entry 
	//
	cont->addfunction(vname,this);

	//
	// Set up the address for the function
	//
	funcaddr=addr;

        //
        // Builtin functions do not have a context, they get run
	// in the context of the function that calls them. By setting
	// func_context to NULL we ensure that callers do not
	// not change contexts.
        //
        func_context=NULL;

}

//
// Destructore for builtin nodes
//
node_builtin::~node_builtin()
{
}
