//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: busy.C,v 2.4 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: busy.C,v $
// Revision 2.4  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.3  1995/06/23  20:57:12  davidsi
// - BUG FIX.  Protect against a possible division by zero error when
//   the total count for the busy dialog has been set to 0.
//
// Revision 2.2  1995/03/08  21:18:12  davidsi
// - Busy::BusyCompleted() now accepts a -ve count as indicating 100% done,
//   but refresh regardless of previous percentage.
//
// Revision 2.1  1995/02/27  22:08:05  davidsi
// - Called to dialog_updatedisplay() are now made to DialogUpdatedisplay(),
//   so d_root no longer needs to be declared.
// - Moved the drawing of the rectangle, and added a clearpage() call, to
//   BusyTotal() so that the same window can be used multiple times.
// - Added percentcomplete member to the Busy class. This caches the previous
//   percentage value that was plotted.  This is used to save screen updates,
//   as an update only occurs when the percentage changes.
// - Added a BusyCompleted() member that also displays a label when updating
//   the completed bar.
//
// Revision 2.0  1995/02/05  18:38:25  davidsi
// - Initial version.
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"

#include "busy.h"
#include "dialog.h"

#define BUSY_HEIGHT	(72.0*1.5)
#define BUSY_WIDTH	(72.0*3.5)

#define BOX_X		(10.0)
#define BOX_Y		(40.0)
#define BOX_WIDTH	(BUSY_WIDTH-20.0)
#define BOX_HEIGHT	(20.0)


//
// This caches the width of "100%" in whatever font we choose.
//
static double widthof100;

//
// This ensures everything is consistent when we refer to 100%!
//
#define TEXT_100 "100%"

//
// BusyCompleted count
//
// Redisplays the bar graph showing how far the process has completed.
// 
// INPUTS :	count		An indication of how far a process has
//				completed.  It should be >0 and less than
//				the value passed to BusyTotal().
// RETURNS:	None
//
void
Busy::BusyCompleted(int count,char *name)
{
	double factor;
	int perc;

	//
	// Prevent the box from overflowing.
	//
	if (count>maxcount)
		count=maxcount;

	if (count>=0 && (fabs(maxcount)>SMALL_NUMBER) )
		perc=(int) ( ((double)count/(double)maxcount)*100.0 );
	else
		perc=100;

	//
	// Only do the display updates if the percentage has changed
	//
	if (perc==percentcomplete && count>=0)
		return;
	percentcomplete=perc;

	factor= ((double)perc)*(BOX_WIDTH/100.0);
	dd->DrawRectangle(BOX_X,BOX_Y,factor,BOX_HEIGHT+0.5);

	if (name)
		BusyMessage(name);

	dd->DrawPage();
	DialogUpdatedisplay();
}

void
Busy::BusyCompleted(int count)
{
	BusyCompleted(count,NULL);
}


//
// BusyMessage msg
//
// Redisplays the message in the busy window.
// 
// INPUTS :	msg		An ascii string.
// RETURNS:	None
//
void
Busy::BusyMessage(char *msg)
{
	dd->DrawSetcolour(BACKGROUND);
	dd->DrawRectangle(BOX_X,BOX_Y-15.0-textheight,BUSY_WIDTH,15.0);
	dd->DrawSetcolour(FOREGROUND);
	dd->DrawMoveto(BOX_X,BOX_Y-15.0);
	dd->DrawText(msg);
}


//
// BusyTotal count
//
// Specifies the largest value that will be passed to BusyCompleted().
// If this function member is not used, the maximum defauls to 100.
// 
// INPUTS :	count		The largest work indicator value
// RETURNS:	None
//
void
Busy::BusyTotal(int count)
{
        dd->DrawClearpage();
	dd->DrawSetcolour(FOREGROUND);
	dd->DrawMoveto(BOX_X,BOX_Y);
	dd->DrawDrawto(BOX_X+BOX_WIDTH,BOX_Y);
	dd->DrawDrawto(BOX_X+BOX_WIDTH,BOX_Y+BOX_HEIGHT);
	dd->DrawDrawto(BOX_X,BOX_Y+BOX_HEIGHT);
	dd->DrawDrawto(BOX_X,BOX_Y);

	dd->DrawMoveto(BOX_X,BOX_Y+BOX_HEIGHT+textheight);
	dd->DrawText("0%");
	dd->DrawMoveto(BOX_X+BOX_WIDTH-widthof100,BOX_Y+BOX_HEIGHT+textheight);
	dd->DrawText(TEXT_100);

	percentcomplete = -1;
	maxcount=count;
}


//
// Constructor 
//
Busy::Busy()
{
	drawd_reg_t drt;

	drt.flags=INITIAL_SIZE_SET|POPUP;
        drt.width=BUSY_WIDTH;
        drt.height=BUSY_HEIGHT;
        drt.title="Busy";
	drt.mb=NULL; //new menu_bar(NULL,NULL,0);
	dd=new Drawdialog();
	dd->DrawdialogCreate(&drt);
        dd->DrawStartpage();
	dd->DrawSetfont("Helvetica",12);
	
	textheight=dd->DrawTextheight();
	widthof100=dd->DrawTextlength(TEXT_100);

	maxcount=100;
	BusyCompleted(0);
}

//
// Destructor 
//
Busy::~Busy()
{
	delete dd;
}
