
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: cmd_file.C,v 2.1 1995/06/07 20:58:00 davidsi Exp $
//
// $Log: cmd_file.C,v $
// Revision 2.1  1995/06/07  20:58:00  davidsi
// - Improved eof() so that it can report files that could not
//   be opened correctly.
//
// Revision 2.0  1995/06/02  20:07:46  davidsi
// - Initial version.
//
//
//


#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>

#include "localmacros.h"
#include "support.h"
#include "security.h"
#include "menu.h"
#include "node.h"
#include "builtin.h"
#include "error.h"

extern "C" {
extern int _filbuf(...);
}

extern errstream cerror;
extern diagstream cinfo;


//
// Variables used throughout this module
//
static int Valuefile_id=0;
static class Valuefile *FILE_VALUE;

//
/////////////////////////////////////////////////////////////////////////////
//
// The File and Valuefile classes provide users with a file type
//
/////////////////////////////////////////////////////////////////////////////
//


//
// Note the file functionality is issolated from the value functionality
// because this needs to inherit from elsewhere and override virtual
// functions.
//
class File
{
	FILE *fptr; 
	char eof;
public:
	void FileOpen(char *name,char *mode);
	int FileWrite(char *txt);
	int FileWrite(char *ptr,int dsize);
	int FileRead(char *ptr,int dsize);
	int FileReadline(char *ptr,int dsize);
	int FileEof();
	File();
	~File();
};


//
// File::FileOpen name mode
//
// Create a file 
//
// INPUTS :	name		The name of the file
//		mode		Standard fopen modes.
// RETURNS:	None
//
void
File::FileOpen(char *name,char *mode)
{
	fptr=fopen(name,mode);
	eof=0;
}

//
// File::FileWrite string
//
// Writes a string to a file 
//
// INPUTS :	string		The string to write to the file
// RETURNS:	None
//
int
File::FileWrite(char *txt)
{
	if (fptr==NULL) {
		cerror << file_not_open << "\n";
		return (-1);
	}
	return (fprintf(fptr,txt));
}

//
// File::FileWrite ptr size 
//
// Writes an area of memory  to a file 
//
// INPUTS :	ptr		Start of data in memory
//		size		Number of bytes to write
// RETURNS:	None
//
int
File::FileWrite(char *ptr,int dsize)
{
	if (fptr==NULL) {
		cerror << file_not_open << "\n";
		return (-1);
	}
	return ( fwrite( ptr,dsize,1,fptr) );
}


//
// File::FileRead ptr size 
//
// Reads an area of memory from a file 
//
// INPUTS :	ptr		Start of data in memory
//		size		Number of bytes to write
// RETURNS:	-1 if error, else 0
//
int
File::FileRead(char *ptr,int dsize)
{
	int ret;

	if (fptr==NULL) {
		cerror << file_not_open << "\n";
		return (-1);
	}
	ret=fread( ptr,dsize,1,fptr);
	if (ret != 1) {
		eof=1;
		return( -1 );
	}
	return( ret );
}


//
// File::FileReadline ptr size 
//
// Read a line from a file 
//
// INPUTS :	ptr		Start of data in memory
//		size		Number of bytes to write
// RETURNS:	-1 if error, else 0
//
int
File::FileReadline(char *ptr,int dsize)
{
	int i;
	char c;

	if (fptr==NULL) {
		cerror << file_not_open << "\n";
		return (-1);
	}
        for(i=0;i<dsize-1;i++) {

                c=getc(fptr);

		//
		// Check for end of file
		//
                if (c == EOF) {
			eof=1;
                        break;
		}

		//
		// Check for end of line
		//
                if (c == '\n') 
                        break;

		ptr[i]=c;
        }
	ptr[i]='\0';

	if (eof) 
		return( -1 );

	return( 0 );
}


//
// File::FileEof
//
// Returns eof condition
//
// INPUTS :	None
// RETURNS:	1 if end of file
//
int
File::FileEof()
{
	if (fptr==NULL) 
		return (2);

	return ( eof );
}



//
// Constructor for the file class
//
File::File()
{
	eof=0;
	fptr=NULL;
}

//
// Destructor for the file class
//
File::~File()
{
	if (fptr)
		fclose(fptr);
}


class Valuefile : public value
{
	File **fileptr;
	//
	// The actual member is only used when building a file.  At 
	// that point no heap space is known (or necessarily allocated) 
	// so we need to point fileptr to somewhere safe until the assignment
	// that usually follows file creation happens.
	//
	File *actual;
public:
	class File *ValuefileGimmeFile() {
		if (fileptr==NULL) 
			DialogPanic("ValuefileGimmeFile - no file");
		return (*fileptr);
	}
	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual class value *clone();
	virtual char *name();
	//
	// These are somewhat unusual for a value derived class
	//
	void ValuefileOpen(char *name,char *mode);

	Valuefile();
	~Valuefile();
};


static class File *
GetFile(class value *val)
{
	class Valuefile *fileval;
	class File *file;

	if ( !(val->type_equal((class value *)FILE_VALUE)) ) {
		cerror << type_mismatch << "\n";
		return(NULL);
	}
        fileval=(class Valuefile *)val;
	file=fileval->ValuefileGimmeFile();
	return(file);
}



int
Valuefile::size()
{
	return( sizeof(*fileptr) );
}

void 
Valuefile::to_mem(char *mem)
{
#ifdef ASSERTS
	if (fileptr==NULL) {
		cerror << internal_address_not_set << " (Valuefile)\n";
		exit(1);
	}
#endif
	*((class File **)mem)=*fileptr;
}

void
Valuefile::from_mem(char *mem)
{
#ifdef ASSERTS
	if (fileptr==NULL) {
		cerror << internal_address_not_set << " (Valuefile)\n";
		exit(1);
	}
#endif
	*fileptr=*((class File **)mem);
}

void
Valuefile::assign_storage(char *mem)
{
	fileptr=(class File **)mem;
}

char *
Valuefile::fetch_storage()
{
#ifdef ASSERTS
	if (fileptr==NULL) {
		cerror << internal_address_not_set << " (Valuefile)\n";
		exit(1);
	}
#endif
	return((char *)fileptr);
}

char *
Valuefile::name()
{
	return( "file" );
}

class value *
Valuefile::clone()
{
	class Valuefile *val=new class Valuefile;


	val->fileptr=fileptr;

	return((class value *)val);
}

void
Valuefile::ValuefileOpen(char *name,char *mode)
{
	class File *w;

	w=*fileptr=new class File();
	w->FileOpen(name,mode);
}


Valuefile::Valuefile()
{
	//
	// At this point no heap space is known (or necessarily allocated)
	// so we need to point fileptr to somewhere safe until the assignment
	// that usually follows file creation happens.
	//
	fileptr=&actual;
	actual=NULL;
	value_flag=Valuefile_id;
}

Valuefile::~Valuefile()
{
}



//
//////////////////////////////////////////////////////////////////////////////
//
// The commands
//
//////////////////////////////////////////////////////////////////////////////
//

//
// file=fileopen("name","mode");
//
// Opens a file
//
static int
fileopen(class node_builtin &bptr)
{
	static char maskset=0;
	static class value *argmask[2];
	class Valuefile *filev;
	class File *file;
	char *name;
	char *mode;

	if (!maskset) {
		argmask[0]=STRING_VALUE;
		argmask[1]=STRING_VALUE;
		maskset=1;
	}

	if (bptr.types_of_arguments(2,argmask)==-1)
		return(-1);

        if ( ( bptr[0]->load(name) == -1 ) ||
	     ( bptr[1]->load(mode) == -1 ) ) {
		cerror << not_string << "\n";
                return(-1);
        }

	filev=new class Valuefile;
	filev->ValuefileOpen(name,mode);

	bptr.SetReturn(filev);
	return(0);
}

//
// Closes a file
//
// fileclose(file);
//
static int
fileclose(class node_builtin &bptr)
{
        class File *file;

	if (bptr.number_of_arguments()!=1) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

        if ( (file=GetFile(bptr[0]))==NULL )
                return(-1);

	delete file;
	return(0);
}


//
// Writes ascii to a file
//
// fileprint(file,args))
//
static int
fileprint(class node_builtin &bptr)
{
        class File *file;
	class value *val;
	char *str;
	int i;

	if (bptr.number_of_arguments()<2) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

	if ( (file=GetFile(bptr[0]))==NULL )
		return(-1);

	i=1;
	while ( (val=bptr[i])!=NULL) {
		if ( val->load(str)== -1 )
			cerror << not_string << "\n";
		else {
			if (file->FileWrite(str) == -1)
				return( -1 );
		}
		i++;
	}

        return(0);
}

//
// Writes binary data to a file
//
// filewrite(file,args))
//
static int
filewrite(class node_builtin &bptr)
{
        class File *file;
	class value *val;
	char *data;
	int size;
	int i;

	if (bptr.number_of_arguments()<2) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

	if ( (file=GetFile(bptr[0]))==NULL )
		return(-1);

	i=1;
	while ( (val=bptr[i])!=NULL) {
		data=val->fetch_storage();
		if (data==NULL)
			return( -1 );
		size=val->size();

		if (file->FileWrite(data,size) == -1)
			return( -1) ;

		i++;
	}

        return(0);
}


//
// Reads ascii from a file
//
// loadfile(file,args))
//
static int
loadfile(class node_builtin &bptr)
{
        class File *file;
	class value *val;
	char *str;
	int i;

	if (bptr.number_of_arguments()<2) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

	if ( (file=GetFile(bptr[0]))==NULL )
		return(-1);

	i=1;
	while ( (val=bptr[i])!=NULL) {
		if ( val->load(str)== -1 )
			cerror << not_string << "\n";
		else {
			if (file->FileWrite(str) == -1)
				return( -1 );
		}
		i++;
	}

        return(0);
}

//
// Reads binary data from a file
//
// fileread(file,args))
//
static int
fileread(class node_builtin &bptr)
{
        class File *file;
	class value *val;
	class value_int *retval;
	char *data;
	int size;
	int i;

	if (bptr.number_of_arguments()<2) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

	if ( (file=GetFile(bptr[0]))==NULL )
		return(-1);

	i=1;
	while ( (val=bptr[i])!=NULL) {
		data=val->fetch_storage();
		if (data==NULL)
			return( -1 );
		size=val->size();

		if (file->FileRead(data,size) == -1) {
			break;
		}

		i++;
	}

	retval=new class value_int;
	retval->assign_storage(NULL);
	retval->store(i-1 );
	bptr.SetReturn(retval);

        return(0);
}

//
// Reads a line(s) from a file
//
// filereadline(file,args))
//
static int
filereadline(class node_builtin &bptr)
{
        class File *file;
	class value *val;
	class value_int *retval;
	char *data;
	int size;
	int i;

	if (bptr.number_of_arguments()<2) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

	if ( (file=GetFile(bptr[0]))==NULL )
		return(-1);

	i=1;
	while ( (val=bptr[i])!=NULL) {
		data=val->fetch_storage();
		if (data==NULL)
			return( -1 );
		if (val->id() != VALUE_STRING) {
			cerror << type_mismatch << "\n";
			return( -1 );
		}
		size=val->size();

		if (file->FileReadline(data,size) == -1)
			break;

		i++;
	}

	retval=new class value_int;
	retval->assign_storage(NULL);
	retval->store( i-1 );
	bptr.SetReturn(retval);

        return(0);
}


//
// Reads file status
//
// fileeof(file)
//
static int
fileeof(class node_builtin &bptr)
{
        class File *file;
	class value *val;
	class value_int *retval;

	if (bptr.number_of_arguments()<1) {
		cerror << incorrect_num_arguments << "\n";
		return(-1);
	}

	if ( (file=GetFile(bptr[0]))==NULL )
		return(-1);

	retval=new class value_int;
	retval->assign_storage(NULL);
	retval->store(file->FileEof() );
	bptr.SetReturn(retval);

        return(0);
}

//
// This function will be called each time a program is loaded and
// "compiled".  It should initialize all the necessary entries
// in the program's symbol table.
//
void
cmd_file(class context *cont)
{
	extern int allocate_type_id();
	static char create_types=1;

	new node_builtin(cont,"fileopen",fileopen);
	new node_builtin(cont,"fileclose",fileclose);
	new node_builtin(cont,"fileprint",fileprint);
	new node_builtin(cont,"filewrite",filewrite);
	new node_builtin(cont,"fileread",fileread);
	new node_builtin(cont,"filereadline",filereadline);
	new node_builtin(cont,"eof",fileeof);

	//
	// New types
	//
	if (create_types) {
		Valuefile_id=allocate_type_id();
		FILE_VALUE=new class Valuefile;

		create_types=0;
	}
	cont->addtype("FILE",new class Valuefile);
}
