
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: cmd_math.C,v 2.2 1995/06/17 20:43:04 davidsi Exp $
//
// $Log: cmd_math.C,v $
// Revision 2.2  1995/06/17  20:43:04  davidsi
// - Removed the pi() function declaration and provide the same
//   functionality through the new PI constant.
//
// Revision 2.1  1995/06/07  20:58:30  davidsi
// - Removed second definition of the sin() function.
//
// Revision 2.0  1995/06/02  20:07:22  davidsi
// - Initial version.
//
//


#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"
#include "support.h"
#include "node.h"
#include "builtin.h"
#include "error.h"

extern errstream cerror;
extern diagstream cinfo;



//
//////////////////////////////////////////////////////////////////////////////
//
// The math functions
//
//////////////////////////////////////////////////////////////////////////////
//

//
// fprepare1 bptr argument
// 
// Unwraps arguments and set return value for math based functions that
// require a single argument.
//
// INPUTS :	bptr		The node information relating to args
// RETURNS:	argument	The argument that was passed to the function
//		NULL if error, else a pointer to value_real that should
//		be assigned the relevant value.
//
class value_real *
fprepare1(class node_builtin &bptr,double &argument)
{
	class value_real *argr;
	char *convstring;

	if (bptr.number_of_arguments()!=1) {
		cerror << incorrect_num_arguments << "\n";
		return(NULL);
	}
	
	//
	// Make the argument into a real 
	//
	if (bptr[0]->load(argument) == -1) {
		cerror << not_real<< "\n";
		return(NULL);
	}

	argr=new class value_real;
	argr->assign_storage(NULL);
	bptr.SetReturn(argr);
	return(argr);
}

//
// fprepare2 bptr arg1 arg2
// 
// Unwraps arguments and set return value for math based functions that
// require two arguments.
//
// INPUTS :	bptr		The node information relating to args
// RETURNS:	argument	The argument that was passed to the function
//		NULL if error, else a pointer to value_real that should
//		be assigned the relevant value.
//
class value_real *
fprepare2(class node_builtin &bptr,double &arg1,double &arg2)
{
	class value_real *argr;
	char *convstring;

	if (bptr.number_of_arguments()!=2) {
		cerror << incorrect_num_arguments << "\n";
		return(NULL);
	}
	
	//
	// Make the arguments into reals 
	//
	if (bptr[0]->load(arg1) == -1) {
		cerror << not_real<< "\n";
		return(NULL);
	}
	if (bptr[1]->load(arg2) == -1) {
		cerror << not_real<< "\n";
		return(NULL);
	}

	argr=new class value_real;
	argr->assign_storage(NULL);
	bptr.SetReturn(argr);
	return(argr);
}


//


//
// MATH_FUNC1 will define a function with the name "functionname"
// that returns the value from the function "function" when that
// function requires a single argument.
//
#define MATH_FUNC1(functionname,function)	\
static int					\
functionname(class node_builtin &bptr)		\
{						\
	double argument;			\
	class value_real *argr;			\
	argr=fprepare1(bptr,argument);		\
	if (argr == NULL)			\
		return( -1 );			\
	argr->store( function(argument) );	\
	return(0);				\
}


//
// MATH_FUNC2 is as MATH_FUNC1 but handles 2 argument functions
//
#define MATH_FUNC2(functionname,function)	\
static int					\
functionname(class node_builtin &bptr)		\
{						\
	double arg1,arg2;			\
	class value_real *argr;			\
	argr=fprepare2(bptr,arg1,arg2);		\
	if (argr == NULL)			\
		return( -1 );			\
	argr->store( function(arg1,arg2) );	\
	return(0);				\
}


//
// Define all 1 argument math functions
//
MATH_FUNC1(facos,acos);
MATH_FUNC1(fasin,asin);
MATH_FUNC1(fatan,atan);
MATH_FUNC1(fcos,cos);
MATH_FUNC1(fsin,sin);
MATH_FUNC1(ftan,tan);
MATH_FUNC1(fcosh,cosh);
MATH_FUNC1(fsinh,sinh);
MATH_FUNC1(ftanh,tanh);
MATH_FUNC1(fexp,exp);
MATH_FUNC1(flog,log);
MATH_FUNC1(flog10,log10);
MATH_FUNC1(fsqrt,sqrt);
MATH_FUNC1(fceil,ceil);
MATH_FUNC1(ffabs,fabs);
MATH_FUNC1(ffloor,floor);

//
// Define all 2 argument math functions
//
MATH_FUNC2(fpow,pow);
MATH_FUNC2(ffmod,fmod);
MATH_FUNC2(fatan2,atan2);



//
// This function will be called each time a program is loaded and
// "compiled".  It should initialize all the necessary entries
// in the program's symbol table.
//
void
cmd_math(class context *cont)
{
	extern int allocate_type_id();
	class value_real *vr;
	static char create_types=1;

	//
	// Create pre-defined variables and types
	//

	vr=new class value_real;
	vr->assign_storage(NULL);
	vr->store(M_PI);
	cont->ContextAddconstant("PI",vr);

	//
	// 1 Argument functions
	//
	new node_builtin(cont,"acos",facos);
	new node_builtin(cont,"asin",fasin);
	new node_builtin(cont,"atan",fatan);
	new node_builtin(cont,"cos",fcos);
	new node_builtin(cont,"sin",fsin);
	new node_builtin(cont,"tan",ftan);
	new node_builtin(cont,"cosh",fcosh);
	new node_builtin(cont,"sinh",fsinh);
	new node_builtin(cont,"tanh",ftanh);
	new node_builtin(cont,"exp",fexp);
	new node_builtin(cont,"log",flog);
	new node_builtin(cont,"log10",flog10);
	new node_builtin(cont,"sqrt",fsqrt);
	new node_builtin(cont,"ceil",fceil);
	new node_builtin(cont,"fabs",ffabs);
	new node_builtin(cont,"floor",ffloor);

	//
	// 2 Argument functions
	//
	new node_builtin(cont,"pow",fpow);
	new node_builtin(cont,"fmod",ffmod);
	new node_builtin(cont,"atan2",fatan2);
}
