
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: context.C,v 2.4 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: context.C,v $
// Revision 2.4  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.3  1995/06/17  20:44:30  davidsi
// - Due to the new statement interpreter, the returnvalue members
//   have been cleaned up since they no longer need to mess with
//   error states.
// - Added ContextAddconstant() to allow constants to be added to
//   the symbol table.
// - Tided up comments.
//
// Revision 2.2  1995/06/02  19:57:06  davidsi
// - BUG FIX. context::addfunction() needs to assign storage to the
//   created value_func member.
//
// Revision 2.1  1995/04/03  20:41:27  davidsi
// - Removed variableaddr() and added ContextSym2addr() and ContextName2sym()
//   functions member so that names can be converted to symbols which in
//   turn can be converted to addresses without a name->symbol lookup.
//   This improves performance.
// - Renamed addvariable() to ContextAddvariable().
//
// Revision 2.0  1995/01/19  21:44:39  davidsi
// - Initial version.
//
//
//
//


#include "stdio.h"
#include "stdlib.h"
#include "memory.h"

#include "context.h"
#include "error.h"

extern errstream cerror;
extern diagstream cinfo;


//
// context::ContextAddvariable varname value
//
// Adds a variable definition to a context.  The type of the variable
// must be specified using a value member but the definition itself
// will use a clone so the original member can be deleted or re-used.
//
// INPUTS :	varname		The name of the variable
//		value		The type of the variable
// RETURNS:	None
//
void
context::ContextAddvariable(char *vname,class value *val)
{
	class psymbol *psym=new psymbol;
	int size;
	class value *val_clone;

	//
	// Variable entries in the symbol table have their own
	// value so that they do not constantly need to create them
	// on the fly.   If they shared a single value then writing
	// to one variable could affect another when the two variables
	// are in simultaneous use, eg during a function call to a 
	// C function.
	//
	val_clone=val->clone();


	//
	// Ensure the underlying value is "fundamental" so that
	// any dispose calls are ignored.  Remember that the
	// cloned value here is the value that ends up being held
	// in the symbol table.
	//
	val_clone->setstatus(fundamental);

	//
	// Note that we store the context pointer, so that
	// given a symbol we know its context - we don't
	// therefore have to search
	//
	psym->value.offset=localoffset;
	psym->val=val_clone;
	psym->flag=cf_offset;
	psym->context=this;

	size=psym->val->size();
	INFO(cinfo << "[CONTEXT " << name << "] adding variable "<< vname;)
	INFO(cinfo <<" "<<"size="<<size << " offset=" << localoffset << "\n";)

	symvar.addsymbol(vname,psym);
	localoffset += size;
}



//
// context::ContextAddconstant constname value
//
// Adding a constant is very similar to adding a variable except 
// we pretend the constant is a function (in truth it is a function
// with no arguments!) so that it has not valid address. The value
// definition passed should be set to the value of the constant
// by the caller prior to this function being called.  The value
// is not cloned and therefore should not be touched or modified
// once passed to this function.
//
// INPUTS :	constname	The name of the constant
//		value		The type and calue of the constant
// RETURNS:	None
//
void
context::ContextAddconstant(char *vname,class value *val)
{
	class psymbol *psym=new psymbol;
	int size;

	//
	// Ensure the underlying value is "fundamental" so that
	// any dispose calls are ignored. 
	//
	val->setstatus(fundamental);

	//
	// Note that we store the context pointer, so that
	// given a symbol we know its context - we don't
	// therefore have to search
	//
	psym->value.offset=localoffset;
	psym->val=val;
	psym->flag=cf_function;
	psym->context=this;

	size=psym->val->size();
	INFO(cinfo << "[CONTEXT " << name << "] adding constant "<< vname;)
	INFO(cinfo << " " << "size=" <<size<< " offset=" <<localoffset<< "\n";)

	symvar.addsymbol(vname,psym);
	localoffset += size;
}



//
// context::ContextName2sym symname
//
// Returns the symbol definition for a variable/constant/function name
// in a context (or any contexts the the current context lies within).
//
// INPUTS :	symname		The name of the symbol to search for
// RETURNS:	The symbol definition or NULL if not found.
//
class psymbol *
context::ContextName2sym(char *vname)
{
	class psymbol *psym;

        psym=(class psymbol *)symvar.findsymbol(vname);
	if (psym==NULL && parentcontext) 
		return(parentcontext->ContextName2sym(vname));

	if (psym==NULL)
		printf("Symbol %s not defined\n",vname);

	return ( psym );
}



//
// context::ContextSym2addr psym c_type
//
// Converts a symbol definition into an active address and value
// descriptor.
//
// INPUTS :	psym		The symbol in question
// RETURNS:	c_type		The type of the symbol
//	 	NULL if error, else the address in the current context
//		of the storage for the specified symbol.
//
char *
context::ContextSym2addr(class psymbol *psym,class value *&c_type)
{
	if (psym== NULL) {	
		c_type=NULL;
		return(NULL);
	}

	c_type=psym->val;
	switch(psym->flag) {
		case cf_offset:
			return (psym->value.offset+psym->context->heap);
		case cf_function:
			return(NULL);
		case cf_addr:
			return( psym->value.addr);
	}
	c_type=NULL;
	return(NULL);
}


//
// context::addtype typename	value
//
// Adds a type definition in the current context
//
// INPUTS : 	typename		The name of the type to add
//		value			The definition of the type
// RETURNS:	None
//
void
context::addtype(char *vname,class value *val)
{
	class psymbol *psym=new psymbol;

	psym->val=val;

	INFO(cinfo << "[CONTEXT " << name << "] adding type "<< vname <<"\n";)

	symtype.addsymbol(vname,psym);
}


//
// context::findtype name
//
// Finds a type in the current context or any context that encloses
// the current context
//
// INPUTS :	name		The name of the type
// RETURNS:	The definition of the type, or NULL if not found.
//
class value *
context::findtype(char *vname)
{
	class psymbol *psym;

        psym=(class psymbol *)symtype.findsymbol(vname);
	if (psym==NULL && parentcontext) 
		return(parentcontext->findtype(vname));
	
	if (psym== NULL) {	
		printf("Cannot find %s\n",vname);
		return(NULL);
	}
	//
	// NOTE: Can only find type definitions in the symtype
	// table so we don't need to check anything else!
	//
	return(psym->val);
}



//
// context::addstructure name value
//
// Adds a structure definition in the current context 
//
// INPUTS :	name		The name of the structure
//		value		The type definition for the structure
// RETURNS:	None
//
void
context::addstructure(char *vname,class value *val)
{
	class psymbol *psym=new psymbol;

	psym->val=val;
	INFO(cinfo <<"[CONTEXT " <<name<< "] adding structure "<< vname <<"\n";)

	symstruct.addsymbol(vname,psym);
}



//
// context::findstructure name
//
// Finds a structure defintion in the current context or any context
// that encloses the current context.
//
// INPUTS :	name		The name of the structure
// RETURNS:	NULL if error, else the definition of the structure
//
class value *
context::findstructure(char *vname)
{
	class psymbol *psym;

        psym=(class psymbol *)symstruct.findsymbol(vname);
	if (psym==NULL && parentcontext) 
		return(parentcontext->findstructure(vname));
	
	if (psym== NULL) {	
		printf("Cannot find %s\n",vname);
		return(NULL);
	}
	//
	// NOTE: Can only find structure definitions in the symstruct
	// table so we don't need to check anything else!
	//
	return(psym->val);
}



//
// context::addfunction fname functionnode
//
// Adds a function to the current context.
//
// INPUTS :	fname		The name of the function
//		functionnode	The definition of the function
// RETURNS:	None
//
void
context::addfunction(char *vname,class node_funcdecl *nodeptr)
{
	class psymbol *psym=new class psymbol;
	class value_func *val=new class value_func();

	//
	// Paranoia!
	//
	val->setstatus(fundamental);
	INFO(cinfo << "Function node is at "<<(address)nodeptr << "\n";)
	val->assign_storage(NULL);
	val->from_mem((char *)&nodeptr);

	psym->flag=cf_function;
	psym->value.nfuncptr=nodeptr;
	psym->val=val;
        symvar.addsymbol(vname,psym);
	INFO(cinfo << "[CONTEXT " << name << "] adding function "<< vname;)
	INFO(cinfo << " " << "node=" << (address)nodeptr << "\n";)
}


//
// context::findfunction name
//
// Finds a function defintion in the current context or any context
// that encloses the current context.
//
// INPUTS :	name		The name of the function
// RETURNS:	NULL if error, else the definition of the function
//
class node_funcdecl *
context::findfunction(char *vname)
{
	class psymbol *psym;

        psym=(class psymbol *)symvar.findsymbol(vname);
	if (psym==NULL && parentcontext) 
		return (parentcontext->findfunction(vname));

        if (!psym) {
                printf("Cannot find function %s\n",vname);
                return(NULL);
        }
        return( psym->value.nfuncptr );
}


//
// context::assign vname addr
//
// Assigns a variable a specific (hard wired!) address.  This is used
// when passing values by reference (rather than value) and the function
// needs to access the variables in the calling function rather than
// local copies.
//
// INPUTS :	vname		The name of the variable
//		addr		The address to be assigned to the variable
// RETURNS:	None
void
context::assign(char *vname,char *addr)
{
        class psymbol *psym;

        psym=(class psymbol *)symvar.findsymbol(vname);

        if (!psym) {
                printf("Cannot find symbol %s for assignment\n",vname);
		exit(1);
        }
	psym->flag=cf_addr;
	psym->value.addr=addr;
}


//
// context::getheap
//
// Returns the heap being used (currently) by the context
//
// INPUTS :	None
// RETURNS:	The current heap for the context
//
char *
context::getheap()
{
	return (heap);
}


//
// context::createheap size
//
// Creates a new heap of given size for the context.
//
// INPUTS :	size		The size of the heap to create
// RETURNS:	-1 if error, else 0
//
int 
context::createheap(int size)
{
	heap=(char *)malloc(size);
	if (heap==NULL) {
		fprintf(stderr,"Could not allocate %d bytes for heap\n",size);
		return (-1);
	}
	//
	// Note that all values on the heap are GUARANTEED to be 0 when
	// they have not yet been set.  This is vital for those data
	// types which maintain pointers on the heap.
	//
	memset(heap,0,size);
	return(0);
}


//
// context::createheap
//
// Creates a new heap for the context.
//
// INPUTS :	None
// RETURNS:	-1 if error, else 0
int
context::createheap()
{
	char *current_heap=heap;
	char **heapptr;

	//
	// The created heap is made sizeof(char *) larger so that
	// the current heap pointer can be stored on the new heap.
	// The heap pointer will be stored at the end of the heap
	// space.
	if (createheap(localoffset+sizeof(char *) )==-1) 
		return (-1 );

	heapptr=(char **)(heap+localoffset);
	*heapptr=current_heap;

	return(0);
}



//
// context::unlock_heap oldheap
//
// Releases the storage that was associated with a heap that was
// previously popped.
//
// INPUTS :	oldheap		The address of the heap - see lock_heap()
// RETURNS:	None
void
context::unlock_heap(char *oldheap)
{
	if (oldheap)
		free(oldheap);
}


//
// context::lock_heap
//
// Locks a heap so that it will NOT be freed when it is poped.  To allow
// values to be returned from functions it is possible to keep a heap
// lying around after the function has completed executing.  Once a copy
// of the value(s) has been obtained the heap can be deleted.
//
// INPUTS : 	None
// RETURNS: 	The address of the heap
//
char *
context::lock_heap()
{
	heapref=1;
	return(heap);
}


//
// context::popheap
//
// Pops the heap in the current context. Remember a context is a description
// of the variables(etc) that are in scope at any one moment in time. A
// heap stores the values of those variables, so there can be multiple heaps
// for a single context (think of a recursive function).
//
// INPUTS :	None
// RETURNS:	None
//
void
context::popheap()
{
        char *previous_heap;
	char **heapptr;

	heapptr=(char **)(heap+localoffset);
	previous_heap=*heapptr;

	
	//
	// Only free the heap if the reference count is 0.  To allow
	// values to be returned from functions it is possible to
	// keep a heap lying around after the function has completed
	// executing.  Once a copy of the value(s) has been obtained
	// the heap can be deleted.
	//
	if (heap && heapref==0)
		free(heap);
	heapref=0;
	heap=previous_heap;
}


//
// returnvalue
//
// Returns the return value for a context. This is typically used when
// a function has completed and the system has returned to the calling
// function.
// 
// INPUTS :	None
// RETURNS:	The value that was set as a return value.  This will
//		require disposal.
//
class value *
context::returnvalue()
{
	return(retval);
}


//
// context::returnvalue value
//
// Set a return value for the current context.  This should only
// be called once for an active context - otherwise a memory
// leak occurs.
// 
// INPUTS :	value		The return value for the context
// RETURNS:	None
//
void
context::returnvalue(class value *val)
{
	retval=val;
}

//
// Constructor for the context class
//
context::context(class context *lcontext,char *s)
{
	name=(char *)malloc(strlen(s)+1);
	strcpy(name,s);
	
	localoffset=0;		// offset for local variables
	heap=NULL;		// no heap space yet allocated
	parentcontext=lcontext;
	heapref=0;
}

//
// Destructor for the context class
//
context::~context()
{
}
