
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dchange.C,v 2.7 1995/05/18 20:56:44 davidsi Exp $
//
// $Log: dchange.C,v $
// Revision 2.7  1995/05/18  20:56:44  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
// - BUG FIX. In change_display::list_change() protect against a div by zero
//   error, when the previous price was 0.
//
// Revision 2.6  1995/04/25  21:13:17  davidsi
// - Added support for a busy cursor.
// - Updated to use ColourSetowner() so that the user's colour prefences for
//   the changes display can be pernamently stored in the preference
//   database.
// - BUG FIX. In list_change() flags can change underneath you, so
//   it is necessary to use lastflags for all checks on the flags
//   set by the user.
// - BUG FIX. In list_change(), the types string that is built for the
//   printer heading was not being correctly initialized.
//
// Revision 2.5  1995/04/16  20:32:41  davidsi
// - Added support for TextdialogSetprintertitle().  Now an appropriate
//   title is generated for printer output.
// - Added callback_sec member to allow price change displays to be updated
//   when securities are added, removed or modified.  This gets registered
//   and removed as appropriate.
//
// Revision 2.4  1995/03/08  21:25:01  davidsi
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that.
//
// Revision 2.3  1995/02/27  22:00:38  davidsi
// - Added full colour support.  This allows losers, gainers and no changers
//   to be readily recognised.
// - Many more (new) menu options including, support for ordering and
//   selection.
// - BUG FIX. Because the Busy class is used, further menu requests can
//   be processed while processing an earlier request. This could
//   cause code dumps, because underlying datastructures could change
//   underneath the display update code.  Restrict access using a
//   Semaphore instance.
//
// Revision 2.2  1995/02/05  17:56:49  davidsi
// - Added support for the new Busy class.  A busy window is now shown while
//   determing what to display.
//
// Revision 2.1  1995/01/31  21:52:24  davidsi
// - Revised to use the textdialog class.  This results in much of
//   the required functionality now being inherited.
// - Added a callback to handle transactions - if a transaction
//   occurs it could potentially require a change to the displayed
//   information.
// - Now supports rises/falls in colour!
//
// Revision 2.0  1995/01/19  22:01:00  davidsi
// - Initial version.
//
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"

#include "support.h"
#include "busy.h"
#include "dchange.h"
#include "security.h"
#include "callback.h"

//
// Convenience defines for colours
//
#define DCOLOUR_NOCHANGE	FOREGROUND
#define DCOLOUR_LOSERS		COLOUR_FUNCTION1
#define DCOLOUR_GAINERS		COLOUR_FUNCTION2

#define DCOLOUR_NOCHANGE_FNC	COLOUR_FUNCTION_FOREGROUND
#define DCOLOUR_LOSERS_FNC	COLOUR_FUNCTION_EXT1
#define DCOLOUR_GAINERS_FNC	COLOUR_FUNCTION_EXT2

//
// Supported menu option on the valuation window
//
enum menu_e {
	MENU_EXIT,			// quit
	MENU_COLOUR,			// Set colours
	MENU_PRINT,			// print
	MENU_PORTFOLIO,			// show only securities in portfolio
	MENU_ALL,			// show all securities
	MENU_NUML,			// ordering - losers first
	MENU_NUMG,			// ordering - gains first
	MENU_ALPHA,			// ordering - alphabetical
	MENU_SELG,			// Show gainers
	MENU_SELL,			// Show losers
	MENU_SELN,			// Show no changers
};


//
// Defines for the database identifiers
//
#define CHGDISP_DB_CLASS "CHANGEDISP"

//
// A change list inherits the list class characteristics but enforces
// an ordering on the members to provide for gainers->losers or
// losers->gainers ordering on price changes.
// 
// The members held on the list are all instances of the change 
// class.
//
class change_list : public list {
	int flags;
	int lastflags;
protected:
	virtual int compare(class list_node *,class list_node *);
public:
	void add(char *str,double amount);
	void fetch(class change_display *,class Busy *b);
	change_list(int);
	~change_list();
};

class change : public list_node {
	double amount;
	char *string;
public:
	change(char *str,double amount);
	virtual ~change();

friend class change_list;
};


//
// change_list::compare l1 l2
//
// compare overrides the standard list compare function. This
// function ensures the list of price changes is ordered as
// requested by the user.
//
// INPUTS :	l1,l2	The two noes to be compared
// RETURNS: 	Value of comparison
//
int
change_list::compare(class list_node *li1,class list_node *li2)
{
	if (flags & DCHG_FLAGS_ALPHA)
		return( alpha_strcmp( ((class change *)li1)->string,
			((class change *)li2)->string) );
	else {
		int a;
		double r1,r2;
		r1=((class change *)li1)->amount;
		r2=((class change *)li2)->amount;
		if (r1>r2)
			a= 1;
		else if (r1<r2)
			a= -1;
		else
			a=0;
		//
		// If Leaders (gains) first then reverse the flag
		//
		if (flags & DCHG_FLAGS_NUMG)
			a= -a;
			
		return(a);
	}
}

//
// change_list::add string amount
//
// Creates a change instance, initialised with the amount and string
// values passed, and inserts it into the main list.
//
// INPUTS :	str		The string to be displayed
// 		amount		The price change held within the string 
// RETURNS:	None 
void
change_list::add(char *str,double amount)
{
	class change *chg=new class change(str,amount);
	insert_order(chg);
}



//
// change_list::fetch changedisplay busy
//
// Creates the price change display on the basis of the price change
// list.  A busy dialog can be used by this member.
//
// INPUTS :	changedisplay		The change_display instance
//		busy			A busy instance or NULL
// RETURNS:	None
//
void
change_list::fetch(class change_display *t,class Busy *busy)
{
	class list_node *ptr;
	class change *cptr;
	int count=0;
	colour_e col;

	if (busy) {
		busy->BusyTotal( length() );
		busy->BusyMessage("Sorting securities");
	}
	ptr=head();
	while (!eol(ptr)) {
		if (busy) {
			count++;
			if (count % 10)
				busy->BusyCompleted(count);
		}

		cptr=(class change *)ptr;
		t->TextAdd(cptr->string);
		if (cptr->amount>0.0) { 
			col=DCOLOUR_GAINERS;
		} else if (cptr->amount<0.0) {
			col=DCOLOUR_LOSERS;
		} else 
			goto noset;

		t->TextSetattr( t->TextCline(), 
			new class attr(FIELDATTR_RIGHT,col));
		t->TextSetattr( t->TextCline(),1,
			new class attr(FIELDATTR_LEFT,col));
noset:
		ptr=ptr->next();
	}
}


//
// Constructor function for change_list
//
change_list::change_list(int fl)
{
	flags=fl;
}

//
// Destructor function for change_list
//
change_list::~change_list()
{
}

//
// Constructor function for change
//
change::change(char *istr,double iamount)
{
	string=(char *)malloc(strlen(istr)+1);
	strcpy(string,istr);
	amount=iamount;
}

//
// Destructor function for change
//
change::~change()
{
	free(string);
}


//
// change_display::list_change
//
// Writes the changes to the display.
//
// CAUTION: list_change may delete the change_display instance. NEVER
// call this function before any other references to the instance.
//
// INPUTS :	None
// RETURNS:	None
//
void
change_display::list_change()
{
	class security *ptr;
	char str[256];			// build up space, hope this is enough
	double number;			// number of securities owned 
	char secname[16];		// security name 
	double perc;			// % growth/decline 
	char *csecname;
	double change;
	double newprice,oldprice;
	class date start_date;
	class date end_date;
	class Busy *busy=NULL;
	int count;
	class change_list *changes;
	class Holding *holding;
	int i;
	char *group,*order,types[256];


	//
	// Don't redraw if the flags haven't changed
	//
	if (lastflags == flags)
		return;

	//
	// If an udpate is already underway don't recurse
	//
	if (update.SemaphoreIslocked() )
		return;
	update.SemaphoreLock();

	if (!setmenuflag) {
		MenuItemaction("Project","Close", MENU_EXIT);
		MenuItemaction("Project","Print", MENU_PRINT);
		MenuItemaction("Project","Colours", MENU_COLOUR);

		MenuItem1ofn("Options","Portfolio",MENU_PORTFOLIO,MENUSTATE_ON);
		MenuItem1ofn("Options","All",MENU_ALL,MENUSTATE_OFF);

		MenuItem1ofn("Order","Alphabetical",MENU_ALPHA,MENUSTATE_ON);
		MenuItem1ofn("Order","Losers First",MENU_NUML,MENUSTATE_OFF);
		MenuItem1ofn("Order","Leaders First",MENU_NUMG,MENUSTATE_OFF);

		MenuItemnofn("Selection","Leaders",MENU_SELG,MENUSTATE_ON);
		MenuItemnofn("Selection","Losers",MENU_SELL,MENUSTATE_ON);
		MenuItemnofn("Selection","No changers",MENU_SELN,MENUSTATE_OFF);

		for(i=2;i<6;i++) {
			class attr *attr;
			attr=new class attr(FIELDATTR_RIGHT,FOREGROUND);
			TextSetdefaultattr(i,attr);
		}


		ColourSetname(DCOLOUR_NOCHANGE_FNC,"No change");
		ColourSetname(DCOLOUR_LOSERS_FNC,"Losers");
		ColourSetname(DCOLOUR_GAINERS_FNC,"Gainers");

		ColourSet(DCOLOUR_NOCHANGE_FNC,BLACK);
		ColourSet(DCOLOUR_LOSERS_FNC,RED);
		ColourSet(DCOLOUR_GAINERS_FNC,BLUE);
		TextdialogSettitle("Changes In Price");
		ColourSetowner(CHGDISP_DB_CLASS);
		setmenuflag=1;
	}

	//
	// If the flags change while we are updating (possible because
	// the Busy class allows window events to be processed), then
	// we may decide to redraw the window before we quit..
	// 
redraw:
	DialogCursor(DCURSOR_NORMAL);
	//
	// Copy the flags that applied when the display was updated
	//
	lastflags=flags;
	changes=new class change_list(lastflags);
	count=0;

	//
	// Set printer title
	//
	if (lastflags & DCHG_FLAGS_ALL)
		group="all securities";
	else if (lastflags & DCHG_FLAGS_PORT)
		group="all securities in portfolio";
	else 
		group="";

	if (lastflags & DCHG_FLAGS_ALPHA)
		order="using alphabetic order";
	else if (lastflags & DCHG_FLAGS_NUML)
		order="using losers first order";
	else if (lastflags & DCHG_FLAGS_NUMG)
		order="using gainers first order";
	else
		order="";

	types[0]='\0';
	if (lastflags & DCHG_FLAGS_SHOWG)
		strcat(types,"gainers");
	if (lastflags & DCHG_FLAGS_SHOWL) {
		if (*types!='\0')
			strcat(types,",");
		strcat(types,"losers");
	}
	if (lastflags & DCHG_FLAGS_SHOWN) {
		if (*types!='\0')
			strcat(types,",");
		strcat(types,"no changers");
	}

	sprintf(str,"Prices changes for %s %s (%s)",group,order,types);
	TextdialogSetprintertitle(str);


	//
	// If we are doing a portfolio update ensure there is
	// no Busy dialog, otherwise ensure there is one.  Note that
	// this might involve deleting a Busy instance.
	//
	if ( !(lastflags & DCHG_FLAGS_PORT) && busy==NULL)
		busy=new class Busy;
	else if ( (lastflags & DCHG_FLAGS_PORT) && busy!=NULL) {
		delete busy;
		busy=NULL;
	}

	TextClear();

	sprintf(str,"Share Name\tNew Price\tLast Price\tChange (p)\t%% Change");
	TextAdd(str);

	end_date.setdate();

	if (busy)
		busy->BusyTotal(main_sec_list->length());
	//
	// For every known security do..
	//
	if (lastflags & DCHG_FLAGS_PORT) 
	        holding=portfolio->head();
	else
        	ptr=main_sec_list->head();

        for ( ;; ){
		//
		// Since updates can take a very (VERY) long time, a flag
		// is set by the Close menu item if an update is in 
		// progress.  That update is aborted and resources are
		// released.
		//
		if (lastflags & DCHG_FLAGS_EXIT) {
			if ( busy )
				delete busy;
			delete changes;
			delete this;
			return;
		}
		csecname=NULL;
		count++;
		if (lastflags & DCHG_FLAGS_PORT)  {
			if (portfolio->eol(holding))
				break;
			holding->HoldingTransactionhead()->extract(&start_date,
					&end_date,&number,NULL,NULL,NULL);
			if (  (fabs(number)<SMALL_NUMBER) ) 
				goto nextiter;	// security never in portfolio
			ptr=holding->HoldingSecurity();
			DialogCursor(DCURSOR_BUSY);
		} else {
			if (main_sec_list->eol(ptr) )
				break;
			if ( busy && (count%10)==0) {
				ptr->SecurityExtract(&csecname,NULL,NULL,
								NULL,NULL);
				busy->BusyCompleted(count,csecname);
			}
		}

		if (ptr->prices->PriceFetchchange(newprice,oldprice)== -1)
			goto nextiter;
		if (newprice<0.0 || fabs(oldprice)<SMALL_NUMBER)
			goto nextiter;
		change=newprice-oldprice;
		perc=(change/oldprice)*100.0;
		if ( (change<0.0 && lastflags&DCHG_FLAGS_SHOWL)
		   ||(change>0.0 && lastflags&DCHG_FLAGS_SHOWG)
		   ||((!(change<0.0)&&!(change>0.0)&&lastflags&DCHG_FLAGS_SHOWN))){
			if (csecname==NULL)
				ptr->SecurityExtract(&csecname,NULL,NULL,
							NULL,NULL);
			strncpy(secname,csecname,15);
			secname[15]='\0';
			sprintf(str,"%s\t%.4f\t%.4f\t%4.0f\t%.4f\n",
				secname,newprice,oldprice,100.0*change,perc);
			changes.add(str,change);
		}
		//
		// What we need to update with may change underneath us,
		// so check and restart if necessary.
		//
		if (lastflags != flags) {
			delete changes;
			goto redraw;
		}

nextiter:
		if (lastflags & DCHG_FLAGS_PORT)
			holding=holding->next();
		else 
			ptr=ptr->next();
	}

	changes.fetch(this,busy);
	if ( busy ) {
		delete busy;
		busy=NULL;
	}
	TextComplete();
        TextdialogOutput();
	delete changes;
	//
	// Check that the flags requested by the user did not change
	// while we were away processing the change information.
	//
	if (lastflags != flags)
		goto redraw;

	DialogCursor(DCURSOR_NORMAL);
	update.SemaphoreUnlock();
}




//
// change_display::MenuCallback client_data onflag
//
// This is the call back used when the user selects something on the 
// menu.  The function overrides the standard function in the
// inheritated menu class.  
//
// INPUTS :	client_data	The id for the menu item thats been selected.
//		onflag		Whether the menu item is on/off.
// RETURNS: 	None
//
void 
change_display::MenuCallback(unsigned long client_data, int onflag)
{

	switch (client_data) {
	case MENU_EXIT:
		//
		// Ignore EXIT calls if the colour dialog is displayed
		//
		if (colourdialog.SemaphoreIslocked())
			return;
		//
		// if we are locked, then set the exit flag 
		//
		if ( update.SemaphoreIslocked() )
			flags |= DCHG_FLAGS_EXIT;
		else
			delete this;
		return;
        case MENU_COLOUR:
		if (colourdialog.SemaphoreClock()!= -1) {
			ColourSelect();
			colourdialog.SemaphoreUnlock();
			TextdialogRedraw();
		}
		break;
	case MENU_PRINT:
		//
		// Don't print if a redisplay is underway - the
		// data structures won't be fully cooked!
		//
		if ( !update.SemaphoreIslocked() )
			TextdialogPrint();
		return;
	case MENU_ALL:
		flags &= ~DCHG_FLAGS_PORT;
		flags |= DCHG_FLAGS_ALL;
		break;
	case MENU_PORTFOLIO:
		flags &= ~DCHG_FLAGS_ALL;
		flags |= DCHG_FLAGS_PORT;
		break;
	case MENU_ALPHA:
		flags &= ~(DCHG_FLAGS_NUML|DCHG_FLAGS_NUMG);
		flags |= DCHG_FLAGS_ALPHA;
		break;
	case MENU_NUML:
		flags &= ~(DCHG_FLAGS_ALPHA|DCHG_FLAGS_NUMG);
		flags |= DCHG_FLAGS_NUML;
		break;
	case MENU_NUMG:
		flags &= ~(DCHG_FLAGS_ALPHA|DCHG_FLAGS_NUML);
		flags |= DCHG_FLAGS_NUMG;
		break;
	case MENU_SELG:		// show gainers
		if (onflag)
			flags |= DCHG_FLAGS_SHOWG;
		else
			flags &= ~DCHG_FLAGS_SHOWG;
		break;
	case MENU_SELL:		// show losers
		if (onflag)
			flags |= DCHG_FLAGS_SHOWL;
		else
			flags &= ~DCHG_FLAGS_SHOWL;
		break;
	case MENU_SELN:		// show non changers
		if (onflag)
			flags |= DCHG_FLAGS_SHOWN;
		else
			flags &= ~DCHG_FLAGS_SHOWN;
		break;
	default:
		DialogPanic("Bad switch: change_display::MenuCallback");
	}
	//
	// If an udpate is already underway don't initiate a further
	// update
	//
	if ( update.SemaphoreIslocked() )
		return;
	list_change();
}


//
// change_display::redisplay this 
//
// This callback is registered to execute whenever prices change or
// a transaction occurs.
//
// INPUTS :	tmpthis		The this ptr defined when the callback
//				was registered.
// RETURNS:	None
//
void
change_display::redisplay(char *tmpthis,char *)
{
	class change_display *rthis=(class change_display *)tmpthis;
	//
	// Ensure that the flags look as if they have changed, thus
	// causing  list_change to perform a real update
	//
	rthis->lastflags = ~rthis->flags;
	rthis->list_change();			// make display consistent
}



//
// ~change_display
//
// Constructor for the change display class.
//
// INPUTS : None
// RETURNS: None
//
change_display::change_display()
{

	//
	// default is to list only securities in portfolio  using
	// alphabetical ordering showing only losers and gainers.
	//
	flags=DCHG_FLAGS_PORT|DCHG_FLAGS_ALPHA|
		DCHG_FLAGS_SHOWL|DCHG_FLAGS_SHOWG;
	lastflags=0;
	setmenuflag=0;
	//
	// Set up call backs, for either transaction changes,
	// price changes or security changes.
	//
	callback_trans=main_trans_list->add_vect.CallbackAdd(
			&change_display::redisplay,
			(char *)this);
	callback_price=main_price_list->change_vect.CallbackAdd(
			change_display::redisplay,
			(char *)this);
	callback_sec=main_sec_list->changed_vect.CallbackAdd(
			change_display::redisplay,
			(char *)this);
	list_change();
}

//
// ~change_display
//
// Destructor for the change display class.
//
// INPUTS : None
// RETURNS: None
//
change_display::~change_display()
{
	//
	// Remove callbacks
	//
        main_trans_list->add_vect.CallbackRemove(callback_trans);
        main_price_list->change_vect.CallbackRemove(callback_price);
	main_sec_list->changed_vect.CallbackRemove(callback_sec);
}



//
// display_change dialog args
//
// Main entry point for creating & managing the window which displays 
// the change.
//
// INPUTS :	dialog	The parent dialog
//		args	To satisfy type checking
// RETURNS:	None.
///
static void
display_change(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
        change_display *newtrans=new change_display;
}


//
// dchange_init parent menu_bar
//
// Initialises the display changes module.
// 
// INPUTS :	None
// RETURNS:	None
//
void
dchange_init(menu_bar *mb)
{
        MENU("Prices","Change in price",display_change,0);
}
