
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: ddiv.C,v 2.6 1995/05/19 18:52:23 davidsi Exp $
//
// $Log: ddiv.C,v $
// Revision 2.6  1995/05/19  18:52:23  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.5  1995/04/25  21:13:35  davidsi
// - Added support for a busy cursor.
// - Updated to use ColourSetowner() so that the user's colour prefences for
//   the dividend display can be pernamently stored in the preference
//   database.
// - BUG FIX. When colours change it is not necessary to rework the
//   text (that can take time) - all that is needed is a quick redraw.
//
// Revision 2.4  1995/04/16  20:30:07  davidsi
// - Added a new menu item "Dates" which allows a range of dates to be
//   specified.  Only those dividends within those dates will be displayed.
// - Added the ability to get dividends on either shares, unit trusts, or
//   both.
// - Added support for TextdialogSetprintertitle().  Now an appropriate
//   title is generated for printer output.
// - Added callback_sec member to allow dividend displays to be updated when
//   securities are added, removed or modified.  This gets registered and
//   removed as appropriate.
// - BUG FIX. Being able to set different date ranges showed up a missing
//   set of parenthesis in create_display() causing incorrect dividends to
//   be displayed.
// - BUG FIX. The optionflags member was being used a bit mask for options,
//   but the enumerated type could not possibly quarantee the necessary
//   bit manipulations would work.  Now uses explicit defines.
//
// Revision 2.3  1995/03/08  21:19:57  davidsi
// - The extract() member of dividend was renamed, so this module modified
//   to reflect that.
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that.
// - Revised to use the new Holding class which contains dividend information
//   rather than the security class.
// - BUG FIX. Added redisplay() to handle callbacks for when dividends are
//   added.  Previously no callback handler was registered.
//
// Revision 2.2  1995/02/27  22:26:12  davidsi
// - Added call to TextComplete() to indicate when processing is complete.
//   This allows the text class to prevent updates occuring while this
//   class is still handling a previous update.
// - Code is now dialog free - tided up menu handling (etc).
// - BUG FIX. When building the text dialog the left/right justification was
//   not being set for each of the fields.  Also now sets colour information.
//
// Revision 2.1  1995/01/31  21:41:44  davidsi
// - Revised to use the textdialog class.  This results in much of
//   the required functionality now being inherited.
//
// Revision 2.0  1995/01/19  21:52:02  davidsi
// - Renamed all calls to date2string() which is now getdate().
// - Changed all references to time to use the new date class.
//
// Revision 1.4  1994/12/11  18:01:33  davidsi
// - In create_display() remove an unnecessary check on whether a dividend
//   is valid.
// - BUG FIX.  When displaying totals for each security, the values displayed
//   were not the totals but the values from the last dividend structure.
//
// Revision 1.3  1994/11/05  21:05:59  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.2  1994/10/17  20:48:21  davidsi
// - All calls to extract() in the security class now require a further
//   argument, so that the type of security can be determined using flags
//   rather than strings.
// - Now that per security dividend lists are circular the termination
//   condition for scanning the list must be updated for having completed
//   the entire list.
// - BUG FIX. Ensure that the menu items are correctly enabled when the window
//   is first built.
//
// Revision 1.1  1994/06/28  21:00:30  davidsi
// Initial revision
//
//
//

/*
 * Include Files
 */
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>

#include "localmacros.h"

#include "ddiv.h"
#include "setdate.h"

//
// Menu identifiers
//
enum menu_e {
	MENU_CLASS,
	MENU_STOCK,
	MENU_CASH,
	MENU_PRINT,
	MENU_EXIT,
	MENU_COLOUR,
	MENU_DATE,
	MENU_SHARES,
	MENU_UNITTRUSTS,
	};

//
// Defines for the database identifiers
//
#define DIVDISP_DB_CLASS "DIVDISP"


//
// Flags for optionflags
//
#define OPTIONS_STOCK		0x1
#define OPTIONS_CASH		0x2
#define OPTIONS_SHARES		0x4
#define OPTIONS_UNITTRUSTS	0x8

//
// dividends_display::MenuCallback client_data onflag
//
// This is the call back used when the user selects something on the 
// menu.  The function overrides the standard function in the
// inheritated menu class.  
//
// INPUTS :	client_data	The id for the menu item thats been selected.
//		onflag		Whether the menu item is on/off.
// RETURNS: 	None
//
void
dividends_display::MenuCallback(unsigned long client_data, int onflag)
{
    	switch (client_data) {
	case MENU_EXIT:
		if (!colourdialog.SemaphoreIslocked())
			delete this;
		return;
	case MENU_COLOUR:
		if (colourdialog.SemaphoreClock()!= -1) {
			ColourSelect();
			colourdialog.SemaphoreUnlock();
			TextdialogRedraw();
		}
		break;
	case MENU_CASH:
		if (onflag)
			optionflags|=OPTIONS_CASH;
		else	
			optionflags&=~OPTIONS_CASH;
		create_display();
		break;
	case MENU_STOCK:
		if (onflag)
			optionflags|=OPTIONS_STOCK;
		else	
			optionflags&=~OPTIONS_STOCK;
		create_display();
		break;
	case MENU_SHARES:
		if (onflag)
			optionflags|=OPTIONS_SHARES;
		else	
			optionflags&=~OPTIONS_SHARES;
		create_display();
		break;
	case MENU_UNITTRUSTS:
		if (onflag)
			optionflags|=OPTIONS_UNITTRUSTS;
		else	
			optionflags&=~OPTIONS_UNITTRUSTS;
		create_display();
		break;
	case MENU_PRINT:
		TextdialogPrint();
		break;
	case MENU_DATE:
		if (DialogSelect2dates("Set date period for dividends",
					"Start",start_date,
					"Finish",end_date)!= -1)
			create_display();
		break;
	default:
		DialogPanic("Bad switch: dividends_display::MenuCallback");
	}
}


//
// dividends_display::redisplay this 
//
// This callback is registered to execute whenever dividends are added.
// It converts the first argument to a class ptr.
//
// INPUTS :	tmpthis		The this ptr defined when the callback
//				was registered.
// RETURNS:	None
void
dividends_display::redisplay(char *tmpthis,char *)
{
        class dividends_display *rthis=(class dividends_display *)tmpthis;
        rthis->create_display();
}


//
// dividends_display::create_display
//
// Generates text giving dividend information in alphanumeric share order.
// This function is private to the class.
//
// INPUTS : None
// RETURNS: None
//
void
dividends_display::create_display()
{
	char str[256];			// build area, hope this is enough 
	double s_tax,s_amount;		// tax/amount per share 
	double t_tax=0.0;		// total tax for all dividends 
	double t_amount=0.0;		// total amount for all dividends 
	dividend *div;
	dividend *head_div;
	char *name;
	class Holding *ptr;
	class security *sec;
	double amount;
	double tax;
	double number;
	char sharename[16];
	char *datebuf;
	divtype_e type;
	sectype_e sectype;
	class date date;
	int i;
	char *paytype,secstr[256];
	char date1[256];
	char date2[256];

	//
	// Create heading for printer
	//
	if ( (optionflags&(OPTIONS_CASH|OPTIONS_STOCK)) == 
		(OPTIONS_CASH|OPTIONS_STOCK) )
		paytype="all dividends";
	else if (optionflags&OPTIONS_CASH)
		paytype="cash dividends";
	else if (optionflags&OPTIONS_STOCK)
		paytype="stock dividends";
	else
		paytype="";
	
	strcpy(secstr,"");
	if (optionflags&OPTIONS_SHARES)
		strcat(secstr,"shares");
	if (optionflags&OPTIONS_UNITTRUSTS) {
		if (*secstr!='\0')
			strcat(secstr," and ");
		strcat(secstr,"unit trusts");
	}
	strcpy(date1,start_date.getdate());
	strcpy(date2,end_date.getdate());
	sprintf(str,"Statement of %s for %s between %s and %s",
			paytype,secstr,date1,date2);
	TextdialogSetprintertitle(str);

	//
	// Set up menus
	//
	if (!setmenuflag) {
		MenuItemaction("Project","Close",MENU_EXIT);
		MenuItemaction("Project","Print",MENU_PRINT);
		MenuItemaction("Project","Dates",MENU_DATE);
		MenuItemaction("Project","Colours",MENU_COLOUR);

		MenuItemnofn("Dividend Type","Stock Dividends",
				 MENU_STOCK,MENUSTATE_ON);
		MenuItemnofn("Dividend Type","Cash Dividends",
				MENU_CASH,MENUSTATE_ON);
		MenuItemnofn("Security Type","Shares",
				 MENU_SHARES,MENUSTATE_ON);
		MenuItemnofn("Security Type","Unit Trusts",
				MENU_UNITTRUSTS,MENUSTATE_ON);
		for(i=1;i<8;i++) {
			class attr *attr;
			if (i==1)
				attr=new class attr(FIELDATTR_LEFT,FOREGROUND);
			else
				attr=new class attr(FIELDATTR_RIGHT,FOREGROUND);
			TextSetdefaultattr(i,attr);
		}
		TextdialogSettitle("Dividends");
		ColourSetowner(DIVDISP_DB_CLASS);
		setmenuflag=1;
	}

	TextClear();

	TextAdd("Share name\tDate\tNumber\tTax\tAmount\tTotal Tax\tTotal");

	for ( ptr=portfolio->head(); !portfolio->eol(ptr); ptr=ptr->next() ) {
		s_tax=0.0;		// total for this share is a big zero
		s_amount=0.0;
		sec=ptr->HoldingSecurity();
		sec->SecurityExtract(&name,NULL,NULL,NULL,&sectype);
		if ( (optionflags&OPTIONS_SHARES)&&(sectype==SEC_TYPE_SHARE) )
			;  
		else if ( (optionflags&OPTIONS_UNITTRUSTS) &&
			  (sectype==SEC_TYPE_UNITTRUST) )
			;
		else
			continue;
		head_div=div=ptr->HoldingDividend();
		while (div!=NULL ) {
			div->DividendExtract(&date,&number,&amount,&tax,&type);

			if ( date.inrange(&start_date,&end_date) &&
			   (((optionflags&OPTIONS_CASH)&&
					(type==DIV_TYPE_CASH) ) ||
		 	   ((optionflags&OPTIONS_STOCK)&&
					(type==DIV_TYPE_STOCK)))){
				if (name) {
					strncpy(sharename,name,15);
					sharename[15]='\0';
				} else
					sharename[0]='\0';
				name=NULL;
				datebuf=date.getdate();
				sprintf(str,"%s\t%s\t%.4f\t%.2f\t%.2f",
					  sharename,datebuf,number,tax,amount);
				TextAdd(str);
  		 		s_tax+=tax;
				s_amount+=amount;
			}
			div=div->dual_next();
			if (head_div==div)
				break;
		}
		if (!name) {	/* written at least 1 value, so give totals */
			sprintf(str,"\t\t\t\t\t%.2f\t%.2f",s_tax,s_amount);
			TextAdd(str);
		}
		t_tax += s_tax;	/* accumulate the per portfolio totals */
		t_amount += s_amount;
		DialogCursor(DCURSOR_BUSY);
    	}
	if (t_tax !=0 || t_amount !=0) {
		sprintf(str,"\t\t\t\t\t%s\t%s","======","======");
		TextAdd(str);
		sprintf(str,"\t\t\t\t\t%.2f\t%.2f",t_tax,t_amount);
		TextAdd(str);
	}
	TextComplete();
	TextdialogOutput();
	DialogCursor(DCURSOR_NORMAL);
}



//
// dividends_display::dividends_display
//
// Constructor for dividends_display class.
// 
// Creates & manages the window which displays dividends for a
// given portfolio.
//
// INPUTS :	None
// RETURNS:	None
//
dividends_display::dividends_display()
{
	// start_data will be set to 0 by the constructor function
	end_date.setdate(); 		// until the date today
	setmenuflag=0;			// menu not yet attached
	optionflags=OPTIONS_STOCK|OPTIONS_CASH|
			OPTIONS_SHARES| OPTIONS_UNITTRUSTS;

	//
	// Set up call backs for dividend changes and security changes
	//
	callback_div=main_div_list->add_vect.CallbackAdd(
			&dividends_display::redisplay,
			(char *)this);
	callback_sec=main_sec_list->changed_vect.CallbackAdd(
			&dividends_display::redisplay,
			(char *)this);
	
	
	create_display();
}

//
// dividends_display::~dividends_display
// 
// Destructor for dividends_display class.
//
dividends_display::~dividends_display()
{
	//
	// Remove the callback node, from the list of callbacks
	// that are made when a dividend is added
	//
	main_div_list->add_vect.CallbackRemove(callback_div);
	main_sec_list->changed_vect.CallbackRemove(callback_sec);
}



//
//=============================================================================
//
// The C "Glue"
//
//=============================================================================
//

//
// display_dividends d a args
//
// C wrapper function that creates a dividends display
//
// INPUTS :	dialog,a,args	All place holders to satify type checking
// RETURNS:	None
//
static void
display_dividends(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	//
	// Simply display the dividends by creating a dividends_display
	// member
	//
        dividends_display *newdiv=new dividends_display;
}


//
// ddiv_init menu_bar
//
// Initialises the display dividends module:
//
// INPUTS :	Menu bar where option to display dividends is to be inserted
// RETURNS:	None
//
void
ddiv_init(menu_bar *mb)
{
	MENU("Dividends","Display",display_dividends,0);
}
