
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: div.C,v 2.8 1995/06/23 20:54:07 davidsi Exp $
//
// $Log: div.C,v $
// Revision 2.8  1995/06/23  20:54:07  davidsi
// - Validate dividend input.  The security must be a real security
//   (not a "group"), a non-zero amount must be entered and there
//   must a positive number of securities.
//
// Revision 2.7  1995/06/07  20:49:26  davidsi
// - Increase width of dividend entry areas to guarantee date will appear
//   in full on the display.
//
// Revision 2.6  1995/06/02  20:00:58  davidsi
// - Renamed "DIVIDEND" menu item to "Receive Dividend".
//
// Revision 2.5  1995/05/18  20:36:05  davidsi
// - The make_callback() function is now CallbackMake().
// - DialogentryCreate() now requires an argument specifying total number
//   of visible columns.
// - BUG FIX. The call to add_load_vect() should occur from within the
//   div_list constructor otherwise the actual list may not exist when
//   the add_load_vect() function is called.
//
// Revision 2.4  1995/04/25  21:09:51  davidsi
// - Added support for a busy cursor.
//
// Revision 2.3  1995/04/16  20:30:40  davidsi
// - BUG FIX. When Dividendedit::ButtonCallback() could not find a holding,
//   the error message stated "could not add transaction" and it should be
//   "dividend".
//
// Revision 2.2  1995/03/08  21:18:57  davidsi
// - Factored the dividend class into two classes, the dividend class provides
//   all the main dividend handling functionality and the Dividendedit class
//   allows dividends to be edited.  This reduces the amount of memory
//   req'd to represent a diviend.
// - The extract() member has been renamed to DividendExtract.
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that.
// - Revised to use the new Holding class which contains dividend information
//   rather than the security class.
// - BUG FIX. dividend::save_record() was incorrectly writing the address
//   of the date string to the file, rather than the date string itself.
//
// Revision 2.1  1995/03/02  21:01:40  davidsi
// - Uses Dialog class. Calls to dialog_xxx functions have been removed.
//
// Revision 2.0  1995/01/19  21:54:22  davidsi
// - Removed the time member of the dividend class.  Dates are now managed
//   using the date class.
// - Renamed all calls to date2string() which is now getdate().
// - div_load must take an additional ptr argument which it can ignore - see
//   changes to the callback class.
// - Added add_vect to the div_list class so that callbacks are handled
//   when dividends are added.
//
// Revision 1.5  1994/12/11  17:59:38  davidsi
// - save_record() and div_load() had the amount and tax fields reversed,
//   resulting in misworked numbers on loading old files.
//
// Revision 1.4  1994/11/05  21:06:35  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.3  1994/10/17  20:44:44  davidsi
// - All calls to extract() in the security class now require a further
//   argument, so that the type of security can be determined using flags
//   rather than strings.
// - When a security name list has been extracted, it no longer should be
//   freed because it is a shared resource and not private to the caller.
// - BUG FIX. dividend::save_record was not writing the type of dividend to
//   the file, so files could not be read back.
// - BUG FIX. The regex used by dividend::div_load was incorrect and failed
//   to match all the intended cases.
//
// Revision 1.2  1994/08/06  16:40:22  davidsi
// - Changes string_list() to extract() for consistancy.
//
// Revision 1.1  1994/06/28  21:01:04  davidsi
// Initial revision
//
//
//


#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"

#include "menu.h"
#include "trans.h"

//
// Variables
//
div_list *main_div_list;

// 
// Defines for entry fields in dividend entry dialog
//
enum dialogid_e {
	ID_DATE,
	ID_NUMBER,
	ID_AMOUNT,
	ID_TAX,
	ID_STOCK,
	BUTTON_CANCEL,
	BUTTON_OK,
	};


//
// Dividendedit::ButtonCallback button state
//
// Handles the call back to create a dividend from the dividend dialog.
// If buttonstring is equal to "OK" then the "OK" button was hit and
// the dividend defined by the user should be added (subject to error
// checks).  If the string is anything else the dialog is removed
// and no dividend is added.
// 
// INPUTS :	button		The button id
//		state		The state of the button
// RETURNS:	None
//
void 
Dividendedit::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char *str;
	char *stockname;
        class date date;        // date dividend was recieved
        double  number;         // number of securities that dividend applies to
        double  amount;         // amount of dividend received
        double  tax;            // amount of tax paid
	class Holding *sec;
	class dividend *newdiv;
	sectype_e sectype;

	//
	// Save lots of identing by using a GOTO!
	//
	if (id!=BUTTON_OK)
		goto done;

	str=d_entrylist.DialogentryGet(ID_DATE);
	if (str) {
		if (date.setdate(str)==-1) {
			DialogMessage("No such date - please correct");
			return;
		}
	} else {
		DialogMessage("A date must be specified");
		return;
	}

	str=d_entrylist.DialogentryGet(ID_NUMBER);
	if (str)
		number=atof(str);
	else
		number=0.0;

	if (fabs(number)<SMALL_NUMBER) {
		DialogMessage("Please enter a complete number of securities");
		return;
	}

	str=d_entrylist.DialogentryGet(ID_AMOUNT);
	if (str) 	
		amount=atof(str);
	else
		amount=0.0;

	if (fabs(amount)<SMALL_NUMBER) {
		DialogMessage("Please enter a full amount for the dividend");
		return;
	}

	str=d_entrylist.DialogentryGet(ID_TAX);
	if (str)
		tax=atof(str);
	else
		tax=0.0;
	
	stockname=d_secname.OptionlistinstanceGet();
	sec=portfolio->HoldinglistFind(stockname);
	if (sec==NULL) {
		DialogMessage("Please correctly specify a security");
		return;
	}

	//
	// Check dividends may be assigned to this security
	//
	sec->HoldingSecurity()->SecurityExtract(NULL,NULL,NULL,NULL,&sectype);
	if (sectype==SEC_TYPE_GROUP) {
		DialogMessage("Dividends cannot be received on this security");
		return;
	}

	newdiv=new class dividend;
	newdiv->DividendCreate(sec->HoldingSecurity(),date,number,amount,tax,type);
	main_div_list->insert_tail(newdiv);
	main_div_list->save();

	//
	// Add it to the list of transactions for the specific security
	//
	sec->HoldingDividendadd(newdiv);

	//
	// Issue the callback
	//
        main_div_list->add_vect.CallbackMake(0,NULL);

	if (type==DIV_TYPE_STOCK)  {
		double numpurch;
		transaction *newtrans;

		str=d_entrylist.DialogentryGet(ID_STOCK);
		if (str)
			numpurch=atof(str);
		else
			numpurch=0.0;
	
		newtrans=new transaction;
                if (newtrans->TransactionCreate(stockname,TRANS_BUY,&date,
					numpurch,amount/numpurch,amount) == -1) 
			DialogMessage("Cannot add transaction");
	}

done:
	delete this;
#undef _L_BUFSIZE
}



//
// Dividendedit::addw_secname parent
//
// Creates a Dialog that allows the selection of a security name. This
// function is private.
// 
// INPUTS :	None
// RETURNS:	None
///
void 
Dividendedit::addw_secname()
{
	//
	// List definition for security names
	//
	static Newoptionlist nolist = {
		20,				// width
		5,				// height 
		NULL,				// list of items 
		"",				// name of list
		"",				// name of item
		OLIST_FLAGS_SEARCH,		// search for matches and...
		NULL,				// default font
	};


	nolist.listitem=main_sec_list->extract();
	nolist.selname="Security Name";

	d_secname=OptionlistCreate(d_horiz,nolist,0);
}


//
// Dividendedit::addw_figures 
//
// Adds the Dialogs to a parent dialog that allow the user to specify
// the date, the number of securities, etc.
// 
// INPUTS :	None - assumes d_horiz set
// RETURNS:	None
///
void 
Dividendedit::addw_figures()
{
	class date tdate;	// temp date for today

	tdate.setdate();
	d_entrylist.DialogentryCreate(d_horiz,12);
	d_entrylist.DialogentrySet("Date",tdate.getdate(),ID_DATE);
	d_entrylist.DialogentrySet("Number","",ID_NUMBER);
	d_entrylist.DialogentrySet("Amount","",ID_AMOUNT);
	d_entrylist.DialogentrySet("Tax","",ID_TAX);
	if (type==DIV_TYPE_STOCK) 
		d_entrylist.DialogentrySet("Stock Obtained","",ID_STOCK);
}


//
// Dividendedit::DividendeditEdit title dividend type
//
// Allows a dividends details to be set up.
//
// INPUTS :	title		The name of the dialog to be created
// 		dtype		The dividend type (must be valid)
// RETURNS:	None
//
void 
Dividendedit::DividendeditEdit(
	char *title,
	divtype_e itype
	)
{

	//
	// Validate the dividend type
	//
	switch ( itype ) {
		case DIV_TYPE_STOCK:
			break;
		case DIV_TYPE_CASH:
			break;
		default:
			DialogPanic("dividend::edit_dividend - bad dividend type");
	}

	type=itype;

	d_shell.DialogshellCreate(title);
	d_vert.DialogverticalCreate(d_shell);
	d_horiz.DialoghorizontalCreate(d_vert);

	//
	// d_horiz users
	//
	addw_secname();
	addw_figures();

	//
	// d_vert users
	//
	d_buttonh.DialoghorizontalCreate(d_vert);
	(void)ButtonPush(d_buttonh," OK ",BUTTON_OK);
	(void)ButtonPush(d_buttonh,"CANCEL",BUTTON_CANCEL);

	d_shell.DialogRealize();
}



//
// Constructor for Dividendedit
//
Dividendedit::Dividendedit()
{
}

//
// Destructor for Dividendedit
//
Dividendedit::~Dividendedit()
{
}


//
// dividend::DividendLoad record 
//
// Creates a dividend from a record structure.  This function is used
// via a callback that occurs when the stored list is loaded from file.
// All security information must be loaded (correctly) for this function
// to operate correctly.
// 
// INPUTS :	record		A valid dividend record.
// RETURNS:	None
///
void
dividend::DividendLoad( 
	class record *rec
	)
{
	char tstr[256];
	char *string;
	class Holding *holding;

	rec->compile("[:|\n]");

	if ( (string=rec->uptomatch()) == NULL) {
		printf("Failed to read name\n");
		goto error;
	}
	holding=portfolio->HoldinglistFind(string);
	if (holding==NULL) {
		sprintf(tstr,"Found dividends for a non-existent security %s",string);
		DialogPanic(tstr);
	}

	sec=holding->HoldingSecurity();

	if ( (string=rec->uptomatch()) == NULL) {
		printf("Failed to read type\n");
		goto error;
	}
	if ( !strcmp("CASH",string ) ) {
		type=DIV_TYPE_CASH;
	} else if ( !strcmp("STOCK",string ) ) {
		type=DIV_TYPE_STOCK;
	} else {
		goto error;
	}

	if ( (string=rec->uptomatch()) == NULL)
		goto error;
	if (date.setdate(string) == -1 ) {
		printf("Failed to compute date\n");
		goto error;
	}

	if ( (string=rec->uptomatch()) == NULL)
		goto error;
	number=atof(string);

	if ( (string=rec->uptomatch()) == NULL)
		goto error;
	tax=atof(string);

	if ( (string=rec->uptomatch()) == NULL)
		goto error;
	amount=atof(string);

	//
	// Add it to the list of all dividends
	//
        main_div_list->insert_tail(this);

	//
	// Add it to the list of transactions for the specific security
	//
	holding->HoldingDividendadd(this);

	//
	// Issue the callback
	//
        main_div_list->add_vect.CallbackMake(0,NULL);

	if ( !strcmp(rec->version(),"VERSION:ODT11:3") ) {
		if ( (string=rec->line()) == NULL) {
			printf("Failed to read class\n");
			goto error;
		}
	}
	return;
error:
 	DialogPanic("Failed during dividend file read.");
	return;
}


//
// div_load args
//
// C wrapper to ensure correct C++ function is called on a callback
//
// INPUTS :	args	The arguments to pass to the C++ function
// RETURNS:	None
//
void
div_load(char *,va_list ap)
{
	class record *rec;
	dividend *newtrans=new dividend;

        rec=va_arg(ap,class record *);
	newtrans->DividendLoad(rec);
}



//
// dividend::save_record record
// 
// Saves the transactions to the appropriate portfolio file.  Note that
// all the transactions are written to the file, it is not a case
// of simply writing out the changes.  At the moment this a bit dubious
// since we don't take the precaution of backing up the old file *TODO*
// 
// INPUTS:	reocrd		The record to save to the file
// RETURNS:	None
///
void
dividend::save_record(class record *rec)
{
char 	string[100];
char *secname;

	sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);
	rec->create(secname);
	switch (type) {
	case DIV_TYPE_CASH:
		rec->append("|CASH|");
		break;
	case DIV_TYPE_STOCK:
		rec->append("|STOCK|");
		break;
	}
	sprintf(string,"%s|",date.getdate());
	rec->append(string);
	sprintf(string,"%f|",number);
	rec->append(string);
	sprintf(string,"%f|",tax);
	rec->append(string);
	sprintf(string,"%f|",amount);
	rec->append(string);
}


//
// dividend::DividendCreate isec idate inumber iamount itax itype
//
// INPUTS :	date 		The date stamp for the dividend
//		inumber		number of securities
//		iamount		amount recieved
//		itax		tax (pre)paid
//		itype		dividend type
// RETURNS:	None
//
void
dividend::DividendCreate(class security *isec,
                        class date &idate,
                        double inumber,
                        double iamount,
                        double itax,
                        divtype_e itype)

{
	sec=isec;
	date=idate;
	number=inumber;
	amount=iamount;
	tax=itax;
	type=itype;
}

//
// DividendExtract cdate_ptr number_ptr amount_ptr tax_ptr type_ptr
//
// Extracts all dividend information. This can be used anywhere
// to determine the information held in a dividend class member.
// The fields cannot be directly accessed.
//
// INPUTS :	None
// RETURNS:	cdate_ptr	Either NULL or the address of a location
//				where the date stamp for the dividend can be
//				stored.
//		number_ptr	As cdate_ptr but number of securities
//		amount_ptr	As cdate_ptr but amount recieved
//		tax_ptr		As cdate_ptr but tax (pre)paid
//		type_ptr	As cdate_ptr but dividend type
//
void
dividend::DividendExtract(class date *cdate,
			double *n,
			double *camount,
			double *ctax,
			divtype_e *ctype)
{
	if (cdate)
		*cdate=date;
	if (n)
		*n=number;
	if (camount)
		*camount=amount;
	if (ctax)
		*ctax=tax;
	if (ctype)
		*ctype=type;
}



//
// dividend
//
// Constructor for the dividend class
//
// INPUTS :	None
// RETURNS:	None
//
dividend::dividend()
{
	number=0.0;
}

//
// ~dividend
//
// Destructor for the dividend class
//
// INPUTS :	None
// RETURNS:	None
//
dividend::~dividend()
{
}



//
// div_list
//
// Constructor for the dividend list class.
//
// INPUTS :	None
// RETURNS:	None
//
//
div_list::div_list()
{
	set_filename("dividends",WRITE_BY_LINE);
	//
        // Register our decoder for building the list of dividends
        //
        add_load_vect(div_load);
}

//
// These are to prevent the need for casts
//
class dividend *
secdiv_list::dual_head()
{
	return( LISTREMAP(dividend,dslist_head,dual_head,this) );
}

class dividend *
dividend::dual_next()
{
	return( LISTREMAP(dividend,dslist_node,dual_next,this) );
}

//
//============================================================================
//========   ALL THE REMAINING FUNCTIONS ARE GLOBAL TO THE PROGRAM    ========
//============================================================================
///

//
// recieve dialog dtype args
//
// C wrapper for callback from menu. Called to create a new dividend
//
// INPUTS :	dialog		dialog of parent (satifies type checking)
//		dtype		type of div to create (DIV_TYPE_XX)
// RETURNS:	None
//
static void
recieve(local_va_dcl)
{
	va_list ap;
	Dividendedit *newdiv=new Dividendedit;
	int d;
	divtype_e dtype;
	dcursortype_e cursor;

	cursor=DialogCursor( DCURSOR_WAIT );

	local_va_start(ap);
        d=va_arg(ap,int);
        dtype=va_arg(ap,divtype_e);

        newdiv->DividendeditEdit("Receive Dividend",dtype);

	DialogCursor( cursor );
}



//
// div_init menu_bar
//
// Initialises the dividends module.  This includes being responsible for
// setting up the decode callback for the dividend file.
// 
// INPUTS :	menu_bar	A handle (dialog) on the main menu bar
// RETURNS:	None
///
void
div_init(menu_bar *mb)
{
	MENU("Dividends","Receive Cash Dividend",recieve,DIV_TYPE_CASH);
	MENU("Dividends","Receive Stock Dividend",recieve,DIV_TYPE_STOCK);


}
