
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dprice.C,v 2.6 1995/05/19 18:51:42 davidsi Exp $
//
// $Log: dprice.C,v $
// Revision 2.6  1995/05/19  18:51:42  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.5  1995/04/25  21:14:31  davidsi
// - Added support for a busy cursor.
//
// Revision 2.4  1995/04/16  20:33:13  davidsi
// - The date_changed() callback now must use class date pointers to
//   match changes to the Setdate class.
//
// Revision 2.3  1995/03/08  21:20:30  davidsi
// - No longer need extern declarations of main_sec_list and main_price_list
//   as the relevant include files provide these.
//
// Revision 2.2  1995/03/01  22:20:28  davidsi
// - Because editprice() has been changed in the price class, modified
//   the call to this particular function member.  The  function member
//   price_display::editprice() also no longer requires a dialog to be
//   passed.
// - Removed all dialog references - use the Dialog class instead.
//
// Revision 2.1  1995/02/05  18:03:11  davidsi
// - Added sdate member to handle the setting of dates.
// - The date_changed() member function is now static to allow it to be
//   used in callbacks.
// - Removed CB_changedate() - its no longer needed since date_changed()
//   can do all the necessary work.
//
// Revision 2.0  1995/01/19  22:05:27  davidsi
// - Revised to use the new date class.
// - Added callback_price node for when prices change and a window is on
//   display.
// - Added redisplay() member to handle the price change callback.
//
// Revision 1.4  1994/11/05  21:07:00  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.3  1994/10/17  20:47:06  davidsi
// - When a security name list has been extracted, it no longer should be
//   freed because it is a shared resource and not private to the caller.
//
// Revision 1.2  1994/08/06  16:40:45  davidsi
// - Changed string_list() to extract() for consistancy.
//
// Revision 1.1  1994/06/29  21:04:15  davidsi
// Initial revision
//
//
//
//


#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dial.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <ctype.h>


#include "localmacros.h"
#include "dprice.h"
#include "menu.h"
#include "support.h"


enum dialoid_e {
	LIST_PRICES,
	LIST_SECURITIES,
	BUTTON_OK,
	MENU_PRICE_EDIT,
	MENU_PRICE_SHOW,
	};

//
// price_display::list_prices
//
// Lists the prices (in ascii form) into an appropriate array for
// the current security.  The number of elements in the array is
// given by the number of prices for the security that lie in
// the current date range.
// 
// INPUTS :	None.
// RETURNS:	A pointer to an array of strings.  The final element
//		in the array will be a NULL pointer to signify array
//		termination.
///
char **
price_display::list_prices()
{
//
// PRICE_PAD_STRING should be a string of spaces which is at least the
// size of the price strings that will be displayed
//
#define PRICE_PAD_STRING "                            "
	char **vlist;
	int iter;
	char buf[256];
	char *datebuf;
	class date *date;
	dateprice *price_array;
	int pnum;
	int num;

	if (!sec)
		return(NULL);

	num=0;
	price_array=sec->prices->PriceFetchgroup(&date_start,&date_end,&pnum);

	vlist=(char **)malloc( sizeof(char *)* (pnum+1) );
	for(iter=0;iter<pnum;iter++) {
		if (price_array[iter].dp_mode==DATEPRICE_END)
			break;
		date=&(price_array[iter].dp_date);

		if (  date->compare(&date_end) == 1 )
			break;
		if (  !date->inrange(&date_start,&date_end))
			continue;

		datebuf=date->getdate();

		if (price_array[iter].dp_mode==DATEPRICE_NONE)
			sprintf(buf,"%s %9s",datebuf,"No price");
		else
			sprintf(buf,"%s %9.4f",datebuf,price_array[iter].dp_price);
		vlist[num]=(char *)malloc(strlen(buf)+1);
		strcpy(vlist[num],buf);
		num++;
    	}
	vlist[num]=NULL;
	free(price_array);

	return(vlist);
}





//
// price_display::price_changed security
//
// This changes the prices being displayed in the window to those for
// the specified security.
//
// INPUTS :	security 	A member of the security class.
// RETURNS:	None.
//
void
price_display::price_changed(security *newsec) 
{
	char **list;
		
	if (newsec==NULL)	// if there is no security don't display 
		return;

	sec=newsec;
	list=list_prices();
	d_prices->OptionlistinstanceRevise(list);
	if (list!=NULL)
		freestringlist(list);
}
    
    

//
// price_display::OptionlistCallback
//
// This handles the callbacks for either the security changing
// or a price entry being "hit"
//
//
void
price_display::OptionlistCallback(unsigned long id)
{
	security *newsec;
	char *name;
	char *str;
	int dmon;
	class date tdate;

	switch(id) {
	case LIST_SECURITIES:
		//
		// If a security has been selected, go grab it and
		// display the new price information.
		//
		name=d_securities.OptionlistinstanceGet();
		newsec=main_sec_list->find_security(name);
		if (newsec!=NULL) 
			price_changed(newsec);
		break;
	case LIST_PRICES:
		//
		// If a price has been hit then check a security has
		// been defined and editing is allowed.  If it is,
		// then put up a dialog to edit the price.
		//
		if (!sec)		// check there is a security
			return;
		if (!editenabled)
			return;
		str=d_prices->OptionlistinstanceGet();
		str[11]='\0';
		if (tdate.setdate(str)==-1) {
			fprintf(stderr,"PANIC: Invalid date %s\n",str);
			exit(1);
		}
		sec->prices->PriceEdit(&tdate);
		price_changed(sec);		// make display consistent 
		break;
	}
}




//
// price_display::ButtonCallback id state
//
// Call back when the user hits the "OK" (remove dialog) button.
// 
// INPUTS :	id		The button id
//		state		The state of the button.
// RETURNS:	None.
//
void
price_display::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	delete this;
}



//
// price_display::date_changed this newdate
//
// Callback for when the date is changed. Updates the display to reflect
// new start and end dates
// 
// INPUTS :	this		The original price_display instance
//		newdate		Pointer to the new revised date
// RETURNS:	None
//
void
price_display::date_changed(char *ths,class date *indate)
{
	class price_display *rthis=(class price_display *)ths;

	rthis->date_start.setdate(*indate);
	rthis->date_end=rthis->date_start;
    	rthis->date_end.add(rthis->date_end.days_in_month()-1 );
	rthis->price_changed(rthis->sec);
}


//
// price_display::create_display mode 
//
// Displays prices within a window, if it is not already on display. Based
// on the 1.5(& earlier) show_prices.  The reason for factoring out this
// function from price_show is to allow for price_edit. 
//
// INPUTS : 	mode 	 TRUE to end
// RETURNS: 	None
///
void
price_display::create_display(int mode)
{
	char *tmp[2];			// pointer to pad string 

	//
	// List definition for security names
	//
	static Newoptionlist nolist = {
		20,			// width 
		5,			// height 
		NULL,			// list of items 
		"Securities",		// name of list 
		"Security",		// name of item 
		OLIST_FLAGS_SEARCH,	// search for matches and... 
		NULL,			// default font
	};

	//
	// List definition for prices
	///
	static Newoptionlist nolist2 = {
		30,			// width 
		31,			// height 
		NULL,			// list of items 
		"Prices",		// name of list 
		"",			// name of item 
		OLIST_FLAGS_NOLABEL,	// flags
		"fixed1"		// default fixed size font
	};


	//
	// Must set the current_share to NULL first because dialog_cal_month()
	// will result in a callback to display the list of prices.
	//

	if (mode)
		d_shell.DialogshellCreate("EDIT SECURITY PRICES");
	else
		d_shell.DialogshellCreate("SECURITY PRICES");

	d_mainv.DialogverticalCreate(d_shell);
	d_pricemain_h.DialoghorizontalCreate(d_mainv);

	d_secmain_v.DialogverticalCreate(d_pricemain_h);

	sdate=new class Setdate;
	sdate->SetdateMonthyear(d_secmain_v,
				&price_display::date_changed,
				(char *)this);

	nolist.listitem=main_sec_list->extract();
	d_securities=OptionlistCreate(d_secmain_v,nolist,LIST_SECURITIES);

	d_pricemain_v.DialogverticalCreate(d_pricemain_h);
	editenabled=mode;
	tmp[0]=PRICE_PAD_STRING;
	tmp[1]=NULL;
	nolist2.listitem=(char **)&tmp;
	d_prices=OptionlistCreate(d_pricemain_v,nolist2,LIST_PRICES);

	d_buttonh.DialoghorizontalCreate(d_mainv);
	(void)ButtonPush(d_buttonh," OK ",BUTTON_OK);

	d_shell.DialogRealize();
}



//
// price_display::redisplay
//
// Callback for when price state changes.
//
void
price_display::redisplay(char *tmpthis,char *)
{
        class price_display *rthis=(class price_display *)tmpthis;
	rthis->price_changed(rthis->sec);	// make display consistent 
}


//
// price_display::price_display
//
// Constructor function for price display class. Defaults are to
// display prices from start of current year.
//
// INPUTS :	None
// RETURNS:	None
//
price_display::price_display()
{
	int d,m,y;

	sec=NULL;
	//
	// Get todays date, then go to the start of the month. The final
	// date is found by add the number of days in the month to the
	// start date.
	//
        date_start.setdate();
	date_start.getdate(d,m,y);
	d=1;
	date_start.setdate(d,m,y);
	date_end=date_start;
    	date_end.add(date_end.days_in_month()-1 );

	callback_price=main_price_list->change_vect.CallbackAdd(
			price_display::redisplay,
			(char *)this);

}


//
// price_display::~price_display
//
// Destructor function for price display class
//
// INPUTS :	None
// RETURNS:	None
//
price_display::~price_display()
{
	//
	// Remove callbacks
	//
	main_price_list->change_vect.CallbackRemove(callback_price);

	d_shell.DialogDestroy();
}


//
// decode dialog code args
//
// C Wrapper to handles menu selections.
// 
// INPUTS :	dialog	Parent dialog for menu call back
//		code	The menu item that was selected
//		...	To satisfy type checking
// RETURNS:	None
//
void
decode(local_va_dcl)
{
	va_list ap;
	dialog d;
	unsigned long code;
	price_display *pd;

	local_va_start(ap);

	d=va_arg(ap,dialog);
	code=va_arg(ap,unsigned long);

	switch(code) {
	case MENU_PRICE_EDIT:
		DialogCursor(DCURSOR_WAIT);
		pd=new price_display;
		pd->create_display(TRUE);
		DialogCursor(DCURSOR_NORMAL);
		break;
	case MENU_PRICE_SHOW:
		DialogCursor(DCURSOR_WAIT);
		pd=new price_display;
		pd->create_display(FALSE);
		DialogCursor(DCURSOR_NORMAL);
		break;
	}
}


//
// dprice_init menu_bar
//
// Initialises the display price module by defining the available menu items 
// 
// INPUTS :	menu_bar	The main menu bar 
// RETURNS:	None
//
void
dprice_init(menu_bar *mb)
{
	MENU("Prices","Display",decode,MENU_PRICE_SHOW);
	MENU("Prices","Edit",decode,MENU_PRICE_EDIT);
}
