
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: draw.C,v 2.6 1995/06/07 20:57:30 davidsi Exp $
//
// $Log: draw.C,v $
// Revision 2.6  1995/06/07  20:57:30  davidsi
// - Improvements to graph plotting to ensure that enough space is always
//   left for price information.
//
// Revision 2.5  1995/05/29  19:08:53  davidsi
// - Revised to use the IS_PRICE() and NOT_PRICE() macros.
// - BUG FIX. DrawGetarea() was not correctly assigning the width argument.
//
// Revision 2.4  1995/05/19  18:54:53  davidsi
// - Added a DrawGetarea() that returns the information previously set by a
//   DrawSetarea() or the defaults if DrawSetarea() has not yet been called.
//
// Revision 2.3  1995/03/01  22:31:57  davidsi
// - All colour handling has been moved to colour.C
// - Added support for the Printerdraw class.  This inherits all
//   draw functionality but provides for specifying a file or printer
//   based output device.
// - DrawPush() and DrawPop() now partially push and pop the state
//   of a draw instance - these could still be improved.
// - Added DrawRef() and DrawUnref() function members to allow multiple
//   references to the same Draw instance with a guarantee it won't get
//   deleted while in use -- provide everyone plays ball of course!
//
// Revision 2.2  1995/02/05  18:01:25  davidsi
// - Added DrawRectangle() and drv_rect() member functions.
//
// Revision 2.1  1995/02/02  21:19:29  davidsi
// - Name change.  All public function members of this class are now prefixed
//   with Draw.  Some names have been further simplified.
// - Added colour handling.
//
// Revision 2.0  1995/01/19  22:06:16  davidsi
// - Revised to use the new date class.
// - Renamed all calls to date2string() which is now getdate().
// - Added drv_setlinewidth() and setlinewidth() functions.
// - Added drv_setcolour.
//
// Revision 1.1  1994/12/11  21:17:05  davidsi
// Initial revision
//
//
//


//
// Include Files
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"

#include "draw.h"
#include "date.h"
#include "dialog.h"
#include "price.h"


#define LEFT_MARGIN		10
#define RIGHT_MARGIN		30
#define TOP_MARGIN		10
#define BOTTOM_MARGIN		10
#define TICK_LENGTH             3       // length in pixels of a tick mark

//
///////////////////////////////////////////////////////////////////////////////
//
// Standard drawing functions. These get replaced by a graphics
// driver that is derived from this class. These functions should never
// be called.
//
///////////////////////////////////////////////////////////////////////////////
//

void
Draw::drv_clearpage()
{
}

void
Draw::drv_drawpage()
{
}

void
Draw::drv_drawto(double x,double y)
{
}

void
Draw::drv_moveto(double x,double y)
{
}

void
Draw::drv_rect(double x1,double y1,double x2,double y2)
{
}

void
Draw::drv_setfont(char *name,double size)
{
}
void
Draw::drv_setgrey(int perc)
{
}

void
Draw::drv_setlinestyle(int count,double *len)
{
}

void
Draw::drv_setlinewidth(double)
{
}

void
Draw::drv_setcolour(colour_e)
{
}

void
Draw::drv_startpage()
{
}

void
Draw::drv_text(char *text)
{
}

double
Draw::drv_textlength(char *txt)
{
}

double
Draw::drv_textheight()
{
}



//
///////////////////////////////////////////////////////////////////////////////
//
// The API that is made available publicly.
//
// These members simply redirect to the virtual driver functions.
//
///////////////////////////////////////////////////////////////////////////////
//

//
// DrawClearpage
//
// Clears the output page.  Not all drivers will necessary implement
// this function.
//
// INPUTS :	None
// RETURNS:	None
//
void
Draw::DrawClearpage()
{
	drv_clearpage();
}


//
// DrawPage
//
// Called when a page is complete. 
//
// INPUTS :	None
// RETURNS:	None
//
void
Draw::DrawPage()
{
	drv_drawpage();
}


//
// DrawText text
//
// Writes the specified text using the current font
//
// INPUTS :	text		The text to be displayed
// RETURNS:	None
//
void
Draw::DrawText(char *text)
{
	drv_text(text);
}



//
// DrawDrawto x,y
//
// Draws a line from the current coordinate to the specified coordinate
// pair.
//
// INPUTS :	x,y		The point to when the line is to be
//				drawn
// RETURNS:	None
//
void
Draw::DrawDrawto(double x,double y)
{
	drv_drawto((x+xbase)*x_conv,(y+ybase)*y_conv);
	xpos=x;
	ypos=y;
}



//
// DrawRect x,y,width,height
//
// Draws a rectangle at the specified coordinates
// pair.
//
// INPUTS :	x,y		The point to when the rect is to be
//				drawn
//		width,height	Size of the rectangle
// RETURNS:	None
//
void
Draw::DrawRectangle(double x,double y,double w,double h)
{
	drv_rect((x+xbase)*x_conv,(y+ybase)*y_conv,
			(x+xbase+w)*x_conv,(y+ybase+h)*y_conv);
	xpos=x+w;
	ypos=y+h;
}


//
// DrawExtract widthptr heightptr
//
// Extracts the current with and height of the drawing area.
//
// INPUTS :	None
// RETURNS:	widthptr	The current width of the drawing area, if
//				pointer not NULL.
// 		heighptr	The current height of the drawing area, if 
//				pointer not NULL.
void
Draw::DrawExtract(double &retwidth,double &retheight)
{
	retwidth=width;
	retheight=height;
}


//
// DrawMoveto x,y
//
// Sets the current drawing position
//
// INPUTS :	x,y		The new position
// RETURNS:	None
//
void
Draw::DrawMoveto(double x,double y)
{
	drv_moveto((x+xbase)*x_conv,(y+ybase)*y_conv);
	xpos=x;
	ypos=y;
}




//
// DrawSetfont
//
// Sets the current font size.
//
// INPUTS :	size		The size of the font to be used for all
//				text calls.  Value specified in points!
// RETURNS:	None
//
void
Draw::DrawSetfont(char *name,double size)
{
	drv_setfont(name,size);
}


//
// DrawSetgrey perc
//
// Sets the grey level for output.  Different drivers will interpret this
// in different ways. 
//
// INPUTS :	perc	Level of darkness. 0=black, 100=white.
// RETURNS:	None
//
void
Draw::DrawSetgrey(int perc)
{
	if (perc>=0 && perc<100)
		drv_setgrey(perc);
	else
		DialogPanic("Draw::DrawSetgrey - Percentage not in valid range");
}


//
// DrawSetlinestyle number array
//
// Sets the lengths of each dash to be used for following line draws
//
// INPUTS :	number		The number of values in 'array'
//		array		The length of each dash.  Note that
//				point to device coordinate is performed on
//				the basis of x coordinate scaling.
// RETURNS:	None
//
void
Draw::DrawSetlinestyle(int count,double *len)
{
        double *ptrs;
        int i;

        ptrs=(double *)malloc(count*sizeof(double));
        for(i=0;i<count;i++)
                ptrs[i]=len[i]*x_conv;
	drv_setlinestyle(count,ptrs);
        free(ptrs);
}

//
// DrawSetlinewidth width
//
// Sets the width to be used for all line draws. Note that the width
// is scaled on the basis of the x scaling value.  This is unlikely 
// to be idea!
//
// INPUTS :	width		The width of the lines to be drawn.
// RETURNS:	None
//
void
Draw::DrawSetlinewidth(double width)
{
	drv_setlinewidth( width*x_conv);
}


//
// DrawSetcolour colour
//
// Sets the colour to be used for all drawing primitives.
//
// INPUTS :	colour		The new colour to be used.
// RETURNS:	None
//
void
Draw::DrawSetcolour(colour_e col)
{
	drv_setcolour(col);
}

//
// DrawTextheight
//
// INPUTS :	None
// RETURNS:	The height of the current font
//
double
Draw::DrawTextheight()
{
	return(drv_textheight());
}



//
// DrawTextlength text 
//
// Reports the length of the specified string in points.  The length
// is converted on the basis of the X scaling factor.  This will not
// always be ideal.
//
// INPUTS :	text		The string whose length, in points, is needed
// RETURNS:	The length of the string in points.
//
double
Draw::DrawTextlength(char *txt)
{
	return((1.0/x_conv)*drv_textlength(txt));
}



//
// DrawStartpage
// 
// Signals the start of a new page.
//
// INPUTS :	None
// RETURNS:	None
//
void
Draw::DrawStartpage()
{
	drv_startpage();
}


//
///////////////////////////////////////////////////////////////////////////////
//
// Second level graphics functions.  These build on the standard public
// functions.  None of the functions in this group have driver specific
// counterparts
//
///////////////////////////////////////////////////////////////////////////////
//



//
// DrawAxis xmin,xmax, ymin,ymax
//
// Draws an axis for a graph
//
// INPUTS :	xmin,xmax		The range of x values to be plotted
// 		ymin,ymax		The range of y values to be plotted
// RETURNS:	None
//
void
Draw::DrawAxis( class date *a_xmin,
		class date *a_xmax, 
		double a_ymin,
		double a_ymax
		)
{
	register int iter;
	int n_ticksx;			// number of ticks on the x axis
	int n_ticksy;			// number of ticks on the y axis 
	int month=-1;			// last month labelled
	int year=-1;			// last year labelled
	double ytextsize;
	char text[20];
	double yticksize;		// num virtual pixels between y ticks
	double xticksize;		// num virtual pixels between x ticks	
	int l1,l2;
	double font_height;

	sprintf(text,"%6.2f",a_ymax);
	l1=strlen(text);
	sprintf(text,"%6.2f",a_ymin);
	l2=strlen(text);

	//
	// Determine largest width necessary to write price information
	//
	iter=max(l1,l2);
	strcpy(text,"88888888888888");
	text[iter]='\0';
	ytextsize=DrawTextlength(text);

	font_height=DrawTextheight();
	yticksize=font_height*1.5;
	xticksize=DrawTextlength("MMMM")*1.5; // always likely to be large size

	//
	// Draw surrounding box
	//
	DrawMoveto(0,0);
	DrawDrawto(width-1,0);
	DrawDrawto(width-1,height-1);
	DrawDrawto(0,height-1);
	DrawDrawto(0,0);


	//
	// Set up values for future scaling
	//
	xmin=a_xmin->daysince1970();
	xmax=a_xmax->daysince1970();
	ymin=a_ymin;
	ymax=a_ymax;


	xb=TICK_LENGTH+LEFT_MARGIN;
	yb=TOP_MARGIN;
	g_width=width-LEFT_MARGIN-RIGHT_MARGIN-TICK_LENGTH;
	g_height=height-TOP_MARGIN-BOTTOM_MARGIN-3*(int)font_height;


	xb+=ytextsize;
	g_width-=ytextsize;

	//
	// Compute number of ticks
	//
	n_ticksy=(int)((g_height-yb)/yticksize);
	n_ticksx=(int)((g_width-xb)/xticksize);

	//
	// Saves us from division by zero and ensures there always 
	// some degree of a scale on the graph.
	//
	if (n_ticksy<2)
		n_ticksy=1;
	if ( n_ticksx<2)	
		n_ticksx=1;

	//
	// Draw all y ticks
	//
	for(iter=0;iter<=n_ticksy;iter++) {
		double ypos=yb+(n_ticksy-iter)*(g_height/n_ticksy);
		double yscale=(ymax-ymin)/(double)n_ticksy;

		sprintf(text,"%6.2f",ymin+(double)iter*yscale);
		DrawMoveto(xb-TICK_LENGTH,ypos);
		DrawDrawto(xb+g_width,ypos);
		DrawMoveto(xb-TICK_LENGTH-ytextsize,ypos);
		DrawText(text);
	}

	//
	// Draw all x ticks
	//
	for(iter=0;iter<=n_ticksx;iter++) { 
		double xpos=xb+iter*(g_width/n_ticksx);
		class date date;
		char *datebuf;
		int nday,nmonth,nyear;

		date.daysince1970((long)(xmin+((double)iter*
					(xmax-xmin)/n_ticksx)));

		datebuf=date.getdate();
		date.getdate(nday,nmonth,nyear);
	
		datebuf[3]='\0';
		datebuf[6]='\0';
		DrawMoveto(xpos,yb+g_height+font_height);
		DrawText(&datebuf[4]);
	 	//
		// if month or year has changed then write them too
		//
		if (month!=nmonth || year!=nyear) {
			month=nmonth;
			DrawMoveto(xpos,yb+g_height+2*font_height);
			DrawText(&datebuf[0]);
			if (year!=nyear) {
				year=nyear;
				DrawMoveto(xpos,yb+g_height+3*font_height);
				DrawText(&datebuf[7]);
			}
		}
		DrawMoveto(xpos,yb+g_height+TICK_LENGTH);
		DrawDrawto(xpos,yb);
	}
	DrawMoveto(xb,yb);
	DrawDrawto(xb,yb+g_height);
	DrawDrawto(xb+g_width,yb+g_height);
}



//
// DrawGraph data sizeofdata startindex endindex
//
// Draws a graph from values held in an array.  It is assumed
// an axis has already been drawn.
//
// INPUTS :	data		The array from where values are to be
//				plotted.
//		sizeofdata	The number of values in the array
//		startindex	The first index value to be plotted
//		endindex	The final index value to be plotted
// RETURNS:	None
//
void
Draw::DrawGraph(double *data,int ndata,int startd,int endd)
{
#define SCALE_PRICE(price) ((yb+g_height-((price)-ymin)*y_scale))

	double y_scale;			// computed scale factor for y
	double x_scale;			// computed scale factor for x
	double lastx,lasty;		// previous values plotted
	register int count;		// offset from startd (ie startd=0)
	double xpos;			// current x position
	double *dptr;			// current ptr to value in array
	double *dptr_b;			// back pointer to value array
	int count_b;			// num of x points back for 1st segment
	char draw_initial_segment=0;	// TRUE when a value is detected before
					// the starting value in the array
	char plot_made=0;		// TRUE when main plot loop set a point
	char firstpoint=1;		// TRUE when nothing (at all) has
					// been plotted

	//
	// If the ending index in the array to be plotted is before
	// the start index then there is nothing to plot.
	//
	if (endd<startd) 
		return;

	//
	// Should not be possible, but lets protect against div by zero
	//
	if ( (fabs(ymax-ymin)<SMALL_NUMBER)||(fabs(endd-startd)<SMALL_NUMBER) )
		return;

	y_scale=g_height/(ymax-ymin);
	x_scale=g_width/(double)(endd-startd);

	lastx=-1;
	dptr_b=dptr=&data[startd];

	//
	// May have to draw initial segment. 
	//
	count_b=0;
	if ( NOT_PRICE(*dptr) ) {

		while (dptr_b>data && NOT_PRICE(*dptr_b) ) {
			count_b++;
			dptr_b--;
		}

		if (dptr_b>=data && IS_PRICE(*dptr_b) ) 
			draw_initial_segment=1;
	}

	//
	// Main loop to plot data on the graph
	//
	for(count=0;dptr<=&data[endd];dptr++,count++) {

		if ( NOT_PRICE(*dptr) )
			continue;
		xpos=(count*x_scale)+xb;
		//
		// If still on same x coordinate, do not attempt
		// to plot anything.
		//
		if ( (int)xpos==(int)lastx)
			continue;

		//
		// If nothing is yet drawn, we need to compute a
		// line segment.
		//
		if (draw_initial_segment) {
			lasty=SCALE_PRICE(((double)count_b/
					((double)(count_b+count)))*
					(*dptr-*dptr_b)+*dptr_b);
			lastx=xb;
			if (firstpoint) {
				DrawMoveto(lastx,lasty);
				firstpoint=0;
			} else
				DrawDrawto(lastx,lasty);
			draw_initial_segment=0;
		}
		lastx=xpos;
		lasty=SCALE_PRICE(*dptr);
		if (firstpoint) {
			DrawMoveto(lastx,lasty);
			firstpoint=0;
		} else
			DrawDrawto(lastx,lasty);
		plot_made=1;
	} 

	//
	// Scan forward - we've already done a back scan!
	//
	dptr=&data[endd];
	while (dptr<&data[ndata] && NOT_PRICE(*dptr) )
		dptr++;

	//
	// Fill in any remaining segment, or possible a complete
	// line if nothing has been plotted so far.
	//
	if ( dptr<&data[ndata] && IS_PRICE(*dptr) && 
	     dptr_b>=data && IS_PRICE(*dptr_b) ) {
		double m,c;
		double v1,v2;

		m=(*dptr_b-*dptr)/(-1.0+(double)(dptr_b-dptr));
		c=*dptr_b-(m*(dptr_b-data));

		v1=m*startd+c;
		v2=m*endd+c;
		if (!plot_made)
			DrawMoveto(xb,SCALE_PRICE(v1));
		DrawDrawto(xb+g_width,SCALE_PRICE(v2));
			
	}
}



//
// DrawPush
//
// Pushes a graphics context to the context stack
//
void
Draw::DrawPush()
{
	struct gfx_context *gfx;

	gfx=(struct gfx_context *)malloc(sizeof(struct gfx_context));

	gfx->width=width;
	gfx->height=height;
	gfx->xbase=xbase;
	gfx->ybase=ybase;
	gfx->x_conv=x_conv;
	gfx->y_conv=y_conv;
	gfx->xpos=xpos;
	gfx->ypos=ypos;
	gfx->next=stack;
	stack=gfx;
}


//
// DrawPop
//
// Restores a graphics context from the context stack
//
void
Draw::DrawPop()
{
	struct gfx_context *gfx=stack;

	if (stack==NULL)
		DialogPanic("Draw::DrawPop - no context to pop");
	width=stack->width;
	height=stack->height;
	xbase=stack->xbase;
	ybase=stack->ybase;
	x_conv=stack->x_conv;
	y_conv=stack->y_conv;
	xpos=stack->xpos;
	ypos=stack->ypos;
	stack=stack->next;

	free(gfx);
}


//
// DrawSetarea xpos ypos width height
// 
// Creates a drawing area that starts at <xpos,ypos> in the current
// drawing area. The size of the area is given by the width and
// height arguments.  
//
// Typically this function will be used after a push_context(), so
// that the pop_context() restores the initial area.
//
// INPUTS :	xpos		start x position
// 		ypos		start y position
//		width		width of new drawing area
//		height		height of area
//
void
Draw::DrawSetarea(double xpos,double ypos,double iwidth,double iheight)
{

	xbase=xpos;
	ybase=ypos;
	width=iwidth;
	height=iheight;
}

//
// DrawGetarea xpos ypos width height
// 
// Returns the current geometry of the drawing area.
//
// RETURNS :	xpos		start x position
// 		ypos		start y position
//		width		width of new drawing area
//		height		height of area
//
void
Draw::DrawGetarea(double &xpos,double &ypos,double &iwidth,double &iheight)
{
	xpos=xbase;
	ypos=ybase;
	iwidth=width;
	iheight=height;
}

// 
// DrawUnref
//
// Indicates one less user of a Draw instance. When no one is using the
// instance it will be deleted. Only call this function, once for
// every call to DrawRef().
//
// INPUTS :	None
// RETURNS:	None
// 
void
Draw::DrawUnref()
{
	refcount--;
	if (refcount==0)
		delete this;
}

//
// DrawRef
//
// Increments the reference count for a Draw instance.  The reference
// count can be used to determine whether its safe to delete a 
// Draw instance.  Use DrawUnref() when a reference is done - the
// Draw instance then is deleted when necessary.
//
// INPUTS :	None
// RETURNS:	None
// 
void
Draw::DrawRef()
{
	refcount++;
}

//
// Construction function
//
Draw::Draw()
{
	refcount=1;
	xbase=0.0;
	ybase=0.0;
	xpos=0.0;
	ypos=0.0;
	greylevel=100;
	stack=NULL;
}

//
// Destruction function
//
Draw::~Draw()
{
}

//
///////////////////////////////////////////////////////////////////////////////
//
// The Printerdraw class
//
///////////////////////////////////////////////////////////////////////////////
//

void
Printerdraw::PrinterdrawSetup()
{
}

void
Printerdraw::PrinterdrawMode(printermode_e mode)
{
	currentmode=mode;
}

Printerdraw::Printerdraw()
{
	//
	// Default output mode is to use the printer
	//
	currentmode=PRTMDE_PRINTER;
}

Printerdraw::~Printerdraw()
{
}
