
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: drawdialog.C,v 2.7 1995/06/07 20:57:06 davidsi Exp $
//
// $Log: drawdialog.C,v $
// Revision 2.7  1995/06/07  20:57:06  davidsi
// - BUG FIX. When constructing draw dialogs ensure that there is no current
//   shell or title - don't wait until creation because creation might not
//   happen but we still need to be able to safely delete the entire thing.
//
// Revision 2.6  1995/05/29  19:09:17  davidsi
// - BUG FIX.  There was a race condition whereby the draw dialog would not
//   have allocated all its underlying structures (graphics contexts etc) even
//   though it was mapped. Now Drawdialog::drv_startpage() explicitly asks
//   whether the underlying drawing area is ready with dialog_drawexposed().
//
// Revision 2.5  1995/05/19  18:54:44  davidsi
// - BUG FIX. Drawdialog::drv_clearpage() must clear the area that is given
//   by <xbase,ybase> to <xbase+w,ybase+h> and not <0,0> to <w,h>.
//
// Revision 2.4  1995/04/16  20:32:23  davidsi
// - BUG FIX. In drv_startpage() even if NOMENU is specified, if the caller
//   has given a menu definition then that should be rendered. NOMENU only
//   refers to the built in menu items.
//
// Revision 2.3  1995/03/02  20:49:52  davidsi
// - Added DrawdialogCreate() so that creation is not linked with
//   specifying how the dialog will be displayed.
// - Modified Drawdialog::drv_clearpage() so that it correctly obtains the
//   background colour that was selected by the user.
// - Support new POPUP window.  This is a simple window with no menus or
//   scroll bars.  Ideal for simple dialogs.
// - BUG FIX. The width and height members in the Draw class are meant to
//   be in device independent units. Unfortunately they were being
//   converted to device dependent units.  This was hard to spot,
//   because provided the area was mapped before width/heights were
//   read everything would tend to look Ok.
//
// Revision 2.2  1995/02/05  18:01:01  davidsi
// - Moved the call to dialog_destroy() from the menu exit handling code to
//   the destructor function.
// - Keep a copy of the original flags in a class data member.
// - Implement the NOSCROLL option.
// - Implement the NOMENU option - drawdialogs no longer always have to
//   have menus on display.
// - Added the drv_rect() member function.
// - BUG FIX.  The constructor was not converting the initial width and height
//   values into device specific values.
// - BUG FIX. In drv_startpage() don't clear the draw area after waiting
//   for it to be mapped - It has already been drawn on.
//
// Revision 2.1  1995/02/02  21:21:35  davidsi
// - The drv_drawto() and drv_moveto() member functions now handle lx,ly
//   variables as the draw class does not.
// - The drv_setfont() function now takes a fontname.
// - Modifications to reflect name changes in the Draw class.
// - This class now called Drawdialog rather than drawdialog to be
//   consistent with other name changes.
//
// Revision 2.0  1995/01/19  22:03:48  davidsi
// - Added the drv_setlinewidth() functions.
// - Added drv_setcolour.
// - Modified drawdialog::drawdialog() so that it now specifies scroll
//   bars on the created drawing area.
//
// Revision 1.1  1994/12/11  21:37:19  davidsi
// Initial revision
//




//
// Include Files
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#define FULL_INCLUDE

#include "localmacros.h"

#include "menu.h"
#include "callback.h"
#include "drawdialog.h"
#include "dialog.h"

extern dialog d_root;

#define GC_BLACK_ID	0
#define GC_WHITE_ID	1

#define SCREEN_WIDTH	10.0
#define SCREEN_HEIGHT	7.0
#define X_PIXELS	1024.0
#define Y_PIXELS	768.0

//
///////////////////////////////////////////////////////////////////////////////
//
// Standard drawing functions.  These override those in the Draw Class.
//
///////////////////////////////////////////////////////////////////////////////
//
void
Drawdialog::drv_clearpage()
{
	int w,h;
	char *colstr;

	colstr=ColourString(ColourBackground());
	dialog_draw_colour(d_area,GC_WHITE_ID,colstr);

	dialog_draw_value(d_area,&w,&h);
	dialog_draw_rect(d_area,GC_WHITE_ID,(int)(xbase*x_conv),
			 (int)(ybase*y_conv),
			 (int)((xbase+(double)w)*x_conv),
			 (int)((ybase+(double)h)*y_conv));
}

void
Drawdialog::drv_drawto(double x,double y)
{
	dialog_draw_line(d_area,GC_BLACK_ID,(int)lx,(int)ly,(int)x,(int)y);
	lx=x;
	ly=y;
}

void
Drawdialog::drv_rect(double x1,double y1,double x2,double y2)
{
	dialog_draw_rect( d_area,GC_BLACK_ID,(int)x1,(int)y1,
				(int)(x2-x1),(int)(y2-y1) );
	lx=x2;
	ly=y2;
}

void
Drawdialog::drv_drawpage()
{
	dialog_draw_refresh(d_area);
}

void
Drawdialog::drv_moveto(double x,double y)
{
	lx=x;
	ly=y;
}

void
Drawdialog::drv_setfont(char *name,double fsize)
{
	int height;

	dialog_draw_setfont(d_area,GC_BLACK_ID,name,(int)fsize);
	height=dialog_draw_fontheight(d_area,GC_BLACK_ID);
	font_height=(1.0/y_conv)*(double)height;
}

void
Drawdialog::drv_setgrey(int perc)
{
}

void
Drawdialog::drv_setlinestyle(int count,double *len)
{
	unsigned char *ptrs;
        int i;

	ptrs=(unsigned char *)malloc(count*sizeof(unsigned char));
        for(i=0;i<count;i++) {
		ptrs[i]=(unsigned char)len[i];
	}

	dialog_draw_lstyle(d_area,GC_BLACK_ID,count,ptrs);
	free(ptrs);
}

void
Drawdialog::drv_setlinewidth(double width)
{
	dialog_draw_lwidth(d_area,GC_BLACK_ID,(int)width);
}

void
Drawdialog::drv_setcolour(colour_e col)
{
	char *colstr;

	colstr=ColourString(col);
	dialog_draw_colour(d_area,GC_BLACK_ID,colstr);
}


void
Drawdialog::drv_text(char *text)
{
        dialog_draw_string(d_area,GC_BLACK_ID,(int)lx,(int)ly,text);
}

double
Drawdialog::drv_textlength(char *txt)
{
	return( (double)dialog_string_length(d_area,GC_BLACK_ID,txt) );
}

double
Drawdialog::drv_textheight()
{
	return(font_height);
}


//
///////////////////////////////////////////////////////////////////////////////
//
// Private functions specific to dialog drawings.
//
///////////////////////////////////////////////////////////////////////////////
//

//
// This may get overridden
//
void
Drawdialog::DrawdialogResize()
{
}

// 
// drv_sizechange
//
// Handles callback when drawing area changes size.
//
// INPUTS :
// RETURNS:
//
void
Drawdialog::drv_sizechange(local_va_dcl)
{
va_list ap;
dialog d;
mcallback *cb;
int w,h;
Drawdialog *rthis;

	local_va_start(ap);
	d=va_arg(ap,dialog);
	rthis=va_arg(ap,Drawdialog *);

	//
	// Get new size of drawing area
	//
	dialog_draw_value(d,&w,&h);

	//
	// Map back into device independent values
	//
	rthis->width=(1/rthis->x_conv)*(double)w;
	rthis->height=(1/rthis->y_conv)*(double)h;

	rthis->DrawdialogResize();
}

enum menuids {
	MENU_EXIT,
};

// 
// drv_menu
//
// Handles menu selections
//
// INPUTS :
// RETURNS:
//
void
Drawdialog::drv_menu(local_va_dcl)
{
va_list ap;
dialog d;
mcallback *cb;
int w,h;
Drawdialog *rthis;

        local_va_start(ap);
        d=va_arg(ap,dialog);
        cb=va_arg(ap,mcallback *);
	
	rthis=(Drawdialog *)(cb->addr);
	switch(cb->action) {
	case MENU_EXIT:
		//
		// In the case of EXIT we don't delete the instance, but
		// we do take it off the display.
		//
		dialog_destroy(rthis->d_shell);
		rthis->d_shell=NULL;
		return;
	default:
		;
	}
}


// 
// drv_startpage
//
// INPUTS :	None
// RETURNS:	None
//
void
Drawdialog::drv_startpage()
{
	position p;
	new_d_draw ng={
		( void (*)(... ))&Drawdialog::drv_sizechange,
		0,
	};
	menu_bar *mb=NULL;
	dialog d_form;
	int mask=0;


	if ( !(oflags & NOSCROLL) ) 
		mask=SHELL_SCROLL;

	if ( oflags & POPUP )
		d_shell=dialog_shell(d_root,title,SHELL_POPUP);
	else {
		d_shell=dialog_shell(d_root,title,SHELL_MAINWINDOW|mask);

		if ( !(oflags & NOMENU) )  {
			if ( !add_mb )
				mb=new menu_bar(NULL,NULL,0);
			else
				mb=add_mb;
			menudata.action=MENU_EXIT;
			menudata.addr=(char *)this;
			MENU("Project","Close",&Drawdialog::drv_menu,
				(unsigned long)&menudata);
		} else
			mb=add_mb;
	}

	d_form=dialog_form(d_shell);
	p.left=0;
	p.right=100;
	p.top=0;
	p.bottom=100;
	p.flags=POSITION_PERC;
   	ng.value=(long)this;

	ng.width=(int)(x_conv*width);
	ng.height=(int)(y_conv*height);
	ng.flags=DRAW_CLEAR_ON_RESIZE;
	d_area=dialog_draw(d_form,&ng,&p);

	if (mb)
		mb->display(d_shell);
	dialog_realise(d_shell);

	dialog_waitformap(d_area);

	//
	// dialog_waitformap() can very occasionally return before
	// the draw area has recieved its first expose event - which
	// may mean its not yet ready to take drawing calls so ensure
	// it has taken an expose
	//
	while (!dialog_drawexposed(d_area) )
		DialogEvent();
}




//
// DrawdialogCreate
//
// Specifies how a draw dialog should be displayed. Must be used
// prior to DrawStartpage().
//
// INPUTS :	drt	The specification for the display
// RETURNS:	None
//
void
Drawdialog::DrawdialogCreate(drawd_reg_t *drt)
{
	x_conv=X_PIXELS/(SCREEN_WIDTH*POINTSPERINCH);
	y_conv=Y_PIXELS/(SCREEN_HEIGHT*POINTSPERINCH);
	
	//
	// If no registration mechanism is specified, the use sensible
	// defaults.
	//
	if (drt==NULL) {
		width=(1.0/x_conv)*600;
		height=(1.0/y_conv)*400;
		title="Untitled Plot";
		add_mb=NULL;
		return;
	}
	
	oflags=drt->flags;
	width=drt->width;
	height=drt->height;
	title=(char *)malloc(strlen(drt->title)+1);
	strcpy(title,drt->title);
	add_mb=drt->mb;
	d_shell=NULL;
}

//
// Constructor function
//
// Construction is easy - its creation thats more work!
//
Drawdialog::Drawdialog()
{
	d_shell=NULL;
	title=NULL;
}


//
// Destructor function
//
Drawdialog::~Drawdialog()
{
	if (d_shell)
		dialog_destroy(d_shell);

	if (title)
		free(title);
}
