
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: drawps.C,v 2.6 1995/06/24 14:35:31 davidsi Exp $
//
// $Log: drawps.C,v $
// Revision 2.6  1995/06/24  14:35:31  davidsi
// - Store the user's preferred output file and the print spooler
//   command line when specified.
//
// Revision 2.5  1995/06/07  20:59:12  davidsi
// - Dialogs now longer need to issue DialogSetmodal() - see dialog.C.
//
// Revision 2.4  1995/05/19  18:54:08  davidsi
// - DialogentryCreate() now requires an argument specifying total number
//   of visible columns.
//
// Revision 2.3  1995/03/01  22:35:29  davidsi
// - Added Setps class to handle printer specific initialization and preference
//   setting.
// - Changed name from drawps to Drawps for consistancy.
// - Added more printers that are supported by this driver.
// - BUG FIX. All functions check the file ptr is valid before using it.
//
// Revision 2.2  1995/02/05  18:00:29  davidsi
// - Added the drv_rect() member function.
//
// Revision 2.1  1995/01/31  22:13:11  davidsi
// - x_conv and y_conv must be set in the constructor, so that they
//   can be propagated by the drawtext (& similar) class.
// - Use the paper.h include file to determine the size of A4 paper.
// - BUG FIX. Fixed font handling so that fonts can now be correctly
//   set. Now use postscript class to handle font sizing (etc).
//
// Revision 2.0  1995/01/19  21:55:14  davidsi
// - Modified to use the new postscript class which accurately determines
//  text lengths.
// - Added the drv_setlinewidth() function.
// - Added drv_setcolour.
//
// Revision 1.1  1994/12/11  21:37:46  davidsi
// Initial revision
//
//
//




//
// Include Files
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"
#include "paper.h"


#include "drawps.h"
#include "button.h"
#include "postscript.h"
#include "menu.h"
#include "printer.h"
#include "dialog.h"
#include "prefs.h"

//
// This appears to be missing from the standard include files
//
extern "C" {
	extern int unlink(char *);
}

//
// All postscript file output will go to this file unless specified 
// otherwise
//
#define DEFAULT_OUTPUTFILE "output.ps"
#define DEFAULT_OUTPUTCMD "lp -oraw -c"
#define PRTPS_CLASS	"PRINTER_PS"
#define PRTPS_CMD	"CMD"
#define PRTPS_FILE	"FILE"


//
// This is the current (real) filename for postscript output
//
static char *filename;
//
// This is the command needed to output to the printer
//
static char *printercmd;

//
///////////////////////////////////////////////////////////////////////////////
//
// The Setps class provides a private mechanism for controlling
// the postscript specific options.
//
///////////////////////////////////////////////////////////////////////////////
//

//
// Identifiers for the different inputs within the dialog
//
enum dialogids {
	BUTTON_CANCEL,		// The cancel button
	BUTTON_OK,		// The ok button
	ID_FNAME,		// The filename text
	ID_CMD			// the cmd text
	};


//
// Setps simply provides an issolate class that inherits the Button
// capabilities.
//
class Setps : private Button
{
	Dialogshell d_shell;		// main shell
	Dialogentry d_entrylist;	// entry list for filename etc
	Dialogvertical d_vert_main;	// main form 
	Dialoghorizontal d_horiz_ok;	// form for buttons

	int preferences_set;		// TRUE when ok/cancel hit
public:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);

	Setps();
};

//
// Handles the button callbacks
//
void
Setps::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char *str;

	switch(id) {
	case BUTTON_CANCEL:
		//
		// Nothing to do for a cancel
		//
		preferences_set=1;
		return;
	case BUTTON_OK:
		//
		// For the OK case, before we can set 
		// "preferences_set" we must load the
		// information in the dialog into the
		// main variables - ready for use.
		//
		str=d_entrylist.DialogentryGet(ID_FNAME);
		if (str) {
			if (filename)
				free(filename);
			filename=(char *)malloc(strlen(str)+1);
			strcpy(filename,str);
		}
		str=d_entrylist.DialogentryGet(ID_CMD);
		if (str) {
			if (printercmd)
				free(printercmd);
			printercmd=(char *)malloc(strlen(str)+1);
			strcpy(printercmd,str);
		}
		preferences_set=1;
		//
		// Update the database
		//
		db_prefs->DatabaseInsert(PRTPS_CLASS,PRTPS_CMD,printercmd);
		db_prefs->DatabaseInsert(PRTPS_CLASS,PRTPS_FILE,filename);
		break;
	}
}


//
// The Setps constructor does all the work needed to build the
// entire dialog.
//
Setps::Setps()
{
	d_shell.DialogshellCreate("Postscript Printer Setup");
	d_shell.DialogSetmodal();
	d_vert_main.DialogverticalCreate(d_shell);

	d_entrylist.DialogentryCreate(d_vert_main,10);
	d_entrylist.DialogentrySet("Filename",filename,ID_FNAME);
	d_entrylist.DialogentrySet("Output Command",printercmd,ID_CMD);

	d_horiz_ok.DialoghorizontalCreate(d_vert_main);
	(void)ButtonPush(d_horiz_ok,"OK",BUTTON_OK);
	(void)ButtonPush(d_horiz_ok,"CANCEL",BUTTON_CANCEL);

	d_shell.DialogRealize();

	preferences_set=0;
	while (preferences_set==0)
		DialogEvent();
}



//
///////////////////////////////////////////////////////////////////////////////
//
// Main postscript functions - override those in the draw class.
//
///////////////////////////////////////////////////////////////////////////////
//

//
// drv_moveto
//
// INPUTS :
// RETURNS: 
void
Drawps::drv_moveto(double x,double y)
{
	if (file==NULL) 
		return;

	if (psflags & PSFLAGS_STROKEPENDING ) {
		fprintf(file,"stroke\n");
		psflags &= ~PSFLAGS_STROKEPENDING;
	}
	fprintf(file,"%d %d moveto\n",(int)(pxbase+x),(int)(pybase-y));
}

//
// drv_drawto
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_drawto(double x,double y)
{
	if (file==NULL) 
		return;

	fprintf(file,"%d %d lineto\n",(int)(pxbase+x),(int)(pybase-y));
	psflags |= PSFLAGS_STROKEPENDING;
}

//
// drv_rect
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_rect(double x1,double y1,double x2,double y2)
{
	drv_moveto(x1,y1);

	if (file==NULL) 
		return;

	fprintf(file,"%d %d lineto\n",(int)(x2),(int)(y1));
	fprintf(file,"%d %d lineto\n",(int)(x2),(int)(y2));
	fprintf(file,"%d %d lineto\n",(int)(x1),(int)(y2));
	fprintf(file,"%d %d lineto\n",(int)(x1),(int)(y1));
	fprintf(file,"closepath\n");
	fprintf(file,"fill\n");
}



//
// drv_text
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_text(char *text)
{
	if (file==NULL) 
		return;

	if (psflags & PSFLAGS_STROKEPENDING ) {
		fprintf(file,"stroke\n");
		psflags &= ~PSFLAGS_STROKEPENDING;
	}
	fprintf(file,"(%s) show\n",text);
}


//
// drv_setfont
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_setfont(char *name,double fsize)
{
	if (file!=NULL) {
		if (psflags & PSFLAGS_STROKEPENDING ) {
			fprintf(file,"stroke\n");
			psflags &= ~PSFLAGS_STROKEPENDING;
		}
	}


	if (current_font)
		free(current_font);
	current_font=(char *)malloc(strlen(name)+1);
	strcpy(current_font,name);
	fontsize=fsize;
	font_height=fsize;

	ps.set(current_font,(int)fsize);

	//
	// The file is not necessarily open yet.  This function can
	// be called at construction time!
	// 
	if (file) {
		fprintf(file,"/%s findfont\n",current_font);
		fprintf(file,"%d scalefont\n",(int)fsize);
		fprintf(file,"setfont\n");
	}
}


//
// drv_textlength
//
// INPUTS :
// RETURNS: 
//
double
Drawps::drv_textlength(char *txt)
{
	double a;

	a=ps.textwidth(txt);
	return(a);
}


//
// drv_clearpage
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_clearpage()
{
	if (file==NULL) 
		return;

}


//
// drv_drawpage
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_drawpage()
{
	if (file==NULL) 
		return;

	if (psflags & PSFLAGS_STROKEPENDING ) {
		fprintf(file,"stroke\n");
		psflags &= ~PSFLAGS_STROKEPENDING;
	}
	fprintf(file,"showpage\n");
	fflush(file);
}

//
// drv_setgrey
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_setgrey(int perc)
{
	if (file==NULL) 
		return;

	if (psflags & PSFLAGS_STROKEPENDING ) {
		fprintf(file,"stroke\n");
		psflags &= ~PSFLAGS_STROKEPENDING;
	}
}

//
// drv_setlinestyle
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_setlinestyle(int count,double *len)
{
	int i;

	if (file==NULL) 
		return;

	if (psflags & PSFLAGS_STROKEPENDING ) {
		fprintf(file,"stroke\n");
		psflags &= ~PSFLAGS_STROKEPENDING;
	}

	
	fprintf(file,"[ ");
	for(i=0;i<count;i++)
		fprintf(file,"%6.2f ",len[i]);
	fprintf(file,"] 0 setdash\n");
}

//
// drv_setlinewidth
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_setlinewidth(double width)
{
	int i;

	if (file==NULL) 
		return;

	if (psflags & PSFLAGS_STROKEPENDING ) {
		fprintf(file,"stroke\n");
		psflags &= ~PSFLAGS_STROKEPENDING;
	}
	
	fprintf(file,"%6.2f setlinewidth\n",width);
}

void
Drawps::drv_setcolour(colour_e col)
{
}

//
// drv_startpage
//
// INPUTS :
// RETURNS: 
//
void
Drawps::drv_startpage()
{
	if (currentmode==PRTMDE_FILE)
		file=fopen(filename,"w");
	else {		// assume it will be PRTMDE_PRINTER
		char *str;

		str=tmpnam(NULL);

		if (tmpfilename)
			free(tmpfilename);
		tmpfilename=(char *)malloc(strlen(str)+1);
		strcpy(tmpfilename,str);
		file=fopen(tmpfilename,"w");
	}
		
	if (file==NULL) {
		DialogMessage("Cannot open output file");
		return;
	}
	fprintf(file,"%%!PS\n");
}

//
//
//
double
Drawps::drv_textheight()
{
        return(font_height);
}


void
Drawps::PrinterdrawSetup()
{
	//
	// This does all the work, when created!
	//
	class Setps ps;
}

//
// Constructor function
//
Drawps::Drawps()
{
	//
	// Postscript works in points so no conversion needed!
	// The x_conv and y_conv values cannot be reset at any other point
	// as this can confuse classes such as the drawtext class, which
	// assumes they are constant for all draw derived classes.
	//
	x_conv=1.0;
	y_conv=1.0;
	psflags=0;

	width=INCHES2POINTS(A4_PAPER_WIDTH);
	height=INCHES2POINTS(A4_PAPER_HEIGHT);
	pybase=height;
	pxbase=0.0;

	file=NULL;
	current_font=NULL;
	tmpfilename=NULL;
}

//
// Destructor function
//
Drawps::~Drawps()
{
	if (file)
		fclose(file);

	if (currentmode==PRTMDE_PRINTER && tmpfilename) {
		char buffer[1024];

		sprintf(buffer,"%s %s",printercmd,tmpfilename);
		system(buffer);
		unlink(tmpfilename);
	}

	if (tmpfilename)
		free(tmpfilename);
}


static class Printerdraw *
pscreate()
{
	return( new class Drawps );
}

void
drawps_init(menu_bar *mb)
{
	char *str;

	printerdevices.PrinterdeviceAdd("Apple Laserwriter",pscreate);
	printerdevices.PrinterdeviceAdd("Apple Laserwriter II",pscreate);
	printerdevices.PrinterdeviceAdd("HP LaserJet 4Si MX",pscreate);
	printerdevices.PrinterdeviceAdd("Generic postscript printer",pscreate);

	str=db_prefs->DatabaseRetrieve(PRTPS_CLASS,PRTPS_FILE);
	if (str==NULL)
		str=DEFAULT_OUTPUTFILE;
	filename=(char *)malloc(strlen(str)+1);
	strcpy(filename,str);

	str=db_prefs->DatabaseRetrieve(PRTPS_CLASS,PRTPS_CMD);
	if (str==NULL)
		str=DEFAULT_OUTPUTCMD;
	printercmd=(char *)malloc(strlen(str)+1);
	strcpy(printercmd,str);
}
