
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: drawtext.C,v 2.6 1995/06/07 20:54:38 davidsi Exp $
//
// $Log: drawtext.C,v $
// Revision 2.6  1995/06/07  20:54:38  davidsi
// - Added DTEXT_FLAG_NODRAWPAGE flag that prevents page throws at the
//   end of writing a text page.
//
// Revision 2.5  1995/05/29  19:08:12  davidsi
// - Removed debug printfs.
//
// Revision 2.4  1995/05/19  18:53:59  davidsi
// - Added a new variant for the DrawtextOutput() function member, where
//   control is provided over exactly what gets drawn.
// - BUG FIX. In DrawtextTitle(), if the new title is NULL don't attempt to
//   strcpy - simply release the old title and set the new title to NULL.
//
// Revision 2.3  1995/04/16  20:33:29  davidsi
// - Added a title mechanism.  The title can be set with DrawtextTitle()
//   and is held in the title member.
// - BUG FIX. Page numbers were overwriting real text because the number of
//   text lines being written to each page failed to exclude those lines
//   req'd for the page number.
//
// Revision 2.2  1995/03/02  20:53:38  davidsi
// - Added DrawtextOutputline() to allow single lines to be output in
//   a controlled fashion.
// - Switch to using Draw reference counting rather than issuing
//   straight deletes on the borrowed Draw instance.
// - BUG FIX. When creating a Drawtext instance, the colour information must
//   also be copied form the originating Draw instance.
//
// Revision 2.1  1995/02/05  18:01:47  davidsi
// - Added the drv_rect() member function.
//
// Revision 2.0  1995/02/02  21:20:47  davidsi
// - Modifications to reflect name changes in the Draw class.
// - Was previously the textps class.
//
//
//
//




//
// Include Files
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"
#include "drawtext.h"

#define NO_PRINT_LINES	5

//
//////////////////////////////////////////////////////////////////////////////
//
// The redirector functions.  Although the Drawtext class inherits from the
// Draw class, the functionality is actually redirected to another Draw
// class member that was given at construction time.
//
//////////////////////////////////////////////////////////////////////////////
//

void 
Drawtext::drv_startpage()
{
	dptr->drv_startpage();
}

void
Drawtext::drv_clearpage()
{
	dptr->drv_clearpage();
}

void
Drawtext::drv_moveto(double x,double y)
{
	dptr->drv_moveto(x,y);
}

void
Drawtext::drv_drawto(double x, double y)
{
	dptr->drv_drawto(x,y);
}

void
Drawtext::drv_rect(double x1, double y1,double x2,double y2)
{
	dptr->drv_rect(x1,y1,x2,y2);
}

void
Drawtext::drv_text(char *string)
{
	dptr->drv_text(string);
}

void
Drawtext::drv_setfont(char *name,double fsize)
{
	dptr->drv_setfont(name,fsize);
}

void 
Drawtext::drv_drawpage()
{
	dptr->drv_drawpage();
}

double
Drawtext::drv_textlength(char *text)
{
	return(dptr->drv_textlength(text));
}

double
Drawtext::drv_textheight()
{
	return(dptr->drv_textheight());
}

void
Drawtext::drv_setgrey(int percentage)
{
	dptr->drv_setgrey(percentage);
}

void
Drawtext::drv_setlinestyle(int count,double *len)
{
	dptr->drv_setlinestyle(count,len);
}

void
Drawtext::drv_setlinewidth(double width)
{
	dptr->drv_setlinewidth(width);
}

void
Drawtext::drv_setcolour(colour_e col)
{
	dptr->drv_setcolour(col);
}




//
//////////////////////////////////////////////////////////////////////////////
//
// The Drawtext specific functions
//
//////////////////////////////////////////////////////////////////////////////
//


//
// Drawtext::DrawtextNpages
//
// Return the number of pages that this Drawtext will output.  The value
// is only valid once DrawtextPrepare() is used, otherwise 0 will be
// returned.
//
int
Drawtext::DrawtextNpages()
{
	return(totalpages);
}



//
// Drawtext::DrawtextPrepare
//
// Outputs a complete text area.
//
// INPUTS : 	None
// RETURNS: 	None
//
void
Drawtext::DrawtextPrepare()
{
	double lwidth;
	int offset;
	int field;
	int lineno;
	int fields;
	int n_lines;

	//
	// If resources are still hanging around from the last call to
	// DrawtextPrepare() release them now.
	//
	DrawtextComplete();

	//
	// Get geometry from the inherited Text class
	//
	fields=TextNfields();
	n_lines=TextNlines();

	charheight=DrawTextheight();

	//
	// Allocate work areas
	//
	strings=(char **)malloc(fields*n_lines*sizeof(char *));
	maxwidths=(int *)malloc((1+fields)*sizeof(int));
	positions=(double *)malloc(fields*sizeof(double));

	for(field=0;field<fields;field++) 
		maxwidths[field]=0;
	
	//
	// Compute the lengths (in pixel terms) of the strings,
	// and determine largest string for each field.
	//
	for(lineno=0;lineno<n_lines;lineno++) {
		for(field=0;field<fields;field++) {
			offset=lineno*fields+field;
			strings[offset]=TextRead(lineno,field);
			lwidth=DrawTextlength(strings[offset]);
			if (lwidth>0.0)
			if ((int)lwidth>maxwidths[field])
				maxwidths[field]=(int)lwidth;
		}
	}

	//
	// Compute start positions (x coordinate offset) for each field.
	//
	for(field=0;field<fields;field++) {
		if (field==0)
			positions[0]=0.0;
		else
			positions[field]=(double)(positions[field-1]+maxwidths[field-1])+10.0; 
	}

	//
	// Determine total number of pages to be displayed
	//
	lines_per_page=(int)(height/charheight);
	text_lines_per_page=(int)(height/charheight)-NO_PRINT_LINES;

	//
	// If there is going to be a title, make room for it.
	//
	if (title)
		text_lines_per_page -= 2;

	//
	// We subtract 1 from n_lines because the first line is 
	// the title line which gets used on everypage and is
	// accounted for within lines_per_page
	//
	totalpages=1+( (n_lines-1) /lines_per_page);
}

//
// Drawtext::DrawtextOutputline
//
// Outputs the specified fields on the given line, updating the x and
// y coordinates appropriately
//
//
void
Drawtext::DrawtextOutputline(double &sx,double &sy,int lineno,int fields)
{
	class attr *attr;
	attr_t attr_flag;
	int field;
	colour_e colour;
	double lwidth;
	int offset;

	for(field=0;field<fields;field++) {
		offset=lineno*fields+field;
		attr=TextReadattribute(lineno,field);
		attr_flag=attr->extract_attr();
		switch (attr_flag&FIELDATTR_TEXTPOS){
		case FIELDATTR_LEFT:
			DrawMoveto(sx+positions[field],sy);
			break;
		case FIELDATTR_RIGHT:
			lwidth=DrawTextlength(strings[offset]);
			DrawMoveto(sx+positions[field]+maxwidths[field]-lwidth,sy);
			break;
		case FIELDATTR_CENTRE:
			lwidth=DrawTextlength(strings[offset]);
			DrawMoveto(sx+positions[field]+maxwidths[field]/2-lwidth/2,sy);
			break;
		}
		DrawText(strings[offset]);
	}
	sy += charheight;
}

//
// Drawtext::DrawtextOutput page cflags
//
// Outputs a single page from a Drawtext instance.
//
// INPUTS : 	page		The page number to output
//		cflags		See DTEXT_XXX in drawtext.h
// RETURNS: 	None
//
void
Drawtext::DrawtextOutput(int pageno,unsigned long cflags)
{
	double sx,sy;
	int lineno;
	int fields;
	int n_lines;
	int line;
	char buffer[10];

	//
	// Check page is valid. Note this also catches the case when
	// DrawtextPrepare() has not been used because in such cases
	// totalpages will be 0 so pageno>totalpages cannot be true.
	//
	// TODO: The underlying text may still have changed if further
	// calls to TextAdd() have been made since DrawtextPrepare().
	// For the moment we hope (pray) that all users of Drawtexts
	// are friendly in this aspect.
	//
	if (pageno<1||pageno>totalpages)
		return;

	//
	// Get sizes
	//
	fields=TextNfields();
	n_lines=TextNlines();

	//
	// Clear the output page. Note that the propagation of the
 	// background colour is a bit unpleasant!  We don't need tricks
	// for the foreground colour because the DrawSetcolour() call
	// is correctly indirected. The background that will be
	// drawn is that of the dptr Draw instance rather than that
	// set in our own draw instance so we must propagate...
	//
	dptr->ColourSet(COLOUR_FUNCTION_BACKGROUND,ColourBackground() );

	if (cflags & DTEXT_FLAG_PAGECLEAR)
		DrawClearpage();

	// DrawSetcolour(RED);

	// DrawMoveto(0,0);
	// DrawDrawto(width-1,0);
	// DrawDrawto(width-1,height-1);
	// DrawDrawto(0,height-1);
	// DrawDrawto(0,0);

	//
	// Set text output colour
	//
	DrawSetcolour(ColourForeground());

	sx=1;
	sy=charheight;

	//
	// Write the title if one has been defined
	//
	if (title) {
		DrawMoveto(sx,sy);
		DrawText(title);
		sy += charheight*2;
	}

	
	// 
	// Print column headings
	//
	DrawtextOutputline(sx,sy,0,fields);
	sy += charheight;

	for( lineno=1+lines_per_page*(pageno-1),line=0;
		lineno<n_lines&&line<text_lines_per_page;
		lineno++,line++) {

		DrawtextOutputline(sx,sy,lineno,fields);
	}
	sy += charheight;
	if (cflags & DTEXT_FLAG_PAGENUMBER) {
		sprintf(buffer,"-%d-",pageno);
		DrawMoveto((width-DrawTextlength(buffer))/2,
				(lines_per_page-1)*charheight);
		DrawText(buffer);
	}

	//
	// Don't render the page yet, if requested to omit 
	// the draw page call.
	//
	if (! (cflags&DTEXT_FLAG_NODRAWPAGE) )
		DrawPage();
}

//
// Drawtext::DrawtextOutput page
//
// Outputs a single page from a Drawtext instance.
//
// INPUTS : 	page		The page number to output
// RETURNS: 	None
//
void
Drawtext::DrawtextOutput(int pageno)
{
	DrawtextOutput(pageno,DTEXT_FLAG_PAGECLEAR|DTEXT_FLAG_PAGENUMBER);
}

//
// Drawtext::DrawtextTitle title
//
// Set the title to be displayed at the top of everypage
//
// INPUTS : 	title	The title to be set
// RETURNS: 	None
//
void
Drawtext::DrawtextTitle(char *ititle)
{
	if (title)
		free(title);

	if (ititle) {
		title=(char *)malloc(strlen(ititle)+1);
		strcpy(title,ititle);
	} else
		title=NULL;
}


//
// Drawtext::DrawtextComplete
//
// Release all resources once outputing is complete
//
// INPUTS : 	None
// RETURNS: 	None
//
void
Drawtext::DrawtextComplete()
{
	if (strings)
		free(strings);
	if (maxwidths)
		free(maxwidths);
	if (positions)
		free(positions);
	strings=NULL;
	maxwidths=NULL;
	positions=NULL;

	//
	// Set totalpages back to 0 to prevent DrawtextOutput() being
	// used without a further DrawtextPrepare() call.
	//
	totalpages=0;
}


//
// Constructor for the Drawtext class.
//
// INPUTS :	dr 	The Draw instance to use for the output. 
// RETURNS:	None
//
Drawtext::Drawtext(class Draw *dr)
{
	dptr=dr;

	//
	// One more reference on the original Draw instance.  It now
	// won't go away until we explicitly delete it.
	//
	dptr->DrawRef();

	//
	// All protected members must be copied upwards so this
	// new text class member appears to have the same properities
	// as the original Draw class member.
	//
	x_conv=dptr->x_conv;
	y_conv=dptr->y_conv;
	xbase=dptr->xbase;
	ybase=dptr->ybase;
	xpos=dptr->xpos;
	ypos=dptr->ypos;
	width=dptr->width;
	height=dptr->height;
	strings=NULL;
	maxwidths=NULL;
	positions=NULL;

	//
	// Colour information also gets copied
	//
	ColourCopy(dptr);

	//
	// Other members are initialized to "safe" values
	//
	totalpages=0;
	title=NULL;
}

//
// Destructor for the Drawtext class.
//
Drawtext::~Drawtext()
{
	if (title)
		free(title);
	//
	// We no longer need a handle on the original Draw instance
	//
	dptr->DrawUnref();
	DrawtextComplete();
}
