
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

// 
// $Id: dtrans.C,v 2.7 1995/06/17 20:40:27 davidsi Exp $
//
// $Log: dtrans.C,v $
// Revision 2.7  1995/06/17  20:40:27  davidsi
// - BUG FIX. The function transactions_display::list_trans_dates() and
//   transactions_display::list_trans_value() could potentially miss
//   some securities because of a missing set of braces.
//
// Revision 2.6  1995/05/19  18:54:32  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.5  1995/04/25  21:10:34  davidsi
// - Added support for a busy cursor.
// - Updated to use ColourSetowner() so that the user's colour prefences for
//   the transaction display can be pernamently stored in the preference
//   database.
// - BUG FIX. When colours change it is not necessary to rework the
//   text (that can take time) - all that is needed is a quick redraw.
//
// Revision 2.4  1995/04/16  20:29:40  davidsi
// - Added a DATE menu option so that the period for transaction display
//   can be specified by the user.
// - Added support for displaying transactions in date order and value order.
// - Added callback_sec member to allow transaction displays to be updated
//   when securities are added, removed or modified.  This gets registered and
//   removed as appropriate.
//
// Revision 2.3  1995/03/08  21:21:13  davidsi
// - Factored the dividend class into two classes, the dividend class provides
//   all the main dividend handling functionality and the Dividendedit class
//   allows dividends to be edited.  This reduces the amount of memory
//   req'd to represent a diviend.
// - The extract() member has been renamed to DividendExtract.
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that.
// - Revised to use the new Holding class which contains transaction
//   information rather than the security class.
// - BUG FIX. dividend::save_record() was incorrectly writing the address
//   of the date string to the file, rather than the date string itself.
//
// Revision 2.2  1995/03/02  20:46:40  davidsi
// - Removed all dialog references - use Dialog class instead.
// - BUG FIX. When  building the text dialog the left/right justification was
//   not being set for each of the fields.  Also now sets colour information.
//
// Revision 2.1  1995/01/31  22:09:20  davidsi
// - Revised to use the textdialog class.  This results in much of
//   the required functionality now being inherited.
//
// Revision 2.0  1995/01/19  21:49:52  davidsi
// - In transactions_display::transactions_display() add a callback
//   so that the display can be refreshed should a transaction be
//   created.
// - Added transactions_display::trans_change for transaction change
//   callbacks.
// - Modified all time based values to use the new date class.
//
// Revision 1.3  1994/11/05  21:07:13  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.2  1994/10/17  20:59:22  davidsi
// - All calls to extract() in the security class now require a further
//   argument, so that the type of security can be determined using flags
//   rather than strings.
// - Now that per security transaction lists are circular the termination
//   condition for scanning the list must be updated for having completed
//   the entire list.
//
// Revision 1.1  1994/07/03  15:55:37  davidsi
// Initial revision
//
//
//
//
 

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>

#include "localmacros.h"

#include "dtrans.h"
#include "menu.h"
#include "callback.h"
#include "setdate.h"


//
// Menu identifiers
//
enum menu_e {
	MENU_EXIT,
	MENU_COLOUR,
	MENU_PRINT,
	MENU_BYDATE,
	MENU_BYSHARE,
	MENU_BYAMOUNT,
	MENU_DATERANGE,
	MENU_DATE,
        MENU_GETRIGHTS,
        MENU_BUY,
        MENU_SELL,
        MENU_SPLIT,
        MENU_BONUS,
	};

//
// Defines for the database identifiers
//
#define TRANSDISP_DB_CLASS            "TRANSDISP"



//
// transactions_display::create_header mode
//
// Creates a text line in the header which can be used to display a
// title for the columns below it.  Code elsewhere is responsible for
// deciding which portion of the string is to be displayed.
// This function is private to the transactions_display class.
//
// INPUTS :	mode 	Either NAMEFIRST or DATEFIRST specifying whether
//			the first column is share name or date
// RETURNS:	None.
//
void
transactions_display::create_header( int mode )
{
	char str[256];		// more build up space, hope this is big enough 

	switch(mode) {
	case NAMEFIRST:
		sprintf(str,"Share name\tType\tTime\tNumber\tPrice\tTotal");
		break;
	case DATEFIRST:
		sprintf(str,"Time\tShare name\tType\tNumber\tPrice\tTotal");
		break;
	default:
		fprintf(stderr,"Bad mode select @ create_header\n");
		exit(1);
  	}
	TextAdd(str);
}



//
// transactions_display::append transaction mode securityname
//
// Appends transaction information to the output list. This function is
// private to the transactions_display class.
//
// INPUTS :	transaction	The transaction description
//		mode		DATEFIRST or NAMEFIRST to specify col order
//		securityname	Name of the security as a C string.	
// RETURNS:	TRUE if something added to the output list, else
//		FALSE.
//
int
transactions_display::append(class transaction *trans,
		int mode,
		char *secname)
{
	char *str;		// build up space, hope this is big enough 

	str=trans->trans2string(mode,trans_displayed,secname);
	if (!str)
		return(FALSE);
	TextAdd(str);
	return(TRUE);
}


//
// transactions_display::list_trans_shares 
//
// Generates an output list containing transaction information for
// the securities in alphanumeric order. This function is private
// to the transactions_display class.
//
// INPUTS :	None.
// RETURNS:	None.
///
void
transactions_display::list_trans_shares()
{
	char mode;			// boolean for writing name or not 
	transaction *trans;		// current transaction
	transaction *head_trans;	// head of transaction list
	class security *sec;		// current security
	class Holding *ptr;		//
	char *secname;			// current security name
	class date tdate;		// date for current transaction

	create_header(NAMEFIRST);
	//
	// Now scan all the securities..
	//
	ptr=portfolio->head();
	while ( !portfolio->eol(ptr) ) {
		mode=TRUE;
		head_trans=ptr->HoldingTransaction();
		sec=ptr->HoldingSecurity();
		sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);
		//
		// For each security scan all its transactions
		//
		trans=head_trans;
		while ( trans!=NULL ) {
			//
			// Check date of transaction
			//
			trans->extract(NULL, &tdate,NULL,NULL,NULL);

			if ( !tdate.inrange(&start_date,&end_date) )
				goto next_transaction;
			//
			// See if name needs to be written
			//
			if (mode) {
				if (append(trans,NAMEFIRST,secname))
					mode=FALSE;	// don't write name next iteration
			} else
				(void)append(trans,NAMEFIRST,""); // is on display
next_transaction:
			//
			// Next transaction
			//
			trans=trans->dual_next();
			if (trans==head_trans)
				break;
		}
		//
		// Next security in holding
		//
		ptr=ptr->next();
		DialogCursor(DCURSOR_BUSY);
	}
	DialogCursor(DCURSOR_NORMAL);
	return;
}


//
// Dated output order
//
// Use a class that inherits ordered insertion from the list class. This
// minimizes the amount of real work needed for date based ordering. The
// transactions are held in datelistnode instances until all transactions
// have been processed.  The datelist class acts as a place holder 
//

//
// Nodes held on date sorted transaction list
//
class datelistnode : public list_node
{
	class date tdate;		// transaction's date
	class transaction *trans;	// ptr to the transaction
	
	//
	// Friends
	//
	friend class datelist;
	friend class transactions_display;
public:
	datelistnode(class date &idate,class transaction *itrans) {
		tdate=idate;
		trans=itrans;
	}
	class datelistnode *next() {
		return( LISTREMAP(datelistnode,list_node,next,this) );
	}
	virtual ~datelistnode()	{ }
};


//
// The date sorted transaction list
//
class datelist : public list 
{
protected:
        virtual int compare(class list_node *n1,class list_node *n2) {
		return( ((class datelistnode *)n1)->
			tdate.compare(&(((class datelistnode *)n2)->tdate)) );
	}

public:
	class datelistnode *head() {
		return( LISTREMAP(datelistnode,list,head,this) );
		}
	datelist() {}
	virtual ~datelist() {}
};


//
// transactions_display::list_trans_dates
//
// Produces an list of transactions using a date ordering.
//
// INPUTS :	None
// RETURNS:	 None
//
void
transactions_display::list_trans_dates()
{
	class datelist transdl;		// date ordered list
	class datelistnode *transnode;	// transaction information on list
	transaction *trans;		// current transaction
	transaction *head_trans;	// head of transaction list
	class security *sec;		// current security
	class security *lastsec;	// last security displayed
	char *secname;			// current security name
	class date tdate;		// date for current transaction
	class Holding *ptr;		//

	create_header(DATEFIRST);
	//
	// Now scan all the securities in the portfolio.  When a 
	// transaction of interest is found insert it into the 
	// date ordered list
	//
	ptr=portfolio->head();
	while ( !portfolio->eol(ptr) ) {
		head_trans=ptr->HoldingTransaction();
		//
		// For each security scan all its transactions
		//
		trans=head_trans;
		while ( trans!=NULL ) {
			//
			// Check date of transaction
			//
			trans->extract(NULL, &tdate,NULL,NULL,NULL);

			if ( tdate.inrange(&start_date,&end_date) ) {
				transnode=new class datelistnode(tdate,trans);
				transdl.insert_order(transnode);
			}
			//
			// Next transaction
			//
			trans=trans->dual_next();
			if (trans==head_trans)
				break;
		}
		//
		// Next security in holding
		//
		ptr=ptr->next();
		DialogCursor(DCURSOR_BUSY);
	}


	//
	// Write the sorted list to the display
	//
	transnode=transdl.head();
	lastsec=NULL;
	while ( !transdl.eol(transnode) ) {
		trans=transnode->trans;
		sec=trans->Security();
		sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);
		if (sec!=lastsec) {
			if (append(trans,DATEFIRST,secname))
				lastsec=sec;
		} else
			(void)append(trans,DATEFIRST,""); 
		transnode=transnode->next();
		DialogCursor(DCURSOR_BUSY);
	}
	DialogCursor(DCURSOR_NORMAL);
	return;
}

//
// Value output order
//
// Use a class that inherits ordered insertion from the list class.
//

//
// Nodes held on value sorted transaction list
//
class valuelistnode : public list_node
{
	double total;			// transaction's value
	class transaction *trans;	// ptr to the transaction
	
	//
	// Friends
	//
	friend class valuelist;
	friend class transactions_display;
public:
	valuelistnode(double itotal,class transaction *itrans) {
		total=itotal;
		trans=itrans;
	}
	class valuelistnode *next() {
		return( LISTREMAP(valuelistnode,list_node,next,this) );
	}
	virtual ~valuelistnode()	{ }
};


//
// The date sorted transaction list
//
class valuelist : public list 
{
protected:
        virtual int compare(class list_node *n1,class list_node *n2) {
		return( ((class valuelistnode *)n1)->total <
			((class valuelistnode *)n2)->total );
	}
public:
	class valuelistnode *head() {
		return( LISTREMAP(valuelistnode,list,head,this) );
		}
	valuelist() {}
	virtual ~valuelist() {}
};


//
// transactions_display::list_trans_value 
//
// Produces a list of transactions using a value ordering.
//
// INPUTS :	None
// RETURNS:	None
///
void
transactions_display::list_trans_value()
{
	class valuelist transvl;	// value ordered list
	class valuelistnode *transnode;	// transaction information on list
	transaction *trans;		// current transaction
	transaction *head_trans;	// head of transaction list
	class security *sec;		// current security
	class security *lastsec;	// last security displayed
	char *secname;			// current security name
	class date tdate;		// date for current transaction
	class Holding *ptr;		//
	double total;			// total value of transaction

	create_header(NAMEFIRST);
	//
	// Now scan all the securities in the portfolio.  When a 
	// transaction of interest is found insert it into the 
	// date ordered list
	//
	ptr=portfolio->head();
	while ( !portfolio->eol(ptr) ) {
		head_trans=ptr->HoldingTransaction();
		//
		// For each security scan all its transactions
		//
		trans=head_trans;
		while ( trans!=NULL ) {
			//
			// Check date of transaction
			//
			trans->extract(NULL, &tdate,NULL,NULL,&total);

			if ( tdate.inrange(&start_date,&end_date) ) {
				transnode=new class valuelistnode(total,trans);
				transvl.insert_order(transnode);
			}
			//
			// Next transaction
			//
			trans=trans->dual_next();
			if (trans==head_trans)
				break;
		}
		//
		// Next security in holding
		//
		ptr=ptr->next();
		DialogCursor(DCURSOR_BUSY);
	}


	//
	// Write the sorted list to the display
	//
	transnode=transvl.head();
	lastsec=NULL;
	while ( !transvl.eol(transnode) ) {
		trans=transnode->trans;
		sec=trans->Security();
		sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);
		if (sec!=lastsec) {
			if (append(trans,NAMEFIRST,secname))
				lastsec=sec;
		} else
			(void)append(trans,NAMEFIRST,""); 
		transnode=transnode->next();
		DialogCursor(DCURSOR_BUSY);
	}
	DialogCursor(DCURSOR_NORMAL);
	return;
}




//
// transactions_display::create_display 
//
// Builds the strings to be displayed for any of the options set by the
// user.  This function is private to the transactions_display class.
//
// INPUTS :	None.
// RETURNS:	None.
//
void
transactions_display::create_display()
{
	int i;

	if (!setmenuflag) {
		MenuItemaction("Project","Close",MENU_EXIT);
		MenuItemaction("Project","Print",MENU_PRINT);
		MenuItemaction("Project","Set Dates",MENU_DATE);
		MenuItemaction("Project","Colours",MENU_COLOUR);

		MenuItem1ofn("Ordering","By security",MENU_BYSHARE,MENUSTATE_ON);
		MenuItem1ofn("Ordering","By date",MENU_BYDATE,MENUSTATE_OFF);
		MenuItem1ofn("Ordering","By amount",MENU_BYAMOUNT,MENUSTATE_OFF);
		MenuItemnofn("Transaction Types","Purchases",
					MENU_BUY,MENUSTATE_ON);
		MenuItemnofn("Transaction Types","Sells",
					MENU_SELL,MENUSTATE_ON);
		MenuItemnofn("Transaction Types","Splits",
					MENU_SPLIT,MENUSTATE_ON);
		MenuItemnofn("Transaction Types","Bonuses",
					MENU_BONUS,MENUSTATE_ON);
		MenuItemnofn("Transaction Types","Rights",
					MENU_GETRIGHTS,MENUSTATE_ON);
		for(i=1;i<7;i++) {
			class attr *attr;
			if (i==1 || i==2)
				attr=new class attr(FIELDATTR_LEFT,FOREGROUND);
			else
				attr=new class attr(FIELDATTR_RIGHT,FOREGROUND);
			TextSetdefaultattr(i,attr);
		}
		TextdialogSettitle("Transactions");
		ColourSetowner(TRANSDISP_DB_CLASS);
		setmenuflag=1;
	}
	TextClear();

	switch (ordering) {
	case MENU_BYSHARE:
 		list_trans_shares();
		break;
	case MENU_BYDATE:
 		list_trans_dates();
		break;
	case MENU_BYAMOUNT:
 		list_trans_value();
		break;
	default:
		DialogPanic("Bad ordering switch value @ create_display\n");
	}
	TextComplete();
	TextdialogOutput();
}


//
// transactions_display::reselect_types transtype value
//
// Allows the user to select different types of transaction (infact allows
// the code that calls this function to allow the user to...!).  This
// function is private to the transactions_display class. 
//
// INPUTS :	transtype	The transaction type to allow/disallow.
//		value		TRUE/FALSE to determine if it is to be
//				allowed.
// RETURNS:	None.
//
void
transactions_display::reselect_types(
			int transtype,		// values of TRANS_XXX 
			int value )
{

	// we only switch off if its on (& vice versa)
	// to save unnecessary work! 
	if (value && !(trans_displayed & transtype) ) {
		trans_displayed |= transtype;
		create_display();
	} else if (!value && (trans_displayed & transtype) ) {
		trans_displayed &= ~transtype;
		create_display();
	}
}





//
// transactions_display::MenuCallback client_data onflag
//
// This is the call back used when the user selects something on the 
// menu.  The function overrides the standard function in the
// inheritated menu class.  
//
// INPUTS :	client_data	The id for the menu item thats been selected.
//		onflag		Whether the menu item is on/off.
// RETURNS: 	None
//
void
transactions_display::MenuCallback(unsigned long client_data, int onflag)
{

	switch (client_data) {
	case MENU_EXIT:
		if (!colourdialog.SemaphoreIslocked())
			delete this;
		break;
	case MENU_COLOUR:
		if (colourdialog.SemaphoreClock()!= -1) {
			ColourSelect();
			colourdialog.SemaphoreUnlock();
			TextdialogRedraw();
		}
		break;
	case MENU_PRINT:
		TextdialogPrint();
		break;
	case MENU_BYSHARE:
		ordering=MENU_BYSHARE;
		create_display();
		break;
	case MENU_BYDATE:
		ordering=MENU_BYDATE;
		create_display();
		break;
	case MENU_BYAMOUNT:
		ordering=MENU_BYAMOUNT;
		create_display();
		break;
	case MENU_DATE:
		if (DialogSelect2dates("Set date period for transactions",
					"Start",start_date,
					"Finish",end_date)!= -1)
			create_display();
		break;
	case MENU_GETRIGHTS:	// bit of a fiddle to enable all rights trans 
		reselect_types(TRANS_GETRIGHTS|TRANS_SELRIGHTS|TRANS_USERIGHTS|
				 TRANS_BUYRIGHTS|TRANS_EXRIGHTS,onflag);	
		break;
	case MENU_BUY:
		reselect_types( TRANS_BUY,onflag);	
		break;
	case MENU_SELL:
		reselect_types( TRANS_SELL,onflag);	
		break;
	case MENU_SPLIT:
		reselect_types( TRANS_SPLIT,onflag);	
		break;
	case MENU_BONUS:
		reselect_types(TRANS_BONUS,onflag);	
		break;
	default:
		DialogPanic("Bad client_data switch value @ transactions_display::MenuCallback");
	}
}



//
// transactions_display::trans_change this 
//
// This callback is registered to execute whenever a transaction occurs.
//
// INPUTS :	tmpthis		The this ptr defined when the callback
//				was registered.
// RETURNS:	None
//
void
transactions_display::trans_change(char *tmpthis,char *)
{
	class transactions_display *rthis=(class transactions_display *)tmpthis;
	rthis->create_display();
}




//
// transactions_display
//
// Constructor for transactions_display class.
//
// Creates & manages the window which displays transactions for a
// given portfolio.
//
// INPUTS :	None
// RETURNS:	None
//
transactions_display::transactions_display()
{
	// start date is set automatically by date constructor to
	// beginning of dates.
	end_date.setdate(); 		// until todays date
	ordering=MENU_BYSHARE;		// corresponds to MENUI_ENABLED 
	trans_displayed=-1;		// everything is initially displayed
	setmenuflag=0;			// no menu yet attached
	callback=main_trans_list->add_vect.CallbackAdd(
			&transactions_display::trans_change,
			(char *)this);
	callback_sec=main_sec_list->changed_vect.CallbackAdd(
			&transactions_display::trans_change,
			(char *)this);


	create_display();
}

//
// ~transactions_display
//
// Destructor for transactions_display class.
//
transactions_display::~transactions_display()
{
	//
	// Remove the callback node, from the list of callbacks
	// that are made when a transaction is added
	//
	main_trans_list->add_vect.CallbackRemove(callback);
	main_sec_list->changed_vect.CallbackRemove(callback_sec);
}


//
// display_transactions 
//
// Main entry point for creating & managing the window which displays 
// transactions performed for a given portfolio.
//
// INPUTS :	parent		The parent for the dialog which is created
// RETURNS:	None.
///
static void
display_transactions(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
	transactions_display *newtrans=new transactions_display;
}


//
// dtrans_init parent menu_bar
//
// Initialises the display transactions module.
// 
// INPUTS :	None
// RETURNS:	None
///
void
dtrans_init(menu_bar *mb)
{
	MENU("Transactions","Display",display_transactions,0);
}
