
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: dval.C,v 2.8 1995/06/07 20:45:38 davidsi Exp $
//
// $Log: dval.C,v $
// Revision 2.8  1995/06/07  20:45:38  davidsi
// - BUG FIX.  Ensure the % sign appears correctly for the % chg column.
//
// Revision 2.7  1995/06/02  19:58:02  davidsi
// - BUG FIX.  In find_change() the value computed when there is less than
//   a year between the average purchase price and the date of the valuation,
//   the system failed to compute a true percentage.
//
// Revision 2.6  1995/05/18  20:49:51  davidsi
// - Callbacks are now stored using opaque handles rather than pointers to
//   the actual nodes (see callback.C for the whole story).
//
// Revision 2.5  1995/04/25  21:11:53  davidsi
// - Added support for a busy cursor.
// - Updated to use ColourSetowner() so that the user's colour prefences for
//   the valuation display can be pernamently stored in the preference
//   database.
// - BUG FIX. When colours change it is not necessary to rework the
//   text (that can take time) - all that is needed is a quick redraw.
//
// Revision 2.4  1995/04/16  20:31:08  davidsi
// - Added callback_sec member to allow valuation displays to be updated when
//   securities are added, removed or modified.  This gets registered and
//   removed as appropriate.
// - BUG FIX.  The valuation_display::~valuation_display() function was not
//   removing the dividend callback handler.
//
// Revision 2.3  1995/03/08  21:18:42  davidsi
// - The extract() member of dividend was renamed, so this module modified
//   to reflect that.
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that.
// - Revised to use the new Holding class which contains dividend information
//   rather than the security class.
//
// Revision 2.2  1995/03/01  22:23:01  davidsi
// - BUG FIX. When  building the text dialog the left/right justification was
//   not being set for each of the fields.  Also now sets colour information.
// - Removed all dialog references - use Dialog class instead.
//
// Revision 2.1  1995/01/31  21:34:03  davidsi
// - Revised to use the textdialog class.  This results in much of
//   the required functionality now being inherited.
//
// Revision 2.0  1995/01/19  21:24:16  davidsi
// - Added redisplay() function which is now used by price and transaction
//   change callbacks.
// - Removed the time members of the valuation_display class.  Dates are now
//   managed using the date class.
//
// Revision 1.4  1994/12/11  20:44:55  davidsi
// - Since the menu function is now declared static, it cannot use
//   the implicit "this", therefore updated to access members via
//   a new pointer.  However, the finish() function, that was not
//   a member, is no longer needed.
//
// Revision 1.3  1994/11/05  21:07:31  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.2  1994/10/17  20:52:27  davidsi
// - All calls to extract() in the security class now require a further
//   argument, so that the type of security can be determined using flags
//   rather than strings.
// - Now that per security dividend lists are circular the termination
//   condition for scanning the list must be updated for having completed
//   the entire list.
// - Support both a current and historic valuation via menu selection.
//
// Revision 1.1  1994/07/03  16:22:59  davidsi
// Initial revision
//
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"
#include "dval.h"
#include "security.h"
#include "menu.h"
#include "callback.h"


//
// Supported menu option on the valuation window
//
enum menu_e {
	MENU_EXIT,
	MENU_COLOUR,
	MENU_PRINT,
	MENU_EVAL,
	MENU_HISTORIC,
	};

//
// Defines for the database identifiers
//
#define VALDISP_DB_CLASS "VALDISP"

//
// apply_perc base per years
//
// Applies a percentage change to a base value for the number of
// years specified.
//
// INPUTS :	base		The base value
//		per		The percentage to apply
//		years		The number of years for which the percentage
//				is to be appplied
// RETURNS:	Amount when percentage has been applied 
///
static double
apply_perc( double base,
		double per,
		unsigned long years
		)
{
	register unsigned long iter;

	for(iter=0;iter<years;iter++)
		base += (base*per);
	
	return(base);
}



//
// find_change startamount avgdate endamount enddate
//
// Finds the percentage change per year for a change over a number
// of years.
//
// INPUTS :	startamount		The amount to start with
//		avgdate			The average date for the start amount
//		endamount		The final amount
//		enddate 		The final date
// RETURNS:	The percentage that applied for that period.
///
double
find_change( double startamount,
		class date *avgdate,
		double endamount,
		class date *enddate
		)
{
	double per=0.0;			// current percentage being considered 
	unsigned long years;		// number of years involved 
	double amount;			// final amount that current % gives  
	double low,high;		// bounds of search 

	if (startamount<SMALL_NUMBER)
		return(100.0);
	//
	// Find initial bounds of search
	//
	if (startamount>endamount) {
		high=0.0;
		low=(endamount-startamount)/startamount;
	} else {
		low=0.0;
		high=(endamount-startamount)/startamount;
	}
	//
	// Start binary search for the correct percentage
	//
	years=(enddate->daysince1970()-avgdate->daysince1970())/365;
	if (startamount==0)
		return(100.0);
	if (years==0)
		return(100.0*(endamount-startamount)/startamount);
	for(;;) {
		per = (high+low)/2.0;			// find mid point 
		amount=apply_perc(startamount,per,years);	// get amount 
		if (SMALL_NUMBER>fabs(amount-endamount)) 	// if close then
			return(100.0*per);		// Done! 
		if (amount>endamount)			// move to upper half
				high=per;
			else				// move to lower half 
				low= per;
	}
}

//
// valuation_display::MenuCallback menuid onflag
//
// This is the call back used when the user selects something on the 
// menu.  The function overrides the standard function in the
// inheritated menu class.  
//
// INPUTS :	menuid		The id for the menu item thats been selected.
//		onflag		Whether the menu item is on/off.
// RETURNS: 	None
//
void 
valuation_display::MenuCallback(unsigned long menuid, int onflag)
{

	switch (menuid) {
	case MENU_EXIT:
		if (!colourdialog.SemaphoreIslocked())
			delete this;
		return;
	case MENU_COLOUR:
		if (colourdialog.SemaphoreClock() != -1 ) {
			ColourSelect();
			colourdialog.SemaphoreUnlock();
			TextdialogRedraw();
		}
		break;
	case MENU_PRINT:
		TextdialogPrint();
		break;
	case MENU_EVAL:
		flags &= ~DVAL_FLAGS_HISTORIC;
		flags |= DVAL_FLAGS_EVAL;
		list_valuation();
		break;
	case MENU_HISTORIC:
		flags &= ~DVAL_FLAGS_EVAL;
		flags |= DVAL_FLAGS_HISTORIC;
		list_valuation();
		break;
	default:
		DialogPanic("Bad switch:valuation_display::MenuCallback");
	}
}


//
// valuation_display::redisplay this 
//
// This callback is registered to execute whenever dividends are added,
// a transaction occurs,  or prices change.  It converts the first
// argument to a class ptr.
//
// INPUTS :	tmpthis		The this ptr defined when the callback
//				was registered.
// RETURNS:	None
void
valuation_display::redisplay(char *tmpthis,char *)
{
        class valuation_display *rthis=(class valuation_display *)tmpthis;
        rthis->list_valuation();
}


//
// valuation_display::list_valuation
//
// Writes a valuation to the display
//
// INPUTS :	None
// RETURNS:	None
//
void
valuation_display::list_valuation()
{
	char str[256];			// build up space, hope this is enough	
	class Holding *ptr;
	class security *sec;
	class dividend *div;
	class dividend *head_div;	// ptr to first dividend on list
	char mode;			// boolean for writing name or not 
	double amount;			// dividend obtained for this security 
	double number;			// number of securities owned 
	double cost;			// amount spent 
	double received;		// amount received 
	double t_amount=0.0;		// total amount for all dividends 
	double t_cost=0.0;		// total amount spent/received 
	double t_value=0.0;		// total value of securities at
					// period end 
	char secname[16];		// security name 
	double price;			// price of security on(or about)
					// end_date 
	double perc;			// % growth/decline 
	class date avgdate;		// avg date of purchases 
	char *csecname;
	int i;


	if (!setmenuflag) {
        	MenuItemaction("Project","Close",MENU_EXIT);
        	MenuItemaction("Project","Print",MENU_PRINT);
		MenuItemaction("Project","Colours",MENU_COLOUR);

	        MenuItem1ofn("Options","Current",MENU_EVAL,MENUSTATE_ON);
       		MenuItem1ofn("Options","Historic",MENU_HISTORIC,MENUSTATE_OFF);
	
		for(i=1;i<8;i++) {
			class attr *attr;
			if (i==1)
				attr=new class attr(FIELDATTR_LEFT,FOREGROUND);
			else
				attr=new class attr(FIELDATTR_RIGHT,FOREGROUND);
			TextSetdefaultattr(i,attr);
		}
		TextdialogSettitle("Valuation");
		ColourSetowner(VALDISP_DB_CLASS);
		setmenuflag=1;
	}

	TextClear();

	sprintf(str,"Share Name\tNumber\tPrice\tValue\tCost\t%% Chg\tDividend");
	TextAdd(str);

        ptr=portfolio->head();
	//
	// For every known security do..
	//
        for ( ; !portfolio->eol(ptr) ; ptr=ptr->next() ) {
		mode=TRUE;		// initially write the security name
		sec=ptr->HoldingSecurity();
		sec->SecurityExtract(&csecname,NULL,NULL,NULL,NULL);
		ptr->HoldingTransactionhead()->extract(&start_date,&end_date,
							&number,&cost,
							&received,&avgdate);
		if (  (fabs(number)<SMALL_NUMBER) 
			&& (fabs(cost)<SMALL_NUMBER) 
			&& (fabs(received)<SMALL_NUMBER) )
			continue;	// security never in persons portfolio
		amount=0.0;		// dividends received=0 

		head_div=ptr->HoldingDividend();
		div=head_div;
		//
		// While there are dividends for the security, scan looking
		// for those in the correct date range.
		//
		while (div!=NULL) {
			class date ddate;
			double dnumber,damount,dtax;

			div->DividendExtract(&ddate,&dnumber,
						&damount,&dtax,NULL);
			if (ddate.inrange(&start_date,&end_date) ) 
				amount += damount;	// get dividend total
			div=div->dual_next();
			if (div==head_div)
				break;
		}
	
		price=sec->prices->PriceFetch(&end_date,TRUE);
		if (mode) {
			strncpy(secname,csecname,15);
			secname[15]='\0';
			mode=FALSE;
		} else
			secname[0]='\0';
		if (price<0) { 			// no price-show reduced info 
			if ( (number<SMALL_NUMBER) && (flags&DVAL_FLAGS_EVAL) )
				continue;
			sprintf(str,"%s\t%.4f\t%s\t%s\t%.3f\t%s\t%.3f"
				,secname,number,"-","-",cost,"-",amount);
		} else if (number<SMALL_NUMBER){ // do we no longer hold security
						// -then show gain/loss % 
			if (flags & DVAL_FLAGS_EVAL)
				continue;
			if (fabs(cost)<SMALL_NUMBER)	// possible div by zero
				perc=100.0;
			else
				perc=100.0*(received-cost)/cost;
			sprintf(str,"%s\t%.4f\t%.3f\t%.3f\t%.3f\t%.1f\t%.3f*",
				secname,number,price,price*number,
				cost,perc,amount);
    		} else {
			double total=cost-received;
			if (fabs(total)<SMALL_NUMBER)	// possible div by zero
				perc=100.0;
			else
				perc=find_change(total,&avgdate,
							price*number,&end_date);
			sprintf(str,"%s\t%.4f\t%.3f\t%.3f\t%.3f\t%.1f\t%.3f",
				secname,number,price,price*number,
				total,perc,amount);
			t_value += price*number;
		}
		TextAdd(str);
		t_amount += amount;
		t_cost += cost-received;
		DialogCursor(DCURSOR_BUSY);
	}
	if (t_cost<SMALL_NUMBER)		// if possible div by zero 
		perc=100.0;
	else
		perc=100.0*(t_value-t_cost)/t_cost;
	sprintf(str,"\t\t\t--------\t--------\t----\t--------");
	TextAdd(str);
	sprintf(str,"%s\t%s\t%s\t%.3f\t%.3f\t%.1f\t%.3f"
			,"","","",t_value,t_cost,perc,t_amount);
	TextAdd(str);
	TextComplete();

	TextdialogOutput();

	DialogCursor(DCURSOR_NORMAL);
}





//
// valuation_display::valuation_display
//
// Constructor for the valuation display class.
//
// INPUTS : None
// RETURNS: None
//
valuation_display::valuation_display()
{

	// default date period is start of unix seconds which will have
	// been set by the date constructor function
	end_date.setdate(); 		// until the date now 
	flags=DVAL_FLAGS_EVAL;		// default is current valuation
	setmenuflag=0;			// no menu yet attached

	//
	// Set up call backs, for either transaction changes,
	// price changes, dividend changes or security changes.
	//
        callback_trans=main_trans_list->add_vect.CallbackAdd(
                        &valuation_display::redisplay,
                        (char *)this);
        callback_price=main_price_list->change_vect.CallbackAdd(
                        &valuation_display::redisplay,
                        (char *)this);
        callback_div=main_div_list->add_vect.CallbackAdd(
                        &valuation_display::redisplay,
                        (char *)this);
        callback_sec=main_sec_list->changed_vect.CallbackAdd(
                        &valuation_display::redisplay,
                        (char *)this);


	list_valuation();
}

//
// valuation_display::~valuation_display
//
// Destructor for the valuation display class.
//
// INPUTS : None
// RETURNS: None
//
valuation_display::~valuation_display()
{
	//
	// Remove callbacks
	//
        main_trans_list->add_vect.CallbackRemove(callback_trans);
        main_price_list->change_vect.CallbackRemove(callback_price);
        main_div_list->add_vect.CallbackRemove(callback_div);
        main_sec_list->changed_vect.CallbackRemove(callback_sec);
}



//
// display_valuation dialog args
//
// Main entry point for creating & managing the window which displays 
// the valuation.
//
// INPUTS :	dialog	The parent dialog
//		args	To satisfy type checking
// RETURNS:	None.
///
static void
display_valuation(local_va_dcl)
{
	va_alist=va_alist;	// stop compiler warnings
        valuation_display *newtrans=new valuation_display;
}


//
// dval_init parent menu_bar
//
// Initialises the display valuations module.
// 
// INPUTS :	None
// RETURNS:	None
//
void
dval_init(menu_bar *mb)
{
	MENU("Analysis","Valuation",display_valuation,0);
}
