
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: interpret.C,v 2.4 1995/06/02 19:55:57 davidsi Exp $
//
// $Log: interpret.C,v $
// Revision 2.4  1995/06/02  19:55:57  davidsi
// - When a parse error occurs display an error message and give the
//   line number for the failure.
//
// Revision 2.3  1995/05/19  18:53:40  davidsi
// - DialogentryCreate() now requires an argument specifying total number
//   of visible columns.
// - All class definitions for macro file handling are now defined in this
//   file. A new extern function has also been added that is used by other
//   modules to load all macro files.
//
// Revision 2.2  1995/04/16  20:35:49  davidsi
// - Modified calls to printf() to use DialogPrintf().
//
// Revision 2.1  1995/03/02  21:02:33  davidsi
// - Uses Dialog class. Calls to dialog_xxx functions have been removed.
//
// Revision 2.0  1995/01/19  21:39:32  davidsi
// - Initial version.
//
//
//


#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <dial.h>

#include "localmacros.h"

#include "interpret.h"
#include "node.h"
#include "record.h"
#include "dialog.h"
#include "button.h"
#include "optionl.h"
#include "support.h"
#include "prefs.h"


#define MAX_MACROS	2048
#define MACRO_DB_CLASS	"macro"
#define MACROLOAD_DB_CLASS "load"

extern int source_line;

//
// The Interpretedit class
//
// Allows macro files to be added or removed.
//
class Interpretedit : private Button,private Optionlist {
	Dialogshell d_shell;         // top most dialog
	Dialogvertical d_vert;
	Optionlistinstance *d_names;

	Dialoghorizontal d_horiz;
	Buttoninstance *rembut;

	Dialoghorizontal d_buthoriz;
	Buttoninstance *okbut;
	Buttoninstance *cancelbut;

private:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);

	//
	// And this is the option list handler.
	//
	void OptionlistCallback(unsigned long id);

public:
	void Refreshdatabase(char **vlist,int number);
	void InterpreteditDialog();
	Interpretedit();
	virtual ~Interpretedit();
};


//
// Local variables
//
static char **macronames;



//
// Enumerated type for dialog ids
// 
enum dialogids {
	BUTTON_OK,
	BUTTON_CANCEL,
	BUTTON_REMOVE,
	OPTIONL_NAMES,
	MENU_ADD_PROG,
	};



//
// BuildMacroNames
//
// Recreates the private cache of macro filenames
//
// INPUTS : 	None
// RETURNS:	None, but the local variable macronames will be specifying
//		a new list that reflects the full list.
//
static
BuildMacroNames()
{
	int i;
	class Interpret *iptr;
	char *name;
	char *macros[MAX_MACROS];
	char *home;
	int j;

	//
	// Free them, if they already exist.
	//
	if (macronames!=NULL)
		freestringlist(macronames);


	//
	// Allocate and fill
	//
	i=0;
	home=db_prefs->DatabaseRetrieve(MACRO_DB_CLASS,
						MACROLOAD_DB_CLASS);
	while ( (home!=NULL) && (i<MAX_MACROS) ) {
		macros[i]=home;
		home=db_prefs->DatabaseRetrievenext(MACRO_DB_CLASS,
						MACROLOAD_DB_CLASS);
		i++;
	}
	macronames=(char **)malloc(sizeof(char **)*(i+1));
	for(j=0;j<i;j++) {
		macronames[j]=(char *)malloc(strlen(macros[j])+1);
		strcpy(macronames[j],macros[j]);
	}
	macronames[j]=NULL;
}


//
// InterpretlistAdd name
//
// Loads the specified program file and parses the text to form
// an executable image.  
//
// INPUTS :	name		The name of the program file
// RETURNS:	None
//
static void
InterpretlistAdd(char *name)
{
	extern FILE *yyin;
	extern int yyparse();
	extern class node_header *basenode;
	char buffer[2048];


	if (!(yyin=fopen(name,"r"))) {
		sprintf(buffer,"Cannot load the macro file %s", name);
		DialogMessage(buffer);
		return;
        }

	DialogPrintf("Loading %s\n",name);

	source_line=0;
        yyparse();
	fclose(yyin);

	if (basenode==NULL) {
		sprintf(buffer,"Cannot parse the macro file %s (line %d)", 
				name,source_line);
		DialogMessage(buffer);
		return;
	}

	DialogPrintf("Executing %s\n",name);
        basenode->execute();
}


//
// Interpretedit::Refreshdatabase namelist num
//
// Rewrites the database entries for which programs should be loaded.
// 
// INPUTS :	namelist		List of files to load
//		num			Number of files to load
// RETURNS:	None
//
void
Interpretedit::Refreshdatabase(char **vlist,int number)
{
	int i;

	//
	// Delete all old entries
	//
	db_prefs->DatabaseDelete(MACRO_DB_CLASS,MACROLOAD_DB_CLASS);
	//
	// Create new entries
	//
	for(i=0;i<number;i++)
		db_prefs->DatabaseInsertmultiple(MACRO_DB_CLASS,
						 MACROLOAD_DB_CLASS,vlist[i]);
	//
	// Rebuild internal cache
	//
	BuildMacroNames();
}

//
// Interpretedit::ButtonCallback  id flag
//
// Handles button callbacks from the Interpretedit dialog.
//
// INPUTS :	id		The id of the button causing the callback
//		flag		The state of the button.
// RETURNS:	None
//
void
Interpretedit::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char *rmmacro;			// The macro file to remove
	char **vlist;			// List of new files
	int number;			// Number of names in the vlist 
	char **mlist;			// Pointer to name in current set
					// of macro files that are loaded 
	char buffer[2048];		// build up area for errors
	int i;
	int iter2;


	switch(id) {
	case BUTTON_CANCEL:
		break;
	case BUTTON_OK:
		//
		// Get list of (possibly) new macro names
		//
		vlist=d_names->OptionlistinstanceGetlist(number);
		//
		// For each of the potentially new names, scan down
		// the current list. If there is no match, it must
		// be a new name.
		// 
		for(i=0;i<number;i++) {
			for(mlist=macronames;*mlist!=NULL;mlist++) {
				if (strcmp(vlist[i],*mlist)==0)
					break;
			}
			//
			// Check for a new macro file and add it.
			//
			if (*mlist==NULL) 
				InterpretlistAdd(vlist[i]);
		}
		Refreshdatabase(vlist,number);
		freestringlist(vlist);
		break;
	case BUTTON_REMOVE:
		rmmacro=d_names->OptionlistinstanceGet();
		sprintf(buffer,
			"Are you sure you no longer wish to load the %s macro file?",
			rmmacro);
		if ( !DialogYesno(buffer,"Yes","No") )
			return;
		vlist=d_names->OptionlistinstanceGetlist(number);
		for(i=0;i<number;i++) {	
			if (strcmp(vlist[i],rmmacro)==0) {
				free(vlist[i]);
				for(iter2=i;iter2<number;iter2++)
					vlist[iter2]=vlist[iter2+1];
				d_names->OptionlistinstanceRevise(vlist);
				Refreshdatabase(vlist,number);
				freestringlist(vlist);
				return;
			}
    		}
		freestringlist(vlist);
		break;
	}
	okbut.ButtoninstanceDestroy();
	cancelbut.ButtoninstanceDestroy();

	d_shell.DialogDestroy();

	delete this;
}

//
// Interpretedit::OptionlistCallback id
//
// Handles optionlist callbacks from the Interpretedit dialog.
//
// INPUTS :	id		The id of the button causing the callback
// RETURNS:	None
//
void
Interpretedit::OptionlistCallback(unsigned long id)
{
	int number;			// number of items already in list
	int iter;			// general interation
	char *newportfolio;		// string containing new alias name
	char **vlist;
	char **newlist;			// completed list
	int iter2;

	if (id!=OPTIONL_NAMES)
		DialogPanic("Interpret::OptionlistCallback - unexpected id");

	newportfolio=d_names->OptionlistinstanceGet();
	vlist=d_names->OptionlistinstanceGetlist(number);

	for(iter=0;iter<number;iter++) {
		//
                // If the portfolio already exists then ignore - the user 
                // must be selecting it..
                //
                if (strcmp(vlist[iter],newportfolio)==0) {
			freestringlist(vlist);
                        return;
		}
        }

	//
	// Update the internal data and the dialog
	//
	newlist=(char **)malloc(sizeof(char **)*(number+2));
	memcpy(newlist,vlist,sizeof(char **)*number);
	newlist[number]=(char *)malloc(strlen(newportfolio)+1);
	strcpy(newlist[number],newportfolio);
	newlist[number+1]=NULL;
	d_names->OptionlistinstanceRevise(newlist);
	freestringlist(newlist);
	freestringlist(vlist);

}




//
// Interpretedit::InterpretAdd 
//
// Allows a interpert details to be set up.
//
// INPUTS :	None
// RETURNS:	None
//
void 
Interpretedit::InterpreteditDialog()
{
	static Newoptionlist nolist = {
		20,			// width
		5,			// height
		NULL,			// list of items
		"Current Files",	// name of list
		"",			// name of item
		0,			// flags
		NULL,			// default font
	};

	d_shell.DialogshellCreate("Macro File Management");
	d_vert.DialogverticalCreate(d_shell);

	BuildMacroNames();
	nolist.listitem=macronames;
	d_names=OptionlistCreate(d_vert,nolist,OPTIONL_NAMES);

	//
	// Removal button
	//
	d_horiz.DialoghorizontalCreate(d_vert);
        rembut=ButtonPush(d_horiz,"REMOVE",BUTTON_REMOVE);

	//
	// Ok and Cancel buttons
	//
	d_buthoriz.DialoghorizontalCreate(d_vert);
        okbut=ButtonPush(d_buthoriz,"OK",BUTTON_OK);
        cancelbut=ButtonPush(d_buthoriz,"CANCEL",BUTTON_CANCEL);

	d_shell.DialogRealize();
}

//
// Interpretedit
//
// Constructor for the Interpretedit class
//
// INPUTS :	None
// RETURNS:	None
//
Interpretedit::Interpretedit()
{
}

//
// ~Interpretedit
//
// Destructor for the Interpretedit clas
//
// INPUTS :	None
// RETURNS:	None
//
Interpretedit::~Interpretedit()
{
}







//
/////////////////////////////////////////////////////////////////////////////
//
// ALL THE REMAINING FUNCTIONS ARE GLOBAL TO THE PROGRAM
//
/////////////////////////////////////////////////////////////////////////////
//


//
// InterpretLoad
//
// Load all programs specified in the preferences database.
//
// INPUTS :	None
// RETURNS:	None
//
void
InterpretLoad()
{
	char **macro;

	BuildMacroNames();
	for(macro=macronames;*macro!=NULL;macro++)
		InterpretlistAdd(*macro);
}


//
// menu dialog dtype args
//
// C wrapper for callback from menu. Called to create a new interpret
//
// INPUTS :	dialog		dialog of parent (satifies type checking)
//		dtype		type of div to create (DIV_TYPE_XX)
// RETURNS:	None
//
static void
menu(local_va_dcl)
{
	va_list ap;
	Interpretedit *newinterpret=new Interpretedit;
	dialog d;
	int dtype;

	local_va_start(ap);
        d=va_arg(ap,dialog);
        dtype=va_arg(ap,int);

        newinterpret->InterpreteditDialog();
}



//
// interpret_init menu_bar
//
// Initialises the interpret module.  This includes being responsible for
// setting up the decode callback for the interpret file.
// 
// INPUTS :	menu_bar	A handle (dialog) on the main menu bar
// RETURNS:	None
///
void
interpret_init(menu_bar *mb)
{
	MENU("Project","Macro files",menu,MENU_ADD_PROG);

	macronames=NULL;
}
