
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: list.C,v 2.2 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: list.C,v $
// Revision 2.2  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.1  1995/02/27  22:01:58  davidsi
// - Removed set_compare() and made compare() a virtual function that
//   can be overidden in derived classes.  The default compare()
//   function member inserts into the list with minimal list traversal.
// - BUG FIX.  A printf() was inadvertently checked in for version 2.0.
//
// Revision 2.0  1995/01/30  21:06:19  davidsi
// - Correct the list destructor so that it now traverses the list deleting
//   all remaining nodes.
//
// Revision 1.4  1994/10/17  20:36:04  davidsi
// - Enhanced the list class to allow for ordered lists.  A node can be
//   inserted in order onto the list using insert_order().  The comparison
//   is set by set_compare().
//
// Revision 1.3  1994/06/28  20:14:11  davidsi
// - BUG FIX.  An exmpty list is when the list ptr is equal to the head
//   node. The forward pointer will equal the back pointer when 1 item
//   is on the list as well when this list is empty.
//
// Revision 1.2  1994/06/27  21:49:27  davidsi
// - BUG FIX.  The list::eol() function must check for being passed a NULL
//   pointer because both the head() and tail() functions return NULL when
//   there are no nodes on the list.
//
// Revision 1.1  1994/06/27  20:55:27  davidsi
// Initial revision
//
//

#include <stdio.h>
#include <dial.h>

#include "list.h"


//
//=============================================================================
//
// List node handling functions
//
//=============================================================================
//


//
// next
//
// Returns the next node on the linked list.  Note that the linked
// lists are circular so the node that follows the tail node is the head
// node.
// 
//
// INPUTS :	None
// RETURNS:	The next node on the linked list
//
list_node *
list_node::next() 
{
	return(forw);
}

//
// previous
//
// Returns the node that is one in front of the current node.  The
// previous node of the head node is the tail node.
//
// INPUTS :	None
// RETURNS:	The previous node on the linked list
//
list_node *
list_node::previous() 
{
	return(back);
}

//
// remove
//
// Returns the node from the linked list.
//
// INPUTS :	None
// RETURNS:	None
//
void
list_node::remove()
{
        back->forw=forw;
	forw->back=back;
	list_head->count--;
}


//
// add newnode
//
// Adds the specified node to the list after the current node.
//
// INPUTS :	newnode		The node to add after the current node
// RETURNS:	None
//
void
list_node::add(list_node *newnode)
{
	newnode->forw=forw;
	forw->back=newnode;
	forw=newnode;
	newnode->back=this;
	list_head->count++;
}

//
// list_node
//
// Constructor function for list nodes
//
// INPUTS :	None
// RETURNS:	None
//
list_node::list_node()
{
	forw=NULL;
	back=NULL;
}

//
// ~list_node
//
// Destructor function for list nodes
//
// INPUTS :	None
// RETURNS:	None
//
list_node::~list_node()
{
}




//
//=============================================================================
//
// List handling functions
//
//=============================================================================
//




//
// eol list_item
//
// Determines where a node is at the end of a linked list
//
// INPUTS :	list_item	Node to be checked
// RETURNS:	TRUE		Node is at end of list
//		FALSE		Node is not at end of list
//
int 
list::eol(list_node *list_item)
{
	if ( list_item==NULL )
		return(TRUE);
	return (list_item==this);
}

//
// head
//
// Returns the head node of the linked list.
//
// INPUTS :	None
// RETURNS:	The node that is at the head of the linked list, or
//		NULL if there is no head.
//
list_node *
list::head() 
{
	if (forw==(list_node *)this)		// Nothing on the list
		return(NULL);
	return(forw);
}

//
// tail
//
// Returns the tail node of the linked list.
//
// INPUTS :	None
// RETURNS:	The node that is at the tail of the linked list, or
//		NULL if there is no tail.
list_node *
list::tail() 
{
	if (forw==(list_node *)this)		// Nothing on the list
		return(NULL);
	return(back);
}

//
// insert_head node
//
// Inserts a node at the the head of the linked list.
//
// INPUTS :	node	The node to insert
// RETURNS:	None
//
void
list::insert_head(list_node *node)
{
	node->list_head=this;
	add(node);
}


//
// insert_tail node
//
// Inserts a node at the the tail of the linked list.
//
// INPUTS :	node	The node to insert
// RETURNS:	None
//
void
list::insert_tail(list_node *node)
{
	node->list_head=this;
	this->back->add(node);
}



//
// compare listnode listnode
//
// Always returns the value which is most likely to ensure an element
// can be inserted into a list with minimal processing.  Typically
// this function is overriden by derived classes that wish to
// control the placement of nodes within the list.
//  
int
list::compare(class list_node *,class list_node *)
{
	//
	// Note that the default is to insert ASAP onto the list
	//
#ifdef SEARCH_FWD
	return(1);
#else
	return(-1);
#endif
}




//
// insert_order node
//
// Inserts a node in a sorted position within the list.  The sort
// position is determine the compare function that is defined for
// the list
//
// INPUTS :	node	The node to insert
// RETURNS:	None
//
void
list::insert_order(list_node *node)
{
list_node *ptr;
list_node *lptr;


	//
	// Point node to head
	//
	node->list_head=this;

#ifdef SEARCH_FWD
	ptr=head();
	lptr=NULL;
	while (!eol(ptr)) {
		if (compare(ptr,node)>=0)
			break;
		lptr=ptr;
		ptr=ptr->next();
	}
	if (lptr==NULL) 
		insert_head(node);
	else
		lptr->add(node);
#else
	/*
	 * It will generally be faster to search backwards.  This is because
	 * the files are written in a sorted order so when data is loaded
	 * most typically it will belong at the end of the list
	 */
	ptr=tail();
	while (!eol(ptr)) {
		if (compare(ptr,node)<=0)
			break;
		ptr=ptr->previous();
	}
	if (ptr==NULL) 
		insert_head(node);
	else
		ptr->add(node);
#endif
}




//
// length
//
// Returns the length of a linked list
//
// INPUTS :	None
// RETURNS:	The length of the list
//
int
list::length()
{
	return(count);
}


//
// list
//
// Constructor function for lists
//
// INPUTS :	None
// RETURNS:	None
//
list::list()
{
	forw=back=this;
	list_head=(list *)this;
	count=0;
}

//
// ~list
//
// Destructor function for lists
//
// INPUTS :	None
// RETURNS:	None
//
list::~list()
{
list_node *ptr;

	ptr=head();
	while  ( !eol(ptr) ) {
		ptr->remove();
		delete ptr;
		ptr=head();
	}
}
