
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: node.C,v 2.4 1995/06/17 20:47:12 davidsi Exp $
//
// $Log: node.C,v $
// Revision 2.4  1995/06/17  20:47:12  davidsi
// - The new node_command class provides a class that is able to
//   execute all command type nodes without shifting execution from
//   one function to another.
// - Teach the node_assign() class to handle the new value_char class since
//   it should be possible to assign characters to ints and vice-versa.
// - The node_expr::evaluate() function now only evaluates left/right
//   subtrees if they exist - the right tree might not if it is a unary
//   operator.
//
// Revision 2.3  1995/05/29  19:10:48  davidsi
// - Added sourceline member which stores the line number in the source
//   where the node was parsed from. NodeGetline() allows the line number
//   to be read.
// - BUG FIX. node_call::evaluate() needs to check that the evaluation of
//   the function address did result in a value.
//
// Revision 2.2  1995/04/25  21:15:09  davidsi
// - Updated error handling in assign::evaluate() so that the names of
//   the mismatched types are displayed.
//
// Revision 2.1  1995/04/03  20:40:28  davidsi
// - Hold symbols as symbol pntrs rather than names that have to be looked
//   up. Both the node_vardecl and node_var have been modified to achieve
//   this - should be a performance win.
//
// Revision 2.0  1995/01/31  22:05:18  davidsi
// - First version.
//
//
//


#include "stdio.h"
#include "stdlib.h"

#include "node.h"

#include "error.h"
#include "dialog.h"

#define DEBUGTRACE

#ifdef DEBUGTRACE
#define TRACEOP(x) x
#else
#define TRACEOP(x)
#endif

extern void builtin(class context *cont);
extern class node_header *basenode;
extern int source_line;



//
//////////////////////////////////////////////////////////////////////////////
//
// node - The generic node class
//
// A generic node cannot be executed, evaluated or have an lvalue so
// set up simple error cases for the times when derived classes don't
// override the functions.
//
//////////////////////////////////////////////////////////////////////////////
// 

char *
node::nodename()
{
	return("base node");
} 

class value *
node::evaluate(class context *)
{
	cerror << this;	
	cerror << "Evaluate executed on "<<this->nodename()<<" but no rule\n";
	return( NULL );
}

node::lvalue(class context *,char *&,class value *&)
{
	cerror << this;	
	cerror << "lvalue executed on "<<this->nodename()<<" but no rule\n";
	return( -1 );
}

node::resolve(class context *)
{
	cerror << this;	
	cerror << "resolve executed on "<<this->nodename()<<" but no rule\n";
	return( 0 );
}

class node_header *
node::base()
{
	if (!basenode) {
		cerror << this;	
		cerror << "basenode not yet set\n";
	}
	return(basenode);
}

node::node()
{
	sourceline=source_line;
	basenode=NULL;
}

node::~node()
{
}




//
//////////////////////////////////////////////////////////////////////////////
//
// The node_declaration class provides the basic functionality needed
// by all classes that have an element of "declaration".
//
//////////////////////////////////////////////////////////////////////////////
//
node_declaration::node_declaration()
{
	//
	// Default is no follow on declaration
	//
	next=NULL;
}

node_declaration::~node_declaration()
{
}

//
// node_declaration::setnext nextdec
//
// Allows a follow on declation to be specified, forming a list of
// declarations.
//
// INPUTS :	nextdec		The next declaration in the declaration
//				list
// RETURNS:	None
//
void
node_declaration::setnext(class node_declaration *c_next)
{
#ifdef ASSERTS
	cinfo << "set next on "<<(address)this<<" to "<<(address)c_next<<"\n";
#endif
	next=c_next;
}


//
// node_declaration::getnext nextdec
//
// Gets the follow on declation.
//
// INPUTS :	None
// RETURNS:	The next declaration in the declaration list
//
class node_declaration *
node_declaration::getnext()
{
	return(next);
}




//
//////////////////////////////////////////////////////////////////////////////
//
// node_header - The node from which all other nodes form a tree for
// 		 a given program. 
//
//////////////////////////////////////////////////////////////////////////////
//

node_header::node_header(class node *cdecl)
{
	heap_built=0;
	decl=cdecl;
}

//
// Destructor for node_header
//
// If the program has a heap for its variables, remove it
//
node_header::~node_header()
{
	if (heap_built)
		prog_context->popheap();
}

char *
node_header::nodename()
{
	return("node_header");
}

//
// node_header::execute fname
//
// Executes the specified function within the program.
//
// INPUTS :	fname 		The name of the function to execute
// RETURNS:	None
//
int
node_header::execute(char *funcname)
{
	class node_funcdecl *fnode;
	class context *newcont;
	class value *val;

	//
	// Initialize basenode to point to this node, so that the
	// base() member can return the right node for the currently
	// executing program.
	//
	basenode=this;

	fnode=prog_context->findfunction(funcname);
	if ( !fnode) {
		printf("Cannot find main to execute\n");
		return ( -1);
	}

	newcont=fnode->getcontext();
	if (!newcont) {
		printf("The function main has no context\n");
		return ( -1);
	}

	if (!heap_built) {
		heap_built=1;
		prog_context->createheap();
	}

	newcont->createheap();

        val=fnode->evaluate(newcont,NULL,NULL);
        if ( val == NULL) {
		prog_context->popheap();
		newcont->popheap();
                return( -1 );
	}
	val->dispose();

	newcont->popheap();
	
	return(0);
}

//
// Executes the main() function within the program.
//
int
node_header::execute()
{
	return( execute("main") );
}

//
// node_header::resolve cont
//
// When a header is resolved we must resolve all nodes below it. This
// means building symbol tables, creating jump information for loops
// and compressing nodes (where possible).   
//
// INPUTS :	cont		The context in which the resolution should
//				take place.
// RETURNS:	0 if OK, -1 for error
//
int
node_header::resolve(class context *cont)
{
	//
	// Create a context for this program
	//
	prog_context=new class context(NULL,"program");

	//
	// Merge all the builtin functions into the symbol table
	//
	builtin(prog_context);

	//
	// Then resolve the declarations that make the program up.
	//
	if (decl->resolve(prog_context) == -1)
		return(-1);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// node_funcdecl - The node that specifies a function declaration. 
// Because function declarations can taken many forms there is a variety
// of constructors but essentially these get combined.
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Contructs <type> <function_name> values
//
node_funcdecl::node_funcdecl(
		class node *ctypedecl,
		char *fname
	)
{
	name=(char *)malloc(strlen(fname)+1);
	strcpy(name,fname);

	typedecl=(class node_typedecl *)ctypedecl;
	func_context=NULL;
}

//
// Constructs <vardecl> <action> values
//
node_funcdecl::node_funcdecl(
		class node *cvardecl,
		class node_command *cactions
	)
{
	vardecl=cvardecl;
	actions=cactions;
	func_context=NULL;
}




//
// Merges "simple" forms of the node_funcdecl instance to produce 
// a fully configured form that specifies return type, function name,
// variables and code body.  Note the two node_fncdecl will be deleted!
//
node_funcdecl::node_funcdecl(
		class node_funcdecl *c_typesym,
		class node_funcdecl *c_varact,
		class node *cargdecl
	)
{
	argdecl=(class node_vardecl *)cargdecl;

	vardecl=c_varact->vardecl;
	actions=c_varact->actions;
	delete c_varact;

	name=c_typesym->name;
	typedecl=c_typesym->typedecl;
	delete c_typesym;

	func_context=NULL;
}

node_funcdecl::~node_funcdecl()
{
	if (func_context)
		delete(func_context);
}

char *
node_funcdecl::nodename()
{
	return("node_funcdecl");
}

//
// node_funcdecl::evaluate cont argcont args
//
// Calls a function during the execution of a macro program. 
//
// INPUTS:	cont		The context for the exection of the function.
//		argcont		The context for the function calling this
//				function, i.e. the context where the
//				arguments can be correctly read. argcont may
//				only be NULL if args is also NULL.
//		args		The arguments
// RETURNS:	A value with the return value of the function, NULL if
//		error.
//
class value *
node_funcdecl::evaluate(class context *cont,
		class context *argcont,
		class node_arg *args
	)
{
	class node_arg *na;
	class node_vardecl *vars;
	class value *val;
	class value *val_lv;
	char *addr;
	char *addr_not_used;
	int ret;

	na=args;
	vars=argdecl;
	

	//
	// For each argument passed
	//
	while (na!=NULL) {
		//
		// If no destination for the argument, break out of the
		// loop.  The check for number of arguments matching
		// the parameters for the function is below (outside
		// this loop).
		// 
		if (vars==NULL)
			break;

		//
		// Passing by value or reference?
		//
		switch (vars->argmode()) {
		case arg_value:
			//
			// Evaluate the argument in the context
			//  of the calling function
			//
			val=na->evaluate(argcont);
			if (!val) {
				printf("Failed to compute argument\n");
				return(NULL);
			}

			if (vars->lvalue(cont,addr,val_lv)==-1) {
				printf("Argument has no lvalue\n");
				exit(1);
                               }

			if ( !val->type_equal(val_lv) ) {
				cerror << this;	
				cerror << type_mismatch <<"\n";
				val->dispose();
				return(NULL);
			}
			val->to_mem(addr);

			val->dispose();
			break;
		case arg_ref:
			if (na->lvalue(argcont,addr,val)==-1) {
				printf("Argument has no lvalue\n");
				exit(1);
			}
			//
			// NB. We only get an lvalue of the
			// argument variable so that we can
			// determine its type
			//
			if (vars->lvalue(cont,addr_not_used,val_lv)==-1) {
				printf("Argument has no lvalue\n");
				exit(1);
			}
			if ( !val->type_equal(val_lv) ) {
				cerror << this;	
				cerror << type_mismatch <<"\n";
				return(NULL);
			}
			if (vars->setaddr(cont,addr)==1) {
				printf("Variable not a ref\n");
				exit(1);
			}
			break;
		default:
			DialogPanic("node_funcdecl::evaluate-bad arg type\n");
		}
		vars=(class node_vardecl *)(vars->getnext());
		na=na->next();
	}

	//
	// Check number of arguments matched number of parameters
	//
	if (vars!=NULL || na!=NULL) {
		printf("Incorrect argument count\n");
		return( NULL );
	}

	//
	// Clear the return value
	//
	cont->returnvalue(NULL);

	//
	// Go and execute the function
	//
	if ( actions->sequence(cont) == -1)
		return( NULL );


	//
	// Check for a return value
	//
	val=cont->returnvalue();
	if (val==NULL) {
		val=new class value;
	} else {
		val->lock_2_heap(cont);
	}
#ifdef ASSERTS
	cinfo << "Function return value is " << (address)val << "\n";
#endif
	return (val);
}


//
// node_funcdecl::resolve context
//
// When a function declaration is resolved, the symbol table
// for the function must be prepared.
//
// INPUTS :	context		The context for the resolution
// RETURNS:	-1 if error, else 0
//
int
node_funcdecl::resolve(class context *cont)
{
	//
	// Resolve previous function, variable or type definition. Note
	// list is in FILO order, hence recurse before rather than
	// after processing this current entry (giving rise to FIFO).
	// Order is important because earlier symbols must be stored
	// before processing later symbols.
	//
	if (next) {
		if ( next->resolve(cont)== -1 )
			return(-1);
	}


#ifdef ASSERTS
	cinfo << "Resolving function=" << name << "\n";
#endif

	if  (!cont) {
		fprintf(stderr,"node_funcdecl::resolve context is NULL\n");
		return(-1);
	}

	//
	// Create a symbol entry for the function name in the current context 
	//
	cont->addfunction(name,this);

	//
	// Now create a (NEW) context for this function
	//
	func_context=new class context(cont,name);

	//
	// Create entries for arguments (if there are any) 
	//
	if (argdecl && (argdecl->resolve(func_context) == -1) ){
		cerror << this;	
		cerror << bad_arg_decl << "\n";
		return (-1);
	}

	//
	// Create entries for all local variables
	//
	if (vardecl && vardecl->resolve(func_context) == -1 ) {
		cerror << this;	
		cerror << bad_var_decl << "\n";
		return (-1);
	}

	//
	// Resolve statements
	//
	if (actions)
		actions=actions->resolve(func_context,NULL,NULL,NULL);
	
#ifdef ASSERTS
	cinfo <<"Completed Resolving " << name << "\n";
#endif


	return (0);
}


//
// node_funcdecl::getcontext
//
// Returns the context being used by the function
//
// INPUTS :	None
// RETURNS:	The current context for the function
//
class context *
node_funcdecl::getcontext()
{
	return(func_context);
}


//
//////////////////////////////////////////////////////////////////////////////
//
// node_vardecl - The node that specifies a variable declaration.
//
//////////////////////////////////////////////////////////////////////////////
//
node_vardecl::node_vardecl(char *fname,
		class node_typedecl *c_type,
		argmode_e am)
{
#ifdef ASSERTS
	cinfo << "creating vardecl for "<<fname<<"\n";
#endif
	
	name=(char *)malloc(strlen(fname)+1);
	strcpy(name,fname);

	type=c_type;
	amode=am;
}

node_vardecl::~node_vardecl()
{
	free(name);
}

char *
node_vardecl::nodename()
{
	return("node_vardecl");
}


//
// node_vardecl::lvalue context addr val
//
// Returns the address being used to store a variable.
//
// INPUTS :	context		The current context
// RETURNS:	addr		The address of the variable for the context
//		val		The type of the variable
//		-1 if error, else 0
//
int
node_vardecl::lvalue(class context *cont,char *&addr,class value *&val)
{
	INFO(cinfo << "Variable (in arglist) assigned\n";)

	if  (!cont) {
		fprintf(stderr,"node_vardecl::setvalue context is NULL\n");
		return(-1);
	}

	//
	// Locate address using symbol table entry
	//
        addr=cont->ContextSym2addr(psym,val);
	if (!val) {
		cerror << this;	
		cerror << "Symbol "<<name<<" not found\n";
		return(-1);
	}
	if (!addr) {
		cerror << this;	
		cerror << "Symbol "<<name<<" does not have a reference address\n";
		return(-1);
	}
	return(0);
}


//
// node_vardecl::resolve context
//
// Resolves a variable declaration. This involves resolving the
// remaining declarations and for this declaration creating a symbol
// table entry and making that entry available for run time execution.
//
// INPUTS :	context			The context for the resolution
// RETURNS:	-1 if error, else 0
int
node_vardecl::resolve(class context *cont)
{
	class value *val;

	//
	// Resolve any further function or variable declarations
	//
	if (next) {
		if ( next->resolve(cont)== -1 )
			return(-1);
	}

	INFO(cinfo << "Resolving  variable name=" <<name<< "\n";)

	if (!cont) {
		printf("node_vardecl::resolve context not defined\n");	
		exit(1);
	}

	val=type->evaluate(cont);
	if (val==NULL) {
		cerror << this;	
		cerror << type_not_defined ;
		cerror << "- variable "<<name<<" cannot be defined\n";
		return(-1);
	}
	INFO(cinfo << "Adding variable " << name << "\n";)
	cont->ContextAddvariable(name,val);
	psym=cont->ContextName2sym(name);

	return (0);
}

//
// Assigns a variable declaration a particular address
//
int
node_vardecl::setaddr(class context *cont,char *addr)
{
        if (!cont) {
                printf("node_vardecl::setaddr context not defined\n");
                exit(1);
        }

        cont->assign(name,addr);
}

//
// node_vardecl::argmode
//
// Returns the mode of the argument (ie byvalue or byreference)
//
// INPUTS :	None
// RETURNS:	The arguments mode
//
argmode_e
node_vardecl::argmode()
{
	return(amode);
}


//
//////////////////////////////////////////////////////////////////////////////
//
// The node_command class binds together all the executable commands. 
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Should not see a resolve() on a pure command class instance because
// it is only ever inherited by other classes.
//
class node_command *
node_command::resolve(class context *cont,
		class node_command *next,
		class node_command *out,
		class node_command *in)
{
	cerror << this;	
	cerror << "resolve executed on "<<this->nodename()<<" but no rule\n";
	return(NULL);
}

//
// node_command::sequence context
//
// The main command/statement execution function. Sets eip to the current
// node then starts walking nodes executing them as it finds them.
//
// * IF YOU A NEW NODE TYPE FOR COMMANDS/STATEMENTS YOU WILL NEED TO 
//   UPDATE THIS FUNCTION *
//
node_command::sequence(class context *cont)
{
	class node_command *eip;

	//
	// Grab execution address
	//
	eip=this;

	//
	// While the address is valid execute the current node
	//
	while (eip!=NULL) switch(eip->nodeid) {
	case NODE_NONE:
		cerror << "Uninitialized node" << eip << "\n" ;
		exit(1);
	case NODE_COMPOUND:
		//
		// Should never find a compound node when executing
		// a function because they are compressed out during
		// the resolution phase.
		//
		cerror << "Compound node was not resolved" << eip << "\n" ;
		exit(1);
	case NODE_WHILE: {
		class value *res;
		int result;
		class node_while *nptr=(class node_while *)eip;

		res=nptr->bexpr->evaluate(cont);
		if (res==NULL) {
			cerror << eip;	
			cerror << not_evaluated;
			return( -1 );
		}
		if (res->load(result)==-1) {
			res->dispose();
			cerror << eip;	
			cerror << not_boolean;
			return( -1 );
		}
		res->dispose();
		if (result) {
			if (!nptr->command) {
				cerror << eip;	
				cerror << empty_body;
				return(-1);
			}
			eip=nptr->command;
		} else
			eip=nptr->nextnode;
		break;
		}
	case NODE_DOWHILE: {
		class value *res;
		int result;
		class node_dowhile *nptr=(class node_dowhile *)eip;

		res=nptr->bexpr->evaluate(cont);
		if (res==NULL) {
			cerror << eip;	
			cerror << not_evaluated;
			return( -1 );
		}
		if (res->load(result)==-1) {
			res->dispose();
			cerror << eip;	
			cerror << not_boolean;
			return( -1 );
		}
		res->dispose();
		if (result) {
			if (!nptr->command) {
				cerror << eip;	
				cerror << empty_body;
				return(-1);
			}
			eip=nptr->command;
		} else
			eip=nptr->nextnode;
		break;
		}
	case NODE_IF: {
		class value *res;
		class node_if *nptr=(class node_if *)eip;
		int result;

		res=nptr->bexpr->evaluate(cont);
		if (res==NULL) {
			cerror << eip;	
			cerror << not_evaluated;
			return( -1 );
		}
		if (res->load(result)==-1) {
			res->dispose();
			cerror << eip;	
			cerror << not_boolean;
			return( -1 );
		}
		res->dispose();
		if (result) {
			if (nptr->truecommand)
				eip=nptr->truecommand;
			else
				eip=nptr->nextnode;
		} else if (nptr->falsecommand)
			eip=nptr->falsecommand;
		else
			eip=nptr->nextnode;
		}
		break;
	case NODE_EVAL: {
		class value *val;
		class node_eval *nptr=(class node_eval *)eip;

		if (nptr->expr==NULL) {
			cerror << eip;	
			cerror << no_expression <<"\n";
			return(-1);
		}
		val=nptr->expr->evaluate(cont);
		if (val	==NULL) {
			cerror << eip;	
			cerror <<"Assignment " << not_evaluated << "\n";
			return(-1);
		}
		val->dispose();
		eip=eip->nextnode;
		}
		break;
	case NODE_RETURN: 
		{
		class value *val=NULL;
		int ret;
		class node_return *nptr=(class node_return *)eip;

		if (nptr->expr) {
			val=nptr->expr->evaluate(cont);
			if (!val) {
				cerror << this;	
				cerror << not_evaluated<<" - return() failed\n";
				return (-1) ;
			}
		} 
		cont->returnvalue(val);
		eip=NULL;
		}
		break;
	case NODE_BREAK: 
	case NODE_CONTINUE: 
		//
		// nextnode gets set appropriately at resolution time
		// for both node_break and node_continue members.
		//
		eip=eip->nextnode;
		break;
	default:
		cerror << "Unknown node type " << eip->nodeid;
		cerror << eip << " " ;
		cerror << eip->nodename() << "\n" ;
		exit(1);
	}
	return(0);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// node_compound allows multiple command/statment nodes to be joined
// in a convenient manner during parsing. However, keeping these nodes 
// at run time would add overhead, so they are removed when resolving
// executable statments.
//
//////////////////////////////////////////////////////////////////////////////
//


//
// Constructor for node_compound
//
node_compound::node_compound(class node_command *caction,class node_command *cnext)
{
	action=caction;
	nextfnc=cnext;
	nodeid=NODE_COMPOUND;
}

//
// Destructor for node_compound
//
node_compound::~node_compound()
{
}

//
// Ascii node identifier
//
char *
node_compound::nodename()
{
	return("node_compound");
}


//
// node_compound::resolve context nextin nextout
//
// At resolution time, we resolve the right hand side of the
// statement tree (i.e those statements that get executed after
// those on the left hand side). This gives a (potentially new)
// follow on node for the left hand side that then gets resolved.
// Finally we delete the node_compound and make good.
// 
// INPUTS :	context		The current context
//		nextin		The next command to execute inside a loop
//		nextout		The next command to execute outside a loop
// RETURNS:	The node (tree) that should be executed.
// 
class node_command *
node_compound::resolve(class context *cont,
		class node_command *nextcmd,
		class node_command *in,
		class node_command *out)
{
	class node_command *ret;
	class node_command *next;

	next=nextfnc->resolve(cont,nextcmd,in,out);
	ret=action->resolve(cont,next,in,out);

	//
	// The compound node has served its purpose so it can be
	// deleted now.
	//
	delete this;

	return(ret);
}



//
//////////////////////////////////////////////////////////////////////////////
//
// The node_if class provides the if..then..else functionality
// 
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_if::node_if(class node *cbexpr,
		class node_command *tc,
		class node_command *fc)
{
	bexpr=cbexpr;
	truecommand=tc;
	falsecommand=fc;
	nodeid=NODE_IF;
}


//
// Destructor
//
node_if::~node_if()
{
}


//
// Ascii identification
//
char *
node_if::nodename()
{
	return("node_if");
}


//
// node_if::resolve context nextin nextout
//
// For an node_if, both the true and (optional) false branches must
// be resolved, updating the pointers to the start nodes in each
// branch.
//
// INPUTS :	context		The current context
//		nextin		The next command to execute inside a loop
//		nextout		The next command to execute outside a loop
// RETURNS:	The node (tree) that should be executed.

class node_command *
node_if::resolve(class context *cont,
		class node_command *nextcmd,
		class node_command *in,
		class node_command *out)
{
	if (truecommand)
		truecommand=truecommand->resolve(cont,nextcmd,in,out);
	if (falsecommand)
		falsecommand=falsecommand->resolve(cont,nextcmd,in,out);
	nextnode=nextcmd;
	return(this);
}



//
//////////////////////////////////////////////////////////////////////////////
//
// The node_while class 
//
//////////////////////////////////////////////////////////////////////////////
//


//
// Constructor
//
node_while::node_while(class node *cbexpr,class node_command *c)
{
	bexpr=cbexpr;
	command=c;
	nodeid=NODE_WHILE;
}

//
// Destructor
//
node_while::~node_while()
{
}

//
// Ascii identification
//
char *
node_while::nodename()
{
	return("node_while");
}


//
// node_while::resolve cont,nextcmd,out
//
// To resolve a while loop, the inside body must be resolved with the
// continue node set to this while node and the break node set to
// the nextcmd node.
//
// INPUTS :	context		The current context
//		nextin		The next command to execute inside a loop
//		nextout		The next command to execute outside a loop
// RETURNS:	The node (tree) that should be executed.
//
class node_command *
node_while::resolve(class context *cont,
		class node_command *nextcmd,
		class node_command *in,
		class node_command *out)
{
	command=command->resolve(cont,this,this,nextcmd);
	nextnode=nextcmd;
	return(this);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// The node_dowhile class. Handles do statement while (expr)
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_dowhile::node_dowhile(class node *cbexpr,class node_command *c)
{
	bexpr=cbexpr;
	command=c;
	nodeid=NODE_DOWHILE;
}

//
// Destructor
//
node_dowhile::~node_dowhile()
{
}


//
// Ascii identification
//
char *
node_dowhile::nodename()
{
	return("node_dowhile");
}

//
// node_dowhile::resolve cont,nextcmd,out
//
// To resolve a while loop, the inside body must be resolved with the
// continue node set to this while node and the break node set to
// the nextcmd node.
//
// INPUTS :	context		The current context
//		nextin		The next command to execute inside a loop
//		nextout		The next command to execute outside a loop
// RETURNS:	The node (tree) that should be executed.
//
class node_command *
node_dowhile::resolve(class context *cont,
		class node_command *nextcmd,
		class node_command *in,
		class node_command *out)
{
	//
	// Any commands that don't have follow ons need to come back
	// to this node - BUT we actually return the command node,
	// because the conditional test should only execute after	
	// the first iteration.
	// 
	command=command->resolve(cont,this,this,nextcmd);
	nextnode=nextcmd;
	return(command);
}



//
//////////////////////////////////////////////////////////////////////////////
//
// The node_return class
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_return::node_return(class node *cexpr)
{
	expr=cexpr;
	nodeid=NODE_RETURN;
}

//
// Destructor
//
node_return::~node_return()
{
}

//
// Ascii identification
//
char *
node_return::nodename()
{
	return("node_return");
}


//
// node_return::resolve context,nextin,nextout
//
// A return node has no follow on. 
//
// INPUTS :	context		The current context
//		nextin		The next command to execute inside a loop
//		nextout		The next command to execute outside a loop
// RETURNS:	The node (tree) that should be executed.
//
class node_command *
node_return::resolve(class context *cont,
		class node_command *nextcmd,
		class node_command *in,
		class node_command *out)
{
	nextnode=NULL;
	return(this);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// The node_break class 
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_break::node_break()
{
	nodeid=NODE_BREAK;
}

//
// Destructor
//
node_break::~node_break()
{
}

//
// Ascii identification
//
char *
node_break::nodename()
{
	return("node_break");
}

//
// node_break::resolve
//
// Resolving the break node requires the nextnode pointer to be set
// to the "jump out of the loop" node.
//
// INPUTS :	context		The current context
//		nextin		The next command to execute inside a loop
//		nextout		The next command to execute outside a loop
// RETURNS:	The node (tree) that should be executed.
//
class node_command *
node_break::resolve(class context *cont,
		class node_command *nextcmd,
		class node_command *in,
		class node_command *out)
{
	nextnode=out;
	return(this);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// The node_continue class 
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_continue::node_continue()
{
	nodeid=NODE_BREAK;
}

//
// Destructor
//
node_continue::~node_continue()
{
}

//
// Ascii identification
//
char *
node_continue::nodename()
{
	return("node_continue");
}

//
// node_continue::resolve cont,in,out
//
// Resolving the continue node requires the nextnode pointer to be set
// to the "start of the loop" node.
//
// INPUTS :	context		The current context
//		nextin		The next command to execute inside a loop
//		nextout		The next command to execute outside a loop
// RETURNS:	The node (tree) that should be executed.
//
class node_command *
node_continue::resolve(class context *cont,
		class node_command *nextcmd,
		class node_command *in,
		class node_command *out)
{
	nextnode=in;
	return(this);
}


//
//////////////////////////////////////////////////////////////////////////////
//
// The node_eval class provides the link between executable statements
// and expressions that require evaluation.
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_eval::node_eval(class node *c_expr)
{
	nodeid=NODE_EVAL;
	expr=c_expr;
}

//
// Destructor
//
node_eval::~node_eval()
{
}


//
// Ascii identification
//
char *
node_eval::nodename()
{
	return("node_eval");
}

//
// node_eval::resolve cont,nextcmd,out
//
//
// INPUTS :	context		The current context
//		nextin		The next command to execute inside a loop
//		nextout		The next command to execute outside a loop
// RETURNS:	The node (tree) that should be executed.
//
class node_command *
node_eval::resolve(class context *cont,
		class node_command *nextcmd,
		class node_command *in,
		class node_command *out)
{
	nextnode=nextcmd;
	return(this);
}
	

//
//////////////////////////////////////////////////////////////////////////////
//
// The node_relop class provides all the relational operators
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_relop::node_relop(operators_e op,class node *lbx,class node *rbx)
{
	operation=op;
	bexpr_left=lbx;
	bexpr_right=rbx;
}

//
// Destructor
//
node_relop::~node_relop()
{
}

//
// Ascii identification
//
char *
node_relop::nodename()
{
	return("node_relop");
}


//
// node_relop::evaluate
// 
// The bulk of the work is performed by the underlying value class(es).
// This function simply needs to extract the two operands, call the
// value class and tidyup.
//
class value *
node_relop::evaluate(class context *cont)
{
	class value *v1,*v2;
	class value *result;
	
	v1=bexpr_left->evaluate(cont);
	v2=bexpr_right->evaluate(cont);

	if (!v1 || !v2) 
		goto error;
	
	result=v1->operation(operation,v2);

	if (result!=v1)
		v1->dispose();
	if (result!=v2)
		v2->dispose();

	if (result==NULL) {
		printf("Boolean operation failed\n");
		return(NULL);
	}
	return(result);
error:
	if (v1)
		v1->dispose();
	if (v2)
		v2->dispose();
	return(NULL);
}



//
//////////////////////////////////////////////////////////////////////////////
//
// The node_expr class 
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_expr::node_expr(operators_e op,class node *lexpr,class node *rexpr)
{
	operation=op;
	expr_left=lexpr;
	expr_right=rexpr;
}

//
// Destructor
//
node_expr::~node_expr()
{
}

//
// Ascii identification
//
char *
node_expr::nodename()
{
	return("node_expr");
}


//
// node_expr::evaluate
//
// The bulk of the work is performed by the underlying value class(es).
// This function simply needs to extract the two operands, call the
// value class and tidyup.
//
class value *
node_expr::evaluate(class context *cont)
{
	class value *v1=NULL;
	class value *v2=NULL;
	class value *result;
	
	if (expr_left) {
		v1=expr_left->evaluate(cont);
		if (!v1)
			goto error;
	}
	if (expr_right) {
		v2=expr_right->evaluate(cont);
		if (!v2)
			goto error;
	}

	result=v1->operation(operation,v2);

	//
	// Throw away the operands
	//
	if (v1 && v1!=result)
		v1->dispose();
	if (v2 && v2!=result)
		v2->dispose();

	if (result==NULL) {
		printf("Arithmetic operation failed\n");
		return(NULL);
	}
	return(result);
error:
	if (v1)
		v1->dispose();
	if (v2)
		v1->dispose();
	return(NULL);
}



//
//////////////////////////////////////////////////////////////////////////////
//
// The node_int class provides for inline constants 
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_int::node_int(int num)
{
	number=num;
	val = new class value_int;
	val->assign_storage((char *)&number);
	//
	// Ensure the underlying value is "fundamental" so that
	// any dispose calls are ignored.
	//
	val->setstatus(fundamental);
}

//
// Destructor
//
node_int::~node_int()
{
	val->setstatus(temporary);
	val->dispose();
}

//
// Ascii identification
//
char *
node_int::nodename()
{
	return("node_int");
}

//
// node_int::evaluate
//
// Simply return the already created value for the constant.
//
class value *
node_int::evaluate(class context *cont)
{
	INFO(cinfo << "Expression evaluated (number=" << number << ")\n";)
	return(val);
}

//
//////////////////////////////////////////////////////////////////////////////
// 
// The node_real class provides for inline floating point constants
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_real::node_real(double num)
{
	number=num;
	val = new class value_real;
	val->assign_storage((char *)&number);
	//
	// Ensure the underlying value is "fundamental" so that
	// any dispose calls are ignored.
	//
	val->setstatus(fundamental);
}

//
// Destructor
//
node_real::~node_real()
{
	val->setstatus(temporary);
	val->dispose();
}

//
// Ascii identification
//
char *
node_real::nodename()
{
	return("node_real");
}

//
// node_real::evaluate
//
// Simply return the already created value for the constant.
class value *
node_real::evaluate(class context *cont)
{
	INFO(cinfo << "Real expression evaluated (number=" << number << ")\n";)
	return(val);
}


//
//////////////////////////////////////////////////////////////////////////////
// 
// The node_string class provides for inline string constants
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_string::node_string(char *s)
{
	INFO(cinfo << "String constant created ("<<s<<") evaluated\n";)
	str=(char *)malloc(strlen(s)+1);
	strcpy(str,s);

	val=new class value_string;
	val->assign_storage(str);
	//
	// Ensure the underlying value is "fundamental" so that
	// any dispose calls are ignored.
	//
	val->setstatus(fundamental);
}

//
// Destructor
//
node_string::~node_string()
{
	val->setstatus(temporary);
	val->dispose();
}

//
// Ascii identification
//
char *
node_string::nodename()
{
	return("node_string");
}


//
// node_string::evaluate
//
// Simply return the already created value for the constant.
//
class value *
node_string::evaluate(class context *cont)
{
	INFO(cinfo << "String constant ("<<str<<") evaluated\n";)
	return(val);
}




//
//////////////////////////////////////////////////////////////////////////////
//
// The node_call provides for function calls.
// 
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_call::node_call(class node *nptr,class node *argl)
{
	funcaddr=(class node_expr *)nptr;
	arglist=(class node_arg *)argl;
}

//
// Destructor
//
node_call::~node_call()
{
}


//
// Ascii identification
//
char *
node_call::nodename()
{
	return("node_call");
}


//
// node_call::evaluate
//
// Most of the function call work is performed by the funcdecl_node. What
// does happen here is, the function is determined, a context is built
// for it and controlled is passed to that functions's funcdecl node
//
class value *
node_call::evaluate(class context *cont)
{
	class node_funcdecl *fnode;
	class context *newcont;
	char	heapcreated=0;
	class 	value *val;

	INFO(cinfo << "Executing function\n";)
	val=funcaddr->evaluate(cont);
	if (val==NULL) {
		cerror << this;	
		cerror << function_not_found << "\n";
		return( NULL );
	}
	fnode=val->funcaddr();
	if (fnode==NULL) {
		cerror << this;	
		cerror << function_not_found << "\n";
		return( NULL );
	}
	
	newcont=fnode->getcontext();
	//
	// If there is no context for the destination function, then
	// assume the function is a builtin function and will run in
	// the current context and on the current heap.
	// Otherwise create a new heap.
	//	
	if (!newcont) 
		newcont=cont;
	else {
		newcont->createheap();
		heapcreated=1;
	}

	//
	// Execute the destination function
	//
	INFO(cinfo << "Evaluation function call\n";)
       	val=fnode->evaluate(newcont,cont,arglist);

	if (heapcreated)
		newcont->popheap();

	return(val);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// The node_assign class handles assignment
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_assign::node_assign(class node *clptr,class node *cexpr)
{
	lptr=clptr;
	expr=cexpr;
}

//
// Destructor
//
node_assign::~node_assign()
{
}

//
// Ascii identification
//
char *
node_assign::nodename()
{
	return("node_assign");
}


//
// node_assign::evaluate
// 
// Computes the value of the RHS and the lvalue of the LHS and then
// assigns the value across.
//
class value *
node_assign::evaluate(class context *cont)
{
	class value *val;
	class value *lv_val;
	char *addr=NULL;

	INFO(cinfo << "Executing assignment (evaluating)\n";)
	val=expr->evaluate(cont);
	if ( val==NULL )
		return( NULL );

	INFO(cinfo << "Executing assignment (assigning)\n";)

	if (lptr->lvalue(cont,addr,lv_val)==-1) {
		printf("Assignment has no lvalue\n");
		return( NULL );
	}
	INFO(cinfo << "Assignment address is " << (address)addr << "\n";)

	//
	// If the types match we can do a copy to copy.  If they don't
	// we are prepared to upgrade/downgrade ints/reals as appropriate
	//
	if (val->type_equal(lv_val))
		val->to_mem(addr);
	else if (lv_val->id()==VALUE_INT && val->id()==VALUE_REAL) {
		class value *newval=lv_val->clone();
		int res;
		//
		// To downgrade the real type, we clone a copy of
		// the lvalue, assign the cloned copy the correct
		// address, and then store the integer value.
		//
		newval->assign_storage(addr);
		val->load(res);
		newval->store(res);
		newval->dispose();
	} else if (lv_val->id()==VALUE_REAL && val->id()==VALUE_INT) {
		class value *newval=lv_val->clone();
		double res;

		newval->assign_storage(addr);
		val->load(res);
		newval->store(res);
		newval->dispose();
	} else if ( (lv_val->id()==VALUE_CHAR && val->id()==VALUE_INT) 
		  || (lv_val->id()==VALUE_INT && val->id()==VALUE_CHAR) ) {
		int arg;
		class value *newval=lv_val->clone();

		val->load(arg);
		//
		// Could add error checking to ensure the value
		// is in a sensible range for a character
		//
		newval->assign_storage(addr);
		newval->store(arg);
		newval->dispose();
	} else {
		cerror << this;	
		cerror << "Assignment failed:" << type_mismatch << "\n";
		cerror << "lvalue = " << lv_val->name() << " and ";
		cerror << "value = " << val->name() << "\n";
		INFO(cinfo << "type1="<< lv_val->id ()<<"type2="<< val->id() <<"\n";)
	}

	return(val);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// The node_arg class is a container class for multiple arguments
// 
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_arg::node_arg(class node *arg,class node *cnext)
{
	expr=arg;
	nextarg=(class node_arg *)cnext;
}

//
// Destructor
//
node_arg::~node_arg()
{
}

//
// Ascii identification
//
char *
node_arg::nodename()
{
	return("node_arg");
}


//
// node_arg::evaluate
//
// Computes the value of the argument
//
class value *
node_arg::evaluate(class context *cont)
{
	return ( expr->evaluate(cont) );
}

//
// node_arg::lvalue
//
// Computes the lvalue of the argument
//
int
node_arg::lvalue( class context *cont,char *&addr,class value *&val)
{
	return ( expr->lvalue(cont,addr,val) );
}


//
// Returns the next argument in the class
//
class node_arg *
node_arg::next()
{
	return(nextarg);
}




//
//////////////////////////////////////////////////////////////////////////////
//
// The node_var class is responsible for all variable references within
// expressions 
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_var::node_var( char *s)
{
        name=(char *)malloc(strlen(s)+1);
        strcpy(name,s);

	//
	// resolve() does not normally get called - we are so deep
	// down in the parse tree, so set psym to NULL and explicitly
	// call resolve() if we hit a NULL psym
	//
	psym=NULL;
}

//
// Destructor
//
node_var::~node_var()
{
}

//
// Ascii identification
//
char *
node_var::nodename()
{
	return("node_var");
}


//
// node_var::resolve
//
// Maps the variable to a symbol table entry. Normally will be called
// the first time a variable is rereferenced.
//
int
node_var::resolve(class context *cont)
{
	class value *val;

	INFO(cinfo << "Resolving variable name=" <<name<< "\n";)

	if (!cont) {
		printf("node_var::resolve context not defined\n");	
		exit(1);
	}
	psym=cont->ContextName2sym(name);
	return (0);
}

//
// Returns the value of the variable
//
class value *
node_var::evaluate(class context *cont)
{
	char *addr;
	class value *val;

	if (psym==NULL)
		resolve(cont);
	

	addr=cont->ContextSym2addr(psym,val);
	if (!val) {
		cerror << this;	
		cerror << "Symbol "<<name<<" not found\n";
		return(NULL);
	}
	//
	// There will be no addr if the symbol was actually a function
	// (NB not a variable storing a function but an actual function).
	//
	if (addr)
		val->assign_storage(addr);
	INFO(cinfo << "Variable " << name << " evaluated (type="<<val->id()<<")";)
	INFO(cinfo << " Address ="<<(address)addr<<"\n";)
	return(val);
}


//
// Returns the lvalue of the variable
//
int
node_var::lvalue(class context *cont,char *&addr,class value *&val)
{
	if (psym==NULL)
		resolve(cont);

	INFO(cinfo << "Variable " << name << " address determined\n";)

	if  (!cont) {
		fprintf(stderr,"node_var::lvalue context is NULL\n");
		return(-1);
	}

        addr=cont->ContextSym2addr(psym,val);
	if (!val) {
		cerror << this;	
		cerror << "Symbol "<<name<<" not found\n";
		return(-1);
	}
	if (!addr) {
		cerror << this;	
		cerror << "Symbol "<<name<<" does not have a reference address\n";
		return(-1);
	}
	return(0);
}



//
//////////////////////////////////////////////////////////////////////////////
//
// node_typedef
// 
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_typedef::node_typedef(class node_typedecl *c_basetype,char *cnewname,int csize)
{
        basetype=c_basetype;
        newname=(char *)malloc(strlen(cnewname)+1);
        strcpy(newname,cnewname);
	size=csize;
}

//
// Destructor
//
node_typedef::~node_typedef()
{
	free(newname);
}

//
// Ascii identification
//
char *
node_typedef::nodename()
{
	return("node_typedef");
}


//
// node_typedef::resolve cont
//
// To resolve a type definition, the base type must be determined and
// then a symbol inserted that aliases the specified name to that base
// type.
//
int
node_typedef::resolve(class context *cont)
{
	class value *val;
	class value_array *val_array;

	if (next) {
		if ( next->resolve(cont)== -1 )
			return(-1);
	}

	INFO(cinfo << "Resolving  typedef  newname=" << newname << "\n";)

        val=basetype->evaluate(cont);
        if (val==NULL) {
		cerror << this;	
                cerror << type_not_defined << " typedef expected" << "\n";
                return(-1);
        }
	if (size==0) {
		INFO(cinfo << "type " << newname << " is an alias\n";)
		cont->addtype(newname,val);
	} else {
		INFO(cinfo << "type " << newname << " is an array\n";)
		INFO(cinfo << "id of array type is "<<val->id()<<"\n";)
		val_array=new class value_array(val,size);
		cont->addtype(newname,val_array);
	}
		

}



//
//////////////////////////////////////////////////////////////////////////////
//
// The node_deref class 
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_deref::node_deref(class node *post,class node *ind)
{
	postfix=post;
	index=ind;
}

//
// Destructor
//
node_deref::~node_deref()
{
}

//
// Ascii identification
//
char *
node_deref::nodename()
{
	return("node_deref");
}


//
// node_deref::evaluate cont
// 
// To dereference the base array must be determined, then the index for
// the dereference. The underlying value class member is responsible
// for the physical dereferencing.
//
class value *
node_deref::evaluate(class context *cont)
{
	class value *val;
	class value *val_index;
	class value *new_val;
	int arr_index;
	int ret;

        val=postfix->evaluate(cont);
        if (val==NULL) {
		cerror << this;	
                cerror << not_evaluated;
                return(NULL);
        }

        val_index =index->evaluate(cont);
        if (val_index==NULL) {
		val->dispose();
		cerror << this;	
                cerror << not_evaluated << "\n";
                return(NULL);
        }

	if (val_index->load(arr_index) == -1) {
		val->dispose();
		val_index->dispose();
		cerror << this;	
                cerror << not_integer << "\n";
                return(NULL);
	}
	
	new_val=val->deref(arr_index);

	val->dispose();
	val_index->dispose();
	
	if (!new_val) {
		cerror << this;	
                cerror << cannot_deref;
                return(NULL);
	}

	return(new_val);
}


//
// node_deref::lvalue
//
// Dereferences a value but returns the address rather than the
// contents.
//
int
node_deref::lvalue(class context *cont,char *&addr,class value *&lval)
{
	class value *val;
	class value *val_index;
	int arr_index;

        val=postfix->evaluate(cont);
        if (val==NULL) {
		cerror << this;	
                cerror << not_evaluated;
                return(-1);
        }

        val_index =index->evaluate(cont);
        if (val_index==NULL) {
		val->dispose();
		cerror << this;	
                cerror << not_evaluated;
                return(-1);
        }

	if (val_index->load(arr_index)==-1) {
		val->dispose();
		val_index->dispose();
		cerror << this;	
                cerror << not_integer << "\n";
                return(-1);
	}
	
	lval=val->deref(arr_index,addr);

	val->dispose();
	val_index->dispose();
	
	if (!lval) {
		cerror << this;	
                cerror << cannot_deref << "\n";
                return(-1);
	}

	return(0);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// The node_typedecl class
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_typedecl::node_typedecl(typedecl_e c_type,char *typename)
{
	identifier=(char *)malloc(strlen(typename)+1);
	strcpy(identifier,typename);
	type=c_type;
	structdef=NULL;
}

//
// Destructor
//
node_typedecl::~node_typedecl()
{
}

//
// Ascii identification
//
char *
node_typedecl::nodename()
{
	return("node_typedecl");
}


//
// node_typedecl::resolve
//
// To resolve a type declaration all following declarations are resolved,
// then the base type is determined and added to the symbol table using
// the identifier specified.
//
int
node_typedecl::resolve(class context *cont)
{
	class value *val;

	//
	// Resolve any further function, variable or type declarations
	//
	if (next) {
		if ( next->resolve(cont)== -1 )
			return(-1);
	}

	INFO(cinfo << "Resolving type name=" <<identifier<< "\n";)

	if (!cont) {
		printf("node_typedecl::resolve context not defined\n");	
		exit(1);
	}

	val=evaluate(cont);
	if (val==NULL) {
		cerror << this;	
		cerror << type_not_defined ;
		cerror << "- type "<<identifier<<" cannot be defined\n";
		return(-1);
	}
	INFO(cinfo << "Adding type " << identifier << "\n";)
	cont->addtype(identifier,val);

	return (0);
}


//
// node_typedecl::setstructure
//
// Sets the structure definition to be used by the type declaration
//
void
node_typedecl::setstructure(class node_structdefn *c_struct)
{
	structdef=c_struct;
}

//
// node_typedecl::evaluate
//
// Although this returns value class DO NOT use the values
// returned for variable info.  They are only valud for
// for type recognition.
//
class value *
node_typedecl::evaluate(class context *cont)
{
	class value *val;
	
	switch (type) {
	case type_symbol: val=cont->findtype(identifier);
		if (val==NULL) {
			cerror << this;	
                               cerror << type_not_defined << "(";
                               cerror << identifier << ")\n";
		return(NULL);
		}
		break;
	case type_struct_defn:
		val=structdef->evaluate(cont);
		if (!val) {
			cerror << this;	
                               cerror << bad_type << "\n";
			return(NULL);
		}
		//
		// If a type name is given then this
		// is truly defining a new type
		//
		if (identifier) 
			cont->addstructure(identifier,val);

		break;
	case type_struct:val=cont->findstructure(identifier);
		if (val==NULL) {
			cerror << this;	
                               cerror << type_not_defined << "(struct ";
                               cerror << identifier << ")\n";
			return(NULL);
		}
		break;
	default:
		cerror << this;	
		cerror << "Complex types not yet supported!\n";
		val=NULL;
	}
	return (val);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// node_structdefn defines structures
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_structdefn::node_structdefn(char *c_fieldname,
		class node_typedecl *c_type
	)
{
	fieldname=(char *)malloc(strlen(c_fieldname)+1);
	strcpy(fieldname,c_fieldname);
	type=c_type;
}

//
// Destructor
//
node_structdefn::~node_structdefn()
{
}

//
// Ascii identification
//
char *
node_structdefn::nodename()
{
	return("node_structdefn");
}


//
// node_structdefn::evaluate
//
// When a structure is "evaluated"  the underlying structure description
// is constructed.
//
class value *
node_structdefn::evaluate(class context *cont)
{
	class value_struct *val=new class value_struct;
	
	build_struct(cont,val);

	return(val);
}

//
// node_structdefn::build_struct
//
// Builds structure descriptions
//
void
node_structdefn::build_struct(class context *cont,class value_struct *val)
{
	class value *val_type;

	if (next)
		((class node_structdefn *)next)->build_struct(cont,val);	

	val_type=type->evaluate(cont);
	if (val_type==NULL) {
		cerror << this;	
		cerror << "Cannot build type\n";
		return;
	}
	INFO(cinfo << "Adding field " << fieldname << "\n";)
	val->addfield(fieldname,val_type);
}

//
//////////////////////////////////////////////////////////////////////////////
//
// The node_field class describes fields within structures
//
//////////////////////////////////////////////////////////////////////////////
//

//
// Constructor
//
node_field::node_field(class node *c_expr,char *c_fieldname)
{
	fieldname=(char *)malloc(strlen(c_fieldname)+1);
	strcpy(fieldname,c_fieldname);
	expr=c_expr;
}

//
// Destructor
//
node_field::~node_field()
{
}

//
// Ascii identification
//
char *
node_field::nodename()
{
	return("node_field");
}

//
// node_field::lvalue
//
// The lvalue of a field is the address of the structure plus the
// offset into the structure of the required field.
//
int
node_field::lvalue(class context *cont,char *&addr,class value *&lval)
{
        class value *val;

        val=expr->evaluate(cont);
        if (!val) {
		cerror << this;	
                cerror << "field has no expression\n";
                return(-1);
        }
        lval=val->lookup(fieldname,addr);
	val->dispose();

	if (!lval) 
		return(-1);

	return(0);
}


//
// node_field::evaluate
//
// The value of a field is the address of the structure plus the
// offset into the structure of the required field dereferenced.
//
class value *
node_field::evaluate(class context *cont)
{
	class value *val;
	class value *ret_val;

	val=expr->evaluate(cont);
	if (!val) {
		cerror << this;	
		cerror << "field has no expression\n";
		return(NULL);
	}
	ret_val=val->lookup(fieldname);
	val->dispose();

	if (!ret_val) {
		cerror << this;	
		cerror << "could not find field\n";
		return(NULL);
	}
	return(ret_val);
}
