
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: portfolio.C,v 2.2 1995/06/07 20:44:54 davidsi Exp $
//
// $Log: portfolio.C,v $
// Revision 2.2  1995/06/07  20:44:54  davidsi
// - Added warning dialogs when a portfolios that should exist cannot be found.
// - Added support for creating the main xsm directory (~/.xsm) when one cannot
//   be found.
// - More support for handling multiple portfolios. Added the functions
//   portfolio_checkremove() and portfolio_remove() to handle the removal
//   of a security - they scan through all known portfolios.  Also added
//   portfolio_rename() to handle security renaming across all portfolios.
// - BUG FIX. Before portfolio_name2path() can set a path on the securities
//   list, it must check it exists.  When starting xsm without any config
//   the list won't exist at the point when the first portfolio is created.
//
// Revision 2.1  1995/05/29  19:06:47  davidsi
// - Move the name of the porfolio down - sometimes Motif clobbers the top
//   most pixels of the name.
// - BUG FIX. In portfolio_reload() don't try to display a portfolio's name
//   if one has not been set.
//
// Revision 2.0  1995/05/19  18:56:20  davidsi
// - Initial version
//
//
//


#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <unistd.h>
#include <dial.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "localmacros.h"

#include "dialog.h"
#include "button.h"
#include "optionl.h"
#include "security.h"
#include "div.h"
#include "trans.h"
#include "portfolio.h"
#include "prefs.h"
#include "val.h"
#include "support.h"


#define SMWB_DIRECTORY 		".xsm"	// name of directory for all portfolio storage
#define MAX_DEFAULT_SIZE	1024	// max string size for port. names
#define DIRLEN			2048	// size for directory paths
//
// Defines for the database identifiers
//
#define PORTFOLIO_DB_CLASS "PORTFOLIO"		// main colour key
#define DEFAULT_DB_CLASS "DEFAULT"		// default
#define DEFINED_DB_CLASS "DEFINED"		// default

#define MAX_PORTFOLIO	128			// max number of portfolios supported

//
// Menu options
//
enum dialogids {
	BUTTON_OK,
	BUTTON_CANCEL,
	BUTTON_SELECT,
	BUTTON_DEFAULT,
	BUTTON_REMOVE,
	SET_PORTFOLIO,
	OPTIONL_NAMES,
	};


//
// The portfolio class
//
class Portfolio : private Button,public Optionlist {
	char defaultp[MAX_DEFAULT_SIZE];// name of proposed default
	char defaultset;		// TRUE when default set 
	char selectp[MAX_DEFAULT_SIZE];	// name of proposed selection
	char selectset;			// TRUE when selection set 
	Dialogshell d_shell;		// top most dialog

	Dialogvertical d_vert;
	Optionlistinstance *d_names;	// lists the names of the portfolios

	Dialoghorizontal d_horiz;
	Dialoghorizontal d_buthoriz;
	Dialoglabel d_labeldefn;
	Dialoglabel d_labeldefault;

private:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);
	
	// 
	// And this is the option list handler.
	//
	void OptionlistCallback(unsigned long id);

	void Display();
	void Refreshdatabase(char **,int);
	void Buildpath(char **,int);
	int Remove(char *rmportfolio);
public:
	void PortfolioSelect();

	Portfolio();
	~Portfolio();
};


//
// Variables
//
static char **portnames;		// all available portfolios
static char *current_portfolio;		// current portfolio name
static Semaphore portdialog;		// only one portfolio dialog is
					// allowed - this controls access
static Dialoglabel *d_portlabel=NULL;
static Dialoglabel *d_blanklabel=NULL;
Holdinglist *portfolio=NULL;



//
// portfolio_name2path path
// 
// Maps a portfolio's name to a full file path
//
// INPUTS :	pname		The name of the portfolio
// RETURNS:	The path name *CAUTION* this is held in static memory
//
static char *
portfolio_name2path(char *pname)
{
	static char path[1024];
	char *home;

	// 
	// If we got it, then use it.  Otherwise it is assumed we will
	// find the files in the current directory!
	//
	if (pname) {
		home=getenv("HOME");
		if (home==NULL)
			home="./";

		//
		// Securities are common to all portfolios
		//
		sprintf(path,"%s/%s",home,SMWB_DIRECTORY);
		if (main_sec_list)
			main_sec_list->SlistSetpath(path);

		//
		// Set the directory path for the portfolio information
		//
		sprintf(path,"%s/%s/%s",home,SMWB_DIRECTORY,pname);
		return(path);
	} 
	return(NULL);
}



//
// portfolio_reload pname
//
// Loads a specified portfolio.
//
// INPUT  : pname		The name of the portfolio to load 
// RETURNS: None
//
void
portfolio_reload(char *pname)
{
	char *path;
	class trans_list *oldtransactions;
	class div_list *olddividends;
	char loadsecurities=0;

	//
	// Get copies of the current lists - we'll delete them later
	// once we've got the new data loaded.
	//
	oldtransactions=main_trans_list;
	olddividends=main_div_list;

	if (portfolio)
		delete portfolio;

	//
	// Only load securities if we have to.
	//
	if (main_sec_list==NULL) {
		main_sec_list=new security_list;
		loadsecurities=1;
	}

	//
	// The new lists
	//
	main_div_list=new div_list;
	main_trans_list=new trans_list;
	portfolio=new class Holdinglist;


	// 
	// If we got it, then use it.  Otherwise it is assumed we will
	// find the files in the current directory!
	//
	path=portfolio_name2path(pname);
	if (path) {
		main_trans_list->SlistSetpath(path);
		main_div_list->SlistSetpath(path);
	}
	
	//
	// Load the securities
	//
	if (loadsecurities)
		main_sec_list->load();

	//
	// Load the transactions
	//
	main_trans_list->load();

	//
	// Load the dividends
	//
	main_div_list->load();

	//
	// Re-establish all consistancy by issuing the callbacks
	// on the old lists.
	//
	if (olddividends)
		main_div_list->add_vect.CallbackCopy(olddividends->add_vect);
	if (oldtransactions)
		main_trans_list->add_vect.CallbackCopy(
						oldtransactions->add_vect);

	if (oldtransactions)
		delete oldtransactions;
	if (olddividends)
		delete olddividends;

	if (main_trans_list)
		main_trans_list->add_vect.CallbackMake(0,NULL);
	if (main_div_list)
		main_div_list->add_vect.CallbackMake(0,NULL);

	//
	// Free the current portfolio name and set up the new one.
	// If no pathname could be computed, assume there is no
	// portfolio.
	//
	if (current_portfolio)
		free(current_portfolio);
	if (pname) {
		current_portfolio=(char *)malloc(strlen(pname)+1);
		strcpy(current_portfolio,pname);
	} else
		current_portfolio=NULL;

	//
	// Create valuation "security" information
	//
	val_init(path);

	//
	// Set label for displaying portfolio name on main window
	//
	if (d_portlabel==NULL) {
		d_blanklabel=new class Dialoglabel;
		d_blanklabel->DialoglabelCreate(*d_mainvert,"");
		d_portlabel=new class Dialoglabel;
		d_portlabel->DialoglabelCreate(*d_mainvert,"");
		d_portlabel->DialogRealize();
	}
	if (pname) {
		sprintf(path,"Current portfolio:%s\n",pname);
		d_portlabel->DialoglabelSet(path);
	} else
		d_portlabel->DialoglabelSet("No portfolio has been set");

}



//
// portfolio_load 
//
// Loads the names of the available portfolios.
//
// INPUT  : None
// RETURNS: None
//
void
portfolio_load()
{
	char *pname;
	char *home;
	char *ports[MAX_PORTFOLIO];
	char **newlist;
	int i,j;

	//
	// If portfolio details not yet loaded, go for them now
	//
	if (portnames==NULL) {
		i=0;
		home=db_prefs->DatabaseRetrieve(PORTFOLIO_DB_CLASS,
							DEFINED_DB_CLASS);
		while ( (home!=NULL) && (i<MAX_PORTFOLIO) ) {
			ports[i]=home;
			home=db_prefs->DatabaseRetrievenext(PORTFOLIO_DB_CLASS,
							DEFINED_DB_CLASS);
			i++;
		}
		portnames=(char **)malloc(sizeof(char **)*(i+1));
		for(j=0;j<i;j++) {
			portnames[j]=(char *)malloc(strlen(ports[j])+1);
			strcpy(portnames[j],ports[j]);
		}
		portnames[j]=NULL;
	}

	//
	// Get default portfolio name
	//
	pname=db_prefs->DatabaseRetrieve(PORTFOLIO_DB_CLASS,DEFAULT_DB_CLASS);


	//
	// If no portfolio name found, then let the user create one.
	//
	if ( pname == NULL ) {
		if (DialogYesno("Welcome to the X Share Manager (XSM)\n\n
Currently there is no portfolio information. It is\n
recommended that you create a portfolio, so as to\n
provide a named area for all your transactions. If\n
you chose not to create a portfolio, information will\n
be stored in your current working directory.\n\n
Do you wish to create a portfolio?",
			"Yes","No") ) {
			class Portfolio *newportfolio=new class Portfolio;
			newportfolio->PortfolioSelect();
			portdialog.SemaphoreLock();
			while (portdialog.SemaphoreIslocked() )
				DialogEvent();

			pname=db_prefs->DatabaseRetrieve(PORTFOLIO_DB_CLASS,
							DEFAULT_DB_CLASS);
		}
	}

	portfolio_reload(pname);
}



//
// portfolio_validate busy
//
// Validates totalized values for all portfolios
//
// INPUT  : 	busy	The busy dialog to display status in
// RETURNS: 	None
//
//
void
portfolio_validate(class Busy &busy)
{
	int i;
	char *realportfolio;

	realportfolio=current_portfolio;
	for(i=0;portnames[i]!=NULL;i++) {
		portfolio_reload(portnames[i]);
		ValuationValidate(busy);
	}
	portfolio_reload(realportfolio);
}

//
// portfolio_checkremove name
// 
// Checks the named security should be removed all portfolios
//
// INPUTS :	name		The name of the security to remove
// RETURNS:	-1 if not to be removed, else 0 
//
int 
portfolio_checkremove(char *name)
{
	char buffer[1024];
	class secdiv_list *dividends;
	class sectrans_list *transactions;
	class Holding *holding;
	int i;
	char *realportfolio;
	int returncode= -1;

	//
	// Make sure the user really wants to do this!
	//
	sprintf(buffer,"Are you sure you want to delete %s",name);
	if (DialogYesno(buffer,"Abort","Delete Security"))
		goto finish;

	realportfolio=current_portfolio;
	for(i=0;portnames[i]!=NULL;i++) {
		
		//
		// Must look at every known portfolio
		//
		portfolio_reload(portnames[i]);

		//
		// There may not be any holding information for this security,
		// in which case go and check if there are any valid prices
		// that may be about to be destroyed.
		//
		holding=portfolio->HoldinglistFind(name);
		if (holding==NULL)
			continue;

		transactions=holding->HoldingTransactionhead();
		dividends=holding->HoldingDividendhead();

		//
		// Check if there are any transactions hanging off the
		// security.  If there are the user must confirm they
 		// wish to continue.
		//
		if (transactions->dual_head()!=NULL) {
			if (DialogYesno(
			 "Valid transactions have been found for this security",
				"Abort","Delete Transactions")) 
				goto finish;
		}
		//
		// Check if there are any dividends hanging off the
		// security.
		//
	        if (dividends->dual_head()!=NULL) {
			if (DialogYesno(
			    "Valid dividends have been found for this security",
			    "Abort","Delete Dividends")) 
			goto finish;
		}
	}

	//
	// Ok to delete the security
	//
	returncode=0;
finish:
	portfolio_reload(realportfolio);
	return(returncode);
}



//
// portfolio_remove name
// 
// Removes the named security from all portfolios
//
// INPUTS :	name		The name of the security to remove
// RETURNS:	-1 if error, else 0 
//
int 
portfolio_remove(char *name)
{
	class transaction *t_ptr;	// current transaction
	class transaction *t_ptr_next;	// next transaction
	class dividend *d_ptr;		// current dividend
	class dividend *d_ptr_next;	// next dividend
	class secdiv_list *dividends;
	class sectrans_list *transactions;
	class Holding *holding;
	int i;
	char *realportfolio;

	realportfolio=current_portfolio;
	for(i=0;portnames[i]!=NULL;i++) {
		
		//
		// Must update every known portfolio
		//
		portfolio_reload(portnames[i]);

		//
		holding=portfolio->HoldinglistFind(name);
		if (holding==NULL)
			continue;

		transactions=holding->HoldingTransactionhead();
		dividends=holding->HoldingDividendhead();

		//
		// Delete transactions if they were found.  
		//
		if (transactions->dual_head()!=NULL) {
			t_ptr=NULL;
			t_ptr_next=transactions->dual_head();
			while ( (t_ptr_next != NULL) && (t_ptr_next!=t_ptr) ) {
				t_ptr=t_ptr_next;
				t_ptr_next=t_ptr->dual_next();
				t_ptr->remove();
				t_ptr->dual_remove();
				delete t_ptr;
			}
			delete transactions;
		}
		main_trans_list->save();

		//
		// Delete dividend information
		//
	        if (dividends->dual_head()!=NULL) {
			d_ptr=NULL;
			d_ptr_next=dividends->dual_head();
			while ( (d_ptr_next != NULL) && (d_ptr_next!=d_ptr) ) {
				d_ptr=d_ptr_next;
				d_ptr_next=d_ptr->dual_next();
				d_ptr->remove();
				d_ptr->dual_remove();
				delete d_ptr;
			}
			delete dividends;
		}
		main_div_list->save();
	}
	portfolio_reload(realportfolio);
	return(0);
}


//
// portfolio_rename oldname newname
// 
// Renames the named security in all portfolios
//
// INPUTS :	oldname		The old name of the security 
// 		newname		The new name of the security 
// RETURNS:	-1 if error, else 0 
//
int 
portfolio_rename(char *oldname,char *newname)
{
	char *realportfolio;
	class security *secptr;
	int i;

	//
	// Get security record
	//
	secptr=main_sec_list->find_security(newname);

	realportfolio=current_portfolio;
	for(i=0;portnames[i]!=NULL;i++) {
		
		//
		// Set security to old name, load the portfolio and
		// reset the security name to the new name.
		//
		secptr->SecuritySetname(oldname);
		portfolio_reload(portnames[i]);
		secptr->SecuritySetname(newname);

		//
		// Now save the information back
		//
		main_trans_list->save();
		main_div_list->save();
	}
	portfolio_reload(realportfolio);
	return(0);
}


//
///////////////////////////////////////////////////////////////////////////////
//
// The Portfolio class
//
///////////////////////////////////////////////////////////////////////////////
//

//
// Portfolio::Remove pname
//
// Deletes portfolios from the disk.
//
// INPUTS :	pname		The name of the portfolio to delete
// RETURNS:	-1 if error, else 0
//
int
Portfolio::Remove(char *rmportfolio)
{
	DIR *dirp;
	struct dirent *dp;
	char path[1024];
	char old_dir[DIRLEN];
	char buffer[1024];
	char *home;
	char **vlist;
	int iter;
	int number;


	//
	// Check name exists
	//
	vlist=d_names->OptionlistinstanceGetlist(number);
	for(iter=0;iter<number;iter++) 
		if (strcmp(vlist[iter],rmportfolio)==0)
			break;
	freestringlist(vlist);
	if (iter==number)
		return(-1);


	//
	// Locate the directory for this portfolio
	//
	home=getenv("HOME");
	if (home==NULL)
		home="./";
	sprintf(path,"%s/%s/%s",home,SMWB_DIRECTORY,rmportfolio);
	getcwd(old_dir,DIRLEN);

	//
	// Check it exists
	//
	if ( (chdir(path)== -1) && (errno!=ENOENT)  ) {
		sprintf(buffer,"Cannot locate directory %s",path);
		DialogMessage(buffer);
		return(-1);
	}
	
	//
	// Never existed, so no point trying to delete it...
	//
	if (errno==ENOENT)
		return(0);

	//
	// Go though the directory and delete *EVERYTHING*.  This function
	// has no knowledge about files that belong to a portfolio, any file
	// found in a portfolio's directory is assumed to belong to that 
 	// portfolio and therefore should be deleted.
	//
	dirp=opendir(".");
	if (dirp==NULL) {
		chdir(old_dir);
		sprintf(buffer,"Cannot access directory %s",path);
		DialogMessage(buffer);
		return(-1);
	}

	//
	// while there are entries
	//
	while ((dp=readdir(dirp))!=NULL) {      
		if (!strcmp(dp->d_name,".") || !strcmp(dp->d_name,".."))
			continue; 
		if (  (unlink(dp->d_name)== -1) && errno!=ENOENT ) {
			sprintf(buffer,"Cannot remove %s/%s",path,dp->d_name);
			DialogMessage(buffer);
			chdir(old_dir);
			return(-1);
		}
	}
	
	//
	// Get back to the original directory and then delete the actual directory
	// used by the portfolio
	//
	chdir(old_dir);
	if ( (rmdir(path)== -1) && (errno!=ENOENT) ) {
		sprintf(buffer,"Cannot remove %s",path);
		DialogMessage(buffer);
		return(-1);
	}
	return(0);
}


//
// Portfolio::Display
//
// Updates labels for current and default portfolio names
//
// INPUTS :	None
// RETURNS:	None
//
void
Portfolio::Display()
{
	char buffer[1024];

	sprintf(buffer,"Current:%s",selectp);
	d_labeldefn.DialoglabelSet(buffer);
	sprintf(buffer,"Default:%s",defaultp);
	d_labeldefault.DialoglabelSet(buffer);
}



//
// Portfolio::Refreshdatabase namelist number
//
// Rewrites the database accoring the to the list of portfolio names
// passed in.
//
// INPUTS :	namelist	The list of all active portfolios
//		number		Number of entries in list
// RETURNS:	None
//
void
Portfolio::Refreshdatabase(char **vlist,int number)
{
	int i;

	//
	// Delete all old entries
	//
	db_prefs->DatabaseDelete(PORTFOLIO_DB_CLASS,DEFINED_DB_CLASS);
	//
	// Create new entries
	//
	for(i=0;i<number;i++) 
		db_prefs->DatabaseInsertmultiple(PORTFOLIO_DB_CLASS,
						DEFINED_DB_CLASS,vlist[i]);
	if (portnames)
		freestringlist(portnames);
	portnames=vlist;
}




//
// Portfolio::Buildpath namelist number
//
// Ensures paths for portfolios exist and creates them if necessary
//
// INPUTS :	namelist	The list of all active portfolios
//		number		Number of entries in list
// RETURNS:	None
//
void
Portfolio::Buildpath(char **vlist,int number)
{
	int i;
	char path[1024];
	char buffer[1024];
	char *home;
	char *ppath;
	struct stat statbuf;
	class trans_list *newtrans;
	class div_list *newdivs;

	home=getenv("HOME");
	if (home==NULL)
		home="./";

	//
	// Check that the .xsm directory exists
	//
	sprintf(path,"%s/%s",home,SMWB_DIRECTORY);
rebuild:
	if ( stat(path,&statbuf) == -1) {
		if (mkdir(path,S_IEXEC|S_IREAD|S_IWRITE)==-1) {
			sprintf(buffer,"Cannot create main xsm directory (%s)",
					path);
			if (DialogYesno(buffer,"Retry","Cancel"))
				goto rebuild;
		}
	}

	//
	// Check all portfolios
	//
	for(i=0;i<number;i++) { 
		sprintf(path,"%s/%s/%s",home,SMWB_DIRECTORY,vlist[i]);
		if ( stat(path,&statbuf) != -1)
			continue;
tryagain:
		if (mkdir(path,S_IEXEC|S_IREAD|S_IWRITE)==-1) {
			sprintf(buffer,
				"Cannot create directory for %s.",vlist[i]);
			if (DialogYesno(buffer,"Retry","Cancel"))
				goto tryagain;
		}
		ppath=portfolio_name2path(vlist[i]);
		if (ppath) {
			newdivs=new div_list;
			newtrans=new trans_list;
			newdivs->SlistSetpath(path);
			newtrans->SlistSetpath(path);
			newtrans->save();
			newdivs->save();
			delete newdivs;
			delete newtrans;
		}
	}
}


//
// Portfolio::ButtonCallback  id flag
//
//
// INPUTS :
// RETURNS:	None
//
void
Portfolio::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	char *str;
	char **vlist;
	char *rmportfolio;		// name of portfolio to remove
	int number;
	int i;
	int iter2;
	dcursortype_e cursor;
	char buffer[1024];

	switch(id) {
	case BUTTON_CANCEL:
		d_shell.DialogDestroy();
		break;
	case BUTTON_OK:
		vlist=d_names->OptionlistinstanceGetlist(number);
		
		//
		// If a default portfolio has been set, find it,
		// and create the database entry
		//
		for(i=0;i<number;i++) {
			if (strcmp(vlist[i],defaultp)==0)
				break;
		}
		//
		// Default not found
		//
		if (i==number) {
			freestringlist(vlist);
			sprintf(buffer,"Cannot find the portfolio %s",defaultp);
			DialogMessage(buffer);
			return;
		}
		if (defaultset) 
			db_prefs->DatabaseInsert(PORTFOLIO_DB_CLASS,
					 DEFAULT_DB_CLASS,defaultp);

		//
		// If portfolio has been selected find it
		//
		for(i=0;i<number;i++) {
			if (strcmp(vlist[i],selectp)==0)
				break;
		}
		//
		// Not found
		//
		if (i==number) {
			sprintf(buffer,"Cannot find the portfolio %s",selectp);
			DialogMessage(buffer);
			freestringlist(vlist);
			return;
		}
		Buildpath(vlist,number);
		Refreshdatabase(vlist,number);
		
		//
		// Destroy the shell before we reload - because
		// reloading can take some time..
		//
		d_shell.DialogDestroy();

		//
		// A portfolio was selected so go to it
		//
		if (selectset) {
			cursor=DialogCursor(DCURSOR_WAIT); 
			portfolio_reload(selectp); 
			DialogCursor(cursor); 
		} 
		break; 
	case BUTTON_DEFAULT:
		strncpy(defaultp,d_names->OptionlistinstanceGet(),
				MAX_DEFAULT_SIZE-1);
		defaultset=1;
		Display();
		return;
	case BUTTON_SELECT:
		strncpy(selectp,d_names->OptionlistinstanceGet(),
				MAX_DEFAULT_SIZE-1);
		selectset=1;
		Display();
		return;
	case BUTTON_REMOVE:
		rmportfolio=d_names->OptionlistinstanceGet();

		if (Remove(rmportfolio) != 0)
			return;

		vlist=d_names->OptionlistinstanceGetlist(number);
		for(i=0;i<number;i++) {	
			if (strcmp(vlist[i],rmportfolio)==0) {
				for(iter2=i;iter2<number;iter2++)
					vlist[iter2]=vlist[iter2+1];
				d_names->OptionlistinstanceRevise(vlist);
				Refreshdatabase(vlist,number-1);
				freestringlist(vlist);
				return;
			}
    		}
		freestringlist(vlist);
		return;
	}

	delete this;
}



// 
// Portfolio::OptionlistCallback (add_alias)
//
// This is a call back used when the user enters a new alias.  It  does
// not cover using a name already in the list, since this gets dealt with
// when the user finally finishes up.  If the name is new then it is inserted
// into the list.  If it is already in the list then it is removed.
//
// INPUTS :	dialog		Dialog causing callback
// RETURNS:	None.
//
void
Portfolio::OptionlistCallback(unsigned long id)
{
	int number;			// number of items already in list
	int iter;			// general interation
	char *newportfolio;		// string containing new alias name
	char **vlist;
	char **newlist;			// completed list
	int iter2;

	if (id!=OPTIONL_NAMES)
		DialogPanic("Portfolio::OptionlistCallback - unexpected id");

	newportfolio=d_names->OptionlistinstanceGet();
	vlist=d_names->OptionlistinstanceGetlist(number);

	for(iter=0;iter<number;iter++) {
		//
                // If the portfolio already exists then ignore - the user 
                // must be selecting it..
                //
                if (strcmp(vlist[iter],newportfolio)==0) {
			freestringlist(vlist);
                        return;
		}
        }

	//
	// Update the internal data and the dialog
	//
	newlist=(char **)malloc(sizeof(char **)*(number+2));
	memcpy(newlist,vlist,sizeof(char **)*number);
	newlist[number]=(char *)malloc(strlen(newportfolio)+1);
	strcpy(newlist[number],newportfolio);
	newlist[number+1]=NULL;
	d_names->OptionlistinstanceRevise(newlist);
	freestringlist(newlist);
	freestringlist(vlist);


	//
	// If a current or default portfolio haven't been set, set them
	// now.
	//
	if (selectp[0]=='\0') {
        	strncpy(selectp,newportfolio,MAX_DEFAULT_SIZE-1);
		selectset=1;
		Display();
	}
	if (defaultp[0]=='\0') {
        	strncpy(defaultp,newportfolio,MAX_DEFAULT_SIZE-1);
		defaultset=1;
		Display();
	}

}




//
// PortfolioSelect  Constructor
// 
void 
Portfolio::PortfolioSelect()
{
	static Newoptionlist nolist = {
		20,			// width
		5,			// height
		NULL,			// list of items
		"Available Portfolios",	// name of list
		"",			// name of item
		0,			// flags
		NULL,			// default font
	};
	int iter;
	static char *fakelist=NULL;
	char *dbres;

	//
	// Set up both current and default portfolio strings
	//
	if (current_portfolio)
		strncpy(selectp,current_portfolio,MAX_DEFAULT_SIZE-1);
	else
		selectp[0]='\0';

	dbres=db_prefs->DatabaseRetrieve(PORTFOLIO_DB_CLASS,DEFAULT_DB_CLASS);
	if (dbres)
		strncpy(defaultp,dbres,MAX_DEFAULT_SIZE-1);
	else
		defaultp[0]='\0';


	//
	// Build the dialog
	//
	d_shell.DialogshellCreate("Select A Portfolio");
	d_vert.DialogverticalCreate(d_shell);

	d_labeldefn.DialoglabelCreate(d_vert,"");
	d_labeldefault.DialoglabelCreate(d_vert,"");
	Display();

	if (portnames!=NULL) 
		nolist.listitem=portnames;
	else
		nolist.listitem=&fakelist;

	d_names=OptionlistCreate(d_vert,nolist,OPTIONL_NAMES);

	d_horiz.DialoghorizontalCreate(d_vert);
        ButtonPush(d_horiz,"SELECT",BUTTON_SELECT);
        ButtonPush(d_horiz,"SET AS DEFAULT",BUTTON_DEFAULT);
        ButtonPush(d_horiz,"REMOVE",BUTTON_REMOVE);

	d_buthoriz.DialoghorizontalCreate(d_vert);
        ButtonPush(d_buthoriz,"OK",BUTTON_OK);
        ButtonPush(d_buthoriz,"CANCEL",BUTTON_CANCEL);

	d_shell.DialogSetmodal();
	d_shell.DialogRealize();
}





//
// Portfolio Destructor
//
Portfolio::Portfolio()
{
	char *defp;

	defaultset=0;
	selectset=0;

	defp=db_prefs->DatabaseRetrieve(PORTFOLIO_DB_CLASS,DEFAULT_DB_CLASS);
	if (defp)
		strncpy(defaultp,defp,MAX_DEFAULT_SIZE-1);
	else
		defaultp[0]='\0';

	if (current_portfolio)
		strncpy(selectp,current_portfolio,MAX_DEFAULT_SIZE-1);
	else
		selectp[0]='\0';
}

//
// ~Portfolio
//
// Destructor for the Portfolio class
//
// INPUTS :	None
// RETURNS:	None
//
Portfolio::~Portfolio()
{
	portdialog.SemaphoreUnlock();
}

//
///////////////////////////////////////////////////////////////////////////////
//
// The Holdinglist class
//
///////////////////////////////////////////////////////////////////////////////
//

//
// sec_compare sec1 sec2
//
// Call back for ordered list node insertions.
//
// INPUTS :	sec1,sec2	The two securities that are to be compared
// RETURNS:	-1,0,1		On basis of the two names of the securities.
//
//
Holdinglist::compare(list_node *sec1,list_node *sec2) 
{
	char *name1,*name2;

	((Holding*)sec1)->security->SecurityExtract(&name1,NULL,NULL,NULL,NULL);
	((Holding*)sec2)->security->SecurityExtract(&name2,NULL,NULL,NULL,NULL);

	return(alpha_strcmp(name1,name2));
}


//
// Holdinglist::HoldinglistFind name
//
// Locates the holding information for a named security
//
// INPUTS :	name		The name of the security
// RETURNS:	The holding information if found, else NULL
//
class Holding *
Holdinglist::HoldinglistFind(char *secname)
{
	class security *secptr;
	class Holding *ptr;

        //
        // Now search for the named security..
        //
        ptr=head();
        while (!eol(ptr)) {
		secptr=ptr->security;
                if ( !(strcmp(secname,secptr->name) ) )
                        return(ptr);
                ptr=ptr->next();
        }
	//
	// Not already on the holding list, so check for
	// existence.
	//
        secptr=main_sec_list->find_security(secname);
	if (secptr==NULL)
	        return(NULL);

	ptr=new class Holding();
	ptr->security=secptr;

	// portfolio->insert_order(ptr);
	insert_order(ptr);

	return(ptr);
}



//
// Holdinglist::HoldinglistExtract
//
// Extracts the names of all securities that are currently/have been
// held.
// 
// INPUTS :	None
// RETURNS:	A null terminated array of character strings containing
//		the stock names.
//
char **
Holdinglist::HoldinglistExtract()
{
	class Holding *ptr;
	char **nameptr;
	static char **namelist=NULL;
	int number=length();

	//if (namelist!=NULL)
	//	freestringlist(namelist);
	namelist=(char **)malloc(sizeof(char *)*(number+1));

	nameptr=namelist;

	ptr=head();
	while (!eol(ptr)) {
		*nameptr =(char *)malloc(strlen(ptr->security->name)+1);
		strcpy(*nameptr,ptr->security->name);
		nameptr++;
		ptr=ptr->next();
	}
	*nameptr = NULL;
	return(namelist);
}


//
// Holdinglist::HoldinglistSelect title
//
// Allows the user to select a security from those that are held
//
// INPUTS :	title		The name for the dialog
// RETURNS:	The holding information if selected, else NULL
//
class Holding *
Holdinglist::HoldinglistSelect(char *title)
{
	char *secname;
	char **stringlist;

	stringlist=HoldinglistExtract();

	secname=DialogOptionlist(title,"Securities:",stringlist);
	if (!secname)
		return(NULL);
	freestringlist(stringlist);
	//
	// Now search for the named security..
	//
	return( HoldinglistFind(secname) );
}


//
// Holdinglist::HoldinglistRevise sec
//
// Given a security this ensures it still is in the correct place
// on the main holding list. Essentially this function is used when
// a securities name may have changed and therefore its position on
// the list moves.  Note that the security may not even be on the list,
// and if this is the case, it won't be added.
//
// INPUTS :	sec		The security to check
// RETURNS:	NULL
//
void
Holdinglist::HoldinglistRevise(class security *dest_sec)
{
	class security *secptr;
	class Holding *ptr;

        //
        // Now search for the specified security instance, could
	// search on the basis of a name but comparing pntrs is
	// faster.
        //
        ptr=head();
        while (!eol(ptr)) {
		secptr=ptr->security;
		if (dest_sec==secptr) {
			//
			// Only want to remove from the list, don't
			// try a full remove because its going straight
			// back on the list
			//
			((class list_node *) ptr)->remove();
			insert_order(ptr);
			return;
		}
                ptr=ptr->next();
        }
}



//
// These save casts
//
class Holding *
Holdinglist::head()
{
	return( LISTREMAP(Holding,list,head,this) );
}



//
// Destructor function.
//
Holdinglist::Holdinglist()
{
}

//
// Destructor function.
//
Holdinglist::~Holdinglist()
{
}

//
///////////////////////////////////////////////////////////////////////////////
//
// The Holding class
//
///////////////////////////////////////////////////////////////////////////////
//

//
// Maps a holding to a security
//
class security *
Holding::HoldingSecurity()
{
	return(security);
}

//
// Returns the dividend information that applies to a holding
//
class secdiv_list *
Holding::HoldingDividendhead()
{
	return( dividends );
}


//
// Returns the list of dividends applying to a holding
//
class dividend *
Holding::HoldingDividend()
{
	return( dividends->dual_head() );
}


//
// Returns the transaction information that applies to a holding
//
class sectrans_list *
Holding::HoldingTransactionhead()
{
	return( transactions);
}

//
// Returns the transactions for the holding
//
class transaction *
Holding::HoldingTransaction()
{
	return( transactions->dual_head() );
}


//
// Adds a dividend to those of a holding
//
void
Holding::HoldingDividendadd(class dividend *newdiv)
{
	dividends->dual_add(newdiv);
}

//
// Adds a transaction to those of a holding
//
void
Holding::HoldingTransactionadd(class transaction *newtrans)
{
	transactions->dual_add(newtrans);
}



//
// Bumps to the next holding in the holding list
//
class Holding *
Holding::next()
{
	return( LISTREMAP(Holding,list_node,next,this) );
}

//
// Constructor function.
//
Holding::Holding()
{
        dividends=new secdiv_list;		// no dividends set
        transactions=new sectrans_list;         // no transaction set
	security=NULL;
}

//
// Destructor function.
//
Holding::~Holding()
{
}



//
///////////////////////////////////////////////////////////////////////////////
//
// ALL THE REMAINING FUNCTIONS ARE GLOBAL TO THE PROGRAM
//
///////////////////////////////////////////////////////////////////////////////
//

//
// menu dialog mtype
//
// C wrapper for callback from menu. 
//
// INPUTS :	dialog		dialog of parent 
//		dtype		menu hit
// RETURNS:	None
//
static void
menu(local_va_dcl)
{
	va_list ap;
	Portfolio *newportfolio;
	dialog d;
	int dtype;

	local_va_start(ap);
        d=va_arg(ap,dialog);
        dtype=va_arg(ap,int);

	if ( portdialog.SemaphoreClock() == -1)
		return;


	newportfolio=new Portfolio;
        newportfolio->PortfolioSelect();
}



//
// portfolio_init menu_bar
//
// Initialises the portfolio module.
// 
// INPUTS :	menu_bar	A handle (dialog) on the main menu bar
// RETURNS:	None
///
void
portfolio_init(menu_bar *mb)
{
	MENU("Project","Select portfolio",menu,SET_PORTFOLIO);
	portnames=NULL;
	current_portfolio=NULL;
}
