
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: price.C,v 2.12 1995/06/21 20:54:57 davidsi Exp $
//
// $Log: price.C,v $
// Revision 2.12  1995/06/21  20:54:57  davidsi
// - When prices cannot be loaded from the automatic input source, but
//   there were no file opening errors - display a warning rather than
//   an error.
// - BUG FIX. If no database entry can be found for the price input
//   source always assume raw rather than teletext.
//
// Revision 2.11  1995/06/17  20:37:19  davidsi
// - PriceNopathname() now remembers if it has already displayed the
//   message and ignores the problem if the user has already decided
//   to do without a properly defined price path.
//
// Revision 2.10  1995/06/07  20:44:34  davidsi
// - BUG FIX.  The validation on price input was rejecting leading spaces.
// - Tided up dialog titles.
//
// Revision 2.9  1995/06/02  20:00:39  davidsi
// - More support for the NOT_A_PRICE define and related macros.
//
// Revision 2.8  1995/05/29  19:04:45  davidsi
// - Revised to use the IS_PRICE() and NOT_PRICE() macros.
// - Bumped price file version id to SMWB_VERSION_FOUR (again!) allowing for
//   the new NOT_A_PRICE definition. From version 4, price files use this
//
// Revision 2.7  1995/05/18  20:16:51  davidsi
// - The call to ValuationValidate() is now to  portfolio_validate()
//   in price_list::validate() because all portfolio valuations must be
//   validated.
// - Added PriceClear() which removes all incore price information.  The
//   disk file is left intact.
// - Added the capability to have price files in multiple directories (as
//   needed).  By default the old method is used, but now a new path
//   can be set with the PriceAssignpath() function member.
// - The make_callback() function is now CallbackMake().
// - Added support for price loading user preferences. The user can now
//   specify teletext or raw input.
// - The price function members price_to_file() and name_to_file() have been
//   removed to make way for cleaner versions PriceConvertname() and
//   PriceFilename().  Multiple price directory support and upgrades are
//   handled by PriceCreatelocation() and PriceChecklocation() function members.
// - Use Getdefault() to locate price files so a system administrator can
//   store a system wide value in /etc/default/smwb. Display a warning if no
//   path can be found - it's highly unlikely that the user will want to
//   continue without full price availability.
//
// Revision 2.6  1995/04/25  21:10:15  davidsi
// - Added support for a busy cursor.
// - BUG FIX. Removed DialogSummaryWait() as it is better to remove the
//   dialog on the callback as that will ensure it goes away as soon as
//   the button is pressed.
// - BUG FIX. In DialogSummary(), don't count non-traded securities as
//   having been updated.
//
// Revision 2.5  1995/04/16  20:29:03  davidsi
// - In DialogSummary() only display the update status for those securities
//   that are marked as being traded.
// - BUG FIX. price_list::fetch() was using a named file and not removing it.
//   Now obtains a temporary filename and unlinks that file.
// - BUG FIX. When reading price information, one must be prepared for
//   handling the weird characters that sources such as teletext can use
//   for fractions.  Without such handling some prices just go missing.
//
// Revision 2.4  1995/04/03  20:39:45  davidsi
// - In price_list::update() maintain the PRICE_UPDATED flag so securities
//   that don't get their prices updated can be tracked. Build a list and
//   summary once the update is complete.
// - To prevent having to confirm price file creations for new securities
//   price::read_share_price() now returns an error code indicating the
//   file could not be opened, as opposed to a general error return.
//   PriceStore() updated to reflect this change.
// - BUG FIX. In price::extract(), the check for a null ptr on the price
//   array was being incorrectly dereferenced prior to the check against null.
//
// Revision 2.3  1995/03/08  21:24:38  davidsi
// - The extract() member of security was renamed to SecurityExtract(), so
//   so this module modified to reflect that.
// - Renamed price_store() to PriceStore().
// - price_list::validate() now calls out to ValuationValidate() once all
//   basic prices have been validated.
// - BUG FIX. The error message in price::PriceStore() when creating
//   a new file was not displaying the correct security name.
// - BUG FIX. The error/warning message for failed renames was not
//   being displayed by price::rename().
// - BUG FIX.  PriceStore() was not handling requests to set prices on dates
//   before the first date in a price file.
//
// Revision 2.2  1995/03/01  22:07:56  davidsi
// - Replaced all calls to dialog_yesno() to DialogYesno() which acts as a
//   C++ wrapper.
// - All calls to dialog_message() are now to DialogMessage().
// - All calls to dialog_panic() are now to DialogPanic().
// - Called to dialog_updatedisplay() are now made to DialogUpdatedisplay().
// - The function member editprice() no longer requires a dialog to be passed.
//
// Revision 2.1  1995/02/05  17:57:08  davidsi
// - Added support for the new Busy class.  A busy window is now shown while
//   prices are updated.
// - Now includes support.h.
//
// Revision 2.0  1995/01/19  21:18:52  davidsi
// - All date references now use the date class.
// - Added price_today and price_yesterday members to allow fast tracking
//   of recent price changes.
// - Renamed all calls to date2string() which is now getdate().
// - Revised price::read_share_price() so that it handles errors.  A dialog
//   is now displayed and the program can continue after the user
//   acknowledges the problem.
// - If price::price_store() cannot read security prices before writing
//   back, it displays a warning.
// - To minimize display lags price_list::validate() now periodically
//   refreshes the display (depending on the dialog implementation).
// - Added an automatic price validation option - When there are lots
//   of problems it is very tiresome to keep hitting the "OK button.
// - Added get_price_change() which reports the price before and
//   after an update.
// - Added change_vect to price_list for callbacks when prices are updated.
// - Added PRICE_HEADER_LOADED flag so we can detect if the header info
//   has been loaded for a price.
// - BUG FIX. Price validation would not work if the price in the file
//   was not a valid fp number.  price::validate() now checks each price
//   for being a NaN and corrects if necessary. Also now checks that the
//   stored max/min values are sane.
// - BUG FIX. The get_price() member could (potentially) search backwards
//   beyond the start of the price array.
// - BUG FIX. The get_price() member now allows one to specify whether
//   a backwards search should occur.  The editprice() member does not
//   want such a search to occur.
// - BUG FIX. The function price::write_share_price() could write prices
//   back even if none had been read.  This now causes a PANIC.
// - BUG FIX. The function price::read_share_price() would not always
//   write back price information for old version files.
// - BUG FIX. In price::read_share_price() price data could be reloaded
//   without releasing previously allocated memory.  Now data that is
//   already in core is not reloaded.
//
// Revision 1.4  1994/12/11  18:04:13  davidsi
// - BUG FIX. price::validate() was not writing back the price data, when
//   0 values had been removed.
//
// Revision 1.3  1994/11/05  21:08:05  davidsi
// - price_list::update() now marks prices that need manual updating so
//   that user entry is only required when all other prices have been
//   updated.
// - Removed printf() that was displaying security names each time prices
//   were updated.
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.2  1994/10/18  20:05:02  davidsi
// - Renamed convert_name() to name_to_file(). This function can now
//   handle names with slash (needed because teletext names may have
//   slashes).
// - Added price_to_file() which maps a price class instance into a
//   file name.
// - Updated all calls to security::extract().
// - Provided an extract() function that allows prices to be obtained
//   in their raw form.  This will not currently force prices to be
//   loaded though!
// - Extensively revised validation so that it now makes a series of
//   checks and warns the user before modifying any data.  A menu option
//   now allows the user to validate their price data.
// - Prices are now read using the ttext class.
// - Added remove() so that prices can be deleted when a security
//   is deleted.
// - Added rename() to allow price files to be renamed when a securities
//   name changes.
// - Added a destructor function - though this is currently a NOP.
//
// Revision 1.1  1994/08/06  16:42:57  davidsi
// Initial revision
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dial.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <ctype.h>
#include <math.h>
#include <nan.h>
#include <signal.h>


#include "localmacros.h"

#include "dialog.h"
#include "support.h"
#include "price.h"
#include "busy.h"
#include "menu.h"
#include "ttext.h"
#include "val.h"
#include "prefs.h"


//
// Positions within price file
//
#define PF_HEADEROFFSET		0	// position of header 
					// don't change this value, otherwise
					// version 1 files can't be read 
#define PF_FURTUREOFFSET	256	// offset of header-future expansion 
#define PF_DATAOFFSET		512	// offset of header 

#define RESIZE_VALUE		365	// number of days price data grows by 
					// when a price for a new day falls 
					// outside the current file 


//
// Variables 
//
price_list *main_price_list;
struct price_desc {
        struct price_desc *next;                // next description
        double price;
        char name[SECNAMELEN];                  // name of share
};

//
// Teletext can use non standard characters for fractions, these
// mappings are defined below. Note this is strictly a workaround
// and really some ctrl codes should be checked!
//
#define TTEXT_ONEHALF		'\\'		// 0x5c
#define TTEXT_THREEQUARTERS	'}'
#define TTEXT_ONEQUARTER	'{'


//
// Define what a numeric character is for teletext
//
#define NUMERIC_CHAR(x) ( isdigit((x)) || 	\
		((x)==TTEXT_ONEQUARTER) ||	\
		((x)==TTEXT_THREEQUARTERS) ||	\
		((x)==TTEXT_ONEHALF))



static struct price_desc 
    *price_info=NULL;		// list of prices loaded from external src 


//----------------------------------------------------------------------------
//
//----------------------------------------------------------------------------


//
// PriceNopathname
//
// Displays warning for when no price information can be located on the
// disk.  The user is given the chance to leave the program and fix the
// problem.
// 
// INPUTS :	None
// RETURNS:	None
//
static void
PriceNopathname()
{
	static char notshown=1;
	
	if (notshown) {
		notshown=0;
		if (DialogYesno("CAUTION: No price information can be found.\n\n
Currently no path for price information has been defined.\n
It is highly unlikely that this is intended. Your system\n
administrator can set a system wide path by defining the\n
variable 'xsmprices' in /etc/default/xsm. Alternatively\n
you may select your own path by defining the environment\n
variable 'xsmprices'.  You are recommended to exit at this\n
point and correct the error.\n\n
Do you wish to exit?",
			"Yes","No") ) 
		exit(1);
	}
}



//
// price::PriceConvertname(src)
//
// Returns the name of the price file for a given security.  
//
// INPUTS :	src		Either the name of the security (if its
//				known or not as given in the price member)
//				or NULL if security name is to be 
//				extracted from the price member. 
// RETURNS:	None
//
char *
price::PriceConvertname(char *src)
{
	static char buffer[SECNAMELEN];	// work area 
	char *cptr;

	if (src==NULL)
		sec->SecurityExtract(&src,NULL,NULL,NULL,NULL);
	
	strncpy(buffer,src,SECNAMELEN);
	buffer[SECNAMELEN-1]='\0';

	//
	// Remap characters that aren't allowed in filenames
	//
	cptr=buffer;
	while (*cptr) {
		if (*cptr=='/')
			*cptr='\\';
		cptr++;
	}

	return(buffer);
}


//
// price::PriceFilename name
// 
// Converts security names into pathnames that hold price information for
// that security.
//
// INPUTS :	name		The name of the security or NULL
//				if unknown
// RETURNS:	A pointer to the pathname. 
//
// CAUTIONS:	The returned pathname is in static memory and WILL be
//		written over if multiple calls are made to this function.
//
char *
price::PriceFilename(char *src)
{
	static char buffer[1024];	// work area 
	char *fname;
	char *env;

	fname=PriceConvertname(src);

	//
	// If no path, then open file using the old fashioned way via
	// the environment variable.
	//
	if (path == NULL) {
		env=Getdefault(VAR_PRICEPATH);
		if (env) 
			sprintf(buffer,"%s/%c/%s",env,fname[0],fname);
		else {
			PriceNopathname();
			strcpy(buffer,fname);
		}
	} else {
		//
		// Otherwise simply bolt the path onto the filename
		//
		sprintf(buffer,"%s/%c/%s",path,fname[0],fname);
	}

	return(buffer);

}

	
//
// price::PriceCreatelocation
//
// Creates the destination directory for a price file.
//
// INPUTS :	None
// RETURNS:	None
//
void
price::PriceCreatelocation()
{
	static char buffer[1024];	// work area 
	static char ebuffer[1024];	// work area 
	char *env;
	char *fname;

	fname=PriceConvertname(NULL);

	//
	// If no path, then open file using the old fashioned way via
	// the environment variable.
	//
	if (path == NULL) {

		env=Getdefault(VAR_PRICEPATH);
		if (env) 
			sprintf(buffer,"%s/%c",env,fname[0]);
		else {
			PriceNopathname();
			return;
		}
	} else {
		//
		// Otherwise simply bolt the path onto the filename
		//
		sprintf(buffer,"%s/%c",path,fname[0]);
	}
	for(;;) {
		if (mkdir(buffer,S_IEXEC|S_IREAD|S_IWRITE)==-1) {
			if (errno==EEXIST)
				break;
			sprintf(ebuffer,"Cannot create directory %s.",buffer);
			if (!DialogYesno(ebuffer,"Retry","Cancel"))
				break;
		} else
			break;
	}
}

//
// price::PriceChecklocation
//
// Ensures the destination directory for a price file exists, creating
// one if necessary.  Also checks to see if the price file is in the
// flat directory and if it is it moves it.
//
// INPUTS :	None
// RETURNS:	None
//
void
price::PriceChecklocation()
{
	char correctpath[1024];
	struct stat statbuf;
	static char buffer[1024];	// work area 
	char *fname;			// name of file (no path)
	char *env;

	strcpy(correctpath,PriceFilename(NULL));
	if ( stat(correctpath,&statbuf) != -1)
		return;

	//
	// File isn't where we expected it. See if its one level	
	// up.
	//
	fname=PriceConvertname(NULL);

	//
	// If no path, then open file using the old fashioned way via
	// the environment variable.
	//
	if (path == NULL) {
		env=Getdefault(VAR_PRICEPATH);
		if (env) {
			sprintf(buffer,"%s/%s",env,fname);
		} else {
			PriceNopathname();
			return;
		}
	} else {
		//
		// Otherwise simply bolt the path onto the filename
		//
		sprintf(buffer,"%s/%s",path,fname);
	}
	//
	// Check if it really is one level up.
	//
	if ( stat(buffer,&statbuf) == -1)
		//
		// No, likely to be a problem but we can't do anything
		// about...
		//
		return;

	PriceCreatelocation();

	while ( ::rename(buffer,correctpath)== -1 ) {
		char *src;

		sec->SecurityExtract(&src,NULL,NULL,NULL,NULL);
		sprintf(buffer,"Cannot move prices for %s",src);
		if (!DialogYesno(buffer,"Retry","Ignore"))
			break;
	}
}



//
// price::PriceWritefile pathname share_ptr mode
//
// Writes the prices for the specified share out to its relevant file.
//
// INPUTS :	mode		TRUE = write entire file
//				FALSE = write header information
// RETURNS:	None.
//
void
price::PriceWritefile( int mode)		
{
	int file;			// file id, note not FILE * ! 
	int neof=1;			// assume not at file end initially 
	char type;			// machine type that wrote the file
	char pad;			// used for reading the pad byte in
	long magic;			// magic value in the file 
	long version;			// version of xsm writing the file
	unsigned long spacereq;		// amount of space required 
	char *name;
	char buffer[1024];
	int baseday;


	//
	// See if the pathname has already been determined, if it
	// has not go and create it now...
	//
	name=PriceFilename(NULL);

	//
	// Go open the file
	//
	if ( (file=open(name, O_RDWR))==-1) {
		if (mode==FALSE) {
			DialogMessage("Price file missing");
			return;
		}
		PriceCreatelocation();
		if ( (file=creat(name, S_IREAD | S_IWRITE))==-1) {
			sprintf(buffer,"Cannot open %s for update",name);
			DialogMessage(buffer);
			return;
		}
	}
	//
	// Seek past header information, and write
	//
	if (lseek(file,PF_HEADEROFFSET,SEEK_SET)==-1) {
		close(file);
		DialogMessage("Having trouble writing prices-Not written");
		perror("lseek failed");
		return;
	}

	type=MACHINE_TYPE;
	if (write(file,&type,1)!=1 && write(file,&pad,1)!=1)
		DialogPanic("Cannot write file. No id");
	magic=SHAREPRICE_MAGIC;
	if (write(file,&magic,4)!=4)
		DialogPanic("Cannot write file. No magic");
	version=SMWB_VERSION;
	if (write(file,&version,4)!=4)
		DialogPanic("Cannot write file. No version");
	baseday=basedate.daysince1970();
	if (write(file,&baseday,4)!=4)
		DialogPanic("Cannot write file. No base date");
	if (write(file,&n_prices,4)!=4)
		DialogPanic("Cannot write file. No number of prices");

	if ( write(file,&pmax,sizeof(double))!=sizeof(double) )
		DialogPanic("Cannot write file. No max");
	if ( write(file,&pmin,sizeof(double))!=sizeof(double) )
		DialogPanic("Cannot write file. No min");

        //
        // Since version 3 files the price_yesterday and price_today
        // is written.
        //
	if ( write(file,&price_yesterday,sizeof(double))!=sizeof(double) )
		DialogPanic("Cannot write file. No price_yesterday");
	if ( write(file,&price_today,sizeof(double))!=sizeof(double) )
		DialogPanic("Cannot write file. No price_today");
	
	//
	// If not simply writing the header information now write
	// the prices
	//
	if (mode) {
		if (!prices) {
			fprintf(stderr,"PANIC:Prices not read for %s\n",name);
			exit(1);
		}
		if (lseek(file,PF_DATAOFFSET,SEEK_SET)==-1) {
			close(file);
			DialogMessage("Having trouble writing prices-Not written");
			return;
		}
		spacereq=sizeof(double) * n_prices;
		if (write(file,prices,(unsigned int)spacereq)!=(unsigned int)spacereq)
			DialogPanic("Cannot write file. No prices");
	}
	close(file);
}




//
// price::PriceReadfile share_ptr mode
//
// Reads the prices for a given share.  The prices are placed into the
// price buffer for that share.
//
// INPUTS :	mode		TRUE = read entire file
//				FALSE = read header information
// RETURNS:	0	= OK.
//		-1	= ERROR.
//		1	= ERROR (file does not exist)
//
int
price::PriceReadfile( int mode)
{
	int file=-1;			// file id, note not FILE * !
					// -1 identifies a file that is not
					// open.
	int neof=1;			// assume not at end of file initially
	char type;			// type of machine that wrote the file 
	char pad;			// used for reading the pad byte in 
	long magic;			// magic value in the file 
	long version;			// version of xsm writing the file 
	unsigned long spacereq;		// amount of space required 
	char *buffer;			// work area
	char *errstring;
	int i;

	buffer=PriceFilename(NULL);
	if ( (file=open(buffer,O_RDONLY))==-1)
		return(1);

	if (version>SMWB_VERSION_ONE && (flags & PRICE_HEADER_LOADED) )
		goto loadprice;

	if (read(file,&type,1)!=1 && read(file,&pad,1)!=1) {
		errstring="File too short. No id";
		goto fatalerror;
	}
	if (read(file,&magic,4)!=4) {
		errstring="File too short. No magic";
		goto fatalerror;
	}
	if (magic!=SHAREPRICE_MAGIC) {
		close(file);
		errstring="File is corrupt, magic does not match";
		goto fatalerror;
	}
	if (read(file,&version,4)!=4) {
		errstring="File too short. No version";
		goto fatalerror;
	}
	if (version==SMWB_VERSION_ONE||version==SMWB_VERSION_TWO) {
		unsigned long secstamp;
		char buffer[64];
		//
		// VERSION_ONE/TWO files have seconds from 1970 stored
		//
		if (read(file,&secstamp,4)!=4){
			errstring="File too short. No base date";
			goto fatalerror;
		}
		sprintf(buffer,"%d",secstamp);

		if (basedate.setdate(buffer) == -1 ) 
			DialogPanic("Bad date read from price file");
		
	} else {
		//
		// > VERSION TWO have days since 1970
		//
		int baseday;

		if (read(file,&baseday,4)!=4) {
			errstring="File too short. No base date";
			goto fatalerror;
		}
		basedate.daysince1970(baseday);
	}
	if (read(file,&n_prices,4)!=4) {
		errstring="File too short. No number of prices";
		goto fatalerror;
	}
	if (read(file,&pmax,sizeof(double))!=sizeof(double)) {
		errstring="File too short. No max";
		goto fatalerror;
	}
	if (read(file,&pmin,sizeof(double))!=sizeof(double)) {
		errstring="File too short. No min";
		goto fatalerror;
	}

	//
	// Since version 3 files the yesterday_price is loaded from
	// the price of the last update.
	//
	if (version>=SMWB_VERSION_THREE) {
		if (read(file,&price_yesterday,sizeof(double))!=sizeof(double)){
			errstring="File too short. No price_yesterday";
			goto fatalerror;
		}
		if (read(file,&price_today,sizeof(double))!=sizeof(double)){
			errstring="File too short. No price_today";
			goto fatalerror;
		}
	} else {
		price_yesterday= NOT_A_PRICE;
		price_today= NOT_A_PRICE;
	}
	//
	// Indicate that the header information has been loaded, so
	// we don't need to load it again!
	//
	flags |= PRICE_HEADER_LOADED;


loadprice:

	//
	// If the data is already, don't load it again!
	//
	if (flags & PRICE_DATA_LOADED) 
		goto closefile;

	//
	// Version 1 had price files where price data followed header info
	// so only seek if not version 1 file.
	// 
	if (version>SMWB_VERSION_ONE)		
		if (lseek(file,PF_DATAOFFSET,SEEK_SET)==-1) {
			close(file);
			DialogMessage("Having trouble reading prices-Not read");
			return(-1);
		}

	//
	// If mode is set or the version of this file is not the current
	// working version, read all price information.  In the later case
	// we will want to write it back, using the current file format.
	//
	if (mode || version!=SMWB_VERSION) {
		spacereq=sizeof(double)*n_prices;
		prices=(double *)malloc((unsigned int)spacereq);
		if (read(file,prices,(unsigned int)spacereq)!=spacereq) {
			errstring="File too short. No prices";
			goto fatalerror;
		}
		//
		// Indicate that the price data has been loaded, so
		// we don't need to load it again!
		//
		flags |= PRICE_DATA_LOADED;

		//
		// Prior to version four, all prices <0 were assumed
		// bad.  Now such prices must lie in a much smaller
		// range and are defined by NOT_A_PRICE.
		//
		// Remap all old prices to new style prices
		//
		if (version<SMWB_VERSION_FOUR) {
			for(i=0;i<n_prices;i++) {
				if (prices[i]<0.0)
					prices[i]=NOT_A_PRICE;
			}
		}
    	}

closefile:
	close(file);
	file= -1;
	//
	// Rewrite old version files in the current format 
	//
	if (version!=SMWB_VERSION) {
		if (unlink(buffer)==-1) {
			errstring="Cannot remove old style price file";
			goto fatalerror;
		}
		PriceWritefile(TRUE);
	}
	return(0);

fatalerror:
	{
		char *src;
		char *buffer;

		sec->SecurityExtract(&src,NULL,NULL,NULL,NULL);

		//
		// The +20 allows for termination and the " ()" that
		// are added in the sprintf
		//
		buffer=(char *)malloc(strlen(errstring)+strlen(src)+20);

		sprintf(buffer,"%s (%s)",errstring,src);
		DialogMessage(buffer);
		free(buffer);
		DialogMessage("You may wish to use Price->Validate to correct this problem");
			
		if ( file!= -1 )
			close(file);
		return(-1);
	}
}




//
// price_getlatest
// 
// Gets the latest price (from a price list - NOT a share price area) for
// a particular name in the price list.
//
// INPUTS :	name		The name of a price in the list
// RETURNS:	The price for the name given or a -ve value if no price was
//		found. 
///
static double
price_getlatest( char *name )
{
	struct price_desc *pricei;

	pricei=price_info;
	while (pricei!=NULL) {
		if (!strcmp(name,pricei->name))
			return(pricei->price);
		pricei=pricei->next;
	}
	return( NOT_A_PRICE );
}


//
// price::PriceStore date price
// 
// For a given share this takes a data and prices, updates the
// price area for that share & other relevant information, and finally
// updates the appropriate file on disk.
//
// INPUTS :	date		The date for which the price applies
//		price		The price.
// RETURNS:
///
void
price::PriceStore( class date *date,
		double price
		)
{
	unsigned long currentoffset;		// offset for the new price 
	char *name;
	int file;
	char *src;
	class date realdate;
	unsigned int todayoffset;
	unsigned int baseoffset;

	sec->SecurityExtract(&src,NULL,NULL,NULL,NULL);


	//
	// If prices are not already loaded in try to update the version
	// on disk with minimal file reads...
	//
	if (prices==NULL) {
		if (PriceReadfile(FALSE)!=0)	// get header information 
			goto normalsave;	// none - try normal way
		if (prices!=NULL)		// if we read the whole lot 
			goto normalsave;	// do normal save 

		currentoffset=date->daysince1970()-basedate.daysince1970();
		//
		// If price falls within the file then update it without
		// reading the file, else read the file and do it the slow way..
		// (must do this because negative values are used to signify
		// no price).	
		//
		if ( (currentoffset+1>n_prices) || (currentoffset<0) ) 
			goto normalsave;
		//
		// Only update max/min if it looks like a real price
		//
		if ( IS_PRICE(price) ) {
			pmax=max(pmax,price);
			pmin=min(pmin,price);
		}
		//
		// If the date for the price is today, then save the price 
		// in the price_today cache and update the datestamp.
		//
		realdate.setdate();
		if ( date->compare(&realdate) == 0) {
			price_yesterday=price_today;
	                price_today=price;
			today= *date;
		}
		PriceWritefile(FALSE);		// write header 
		name=PriceFilename(NULL);
		if ( (file=open(name, O_RDWR ))==-1) {
			DialogMessage("Cannot open price file for writing");
			return;
		}
		if (lseek(file,PF_DATAOFFSET+sizeof(double)*currentoffset,
					SEEK_SET)==-1) {
			close(file);
			DialogMessage("Having trouble writing prices-Not written");
			return;
		}
		if (write(file,&price,sizeof(double))!=sizeof(double))
			DialogPanic("Cannot write file. No prices");
		close(file);
		return;
	}

normalsave:

	if (prices==NULL) {		// try to read all the prices 
		//
		// If error give warning.  If file does not exist
		// no warning will be given, because PriceReadfile()
		// will return 1 (not -1).
		//
		if (PriceReadfile(TRUE) == -1 ) {
			char *buf;
			
			buf=(char *)malloc(strlen(src)+256);
			sprintf(buf,"Price file for %s will be created, old data will be lost",src);
			DialogMessage(buf);
			free(buf);
		}
	}

	if (prices==NULL) {			// no old prices 
		n_prices=1;
		pmax=price;
		pmin=price;
		basedate.setdate();
		prices=(double *)malloc(sizeof(double)*(unsigned int)n_prices);

		if (prices==NULL) 
			DialogPanic("price::PriceStore - malloc failed!");
		
		*(prices)=price;
	} else {				// add to old prices 
		unsigned long oldsize;

		//
		// Only update max/min if it looks like a real price
		//
		if ( IS_PRICE(price) ) {
			pmax=max(pmax,price);
			pmin=min(pmin,price);
		}
		oldsize=n_prices;

		todayoffset=date->daysince1970();
		baseoffset=basedate.daysince1970();
		currentoffset=date->daysince1970()-basedate.daysince1970();
		if (todayoffset<baseoffset) {
			//
			// Date is for date before first date in file
			//
			double *oldprices=prices;
			unsigned int newsize;
			unsigned int i;

			currentoffset=basedate.daysince1970()-date->daysince1970();
#ifdef DEBUG
			printf("%s reallocate early price space\n",src);
#endif
			newsize= currentoffset + n_prices; 
			prices=(double *)malloc(sizeof(double)*newsize);
			if (prices==NULL) 
				DialogPanic("malloc failed PriceStore");
			for(i=0;i<currentoffset;i++)
				(prices)[i]=NOT_A_PRICE;
			for(;i<currentoffset + n_prices;i++)
				(prices)[i]=(oldprices)[i-currentoffset];
			free(oldprices);
			basedate.daysince1970(basedate.daysince1970()-currentoffset);
			currentoffset=0;
			n_prices=newsize;
		} else if (currentoffset+1>n_prices) {
			//
			// Grow the data area if the current date offset falls
			// outside the area, setting all non used values to -1
			// to indicate that they are not in use.
			//
#ifdef DEBUG
			printf("%s reallocate price space\n",src);
#endif
			n_prices=currentoffset+RESIZE_VALUE;
			prices=(double *)realloc(prices,
					sizeof(double)*(unsigned int)n_prices);

			if (prices==NULL) {
				printf("currentoffset = %d\n",currentoffset);
				printf("nprices = %d\n",n_prices);
				DialogPanic("realloc failed PriceStore");
			}
			
			for(;oldsize<n_prices;oldsize++)
				(prices)[oldsize]=NOT_A_PRICE;
		}
		(prices)[currentoffset]=price;
	}
	//
	// If the date for the price is today, then save the price in the
	// price_today cache and update the datestamp.
	//
	realdate.setdate();
        if ( date->compare(&realdate) == 0) {
		price_yesterday=price_today;
                price_today=price;
		today= *date;
	}
	PriceWritefile(TRUE);
}


//
// price::PriceUpdated
//
// Returns whether a price for a security was updated when the
// last update request was issued.
//
// INPUTS :	None
// RETURNS:	TRUE if the price has been updated else FALSE.
// 
int
price::PriceUpdated() 
{
	return ( (flags & PRICE_UPDATED) == PRICE_UPDATED);
}




//
// price_entermanually date share
//
// This is the callback that is called when the user has selected 
// a particular days price to be edited.
//
// INPUTS :	date		Date for when price will be stored
//		share		The share desc for the share whose price
//				is to be entered.
// RETURNS:	None.
///
static double
price_entermanually( class date *date, char *name)
{
#define STR_SIZE 10
	char str[STR_SIZE];			// users entered value 
	char *s;
	char *datebuf;
	double amount;

	datebuf=date->getdate();

tryagain:
	str[0]='\0';
	if (DialogString("Enter Price Of Security","New Price:",str,STR_SIZE,
				name,datebuf,NULL))
		return(NOT_A_PRICE);
	//
	// Should be better validation than this..
	//
	if ( str[0]=='\0') {
		if (DialogYesno("Not a valid price","Retry","Cancel"))
			goto tryagain;
		return(NOT_A_PRICE);
	}
	s=remove_lt_spaces(str);
	if (!(alpha_strcmp("no price",s)) || !(alpha_strcmp("noprice",s))
		||  !(alpha_strcmp("bad",s)) ||  !(alpha_strcmp("badprice",s)) )
		return(NOT_A_PRICE);

	amount=atof(s);
	if ( NOT_PRICE(amount) ) {
		if (DialogYesno("Not a valid price","Retry","Cancel"))
			goto tryagain;
		return(NOT_A_PRICE);
	}

	return(atof(s));

#undef STR_SIZE
}






//
// price::PriceFetch share date 
//
// Returns the price for a given share at the specified date.
//
// INPUTS 	date		The date for which the price is required.
//		searchback	TRUE if function should search back in
//				time for the most recent price, else
//				FALSE.
// RETURNS:	An appropriate price
//		<0 error
//
double
price::PriceFetch(class date *date,int searchback)
{
	long date_offset;
	register double *price_ptr;

	//
	// If the requested date is todays then use private cache
	// if already loaded.
	//
	if ( date->compare(&today) == 0) {
		if ( IS_PRICE(price_today) )
			return(price_today);
	}

	//
	// If share prices have not yet been loaded, load them..
	//
	if (prices==NULL) 
		PriceReadfile(TRUE);

	date_offset=date->daysince1970()-basedate.daysince1970();
	if (date_offset>=n_prices)			// adjust if overshot
		date_offset=n_prices-1;
	if (date_offset<0)				// no prices loaded 
		return(NOT_A_PRICE);			// error 
	price_ptr=(double *)&(prices[date_offset]);	// find mem loc.

	if (searchback)
		while (price_ptr>prices && *price_ptr<0)
			price_ptr--;			// locate real price 
	if (*price_ptr<0)
		return(NOT_A_PRICE);			// error 
	else
		return(*price_ptr);
}



//
// price::PriceFetchchange newprice oldprice
//
// Returns the price after and before the last update.
//
// INPUTS:	None
// RETURNS: 	Either -1 is returned or 0 is returned and:
//		newprice	- The price after the last update
//		oldprice	- The price before the last update
//
int
price::PriceFetchchange(double &newp,double &oldp)
{
	register double *price_ptr;

	//
	// If share update prices have not yet been loaded, load them..
	// DO NOT load all the price information!
	//
	if ( !(flags & PRICE_HEADER_LOADED) )
		PriceReadfile(FALSE);

	if ( NOT_PRICE(price_today) || NOT_PRICE(price_yesterday) ) 
		return(-1);

	newp=price_today;
	oldp=price_yesterday;
	return(0);
}



//
// price::PriceFetchgroup startdate enddate number
//
// Returns an array of prices for between the two dates - though
// there is a maximum number of prices that can be held specified
// by MAX_LIST_SIZE. 
//
// INPUTS :	startdate  	The first date that a price should be
//				obtained for.
//		enddate		The last date.
// RETURNS:	A pointer to an array of prices - this can be
//		released with free().
//		num gives the number of prices stored.
//
dateprice *
price::PriceFetchgroup( class date *startdate, class date *enddate,int *num)
{
#define MAX_LIST_SIZE	64

	dateprice *list;
	int iter;
	dateprice priceret[MAX_LIST_SIZE];		// staging area
	int dayoffset;
	class date tdate;

	if (prices==NULL) {		// if no prices, read now 
		PriceReadfile(TRUE);
	}
	*num=0;
	//
	// Typically this will be used to pull a months worth of prices so
	// so MAX_LIST_SIZE should be enough
	//
	dayoffset=basedate.daysince1970();
	for(iter=0;iter<n_prices;iter++) {
		tdate.daysince1970(iter+dayoffset);

                if (  tdate.compare(enddate) == 1 )
                        break;
		if ( !tdate.inrange(startdate,enddate) )
			continue;

		if ( NOT_PRICE(prices[iter]) ) {
			priceret[*num].dp_price=0.0;
			priceret[*num].dp_mode=DATEPRICE_NONE;
		} else  {
			priceret[*num].dp_price=prices[iter];
			priceret[*num].dp_mode=0;
		}
		priceret[*num].dp_date=tdate;
		if ( (*num)++==MAX_LIST_SIZE-1)
			DialogPanic("Exceeded maximum number of prices in a month");
	}

	//
	// Copy to freshly allocated memory.
	//
	priceret[(*num)++].dp_mode=DATEPRICE_END;
	list=(dateprice *)malloc(sizeof(dateprice)* (*num) );
	memcpy(list,priceret,sizeof(dateprice)* (*num) );
	return(list);
}



//
// price::PriceEdit date
//
// Allows a price to be edited for a given date.
//
// INPUTS :	date		The date for the price edit.
// RETURNS:	None.
//
void
price::PriceEdit(class date *date)
{
#define STR_SIZE	10

	static int semaphore=0;		// flag to only allow one edit 
	double newprice;
	char *s;
	char str[STR_SIZE];
	char *datestring;
	double cprice;
	char *secname;

	sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);

	if (semaphore)		// if edit running then don't run another 
		return;

	semaphore=1;			// prevent another until completed 
	cprice=PriceFetch(date,FALSE);
tryagain:
	datestring=date->getdate();
	if ( IS_PRICE(cprice) )
		sprintf(str,"%9.4f",cprice);
	else
		str[0]='\0';

	if (DialogString("Edit Price Of Security","New Price:",str,STR_SIZE,
				secname,datestring,NULL)) {
		semaphore=0;			// completed 
		return;
	}
	//
	// Should be better validation than this..
	//
	if ( str[0]=='\0') {
		if (DialogYesno("Not a valid price","Retry","Cancel"))
			goto tryagain;
		semaphore=0;			// completed 
		return;
	}
	s=remove_lt_spaces(str);
	if (!(alpha_strcmp("no price",s)) || !(alpha_strcmp("noprice",s))
		||  !(alpha_strcmp("bad",s)) ||  !(alpha_strcmp("badprice",s)) )
	    	newprice=NOT_A_PRICE;
	else {
		newprice=atof(s);

		if (newprice < SMALL_NUMBER) {
			if (DialogYesno("Not a valid price","Retry","Cancel"))
				goto tryagain;
			return;
		}
	}

	PriceStore(date,newprice);

	// 
	// Issue price changed callback
	//
	main_price_list->change_vect.CallbackMake(0,NULL);

	semaphore=0;			// completed 

#undef STR_SIZE
}


//
// price::extract pricedata startingdate number
//
// Extracts all price data.  Overwriting the price data returned from this
// call is *DANGEROUS* and should not be undertaken.
//
// INPUTS :	None
// RETURNS:	pricedata		A pointer to the price data
//		sdate			The date for the first price
//		number			Number of prices in the data area
//
void
price::extract(double **pricedata,
		class date *sdate,
		unsigned long *nprices) 
{
	//
	// If prices have not yet been loaded, load them now
	//
	if (prices==NULL) 
		PriceReadfile(TRUE);

	if (pricedata!=NULL)
		*pricedata=prices;
	if (sdate!=NULL)
		*sdate=basedate;
	if (nprices!=NULL)
		*nprices=n_prices;
}


//
// Used to detect the fperror() signal handler executing
//
static int sigfail;


//
// Signal handler for price validation.  An fp error may hit this
// handler.
//
static void
fperror(int) 
{
	sigfail=1;
	signal(SIGFPE,fperror);
}



//
// price::validate flag
//
// Validates the prices for a security.
//
// INPUTS :	flag		TRUE if errors should only be fixed after
//				prompting the user, FALSE if automatic
// RETURNS:
//
void
price::validate(int askflag)
{
	double new_min,new_max;
	int iter;
	double price;
	double lprice;
	char *secname;
	char buffer[1024];
	int fixprice=0;
	int badnumber=0;
	int lastrepair= -1;
	int maxset;
	sectype_e type;

	sec->SecurityExtract(&secname,NULL,NULL,NULL,&type);
	
	//
	// If its a group based security, nothing can be
	// validated using this function
	//
	if (type==SEC_TYPE_GROUP) 
		return;

	//
	// Check price file is in the right place
	//
	PriceChecklocation();


	//
	// Get the prices an access each price and attempt to
	// decifer whether its a good fp number.
	//
	PriceReadfile(TRUE);
	new_min=NOT_A_PRICE;
	new_max=NOT_A_PRICE;
	lprice=NOT_A_PRICE;
	signal(SIGFPE,fperror);
	for(iter=0;iter<n_prices;iter++) {
		price=prices[iter];
		sigfail=0;
		//
		// Check stored max and min are sane fp numbers
		//
		if (pmax<SMALL_NUMBER || pmin<SMALL_NUMBER)
			;
		if (sigfail)	
			pmax=pmin=0.0;
		sigfail=0;
		if ( IS_PRICE(price) )
			;
		if (isnan(price) || sigfail ) {
			if (!askflag)
				badnumber=1;
			else if (!badnumber) {
                                sprintf(buffer,"%s has prices that are not valid numbers",secname);
                                DialogMessage( buffer);
				badnumber=1;
			}
			lastrepair=iter;
		}
#ifdef NEVER
		if ( lprice>=0 ) {
			// code to fix up price goes here! 
		}
		else
#endif
		if (price<SMALL_NUMBER && price>=0.0) {
			if (!askflag)
				fixprice=1;
			else if (fixprice==0) {
				sprintf(buffer,"%s has prices equal to 0",secname);
				if (DialogYesno(buffer,"Correct","Ignore"))
					fixprice=1;
				else
					fixprice=2;
			}
			if (fixprice==1) 
				lastrepair=iter;
		}
		else if (price>LARGEST_PRICE) {
			if (!askflag)
				fixprice=1;
			else if (fixprice==0) {
				sprintf(buffer,"%s has prices > %6.2f",secname,LARGEST_PRICE);
				if (DialogYesno(buffer,"Correct","Ignore"))
					fixprice=1;
				else
					fixprice=2;
			}
			if (fixprice==1) 
				lastrepair=iter;
		}

		if (iter==lastrepair) {
			price = NOT_A_PRICE;
			prices[iter] = NOT_A_PRICE;
		}

		lprice=price;
		if (IS_PRICE(price) && (price<new_min || new_min<0.0) )
			new_min=price;
		if (IS_PRICE(price) && (price>new_max || new_max<0.0) ) {
			new_max=price;
			maxset=iter;
		}
	}
	signal(SIGFPE,NULL);
	if (fabs(new_max-pmax)>SMALL_NUMBER ||
	    fabs(new_min-pmin)>SMALL_NUMBER) {
		sprintf(buffer,"%s was (%6.2f-%6.2f) now (%6.2f-%6.2f)",
			secname,
			pmin,pmax,new_min,new_max);
		if (!askflag) {
			pmax=new_max;
			pmin=new_min;
			fixprice=3;
		} else if (DialogYesno(buffer,"Correct","Ignore")) {
			pmax=new_max;
			pmin=new_min;
			fixprice=3;
		}
	}
	if ( fixprice==1 || fixprice==3 || badnumber ) 
		PriceWritefile(TRUE);
}




//
// price_list::update mode
// 
// Obtains the latest prices from the price server, and then updates
// each securities price.  Any alias names for a stock are used when 
// searching the price list.
//
// INPUTS :	mode		TRUE = get prices from server or user entry
//				FALSE = only get prices from the server
// RETURNS:	None.
///
void
price_list::update( int mode )
{
	double price;
	class price *priceptr;
	class security *sec;
	char **al_names;
	int al_count;
	int iter;
	char *secname;
	unsigned long secflags;
	class date tdate,checkdate;
	class Busy busy;
	int count=0;
	int fullcount=0;

	busy.BusyTotal(main_sec_list->length()/5);
	busy.BusyCompleted(0,"Connecting to price server");

	if (fetch(&tdate)==-1)		// go get prices 
		//
		// Price server had problems, so bail out here
		//
		return;

	checkdate.setdate();

	if ( tdate.compare(&checkdate)!= 0)   // warn if not for today 
		if (!DialogYesno("Prices loaded are not for today",
				  "Continue","Abort"))
			return;

	//
	// Mark all securities as not having had their prices updated
	//
	sec=main_sec_list->head();
	while ( !main_sec_list->eol(sec) ) {
		sec->prices->flags &= ~PRICE_UPDATED;
		sec=sec->next();
	}

	//
	// Perform the automatic update first
	//
	sec=main_sec_list->head();
	while ( !main_sec_list->eol(sec) ) {
		price=NOT_A_PRICE;
		priceptr=sec->prices;
		sec->SecurityExtract(&secname,&secflags,&al_names,
					&al_count,NULL);
		//
		// Since processing takes such a long time, update
		// the display after every 5 updates.
		//
		count++;
		if (count>=5) {
			fullcount++;
			busy.BusyCompleted(fullcount,secname);
			count=0;
		}

		if ( secflags & SEC_FLAGS_AUTO) {
			price=price_getlatest(secname);
			if (NOT_PRICE(price))	{	// look for an alias 
				for(iter=0;iter<al_count && price<0.0;iter++)
					price=price_getlatest(al_names[iter]);
			}
		}
		// 
		// Notice how we ask the user for the price for the date
		// reported by the price server & not the current date!
		//
		 
		if (mode && NOT_PRICE(price) &&  (secflags &SEC_FLAGS_MANUAL) ) 
			priceptr->flags |= PRICE_MANUAL_UPDATE;
		else
			priceptr->flags &= ~PRICE_MANUAL_UPDATE;
			
		if ( IS_PRICE(price) ) {
			priceptr->PriceStore(&tdate,price);	
			priceptr->flags |= PRICE_UPDATED;
		}
		sec=sec->next();
	}

	//
	// Scan again, this time looking for those stocks that have
	// yet to be updated and are set as being manually updated.
	//
	sec=main_sec_list->head();
	while ( !main_sec_list->eol(sec) ) {
                priceptr=sec->prices;
		sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);
		if ( priceptr->flags&PRICE_MANUAL_UPDATE) {
			price=price_entermanually(&tdate,secname);
			if ( IS_PRICE(price) ) {
				priceptr->PriceStore(&tdate,price);	
				priceptr->flags |= PRICE_UPDATED;
			}
		}
		sec=sec->next();
	}

	// 
	// Issue price changed callback
	//
	change_vect.CallbackMake(0,NULL);
}



//
// price_list::validate askflag
//
// Validates all known stock prices.
//
// INPUTS :	askflag		TRUE if the user should be prompted, else
//				FALSE.
// RETURNS:	None
void
price_list::validate(int askflag)
{
	class security *sec; 
	int count=0;
	int fullcount=0;
	class Busy busy;
	char *secname;

	sec=main_sec_list->head();
	busy.BusyTotal(main_sec_list->length()/5);
	while ( !main_sec_list->eol(sec) ) {
		sec->prices->validate(askflag);

		//
		// Since validation takes such a long time, update
		// the display after every 5 validates.
		//
		count++;
		if (count>=5) {
			fullcount++;
			sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);
			busy.BusyCompleted(fullcount,secname);
			count=0;
		}

		sec=sec->next();
	}
	portfolio_validate(busy);
	// 
	// Issue price changed callback
	//
	change_vect.CallbackMake(0,NULL);
}



//
// price_list::fetch date
// 
// INPUTS :	None
// RETURNS:	date		The date for the update
//
int
price_list::fetch( class date *date)
{
#define _L_BUFFERSIZE	256		// size of work buffer - price server
					// should not wrt a string out larger 
					// than this 
	FILE *ptr;			// pointer to file 
	char buffer[_L_BUFFERSIZE];	// input area 
	char buffer1[_L_BUFFERSIZE];	// secondary input area
	char *option;			// ptr to price server option string 
	struct price_desc *pricei;	// ptr to new price description 
	struct price_desc *price_ptr;	// ptr to price desc in current list
	char *price_start;		// position of price in buffer
	int factor;			// division factor, eg if server
					// gives prices in pence then this 
					// must be 100 to give pounds
	char *fname;
	class ttext tt;
	int starts[10];
	int ends[10];
	char ttextload=0;
	char *str;


#ifdef NEVER
	option=XGetDefault(display,PROGRAM_NAME,XDEFAULT_PS);
	if (!option)
#endif
	option=DEFAULT_PS;

	str=db_prefs->DatabaseRetrieve(INPUT_DB_CLASS,INPUTFEED_DB_CLASS);
	if (str==NULL || !strcmp(str,FEED_ASCII))
		ttextload=0;			// don't use teletext


	//
	// If teletext price loading, initialize, load the pages,
	// and get the price file
	//
	if (ttextload) {
		starts[0]=1; ends[0]=15;
		starts[1]=21; ends[1]=35;
		tt.set(24,2,6,20,starts,ends);

		fname=tmpnam(NULL);
		tt.read_pagelist(option,fname);

		if ((ptr=fopen(fname,"r"))==NULL) {
			DialogMessage("Failed to load prices (no server)");
			price_info=NULL;
			return(-1);
		}
		//
		// Ensure the file is deleted when we are done.
		//
		unlink(fname);
	} else {
		//
		// Simple load - just open a pipe to the price loader
		//
		if ((ptr=popen(option,"r"))==NULL) {
			DialogMessage("Failed to start page loader)");
			return(-1);
		}
        }

	if (fgets(buffer,_L_BUFFERSIZE,ptr)==NULL) {
		DialogMessage("Failed to load prices (no header)");
		price_info=NULL;
		return(-1);
	}
	sscanf(buffer,"%s",buffer1);
	if (!strcmp("days",buffer1)) {	// amiga or equivalent wrote the file 
		int days,year;
		sscanf(buffer,"%s %d %d",buffer1,&year,&days);
		if (year!=1980) {
			DialogMessage("Base year must be 1980, update ignored\n");
			return(-1);
		}
		date->daysince1970(days+365*10+2);
		factor=1;
	} else {
		date->setdate();
		factor=100;
	}

	//
	// Read each line of the input which is assumed to be of the
	// form "stockname price".  For each entry create a price
	// information structure that can be traversed later.
	//
	price_info=NULL;
	while(fgets(buffer,_L_BUFFERSIZE,ptr)!=NULL) {
		option=&buffer[strlen(buffer)-1];
		//
		// Locate the final digit of the price field 
		//
		while (option>=buffer && !(NUMERIC_CHAR(*option)))
			option--; 
		if (option==buffer)	// did not find a digit-try next line 
			continue;
		while (NUMERIC_CHAR(*option)|| *option=='.' )
			--option;
		price_start=option+1;	// price found, now remove spaces! 
		while (*option==' ' || *option=='\t')
			--option;
		*(option+1)='\0';
		pricei=(struct price_desc *)malloc(sizeof(struct price_desc));
		if (*price_start==TTEXT_ONEQUARTER)
			pricei->price=0.25/(float)factor;
		else if (*price_start==TTEXT_THREEQUARTERS)
			pricei->price=0.75/(float)factor;
		else if (*price_start==TTEXT_ONEHALF)
			pricei->price=0.5/(float)factor;
		else
			pricei->price=atof(price_start)/(float)factor;
		strncpy(pricei->name,buffer,SECNAMELEN);
		pricei->name[SECNAMELEN-1]='\0';
		price_ptr=price_info;
		if (price_ptr==NULL) {
			price_info=pricei;
			pricei->next=NULL;
		} else {
			while (price_ptr->next !=NULL 
				&& strcmp(price_ptr->next->name,pricei->name)<0)
				price_ptr=price_ptr->next;
			pricei->next=price_ptr->next;
			price_ptr->next=pricei;
		}
    	}
	fclose(ptr);
	if (price_info==NULL)		// assume popen did'nt work 
		DialogMessage("No prices were loaded automatically");
	return(0);
#undef _L_BUFFERSIZE
}



//
// price::remove
//
// Removes price information both from memory and disk.  It should
// only be used when a security is being deleted from the system.
//
// INPUTS:	None
// RETURNS:	None
//
void
price::remove()
{
	char *name;
	char buffer[1024];

        name=PriceFilename(NULL);
	while ( unlink(name) == -1) {
		char *secname;
		
		//
		// If the file does not exist, then we cannot 
		// expect to delete it!
		//
		if (errno==ENOENT)
			break;

	        sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);
		sprintf(buffer,"Cannot delete prices for %s",secname);
		if (!DialogYesno(buffer,"Retry","Ignore"))
			break;
	}
	delete this;
}

//
// price:::rename oldname
//
// Renames price information on the disk.
//
// INPUTS:      oldname		The old name for the security. The new
//				name is assumed to have already been 
//				stored in the member of the security class.
// RETURNS:     None
//
void
price::rename(char *oldname)
{
        char *name;
        char oldpathname[1024];
        char buffer[1024];

        name=PriceFilename(oldname);
	strcpy(oldpathname,name);
        name=PriceFilename(NULL);
	while ( ::rename(oldpathname,name)== -1 ) {
		char *secname;
	
                sec->SecurityExtract(&secname,NULL,NULL,NULL,NULL);
                sprintf(buffer,"Cannot rename prices for %s",secname);
                if (!DialogYesno(buffer,"Retry","Ignore"))
                        break;
        }
}



//
// Constructor for price_list 
//
price_list::price_list()
{
}

//
// price::PriceAssignpath path
//
// assigns a path to a price file
//
// INPUTS : 	path 		The path to use.
// RETURNS:	None
//
void
price::PriceAssignpath(char *newpath)
{
	if (path)
		free(path);
	if (newpath) {
		path=(char *)malloc(strlen(newpath)+1);
		strcpy(path,newpath);
	} else {
		path=NULL;
	}
}

//
// price::PriceClear
//
// Removes all incore price information but leaves the disk file intact.
// This can be used when a "security" such as valuation changes and the
// in core data is stale.
//
// INPUTS : 	None
// RETURNS:	None
//
void
price::PriceClear()
{
	if (prices)
		free(prices);
	prices=NULL;

	if (path)
		free(path);
	path=NULL;


	flags=0;
	n_prices=0;
	pmax= NOT_A_PRICE;
	pmin= NOT_A_PRICE;
	price_today= NOT_A_PRICE;
	price_yesterday= NOT_A_PRICE;
}


//
// Constructor for price
//
price::price(class security *secptr)
{
	sec=secptr;
	flags=0;
	n_prices=0;
	prices=NULL;
	path=NULL;
	pmax= NOT_A_PRICE;
	pmin= NOT_A_PRICE;
	price_today= NOT_A_PRICE;
	price_yesterday= NOT_A_PRICE;
}


//
// Destructor for price
//
price::~price()
{
}


enum menuid {
	MENU_PRICE_AUTO,
	MENU_PRICE_MANUAL,
	MENU_PRICE_VALIDATE,
	MENU_PRICE_VALIDATE_AUTO,
	MENU_PRICE_STATUS,
};

enum buttonid {
	BUTTON_OK,
	BUTTON_CANCEL, 
	BUTTON_NONE
};


//
// The DialogSummary displays which stocks did not have their prices updated
//
class DialogSummary : private Button, private Optionlist {

	Dialogshell d_shell;			// top most dialog
	Dialogvertical  d_mainvert;
	Dialoglabel l1;
	Dialoglabel l2;
	Optionlistinstance *d_optionlist;
	Dialoghorizontal d_buttonhoriz;		// button horizontal form

	int buttonid;
	
private:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);

public:
	DialogSummary();
	~DialogSummary();
};


//
// DialogSummary::ButtonCallback buttonid state
//
// Handle button callbacks
//
// INPUTS :	buttonid		Identifier for the button that
//					was pressed
//		state			Button up/down flag
// RETURNS:	None
//
void
DialogSummary::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	switch(id) {
	case BUTTON_CANCEL: 
		buttonid=BUTTON_CANCEL;
		delete this;
		break;
	case BUTTON_OK: 
		buttonid=BUTTON_OK;
		delete this;
		break;
	}
}

//
// Constructor for DialogSummary.  
//
// Does all the work of finding the stocks that did not get updated
// and setting up the dialog
//
DialogSummary::DialogSummary()
{
	char *secname;
	char **secnames;
	class security *sec;
	char buffer[1024];
        static Newoptionlist nolist = {
                12,			// width
                10,			// height
                NULL,			// list of items
                "",			// name of list
                "",			// name of item
                OLIST_FLAGS_SEARCH,	// flags
		NULL,			// default font
        };
	security **secnotupdatelist;
	int secnotupdatecnt=0;
	int secupdatecnt=0;
	unsigned long flags;
	dcursortype_e cursor;

	cursor=DialogCursor(DCURSOR_BUSY);

	buttonid=BUTTON_NONE;

        d_shell.DialogshellCreate("Price Update Summary");

        d_mainvert.DialogverticalCreate(d_shell);


	secnames=allocstringlist( main_sec_list->length() );
	sec=main_sec_list->head();
	while ( !main_sec_list->eol(sec) ) {
               	sec->SecurityExtract(&secname,&flags,NULL,NULL,NULL);
		if ( flags & SEC_FLAGS_TRADED ) {
			if ( !(sec->prices->PriceUpdated()) )
				secnames[secnotupdatecnt++]=secname;
			else
				secupdatecnt++;
		}
		sec=sec->next();
		DialogCursor(DCURSOR_BUSY);
	}
	secnames[secnotupdatecnt]=NULL;
	nolist.listitem=secnames;
	nolist.listname="Securities not updated:";

	sprintf(buffer,"securities updated %d",secupdatecnt);
	l1.DialoglabelCreate(d_mainvert,buffer);
	sprintf(buffer,"securities not updated %d",secnotupdatecnt);
	l2.DialoglabelCreate(d_mainvert,buffer);

	DialogCursor(DCURSOR_WAIT);
	d_optionlist=OptionlistCreate(d_mainvert,nolist,0);

        d_buttonhoriz.DialoghorizontalCreate(d_mainvert);
        (void)ButtonPush(d_buttonhoriz," DONE ",BUTTON_CANCEL);

	d_shell.DialogRealize();
	//
	// Although secnames is a stringlist we didn't allocate memory
	// for each individual string so it cannot be released using the
	// normal freestringlist() call.
	//
	free(secnames);
	DialogCursor(cursor);
}




//
// Destructor for DialogSummary
//
DialogSummary::~DialogSummary()
{
}



//
// decode dialog code
//
// Handle main menu requests
//
// INPUTS :	dialog		The dialog causing the callback
//		code		Which menu item was selected
// RETURNS:	None
//
static void
decode(local_va_dcl)
{
va_list ap;
dialog d;
unsigned long code;

	local_va_start(ap);
	d=va_arg(ap,dialog);
	code=va_arg(ap,unsigned long);

	switch(code) {
	case MENU_PRICE_AUTO:
		main_price_list->update(FALSE);
		break;
	case MENU_PRICE_MANUAL:
		main_price_list->update(TRUE);
		break;
	case MENU_PRICE_VALIDATE:
		main_price_list->validate(TRUE);
		break;
	case MENU_PRICE_VALIDATE_AUTO:
		main_price_list->validate(FALSE);
		break;
	case MENU_PRICE_STATUS: {
		DialogSummary *ds=new class DialogSummary;
		break;
		}
	}
}


//
// price_init menu_bar
//
// Initialises the price module:
//	- Sets up the menu for the main menu bar
// 
// INPUTS :	menu_bar	The main menu bar dialog
// RETURNS:	None
///
void
price_init(menu_bar *mb)
{
	MENU("Prices","Auto update prices",decode, MENU_PRICE_AUTO);
	MENU("Prices","Full update prices",decode, MENU_PRICE_MANUAL);
	MENU("Prices","Validate",decode, MENU_PRICE_VALIDATE);
	MENU("Prices","Auto validate",decode, MENU_PRICE_VALIDATE_AUTO);
	MENU("Prices","Update Status",decode, MENU_PRICE_STATUS);
}
