
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: price.h,v 2.5 1995/05/29 19:04:26 davidsi Exp $
//
// $Log: price.h,v $
// Revision 2.5  1995/05/29  19:04:26  davidsi
// - Added IS_PRICE() macro for quick recognition of valid prices.
// - Added NOT_A_PRICE define to give a value that cannot be a price - don't
//   rely on -ve numbers anymore!
// - Bumped price file version id to SMWB_VERSION_FOUR (again!) allowing for
//   the new NOT_A_PRICE definition. From version 4, price files use this
//   value for invalid prices, not simply -ve numbers.
//
// Revision 2.4  1995/05/18  20:25:13  davidsi
// - Added PriceClear() to the price class.  This removes all incore price
//   information.  The disk file is left intact.
// - Added the capability to have price files in multiple directories (as
//   needed).  By default the old method is used, but now a new path
//   can be set with the PriceAssignpath() function member.
// - The price function members price_to_file() and name_to_file() have been
//   removed to make way for cleaner versions PriceConvertname() and
//   PriceFilename().
// - Multiple price directory support and upgrades are handled by
//   PriceCreatelocation() and PriceChecklocation() function members.
//
// Revision 2.3  1995/04/03  20:39:30  davidsi
// - Added PRICE_UPDATED flag so securities that don't get their prices
//   updated can be tracked.
//
// Revision 2.2  1995/03/08  21:22:42  davidsi
// - Made main_price_list externally visible.
// - Renamed price_store() to PriceStore().
//
// Revision 2.1  1995/03/01  22:08:19  davidsi
// - The function member editprice() no longer requires a dialog to be passed.
// - Removed the suplib.H include.
//
// Revision 2.0  1995/01/19  21:17:47  davidsi
// - All date references now use the date class.
// - Added price_today and price_yesterday members to allow fast tracking
//   of recent price changes.
// - Added PRICE_HEADER_LOADED flag so we can detect if the header info
//   has been loaded for a price.
// - Added change_vect to price_list for callbacks when prices are updated.
// - BUG FIX. The get_price() member now allows one to specify whether
//   a backwards search should occur.
//
// Revision 1.4  1994/12/11  18:08:02  davidsi
// - Updated declaration of extract().  This function now reports the
//   start time and number of prices loaded.
//
// Revision 1.3  1994/11/05  21:11:26  davidsi
// - Made the price_list class  a friend of the price class to allow direct
//   access to the flags field.
// - Added PRICE_MANUAL_UPDATE flag.  This is used during price updating to
//   mark those securities which need a manual (or possible manual) update
//   once all auto updates are complete.
//
// Revision 1.2  1994/10/17  20:38:39  davidsi
// - Added new extract function.
// - Added new remove function to allow prices to be pernamently deleted
//   from both disk and memory.
// - Added new rename function to allow price files to be renamed.
// - Renamed convert_name to price_to_file() and added name_to_file() to
//   separate the two different pieces of functionality.
// - Added destructor function.
//
// Revision 1.1  1994/08/06  16:42:57  davidsi
// Initial revision
//
//
#ifndef __PRICE_H
#define __PRICE_H

#include "localmacros.h"
#include "dslist.h"
#include "record.h"
#include "security.h"
#include "date.h"




//
// Processor types
//
#define i486            1
#define mc680x0         2

//
// values for type field
//
#define MACHINE_TYPE    i486                    // i486 or i386 and i387 

//
// value for magic field 
//
#define SHAREPRICE_MAGIC        1234567         // might do better 
#define SMWB_VERSION_ONE	1
#define SMWB_VERSION_TWO	2
#define SMWB_VERSION_THREE	3
#define SMWB_VERSION_FOUR	4
#define SMWB_VERSION   		SMWB_VERSION_FOUR

//
// Default binary for loading prices
//
#define DEFAULT_PS      "priceldr"


//
// Defines for dp_mode field
//
#define DATEPRICE_VALID	0
#define DATEPRICE_NONE	1
#define DATEPRICE_END	2

class dateprice  {
public:
	double dp_price;
	class date dp_date;
	char dp_mode;
};


class secprice_list : public dslist_head {
public:
	secprice_list(class security *secptr);
	//
	// These are to prevent needing casts
	//
        class price *dual_head();
};


class price_list 
{
public:
        class callback change_vect;	// callbacks for when prices
					// are changed

	price_list();
	int fetch(class date * );
	void validate(int);
	void update(int mode);
};

class price {

	unsigned long	n_prices;	// number of prices in below array
	double		*prices;	// prices
	class date	basedate;	// date of first price
	double		pmax,pmin;	// range of prices

	class date	today;		// date for price_today
	double		price_today;	// price today - once updated
	double		price_yesterday;// price on last update - when the
					// price has been updated today

	unsigned long flags;
	class security *sec;
	char *path;			// path if for price file if not
					// NULL else use default.

friend class price_list;

private:

public:
	char *PriceConvertname(char *name);
	char *PriceFilename(char *name);
	void PriceCreatelocation();
	void PriceChecklocation();
	void PriceWritefile(int mode);
	int PriceReadfile(int mode);
	void PriceStore( class date *date, double price);
	int PriceUpdated();
	dateprice *PriceFetchgroup(class date *,class date *,int *);
	double PriceFetch(class date *date,int searchback);
	int PriceFetchchange(double &newp,double &oldp);
	void PriceEdit(class date *);
	void validate(int);
	void extract(double **pricedata,class date *date,unsigned long *nprices);
	void remove();
	void rename(char *oldname);
	void PriceClear();
	void PriceAssignpath(char *pname);

	price(class security *);
	~price();
};

//
// Values for price flags
//
#define PRICE_MANUAL_UPDATE	0x1
#define PRICE_HEADER_LOADED	0x2	// price header has been loaded
#define PRICE_DATA_LOADED	0x4	// actual prices have been loaded
#define PRICE_UPDATED		0x8	// prices were updated during
					// last price update.


//
// Largest possible price
//
#define LARGEST_PRICE 1000000.0

//
// Not a price
//
#define NOT_A_PRICE	(-VERY_SMALL_NUMBER)

//
// Price validity macros
//
#define IS_PRICE(x)	( ((x)>=0.0) || ((x)<-SMALL_NUMBER) )
#define NOT_PRICE(x)	(!(IS_PRICE(x)))


//
// Externally available variables
//
extern price_list *main_price_list;

#endif
