
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: record.C,v 2.2 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: record.C,v $
// Revision 2.2  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.1  1995/03/08  21:19:42  davidsi
// - In record::write() don't write the record if it is of zero length.
// - BUG FIX. If the input file is empty record::readbytes() was able to
//   break out of the processing loop and reference a string that had
//   not been allocated due to nothing having been read from the file.
//
// Revision 2.0  1995/01/19  21:14:22  davidsi
// - Added extern declarations for regcmp & friends to stop compiler
//   warnings.
//
// Revision 1.2  1994/11/05  21:08:32  davidsi
// - Various modifications to allow this file to compile using g++.
//
// Revision 1.1  1994/07/03  20:31:47  davidsi
// Initial revision
//
//
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <dial.h>

#include "record.h"

extern char *__loc1;			// used by regex/regcmp functions
extern "C" {
extern char *regex(char *, char *);
extern char *regcmp(char *,char *);
extern int _filbuf(...);
}

#define DEFAULT_SIZE 10			// default size of string buffers


//
//=============================================================================
//
// Private Functions
//
//=============================================================================
//


//
// check_length size_in_bytes
//
// Ensures that both the main input/output buffer (string) and the
// copy buffer (buffer) are capable of storings strings whose size
// is less than the value given by size_in_bytes.
//
void
record::check_length(int len)
{
	if (length<len) {
		if (string)
			string=(char *)realloc(string,len*2);
		else
			string=(char *)malloc(len*2);
		if (buffer)
			buffer=(char *)realloc(buffer,len*2);
		else
			buffer=(char *)malloc(len*2);
		length=len*2;
	}
}

//
// readbytes termination_character append_mode
//
// Reads a sequence of bytes into the input buffer (string) until either
// the end of file is read or the termination character is found.  If
// append_mode is TRUE the terminating character will be added to the
// input buffer, otherwise it is ignored.  A data read from the file
// will be null byte terminated.
// 
// RETURNS:	0 	Termination character read
//		1	EOF reached (some characters may also be present
//			in the input buffer.
//
int
record::readbytes(char terminator,int append)
{
	char loaded=0;		// TRUE if terminator character read
	int position=0;		// position into input buffer
	char c;			// character read

	for(;;) {
		c=getc(file);
		if (c == EOF)
			break;
		if (c == terminator) {
			loaded=1;
			break;
		}
		check_length(position+1);
		string[position++]=c;
		check_length(position+1);
	}
	if (string!=NULL) {
		if (append) {
			string[position++]=c;
			check_length(position+1);
		}
		string[position]='\0';
	}
	return(loaded);
}



//
//=============================================================================
//
// Functions for pattern matching input records
//
//=============================================================================

//
// reset
// 
// Resets the point where matching starts to the beginning of the input
// record.
//
void
record::reset()
{
	text_ptr=string;
}

//
// skipover pattern
// 
// Looks for the current text matching that specified by pattern.  Unlike
// the match() and uptomatch() functions, the text must match starting
// with the current character (ie. the search is strictly anchored to the
// current point in the record.)
//
// RETURNS:	TRUE	match was found and the matched string has been
//			skipped over.
// 		FALSE	no match.  No characters have been skipped over.
//
int
record::skipover(char *pattern)
{
	char *buf_ptr=buffer;
	int len=strlen(pattern);

	//
	// If no record set or next character to be read is null then
	// return nothing.
	//
	if (text_ptr==NULL || *text_ptr=='\0' ) 
		return(FALSE);

	//
	// Otherwise simply check the following characters
	//
	if ( !(strncmp(pattern,text_ptr,len) ) ) {
		text_ptr += len;
		return(TRUE);
	}
	return(FALSE);
}



//
// match
// 
// Performs a complete string match/search based on regex.
//
// RETURNS:	NULL	No match made.  No characters skipped over.
//		!NULL	Pointer to the all text upto (but not including)
//			the first character that DID NOT match the compiled
//			pattern.  All these characters will be skipped over.
//
char *
record::match()
{
	char *temp=text_ptr;		// ready for backtracking

	if (!cpattern) { 		// check for compiled pattern
#ifdef INLINE_CHECK
		fprintf(stderr,"uptomatch - no patten compiled\n");
#endif
		return(NULL);
	}

	//
	// Attempt to match the current string stream with the 
	// compiled pattern.
	//
	text_ptr=::regex(cpattern, text_ptr);

	//
	// No match.  Back off and reset the string pointer in the current
	// record.
	//
	if (text_ptr==NULL) {
		text_ptr=temp;
		return(NULL);
	}

	//
	// Next case shouldn't happen...but lets make sure there is 
	// something to copy!
	//
	if (temp>text_ptr) {
		text_ptr=temp;
		return(NULL);
	}
	strncpy(buffer,temp,text_ptr-temp);
	buffer[text_ptr-temp]='\0';
	return(buffer);
}

//
// uptomatch
// 
// Performs a complete string match/search based on regex. Unlike
// match() this function returns the characters up to (but not 
// including) the first matching character and skips over all those
// that matched. It is ideal for extracting fields from a record.
//
// RETURNS:	NULL	No match made.  No characters skipped over.
//		!NULL	Pointer to the all text that was before the matching
//			pattern.  These and all the characters that matched
//			are skipped.
//
char *
record::uptomatch()
{
	char *temp=text_ptr;		// ready for backtracking

	if (!cpattern) { 		// check for compiled pattern
#ifdef INLINE_CHECK
		fprintf(stderr,"uptomatch - no patten compiled\n");
#endif
		return(NULL);
	}

	//
	// Attempt to match the current string stream with the 
	// compiled pattern.
	//
	text_ptr=::regex(cpattern, text_ptr);

	//
	// No match.  Back off and reset the string pointer in the current
	// record.
	//
	if (text_ptr==NULL) {
		text_ptr=temp;
		return(NULL);
	}

	//
	// Next case shouldn't happen...but lets make sure there is 
	// something to copy!
	//
	if (temp>__loc1) {
		text_ptr=temp;
		return(NULL);
	}
	strncpy(buffer,temp,__loc1-temp);
	buffer[__loc1-temp]='\0';
	return(buffer);
}


//
// compile input_string_pattern
//
// Compiles a pattern matching string that can then be used by match()
// and/or uptomatch().
//
void
record::compile(char *string)
{
	cpattern=regcmp(string,NULL);
}


//
// line
//
// Reads the next line (or null terminated string) from the record.  This
// is strictly a convenience function.
//
// RETURNS:	A pointer to the characters that had to be skipped in
//		order to reach the end of line or record.
//
char *
record::line()
{
	char *buf_ptr=buffer;

	//
	// If no record set or next character to be read is null then
	// return nothing...
	//
	if (text_ptr==NULL || *text_ptr=='\0' ) {
		buffer[0]='\0';
		return(buffer);
	}


	buf_ptr=buffer;
	while (*text_ptr!='\0' && *text_ptr!='\n') {
		*buf_ptr++=*text_ptr++;
	}
	*buf_ptr++='\0';
	text_ptr++;
	return(buffer);
}


//
//=============================================================================
//
// Functions for basic file handling
//
//=============================================================================
//


//
// create string
//
// Places the specified string in the output buffer ready for being
// written
//
//
void
record::create(char *str)
{
	int len=strlen(str)+1;

	check_length(len);
	strcpy(string,str);
}


//
// append string
//
// Appends the specified string to the output buffer ready for being
// written
//
//
void
record::append(char *str)
{
	int len=strlen(str)+strlen(string)+1;

	check_length(len);
	strcat(string,str);
}


//
// write
// 
// Writes the output buffer to the file using the format the file
// was opened with.
//
void
record::write()
{
	int len;

	len=strlen(string);
	if (len==0)
		return;

	if (format==OPEN_WRITE_SIZE)
		fprintf(file,"%d:",len);
	fprintf(file,"%s\n",string);
	if (format==OPEN_WRITE_PERC)
		fprintf(file,"%%\n");
}



//
// read
//
// Reads a record from the file.
//
// RETURNS:	TRUE	A record was read from the file
//		FALSE	No record was read from the file
//
int
record::read()
{

	int position;
	char c;
	int num;
	char loaded=0;

	check_length(DEFAULT_SIZE);

	position=0;

	//
	// If file is organized on a record per line basis...
	//
	if ( format==FORMAT_BY_LINE) {
		loaded=readbytes('\n',TRUE);
		goto done;
	}

	//
	// If records in file are delimeted by %
	//
	if ( format==FORMAT_BY_PERC) {
		loaded=readbytes('%',FALSE);

		do 
			c=getc(file);
		while ( c != '\n' && c!=EOF );

		goto done;
	}

	//
	// File is organized using arbitrary record lengths
	// FORMAT_BY_SIZE
	//
	for(;;) {
		c=getc(file);
		if (c == EOF) {
			string[0]='\0';
			goto done;
		}
		if (c == ':')
			break;
		string[position++]=c;
		check_length(position+1);
	}
	string[position]='\0';
	num=atoi(string);

	check_length(num+1);

	for(position=0;position<num;position++) 
		string[position]=getc(file);
	string[num]='\0';
	//
	// Jump over terminating the forced carriage return
	//
	(void)getc(file);
	loaded=1;

done:
	text_ptr=string;
	if (!loaded) {
		return (FALSE);
	}
	return (TRUE);
}


//
// open filename mode
//
// Opens a file ready for reading or writing records as determined
// by the mode argument. When files are opened for reads, a file
// header is read so as to determine file layout and version 
// information.  
//
// RETURNS:	-1	Error in opening file (errno set)
//		1	File open
//
int
record::open(char *name,unsigned long mode)
{
char modestring[2];

	if (mode==OPEN_WRITE_LINE || mode==OPEN_WRITE_SIZE ) 
		strcpy(modestring,"w");
	else if (mode==OPEN_READ)
		strcpy(modestring,"r");
	else
		return(-1);
	
	if ( (file=fopen(name,modestring))==NULL) {
                return(-1);
        }
	if (mode==OPEN_READ) {
		//	
		// Now read the first line which will descibe the format used
		// by this file.
		//
		if (!readbytes('\n',FALSE)) 
			return(-1);
		if ( !(strcmp("BYSIZE",string)) ) {
			format=FORMAT_BY_SIZE;
		} else if ( !(strcmp("BYPERC",string)) ) {
			format=FORMAT_BY_PERC;
		}
		if (!readbytes('\n',FALSE)) 
			return(1);
		versionstring=(char *)malloc(strlen(string)+1);
		strcpy(versionstring,string);
		
	} else {
		if (mode==OPEN_WRITE_LINE) {
			fprintf(file,"BYLINE\n");
			format=FORMAT_BY_LINE;
		} else if (mode==OPEN_WRITE_SIZE) {
			fprintf(file,"BYSIZE\n");
			format=FORMAT_BY_SIZE;
		} else if (mode==OPEN_WRITE_PERC) {
			fprintf(file,"BYPERC\n");
			format=FORMAT_BY_PERC;
		}
		
	}
	
	return(1);
}

//
// close
//
// Closes a file
// RETURNS:	1	File now closed
//
int
record::close()
{
	fclose(file);
	return(1);
}

//
//=============================================================================
//
// Support functions
//
//=============================================================================
//

//
// version
//
// Returns the version string from a file.   The version string is
// a free format string and is always the second line of the file on
// disk.
//
// RETURNS:	A pointer to the string containing version information.
//		Potentially can be NULL.
char *
record::version()
{
	return(versionstring);
}

//
//=============================================================================
//
// Class creation
//
//=============================================================================
//


record::record()
{

	//
	// Ensure no strings are considered as being allocated
	//
	string=NULL;			// no main input/output buffer
	buffer=NULL;			// no copy buffer
	cpattern=NULL;			// no compiled string pattern
	length=0;
	versionstring=NULL;		// no version
	format=FORMAT_BY_LINE;		// assume line by line records
}

record::~record()
{
	if (string)
		free(string);
	if (buffer)
		free(buffer);
	if (cpattern)
		free(cpattern);
	if (versionstring)
		free(versionstring);
}
