
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: security.C,v 2.8 1995/06/07 20:44:05 davidsi Exp $
//
// $Log: security.C,v $
// Revision 2.8  1995/06/07  20:44:05  davidsi
// - More support for handling multiple portfolios. Use functions
//   in portfolio.C to handle security renaming and removal.
//
// Revision 2.7  1995/05/19  18:50:11  davidsi
// - Moved the declaration of portfolio to portfolio.C.
// - The entire implementation of Holdinglist and Holding classes have been
//   moved to portfolio.C.
// - The make_callback() function is now CallbackMake().
//
// Revision 2.6  1995/04/25  21:09:13  davidsi
// - Added support for a busy cursor that is displayed when editing
//   a security (SecurityeditEdit) and when extracting security names.
//
// Revision 2.5  1995/04/16  20:29:22  davidsi
// - Added a unit trust security type (SEC_TYPE_UNITTRUST).  This
//   required several modifications to the dialogs and to the securities
//   file.
// - Added a changed_vect member, that provides callbacks when securities
//   are added, removed or modified.
// - Added HoldinglistRevise(), so that securities can be reposition on
//   the holding list when(if) their name changes.
// - BUG FIX. In SecurityLoad() the flags that may be revised due to the
//   security load must be cleared because the constructor will set default
//   flag values.
//
// Revision 2.4  1995/04/03  20:38:39  davidsi
// - Added HoldingDividendhead() to the holding class.
// - Updated security::remove() to correctly location the appropriate
//   holding information.
// - Added a SecuritylistFindname() member to the Securitylist class. This
//   allows alias names to be used when searching for securities as well
//   as the key (prime) names.
// - Removed unnecessary casts.
// - BUG FIX. When adding a new security it should have been possible to
//   change the key name of the security and make the old key name an alias.
//   Unfortunately the checking was to rigorous and prevented this.
// - BUG FIX. It was possible to add a security with a name that had already
//   been used for an alias elsewhere.
// - BUG FIX. Added a new callback semaphore to the Securityedit to prevent
//   multiple callbacks - these could result in serious system failures as
//   many  "are you sure" dialogs could be created.
//
// Revision 2.3  1995/03/08  21:25:51  davidsi
// - Factored the security class into two classes, the security class provides
//   all the main security handling functionality and the Securityedit class
//   allows securities to be edited.  This reduces the amount of memory
//   req'd to represent a security.
// - The SEC_TYPE_XX defines are now a enumerated types - function members
//   changed to reflect this.
// - Renamed extract() to SecurityExtract().
// - Added support for SEC_TYPE_GROUP: don't write such "securities" to
//   the disk, as they are dynamically created.
// - main_sec_list is now declared extern.
// - Revised to use the new Holding class which contains dividend and
//   transaction information rather than the security class.
// - BUG FIX.  The default type of a security is now a share. Previously it
//   was a security, but this is not helpful in terms of being a useful
//   default.
// - BUG FIX. Securityedit::ButtonCallback() was incorrectly using the
//   previous alias information when verifying no conflicts, it was also
//   failing to initialize the new alias information.
//
// Revision 2.2  1995/02/27  22:30:51  davidsi
// - The function member sec_compare() has become compare() overriding
//   the compare() in the inherited list class.
// - Many changes to support the new Dialog class, rather than issuing
//   direct calls to the c "dialog" functions.
// - BUG FIX. security::addw_namealias() was only allocating memory if
//   the alias count was non zero, but it would still attempt to write
//   a terminating NULL entry into the unallocated array.
//
// Revision 2.1  1995/02/05  17:59:32  davidsi
// - Support functions have now moved from the system specific suplib so
//   now include support.h.
//
// Revision 2.0  1995/01/19  21:13:50  davidsi
// - security_load() now must take an initial argument which is ignored.
//
// Revision 1.4  1994/12/11  18:06:26  davidsi
// - In security::remove() do not call get_price(), since price::extract()
//   will now automatically do this.
//
// Revision 1.3  1994/11/05  21:10:11  davidsi
// - Added set() function to the security list class.  This allows
//   the flags to be modified.
// - BUG FIX.  When the name of security changes the cached list must be
//   invalidated.  Ditto for when a security is deleted.
// - Renamed typeof() to typename() so that no errors are generated
//   when compiling with g++.
// - Added set() function to the security list class.  This allows
//   the flags to be modified.
// - Renamed typeof() to typename() so that no errors are generated
//   when compiling with g++.
// - Modified ~security_list() to be a virtual function so that g++
//   does not issue warnings.
//
// Revision 1.2  1994/10/18  19:52:30  davidsi
// - Added support for a new flag (SEC_FLAGS_DUMMY). This allows a dmmy
//   security record to be created - it won't be stored on disk nor will
//   it be added to main security lists within the program.
// - To allow dummy securities to be set up added a new constructor function
//   that sets flags as desired.
// - The security class now handles security types.  This solves the problem
//   of trying to dynamically change a security from one type to another.
// - Added extract_action().  This allows a function to poll waiting for
//   a security to be updated.
// - Added remove() function to check for a security having transactions,
// - Added remove() function to check for a security having transactions,
//   dividends, prices(etc) before removing it.  Also now ensures all
//   data remains consistant.
// - Changing a securities name is now correctly handled. Dividends,
//   prices and transactions will all be made consistant.
// - Securities now use the insert_order() operation to ensure securities
//   are held in a name sorted fashion.
// - The extract() function for security lists, now only builds name lists
//   when necessary.  The saves time. It also removes a memory leak.
// - BUG FIX. Fixed create_security() so that it asserts d_shell is zero, it
//   was asserting non-zero which meant you could not edit securities!
// - BUG FIX. In addw_update_mode() the setting of the update type buttons
//   was inverse to the real state of affairs.
// - BUG FIX. edit_security() always set SEC_ACTION_ADD so it was not
//   possible to edit securities once they had been created.  Updated many
//   of the messages to make it clear the security is being modified rather
//   than created.
// - BUG FIX. In add_alias(), the new array of strings was being incorrectly
//   constructed leading to core dumps.
// - BUG FIX. Just because a security is being modified does not mean
//   you can duplicate a name.  Still perform the check, but ensure
//   that the check does not consider the current name of the security.
//
// Revision 1.1  1994/08/06  16:38:26  davidsi
// Initial revision
//
//
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <dial.h>

#include "localmacros.h"

#include "support.h"
#include "security.h"


#define NAME_WIDTH 	16	// number of characters shown for sec names
#define ALIAS_HEIGHT	10	// height of alias dialog

//
// Variables
//
security_list *main_sec_list;



//
//=============================================================================
//
// Handles security creation callbacks
//
//=============================================================================
//

enum buttonid_e {
	BUTTON_OK,
	BUTTON_CANCEL,
	BUTTON_AUTO,
	BUTTON_MANUAL,
	BUTTON_TRADED,
	BUTTON_NTRADED,
	BUTTON_SHARE,
	BUTTON_WARRANT,
	OPTIONL_ALIAS,
};


//
// ButtonCallback
//
// Handles button callbacks.  If add/modify a security then creates
// a new structure on the basis of what has been entered into
// the security creation dialog.
//
// INPUTS :	id		The button identifier
//		flag		The state of the button
// RETURNS:	None
//
void
Securityedit::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	register int iter;		// iteration counter
	char *sec_name;			// ascii name of the security
	char buffer[SECNAMELEN];	// buffer for name of security
	char oldname[SECNAMELEN];	// buffer for old name
	char name_changed=0;		// name changed flag
	int newaliascount;
	char **newaliases;
	class security *possec;		// possible security found with name

	//
	// If a callback is already underway then don't start another.
	//
	// NOTE: We only get away with this because the BUTTON_AUTO 
	// and friends don't actually need the callback to do anything!
	//
	if ( callback.SemaphoreClock() == -1)
		return;

	switch(id) {
	case BUTTON_OK:
		break;
	case BUTTON_CANCEL:
		goto tidyup;
	case BUTTON_AUTO:
	case BUTTON_MANUAL:
	case BUTTON_TRADED:
	case BUTTON_NTRADED:
	case BUTTON_SHARE:
	case BUTTON_WARRANT:
		callback.SemaphoreUnlock();
		return;
	default:
		DialogPanic("security::ButtonCallback-Bad switch on button id");
	}

	newaliases=d_namealias->OptionlistinstanceGetlist(newaliascount);

	if ( !(secptr->flags & SEC_FLAGS_DUMMY)  ) {
		//
		// Copy and truncate name
		//
		sec_name=d_nameinput.DialoginputGet();
		strncpy(buffer,sec_name,SECNAMELEN-1);
		buffer[SECNAMELEN-1]='\0';

		//
		// Check for possible errors before updating the stucture
		//
		for(iter=0;iter<newaliascount;iter++) {
			//
  			// Check the aliases are unique.  If an alias gives
			// the same security as we are editing then no
			// problem.
			//
			possec=main_sec_list->SecuritylistFindname(newaliases[iter]);
			if (possec!=NULL && possec!=secptr) {
				DialogMessage("Alias is not unique");
				freestringlist(newaliases);
				callback.SemaphoreUnlock();
				return;
			}
		}
		if (action==SEC_ACTION_ALT) { 
			if (DialogYesno(
				"Are you sure you want to modify this security",
				"NO","YES")) {
				freestringlist(newaliases);
				callback.SemaphoreUnlock();
				return;
			}
		} else {
			if (DialogYesno(
				"Are you sure you want to add this security",
				"NO","YES")) {
				freestringlist(newaliases);
				callback.SemaphoreUnlock();
				return;
			}
		}
	} else {	
		// 
		// Security is a dummy security so no name(etc).
		//
		buffer[0]='\0';
		newaliases=NULL;
		newaliascount=0;
	}


	//
	// If name is changing (or being set for the first time) 
	// then ensure the (new) name is unique.  Note an alias
	// and key name may be identical if they reference the
	// same security.
	//
	possec=main_sec_list->SecuritylistFindname(buffer);
	if ( (strcmp(secptr->name,buffer)!=0 )  &&
 	     (possec!=NULL && possec!=secptr) ) {
		if (newaliases)
			freestringlist(newaliases);
		if (!DialogYesno("Security name is not unique","Retry","Abort"))
			goto tidyup;
		callback.SemaphoreUnlock();
		return;
	}
	//
	// If updating, then release the resource previously used and
 	// re-allocated on as needed basis. 
	//
	if (action==SEC_ACTION_ALT) 
		freestringlist(secptr->aliases);

	//
	// Detect the name of a security changing
	//
	if ( (strcmp("",secptr->name)!=0) 
		&& (strcmp(secptr->name,buffer)!=0) ) {
		name_changed=1;
		strcpy(oldname,secptr->name);
	}

	//
	// Copy the new name
	//
	strcpy(secptr->name,buffer);

	//
	// Now update the structure
	//
	secptr->alias_count= newaliascount;
	secptr->aliases=newaliases;

	secptr->flags &= ~(SEC_FLAGS_TRADED|SEC_FLAGS_NOTTRADED| 
			SEC_FLAGS_AUTO|SEC_FLAGS_MANUAL);

	if (but_status[0]->ButtoninstanceGet() == BUTTON_ENABLED)
		secptr->flags|=SEC_FLAGS_TRADED;
	else if (but_status[1]->ButtoninstanceGet() == BUTTON_ENABLED)
		secptr->flags|=SEC_FLAGS_NOTTRADED;
	else
		DialogPanic("security::ButtonCallback - neither traded or not traded");
	
	if (but_up[0]->ButtoninstanceGet()==BUTTON_ENABLED)	// AUTOMATIC ??
		secptr->flags |= SEC_FLAGS_AUTO;
	if (but_up[1]->ButtoninstanceGet()==BUTTON_ENABLED)	// MANUAL ??
		secptr->flags |= SEC_FLAGS_MANUAL;

	//
	// Determine type of security
	//
	if (but_types[0]->ButtoninstanceGet()==BUTTON_ENABLED)
		secptr->type_flags=SEC_TYPE_SHARE;
	else if (but_types[1]->ButtoninstanceGet()==BUTTON_ENABLED)
		secptr->type_flags=SEC_TYPE_WARRANT;
	else if (but_types[2]->ButtoninstanceGet()==BUTTON_ENABLED)
		secptr->type_flags=SEC_TYPE_UNITTRUST;
	else
		DialogPanic("security::ButtonCallback - neither share nor warrant");

	if (secptr->flags & SEC_FLAGS_DUMMY ) 
		goto tidyup;

	secptr->add();
        //
        // Save the revised list
        //
        main_sec_list->save();
	//
	// If the security was being altered then its name may have 
	// changed so we must now resave dividend and transaction 
	// information.  Since they get their information from this
	// record, the new name will be correctly recorded!
	//
	if (name_changed) { 
		//
		// Note the add() above won't have inserted anything
		// on the list (its already on the list), so we
		// must mark the list as stale here.
		//
	        main_sec_list->set(SECLST_FLAGS_STALE);
		(void)portfolio_rename(oldname,buffer);
		secptr->prices->rename(oldname);
	}

	//
	// Reposition on the holding list (if its on the list) so
	// that the holding list retains alphanumeric ordering
	//
	portfolio->HoldinglistRevise(secptr);

	//
	// Issue callbacks for "security changed"
	//
	main_sec_list->changed_vect.CallbackMake(0,NULL);
tidyup:
	secptr->edited.SemaphoreUnlock();
	delete this;
}



// 
// OptionlistCallback (add_alias)
//
// This is a call back used when the user enters a new alias.  It  does
// not cover using a name already in the list, since this gets dealt with
// when the user finally finishes up.  If the name is new then it is inserted
// into the list.  If it is already in the list then it is removed.
//
// INPUTS :	dialog		Dialog causing callback
// RETURNS:	None.
//
void
Securityedit::OptionlistCallback(unsigned long id)
{
	int number;			// number of items already in list
	int iter;			// general interation
	char *new_alias;		// string containing new alias name
	char **vlist;
	char **newlist;			// completed list
	int iter2;

	if (id!=OPTIONL_ALIAS)
		DialogPanic("security::OptionlistCallback - unexpected id");


	new_alias=d_namealias->OptionlistinstanceGet();
	vlist=d_namealias->OptionlistinstanceGetlist(number);

	for(iter=0;iter<number;iter++) {	
		//
		// If the alias already exists then assume the user wishes to
		// delete it..
		//
		if (strcmp(vlist[iter],new_alias)==0) {
			for(iter2=iter;iter2<number;iter2++)
				vlist[iter2]=vlist[iter2+1];
			d_namealias->OptionlistinstanceRevise(vlist);
			freestringlist(vlist);
			return;
		}
    	}
	//
	// Update the internal data and the dialog
	//
	newlist=(char **)malloc(sizeof(char **)*(number+2));
	memcpy(newlist,vlist,sizeof(char **)*number);
	newlist[number]=(char *)malloc(strlen(new_alias)+1);
	strcpy(newlist[number],new_alias);
	newlist[number+1]=NULL;
	d_namealias->OptionlistinstanceRevise(newlist);
	freestringlist(newlist);
	freestringlist(vlist);
}



//
//=============================================================================
//
// Functions to build the security creation dialog
//
//=============================================================================
//


//
// addw_buttons parent 
//
// Creates a dialog with two buttons that allow the user to create
// (when data entered) the new security, or to cancel its creation.
//
// INPUTS :	None
// RETURNS:	None	
//
void
Securityedit::addw_buttons()
{
	char *name;

	d_buttonhoriz.DialoghorizontalCreate(d_mainvert);
	//
	// Are we altering the security rather than creating ?
	//
	if (action==SEC_ACTION_ALT)
		name="ALTER";
	else if (action==SEC_ACTION_ADD)
		name="CREATE";
	else {
		DialogPanic("security::addw_buttons - bad action");
	}
	(void)ButtonPush(d_buttonhoriz,name,BUTTON_OK);
	(void)ButtonPush(d_buttonhoriz,"CANCEL",BUTTON_CANCEL);
}



// 
// addw_namealias parent
//
// Provides a dialog that allows the user to specify aliases for the
// new security OR allows the current aliases to be modified.
// How useful these aliases are is another matter...
//
// INPUTS :	parent		parent of the dialog to be displayed
// RETURNS:	None.	
//
void
Securityedit::addw_namealias()
{
	static Newoptionlist nolist = {
		NAME_WIDTH,		// width
		ALIAS_HEIGHT,		// height
		NULL,			// list of items
		"Aliases",		// name of list
		"",			// name of item
		0,			// flags
		NULL,			// default font
	};
	char **xms;
	int iter;

	nolist.listname=new char[strlen(secptr->typename())+strlen(" Aliases")+1];
	strcpy( nolist.listname,secptr->typename() );
	strcat( nolist.listname," Aliases");

	xms=(char **)malloc(sizeof(char **)*(secptr->alias_count+1));
	for(iter=0;iter<secptr->alias_count;iter++) {
		xms[iter]=(char *)malloc(strlen(secptr->aliases[iter])+1);
		strcpy(xms[iter],secptr->aliases[iter]);
	}

	xms[secptr->alias_count]=NULL;
	nolist.listitem=xms;

	d_namealias=OptionlistCreate(d_vert2,nolist,OPTIONL_ALIAS);
	if (secptr->alias_count!=0)
		freestringlist(xms);
	delete nolist.listname;
}




// 
// addw_name parent
//
// This creates a dialog where the user can enter the name of the new security
// or see the name of the security being altered.
//
// INPUTS :	parent		The parent dialog
// RETURNS:	None.	
//
void
Securityedit::addw_name()
{
	char *lname;

	d_namevert.DialogverticalCreate(d_vert1);

	lname=new char[ strlen(secptr->typename())+strlen(" Name")+1];
	strcpy( lname,secptr->typename() );
	strcat( lname," Name");

	d_namelabel.DialoglabelCreate(d_namevert,lname);
	d_nameinput.DialoginputCreate(d_namevert,NAME_WIDTH,secptr->name);

	delete lname;
}


//
// addw_update_mode parent
//
// Creates a frame with two buttons which allow the user to specify
// the update mode of a security
//
// INPUTS :	parent		parent of the dialog to be displayed
// RETURNS:	None.	
//
void
Securityedit::addw_update_mode()
{
	buttonstate_e bflags[2];

	if (secptr->flags & SEC_FLAGS_AUTO) 
		bflags[0]=BUTTON_ENABLED;
	else
		bflags[0]=BUTTON_DISABLED;

	if (secptr->flags & SEC_FLAGS_MANUAL) 
		bflags[1]=BUTTON_ENABLED;
	else
		bflags[1]=BUTTON_DISABLED;

	d_upframe.DialogframeCreate(d_vert1);
	d_upvert.DialogverticalCreate(d_upframe);
	d_uplabel.DialoglabelCreate(d_upvert,"Update Modes");

	but_up[0]=ButtonToggle(d_upvert,"Automatic",BUTTON_AUTO,bflags[0]);
	but_up[1]=ButtonToggle(d_upvert,"Manual",BUTTON_MANUAL,bflags[1]);
}





//
// addw_status parent
//
// Creates a frame with two buttons which allow the type of security
// status to be specified by the user.
//
// INPUTS :	parent		parent of the dialog to be displayed
// RETURNS:	None.	
//
void
Securityedit::addw_status()
{
	char *lname;
	buttonstate_e bflags[2];

	if ( secptr->flags&SEC_FLAGS_TRADED ) {
		bflags[0]=BUTTON_ENABLED;
		bflags[1]=BUTTON_DISABLED;
	} else {
		bflags[0]=BUTTON_DISABLED;
		bflags[1]=BUTTON_ENABLED;
	}

        lname=new char[ strlen(secptr->typename())+strlen(" Status")+1];
        strcpy( lname,secptr->typename() );
        strcat( lname," Status");

	d_statusframe.DialogframeCreate(d_vert1);
	d_statusvert.DialogverticalCreate(d_statusframe);
	d_statuslabel.DialoglabelCreate(d_statusvert,lname);

	but_status[0]=ButtonTogglemutex(d_statusvert,"Traded",
						BUTTON_TRADED,bflags[0]);
	but_status[1]=ButtonTogglemutex(d_statusvert,"Not traded",
						BUTTON_NTRADED,bflags[1]);
	delete(lname);
}

//
// addw_type parent
//
// Creates a frame with two buttons which allow the type of security
// status to be specified by the user.
//
// INPUTS :	parent		parent of the dialog to be displayed
// RETURNS:	None.	
//
void
Securityedit::addw_type()
{
	buttonstate_e bflags[3];

	bflags[0]=BUTTON_DISABLED;
	bflags[1]=BUTTON_DISABLED;
	bflags[2]=BUTTON_DISABLED;

	if ( secptr->type_flags==SEC_TYPE_SHARE )
		bflags[0]=BUTTON_ENABLED;
	else if ( secptr->type_flags==SEC_TYPE_WARRANT )
		bflags[1]=BUTTON_ENABLED;
	else if ( secptr->type_flags==SEC_TYPE_UNITTRUST )
		bflags[2]=BUTTON_ENABLED;

	d_typeframe.DialogframeCreate(d_vert1);
	d_typevert.DialogverticalCreate(d_typeframe);
	d_typelabel.DialoglabelCreate(d_typevert,"Security Type");

	but_types[0]=ButtonTogglemutex(d_typevert,"Share",
						BUTTON_SHARE,bflags[0]);
	but_types[1]=ButtonTogglemutex(d_typevert,"Warrant",
						BUTTON_WARRANT,bflags[1]);
	but_types[2]=ButtonTogglemutex(d_typevert,"Unit Trust",
						BUTTON_WARRANT,bflags[2]);
}


//
// SecurityeditEdit
//
// Displays a dialog that allows the details of a security to be
// created or modified.
//
void
Securityedit::SecurityeditEdit(class security *isecptr)
{
	char buffer[256];
	dcursortype_e cursor;

	secptr=isecptr;

	// dialog already displayed
	if (secptr->edited.SemaphoreClock()==-1) {
		DialogMessage("Security is already being edited");
		delete this;
		return;
	}

	cursor=DialogCursor( DCURSOR_WAIT );

	//
	// If the security is already on the main security list and
	// the DUMMY flag is not set then we are altering a security
	// that already exists
	//
	if (  (secptr->flags & SEC_FLAGS_INSERTED) &&
		 !(secptr->flags & SEC_FLAGS_DUMMY ) )
		action=SEC_ACTION_ALT;
	else
		action=SEC_ACTION_ADD;
	strcpy(buffer,"Set details for ");
	strcat(buffer,secptr->typename());

	d_shell.DialogshellCreate(buffer);
	d_mainvert.DialogverticalCreate(d_shell);

	d_form.DialogformCreate(d_mainvert);
	d_vert1.DialogverticalformCreate(d_form,0,50,0,100);
	d_vert2.DialogverticalformCreate(d_form,51,100,0,100);

	//
	// vert1 users
	//
	addw_name();
	addw_type();
	addw_status();
	addw_update_mode();

	//
	// vert2 users
	//
	addw_namealias();

	//
	addw_buttons();
	d_shell.DialogRealize();

	DialogCursor( cursor );
}

//
// Constructor function.
//
Securityedit::Securityedit()
{
	action=SEC_ACTION_NONE;
}

//
// Destructor function.
//
Securityedit::~Securityedit()
{
}


//
//=============================================================================
//
// Functions to handle the security class
//
//=============================================================================
//


//
// typename
//
// This simply returns "Security" but note that the function is
// a virtual function so that classes can inherit security like
// functionality but have different names, i.e "Share" or "Warrant".
//
// INPUTS:	None
// RETURNS:	A string specifying the type of security
//
char *
security::typename()
{
	switch (type_flags) {
		case SEC_TYPE_SHARE:
			return("Share");
		case SEC_TYPE_WARRANT:
			return("Warrant");
		case SEC_TYPE_GROUP:
			return("Group");
		case SEC_TYPE_UNITTRUST:
			return("Unit Trust");
		default:
			return("Security");
	}
}


//
// SecurityExtract  c_name c_flags c_names c_count type
//
// Extracts information about securities for classes that cannot
// see private security information.
//
// INPUTS :	None
// RETURNS:	c_name		The name of the security
//		c_flags		The flags applying to the security
//		c_names		An array of alias names
//		c_count		The number of aliases
//
void
security::SecurityExtract(char **c_name, 
		unsigned long *c_flags,
		char ***c_names,
		int *c_count,
		sectype_e *c_typef
		)
{
	if (c_name)
		*c_name=name;
	if (c_flags)
		*c_flags=flags;
	if (c_names) {
		*c_names=aliases;
		*c_count=alias_count;
	}
	if (c_typef) 
		*c_typef=type_flags;
}


//
// create record
//
// Provides a generic record to security class member mapping function.
// In general classes that are based on the security class can call
// this function with an uninitialized member of that class.
//
// INPUTS :	record		The record to be read and processed
// RETURNS:	0 if record mapped correctly otherwise -1.
//
void
security::SecurityCreate(char *c_name, 
		unsigned long c_flags,
		char **c_names,
		int c_count,
		sectype_e c_typef
		)
{
	char buffer[SECNAMELEN];		// buffer for name of security
	int iter;
	unsigned long new_flags;


	if (c_name) 
		strncpy(buffer,c_name,SECNAMELEN-1);
	else
		buffer[0]='\0';
	buffer[SECNAMELEN-1]='\0';
	strcpy(name,buffer);

	new_flags=0;
	if (c_flags & SEC_FLAGS_TRADED) 
		new_flags |= SEC_FLAGS_TRADED;
	else 
		new_flags |= SEC_FLAGS_NOTTRADED;
			
	if (c_flags & SEC_FLAGS_AUTO) 
		new_flags |= SEC_FLAGS_AUTO;	
	if (c_flags & SEC_FLAGS_MANUAL) 
		new_flags |= SEC_FLAGS_MANUAL;

	flags=new_flags;
		
       	alias_count=c_count;
	if (!c_names)
	       	alias_count=0;
	if (alias_count) {
		aliases=(char **)malloc(100 * sizeof(char *));
		for (iter=0;iter<alias_count;iter++ ) {
			aliases[iter]=(char *)malloc(strlen(c_names[iter])+1);
			strcpy(aliases[iter],c_names[iter]);
		}
	}

	type_flags=c_typef;
}

//
// security::add
//
// Adds a security to the master list iff the security is not already
// on that list.  The securities file will be updated so any changes
// are pernament.
//
// INPUTS :	None
// RETURNS:	None
//
void
security::add()
{
        //
        // If not already on the main list of securities insert it now,
	// otherwise remove it and then re-insert. This is necessary
	// because its name may have changed.
        //
        if (  !(flags & SEC_FLAGS_INSERTED) ) {
                flags |= SEC_FLAGS_INSERTED;
                main_sec_list->insert_order(this);
        } else {
		//
		// Remove this security node from the main stored list for
 		// securities. Note that we cannot call remove() because
		// that permanently removes securities and all we want to
		// do is ensure it is in the right position on the list
		//
		((class list_node *) this)->remove();
                main_sec_list->insert_order(this);
	}
}

//
// security::remove
//
// Removes an entry for a security including (if the user confirms)
// all transactions, all dividends and all price information.  The
// master disk files will be updated by this function so any change
// is pernament and cannot be undone.
//
// INPUTS :	None
// RETURNS:	None
//
void
security::remove()
{
	unsigned long n_prices;		// number of prices for security
	char buffer[1024];

	//
	// Security removal is not straightforward because
	// we must check that removal of the security will
 	// not cause consistancy problems!
	//
	if ( flags & SEC_FLAGS_DUMMY) 
		goto delete_it;

	//
	// See if there will be any consistancy problems, that the 
	// user may not be happy about..
	//
	if (portfolio_checkremove(name) == -1)
		return;

	prices->extract(NULL,NULL,&n_prices);
	if (n_prices!=0) {
		if (DialogYesno("Valid prices have been found for this security","Abort","Delete Prices")) 
			return;
	}

	if (portfolio_remove(name) == -1)
		return;

	//
	// The price class has a remove function because all the data
	// is self contained.
	//
	prices->remove();

delete_it:
	//
	// Remove this security node from the main stored list for
 	// securities.
	//
	((class list_node *) this)->remove();

        main_sec_list->set(SECLST_FLAGS_STALE);
	return;
}

//
// SecurityLoad record
//
// Provides a generic record to security class member mapping function.
// In general classes that are based on the security class can call
// this function with an uninitialized member of that class.
//
// INPUTS :	record		The record to be read and processed
// RETURNS:	0 if record mapped correctly otherwise -1.
//
int
security::SecurityLoad(class record *rec)
{
	char buffer[SECNAMELEN];		// buffer for name of security
	char *string;
	char *type;
	int iter;
	
	string=rec->line();
	strncpy(buffer,string,SECNAMELEN-1);
	buffer[SECNAMELEN-1]='\0';
	//
	// Clear any flags that may get or'd on
	//
	flags &= ~(SEC_FLAGS_TRADED|SEC_FLAGS_NOTTRADED|
		   SEC_FLAGS_MANUAL|SEC_FLAGS_AUTO);
	//
	// Look for a version 2 file
	//
	if ( !strcmp(rec->version(),"VERSION:2") ) {
		type=rec->line();
		if ( !strcmp(type,"WARRANT") )
			type_flags=SEC_TYPE_WARRANT;
		else if ( !strcmp(type,"SHARE") )
			type_flags=SEC_TYPE_SHARE;
		else {
			printf("Unknown security type (%s)\n",type);
			exit(1);
		}


		//
		// Don't support classes....at least yet..
		//
		type=rec->line();

		type=rec->line();
		if ( !strcmp("SHARE_TRADED",type) )
			flags |= SEC_FLAGS_TRADED;
		else 
			flags |= SEC_FLAGS_NOTTRADED;
			
		type=rec->line();
		if (!strcmp("SHARE_UPDATE_AUTO",type))
			flags |= SEC_FLAGS_AUTO;	
		if (!strcmp("SHARE_UPDATE_MANUAL",type))
			flags |= SEC_FLAGS_MANUAL;
		
		string=rec->line();
        	alias_count=atoi(string);
		if (alias_count) {
			aliases=(char **)malloc(100 * sizeof(char *));
			for (iter=0;iter<alias_count;iter++ ) {
       	         		string=rec->line();
		                aliases[iter]=(char *)malloc(strlen(string)+1);
				strcpy(aliases[iter],string);
			}
		}
		goto done;
	}

	type=rec->line();
	if ( !strcmp(type,"Warrant") ) 
		type_flags=SEC_TYPE_WARRANT;
	else if ( !strcmp(type,"Share") ) 
		type_flags=SEC_TYPE_SHARE;
	else if ( !strcmp(type,"Unit Trust") ) 
		type_flags=SEC_TYPE_UNITTRUST;
	else {
		printf("Share = %s\n",string);
		printf("Unknown security type (%s)\n",type);
		exit(1);
	}

        aliases=(char **)malloc(100 * sizeof(char *));
	alias_count=0;
	while (rec->skipover("__ALIAS:") ) {
		string=rec->line();
                aliases[alias_count]=(char *)malloc(strlen(string)+1);
                strcpy(aliases[alias_count],string);
		alias_count++;
	}
	if (!alias_count) {
		free(aliases);
		aliases=NULL;
	} else
		aliases[alias_count]=NULL;

	
	//
	// Flags are separated by '|' and there must be a '|' at the end
	// of the line...
	//
	rec->compile("|");
	while ( (string=rec->match()) != NULL) {
		if (!strcmp("traded|",string))
			flags |= SEC_FLAGS_TRADED;
		else if (!strcmp("nottraded|",string ))
			flags |= SEC_FLAGS_NOTTRADED;
		else if (!strcmp("update_manual|",string))
			flags |= SEC_FLAGS_MANUAL;
		else if (!strcmp("update_auto|",string))
			flags |= SEC_FLAGS_AUTO;	
	}
done:
	strcpy(name,buffer);
	//
	// Cannot already be on the list of active securities, so
	// add it now.
	//
	add();
	return 0;
}


//
// security_load unused_ptr rec_ptr
//
// Call back function for loading a security
//
// INPUTS :	rec_ptr 	record pointer for where data is to be
//				loaded from
// RETURNS:	None
//
static void
security_load(char *,va_list ap)
{
	security *newsec=new security;
	class record *rec;

        rec=va_arg(ap,class record *);

        if (newsec->SecurityLoad(rec)== -1 )
                delete newsec;
}


//
// SecurityIslocked
//
// Returns the state of the securities semaphore
//
// INPUTS :	None
// RETURNS:	TRUE if the semaphore is locked, else FALSE
//
int
security::SecurityIslocked()
{
	return (edited.SemaphoreIslocked() );
}


//
// SecuritySetname name
//
// Sets the name of a security.  EXTREME CARE must be exercised
// with this function - generally it should only be used when
// temporarily setting the name of a security.  There are no 
// checks.
//
// INPUTS :	Name		The new name
// RETURNS:	None
//
void
security::SecuritySetname(char *newname)
{
	strcpy(name,newname);
}

//
// save_security record
//
// Provides a generic security class member to record mapping function.
//
// INPUTS :	record		The record to be created from the security
//				data.
// RETURNS:	None
//
void 
security::save_record(class record *rec)
{
	int iter;
	char *type;

	if (type_flags==SEC_TYPE_GROUP) {
		rec->create("");
		return;
	}

	type=typename();

	rec->create(name);
	rec->append("\n");
	rec->append(type);
	rec->append("\n");

	for(iter=0;iter<alias_count;iter++) {
		rec->append("__ALIAS:");
		rec->append(aliases[iter]);
		rec->append("\n");
	}

	if (flags & SEC_FLAGS_TRADED)	
		rec->append("traded|");
	if (flags & SEC_FLAGS_NOTTRADED)	
		rec->append("nottraded|");

	if (flags & SEC_FLAGS_MANUAL)	
		rec->append("update_manual|");
	if (flags & SEC_FLAGS_AUTO)	
		rec->append("update_auto|");
}


//
// These save casts
//
class security *
security_list::head()
{
	return( LISTREMAP(security,slist,head,this) );
}

class security *
security::next()
{
	return( LISTREMAP(security,slist_node,next,this) );
}


//
// Constructor function.
//
security::security()
{
	name[0]='\0';				// no name until set otherwise
	aliases=NULL;
	alias_count=0;
	flags=SEC_FLAGS_TRADED|SEC_FLAGS_AUTO;
	type_flags=SEC_TYPE_SHARE;
	prices=new price(this);

}

security::security(unsigned long initflags)
{
	name[0]='\0';				// no name until set otherwise
	aliases=NULL;
	alias_count=0;
	type_flags=SEC_TYPE_SHARE;
	flags=initflags;
        prices=new price(this);
}

//
// Destructor function.
//
security::~security()
{
}


//
//=============================================================================
//
// Functions for the security_list class
//
//=============================================================================
//

//
// sec_compare sec1 sec2
//
// Call back for ordered list node insertions.
//
// INPUTS :	sec1,sec2	The two securities that are to be compared
// RETURNS:	-1,0,1		On basis of the two names of the securities.
//
//
security_list::compare(list_node *sec1,list_node *sec2) 
{
	char *name1,*name2;

	((security *)sec1)->SecurityExtract(&name1,NULL,NULL,NULL,NULL);
	((security *)sec2)->SecurityExtract(&name2,NULL,NULL,NULL,NULL);

	return(alpha_strcmp(name1,name2));
}

//
// insert_order new_security
//
// The security list is a dslist, but before anything can be
// added we must invalidate any cached name lists that may have
// been build by extract().
//
void
security_list::insert_order(security *sec)
{
	main_sec_list->flags|=SECLST_FLAGS_STALE;
	((dslist *)main_sec_list)->insert_order(sec);
}


//
// security_list::set newflags
//
// Sets flags on a security list. 
//
//
// INPUTS :	newflags	The new flags value
// RETURNS:	None
void
security_list::set(unsigned long newflags)
{
	flags = newflags;
}

//
// version
//
// Replace the standard slist version function with one which will
// write the version id for security lists that is being currently
// used.
//
// INPUTS :	None
// RETURNS:	Pointer to version id string
//
char *
security_list::version()
{
	return("VERSION:3");
}

//
// find_security secname
//
// Finds the (explicitly) named security and returns a pointer to the
// appropriate member of the security class.
//
// INPUTS :	secname			The name of the security
// RETURNS:	Pointer the security member or NULL if not found.
//
class security *
security_list::find_security(char *secname)
{
	class security *ptr;

        //
        // Now search for the named security..
        //
        ptr=head();
        while (!eol(ptr)) {
                if ( !(strcmp(secname,ptr->name) ) )
                        return(ptr);
                ptr=ptr->next();
        }
        return(NULL);
}

//
// SecuritylistFindname secname
//
// Finds the (explicitly) named security using either the key names
// or aliases and returns a pointer to the appropriate member of the
// security class.
//
// INPUTS :	secname			The name of the security
// RETURNS:	Pointer the security member or NULL if not found.
//
class security *
security_list::SecuritylistFindname(char *secname)
{
	class security *ptr;
	int i;

        //
        // Search for the named security..
        //
        ptr=head();
        while (!eol(ptr)) {
                if ( !(strcmp(secname,ptr->name) ) )
                        return(ptr);
		for(i=0;i<ptr->alias_count;i++)
			if ( !(strcmp(secname,ptr->aliases[i])) )
                        	return(ptr);
                ptr=ptr->next();
        }
        return(NULL);
}

//
// extract 
//
// Extracts from a security list the array of names of current securities
//
// INPUTS :	None
// RETURNS:	Pointer an array of pointers to names of securities.  The
// 		final element in the array will be NULL.
//
char **
security_list::extract()
{
	class security *ptr;
	char **nameptr;
	static char **namelist=NULL;
	int number=length();
	dcursortype_e cursor;

	cursor=DialogCursor(DCURSOR_BUSY);
	if ( (flags&SECLST_FLAGS_STALE) || (namelist==NULL) ) {
		if (namelist!=NULL)
			freestringlist(namelist);
		namelist=(char **)malloc(sizeof(char *)*(number+1));

		nameptr=namelist;

		ptr=head();
		while (!eol(ptr)) {
			*nameptr =(char *)malloc(strlen(ptr->name)+1);
			strcpy(*nameptr,ptr->name);
			nameptr++;
			ptr=ptr->next();
			DialogCursor(DCURSOR_BUSY);
		}
		*nameptr = NULL;

		//
		// Clear stale flag.  The cached list is back up to date
		//
		flags &= ~SECLST_FLAGS_STALE;
	}

	DialogCursor(cursor);

	return(namelist);
}


//
// select_security 
//
// Allows user selection of a security from a dialog.
//
// INPUTS :	title		A title for the dialog
// RETURNS:	Pointer to the member of the security class or NULL 
// 		if the security cannot be found.
//
class security *
security_list::select_security(char *title)
{
	char *secname;
	char **stringlist;

	stringlist=extract();
	
	secname=DialogOptionlist(title,"Securities:",stringlist);

	if (!secname)
		return(NULL);
	//
	// Now search for the named security..
	//
	return( find_security(secname) );
}



//
// Constructor function.
//
security_list::security_list()
{
	set_filename("securities",WRITE_BY_SIZE);
	add_load_vect( security_load );
	//
	// No cached name list can be valid at this point
	//
	flags=SECLST_FLAGS_STALE;
}

//
// Destructor function.
//
security_list::~security_list()
{
}

