
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: setdate.C,v 2.5 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: setdate.C,v $
// Revision 2.5  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.4  1995/06/17  20:34:43  davidsi
// - DialogSelect2dates::ButtonCallback() now checks that the start date
//   is before the end date prior to allowing the dialog to be removed with
//   the 'OK' button.
//
// Revision 2.3  1995/04/25  20:59:27  davidsi
// - Added support for a busy cursor.
// - Updated the comments.
//
// Revision 2.2  1995/04/16  20:27:02  davidsi
// - SetdateMonthyear() now requires a function that uses pointers to instances
//   of the date class.  The callback member has also been changed to handle this.
// - Added SetdateDaymonthyear() to allow days to also be set.
// - Added Display() function member to update the display rather than
//   doing it inline.
// - Inherit from the date class to make date manipulations simpler - though
//   some care is needed to keep all the values in sync.
// - Added a new constructor function that allows the date to be
//   initialized to a given date.
// - Added day checking so that a day number cannot be set beyond the total
//   number of days in the month - applies when either months or years are
//   changed.
// - Added a mechanism for tayloring the title of the dialog.  The title has
//   a default, but can be set with SetdateSettitle().
//
// Revision 2.1  1995/03/02  21:06:47  davidsi
// - Removed all dialog references - now uses the Dialog class.
// - Added a label, so that for the SetdateMonthyear() dialog
//   a user can see what the current setting is.
// - Trimmed the visible portion of the month list from 12 to 6
//   entries.
//
// Revision 2.0  1995/02/05  18:43:09  davidsi
// - Initial version.
//
//
//
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"
#include "support.h"
#include "date.h"
#include "setdate.h"


//
// Total number of months
//
#define MAX_MONTHS 12

//
// Names of the months
//
static char *months_full[] = {
    { "January" }, { "February" }, { "March" }, { "April" },
    { "May" }, { "June" }, { "July" }, { "August" }, { "September" },
    { "October" }, { "November" }, { "December" }
};



//
// Button identifiers
//
enum idents_e {
	BUT_NEXTYEAR,
	BUT_LASTYEAR,
	BUT_NEXTDAY,
	BUT_LASTDAY,
	};

//
// month_name month
// 
// Returns a month given a month number
//
// INPUTS :	month		Month number 0-11
// RETURNS:	ascii string of the req'd month.
//
static char *
month_name(int month)
{
    return(months_full[month]);
}

//
// Display
//
// Brings the display up to date.
//
// INPUTS :	None - But assumes day/month/year are all valid!
// RETURNS:	None
//
void
Setdate::Display()
{
	char buffer[256];
	int day,month,year;

	getdate(day,month,year);
	switch(displaytype) {
	case DISPLAY_MONTHYEAR:
		sprintf(buffer," %s: %s %d   ",title,month_name(month-1),year);
		d_label.DialoglabelSet(buffer);
		break;
	case DISPLAY_DAYMONTHYEAR:
		sprintf(buffer," %s: %d %s %d   ",title,day,
						month_name(month-1),year);
		d_label.DialoglabelSet(buffer);
		break;
	}
}



//
// ButtonCallback id flag
// 
// The callback for buttons within the Setdate class
//
// INPUTS :	id		The id for the button.
//		flag		The button's state
// RETURNS:	None
//
void
Setdate::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	int day,month,oldmonth,year;

	switch(id) {
	case BUT_NEXTYEAR:
		getdate(day,month,year);
		//
		// If februrary and on day 29 bring it back to 28
		//
		if (month==2 && day==29)
			day=28;
		year++;
		//
		// If we don't succed in setting the date, read back
		// the date to ensure consistancy.
		//
		if ( setdate(day,month,year) == -1 )
			getdate(day,month,year);
		break;
	case BUT_LASTYEAR:
		getdate(day,month,year);
		//
		// If februrary and on day 29 bring it back to 28
		//
		if (month==2 && day==29)
			day=28;
		year--;
		//
		// If we don't succed in setting the date, read back
		// the date to ensure consistancy.
		//
		if ( setdate(day,month,year) == -1 )
			getdate(day,month,year);
		break;
	case BUT_NEXTDAY:
		getdate(day,oldmonth,year);
		add(1);
		getdate(day,month,year);
		if (month!=oldmonth)
			d_monthlist->OptionlistinstanceSet(month_name(month-1));
		break;
	case BUT_LASTDAY:
		getdate(day,oldmonth,year);
		sub(1);
		getdate(day,month,year);
		if (month!=oldmonth)
			d_monthlist->OptionlistinstanceSet(month_name(month-1));
		break;
	}
	Display();
	if (callback)
		callback(callbackval,(class date *)this);
}


//
// OptionlistCallback (SetdateMonthchangecb) dialog this
//
// Handles the callback when a month is selected from the month list
// widget.
//
// INPUTS :	dialog		The list dialog
//		this		The callback value that is the correct this
//				value.
//
void
Setdate::OptionlistCallback(unsigned long id)
{
	char *monthstr;
	int iter;
	int day,month,year;
	int daysinmonth;

	getdate(day,month,year);
        monthstr=d_monthlist->OptionlistinstanceGet();
        for(iter=0;iter<MAX_MONTHS;iter++) {
		if (!alpha_strcmp(monthstr,month_name(iter)) ) {
			month=iter+1;
			break;
		}
	}
	//
	// If new month has less days than previous month then 
	// set the day to the number of days in the new month
	//
	setdate(1,month,year);
	daysinmonth=days_in_month();
	if (day>daysinmonth)
		day=daysinmonth;
	setdate(day,month,year);

	Display();
	if (callback!=NULL)
		callback(callbackval,(class date *)this);
}



//
// SetdateMonthyear dialog func value
//
// Provides a callback driven dialog for change year and month information.
// 
// INPUTS :	dialog		The parent dialog
//		func		The function to be called when a month/year
//				is selected
//		value		The value to be passed to func.
// RETURNS:	None
//
void
Setdate::SetdateMonthyear( Dialog d_parent,
		void (*f)(char *,class date *),
		char *cbv)
{
	int iter;
	long t;				// used to read the time
	char **strs;
	static Newoptionlist nolist = {
		12,			// width
		MAX_MONTHS/2,		// height
		NULL,			// list of items
		"",			// name of list
		"",			// name of item
		OLIST_FLAGS_NOTEXT,	// flags
		NULL,			// default font
	};
	int dd,mm,yy;

	nolist.listname="Months";

	strs=(char **)malloc(sizeof(char **)*(MAX_MONTHS+1));
	for (iter = 0; iter < 12; iter++)
		strs[iter] =month_name(iter);
	strs[MAX_MONTHS]=NULL;
	nolist.listitem=strs;

	//
	// SetdateMonthyear only allows months and years to be messed
	// with, so set day to the first day of the month.
	//
	getdate(dd,mm,yy);
	setdate(1,mm,yy);
	displaytype=DISPLAY_MONTHYEAR;

        d_vert.DialogverticalCreate(d_parent);

	d_label.DialoglabelCreate(d_vert,"");
	Display();

	d_monthlist=OptionlistCreate(d_vert,nolist,0);
        d_horiz.DialoghorizontalCreate(d_vert);

	(void)ButtonPush(d_horiz,"FORW YEAR",BUT_NEXTYEAR);
	(void)ButtonPush(d_horiz,"BACK YEAR",BUT_LASTYEAR);

	free(strs);
	callbackval=cbv;
	callback=f;
}

//
// SetdateDaymonthyear dialog func value
//
// Provides a callback driven dialog for change year and month information.
// 
// INPUTS :	dialog		The parent dialog
//		func		The function to be called when a month/year
//				is selected
//		value		The value to be passed to func.
// RETURNS:	None
//
void
Setdate::SetdateDaymonthyear( Dialog d_parent,
		void (*f)(char *,class date *),
		char *cbv)
{
	int iter;
	long t;				// used to read the time
	char **strs;
	static Newoptionlist nolist = {
		12,			// width
		MAX_MONTHS/2,		// height
		NULL,			// list of items
		"",			// name of list
		"",			// name of item
		OLIST_FLAGS_NOTEXT,	// flags
		NULL,			// default font
	};

	nolist.listname="Months";

	strs=(char **)malloc(sizeof(char **)*(MAX_MONTHS+1));
	for (iter = 0; iter < 12; iter++)
		strs[iter] =month_name(iter);
	strs[MAX_MONTHS]=NULL;
	nolist.listitem=strs;

	displaytype=DISPLAY_DAYMONTHYEAR;

        d_vert.DialogverticalCreate(d_parent);

	d_label.DialoglabelCreate(d_vert,"");
	Display();

        d_horiz2.DialoghorizontalCreate(d_vert);

	(void)ButtonPush(d_horiz2,"FORW DAY",BUT_NEXTDAY);
	(void)ButtonPush(d_horiz2,"BACK DAY",BUT_LASTDAY);

	d_monthlist=OptionlistCreate(d_vert,nolist,0);
        d_horiz.DialoghorizontalCreate(d_vert);

	(void)ButtonPush(d_horiz,"FORW YEAR",BUT_NEXTYEAR);
	(void)ButtonPush(d_horiz,"BACK YEAR",BUT_LASTYEAR);

	free(strs);
	callbackval=cbv;
	callback=f;
}

//
// SetdateSettitle title
//
// Sets an alternative title on a date dialog. This should be called
// before the dialog hierachy is realised. If its called after
// realisation the display won't be updated until the next event.
//
// INPUTS :	title		The new title
// RETURNS:	None
//
void
Setdate::SetdateSettitle(char *newtitle)
{
	free(title);
	title=(char *)malloc(strlen(newtitle)+1);
	strcpy(title,newtitle);
}


//
// Constructor
//
Setdate::Setdate()
{
#define DEFAULT_TITLE "Date"

	setdate();

	title=(char *)malloc(strlen(DEFAULT_TITLE)+1);
	strcpy(title,DEFAULT_TITLE);
	
	callbackval=0;
	callback=NULL;
}

//
// Destructor
//
Setdate::~Setdate()
{
	free(title);
}


//
///////////////////////////////////////////////////////////////////////////////
//
// Publicly available functions
//
///////////////////////////////////////////////////////////////////////////////
//


enum setdate_e {
	BUTTON_NONE,
	BUTTON_OK,
	BUTTON_CANCEL,
	};

//
// The DialogSelect2dates class provides the basic functionality
// for handling dialogs that request two dates from the user.  We
// use a class so that we can manage the button callbacks.
// 
//
class DialogSelect2dates : private Button
{
        Dialogshell d_shell;                    // top most dialog
        Dialogvertical d_mainv;
        Dialoghorizontal d_mainh;
        Dialogvertical d_sub1v;
        Dialogvertical d_sub2v;
        Dialoghorizontal d_buttonh;		// button horizontal form

	class Setdate date1;
	class Setdate date2;
	setdate_e buttonid;

private:
	//
	// This overrides the pure virtual button handler in the Button class
	//
	virtual void ButtonCallback(unsigned long id,buttonstate_e flag);
        static void date_changed(char *rthis,class date *);
public:
	int DialogSelect2datesWait();
	DialogSelect2dates(char *title,
				char *date1,date &start_date,
                                char *date2,date &end_date);
	~DialogSelect2dates();
};


//
// date_changed dateptr newdate
//
// Handles date change callbacks.
//
// INPUTS :	dateptr		The instance of the date class
//		newdate		The new date for that instance
// RETURNS:	None 
//
void 
DialogSelect2dates::date_changed(char *dateptr,class date *newdate)
{
	class date *dptr=(class date *)dateptr;
	dptr->setdate(*newdate);
}



//
// Handles button callbacks.
//
void
DialogSelect2dates::ButtonCallback(unsigned long id,buttonstate_e flag)
{
	switch(id) {
	case BUTTON_OK:
		if (date1.compare(&date2)>0) {
			DialogMessage("Start date is after the end date.");	
			return;
		}
		buttonid=BUTTON_OK;
		break;
	case BUTTON_CANCEL:
		buttonid=BUTTON_CANCEL;
		break;
	}
}


//
// DialogSelect2datesWait
//
// Waits until a OK or CANCEL button is hit
// 
// INPUTS : 	None
// RETURNS:	None
//
int
DialogSelect2dates::DialogSelect2datesWait()
{
	while (buttonid==BUTTON_NONE)
		DialogEvent();

	if (buttonid==BUTTON_OK)
		return(0);
	
	return(-1);
}

// 
// Constructor
//
// INPUTS :	title		The title for the shell dialog
// 		strdate1	The heading for the 1st date to be entered
//		start_date	The 1st date that will be modified. 
// 		strdate2	The heading for the 2nd date to be entered
//		end_date	The 2nd date that will be modified. 
//
DialogSelect2dates::DialogSelect2dates(char *title,
				char *strdate1,date &start_date,
				char *strdate2,date &end_date)
{
	dcursortype_e cursor;

	cursor=DialogCursor( DCURSOR_WAIT );

	d_shell.DialogshellCreate(title);

	d_mainv.DialogverticalCreate(d_shell);

	d_mainh.DialoghorizontalCreate(d_mainv);

	d_sub1v.DialogverticalCreate(d_mainh);
	date1.setdate(start_date);
	date1.SetdateSettitle(strdate1);
	date1.SetdateDaymonthyear(d_sub1v,&DialogSelect2dates::date_changed,
					(char *)&start_date);

	d_sub2v.DialogverticalCreate(d_mainh);
	date2.setdate(end_date);
	date2.SetdateSettitle(strdate2);
	date2.SetdateDaymonthyear(d_sub2v,&DialogSelect2dates::date_changed,
					(char *)&end_date);

	d_buttonh.DialoghorizontalCreate(d_mainv);

	buttonid=BUTTON_NONE;
	(void)ButtonPush(d_buttonh," OK ",BUTTON_OK);
	(void)ButtonPush(d_buttonh," CANCEL ",BUTTON_CANCEL);

	d_shell.DialogRealize();
	DialogCursor( cursor );
}


// 
// Destructor
//
DialogSelect2dates::~DialogSelect2dates()
{
}


//
// DialogSelect2dates title date1 start_date date2 end_date
//
// Allows two dates to be modified by the user.
//
// INPUTS :	title		The title for the shell dialog
// 		strdate1	The heading for the 1st date to be entered
//		start_date	The 1st date that will be modified. 
// 		strdate2	The heading for the 2nd date to be entered
//		end_date	The 2nd date that will be modified. 
// RETURNS:	-1 if cancel was hit else 0
//
int 
DialogSelect2dates(char *title,
			char *date1,date &start_date,
			char *date2,date &end_date)
{
	class DialogSelect2dates *dateset;
	int ret;

	dateset=new class DialogSelect2dates(title,
						date1,start_date,
						date2,end_date);
	ret=dateset->DialogSelect2datesWait();
	delete dateset;
	
	return(ret);
}
