
/*
 *    X SHARE MANAGER (XSM)
 *
 *    Copyright (C) 1995  David Simons
 *    Copyright 1993, David Koblas (koblas@netcom.com)  
 *    Portions by David Brooks of the OSF
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * $Id: d_base.c,v 2.7 1995/05/29 19:00:44 davidsi Exp $
 *
 * $Log: d_base.c,v $
 * Revision 2.7  1995/05/29  19:00:44  davidsi
 * - BUG FIX.  In dialog_root() the memory used for the root dialog must be
 *   cleared out.  Usually dialog_init() performs this task but for the root
 *   dialog, the dialog_init() function is not called.
 *
 * Revision 2.6  1995/05/19  18:58:57  davidsi
 * - Added dialog_setfont() to set fonts so they can be set independently
 *   of any initialization. dialog_root() no longer manages any font
 *   initialization.
 *
 * Revision 2.5  1995/04/25  22:19:20  davidsi
 * - Moved bitmaps within source file.
 *
 * Revision 2.4  1995/04/25  21:04:52  davidsi
 * - Moved dialog_updatedisplay(d) from d_draw.c to this file.
 * - Support for different cursor types.
 * - BUG FIX. dialog_destroy2() was writing NULL into the parent's child
 *   pointer, but in many cases this will result in the siblings being
 *   lost to the parent (and being untraceable) so make the parent point
 *   to the next sibling when a dialog is removed.
 *
 * Revision 2.3  1995/04/16  19:53:41  davidsi
 * - Removed dialog_wait() as it was no longer being used.
 *
 * Revision 2.2  1995/02/25  21:11:03  davidsi
 * - Removed references to dialog_panic().  This is no longer supported.
 * - Added wid_selbox_remove - previously in widget.c (no longer available).
 * - Added handle_event - previously in dialogs0.c (no longer available).
 *
 * Revision 2.1  1995/02/05  17:52:18  davidsi
 * - BUG FIX. When destroying a shell dialog, first unmanage it.  This
 *   ensures it disappears from the display regardless of whether
 *   the XtDestroyWidget stalls.
 *
 * Revision 2.0  1995/01/19  22:17:18  davidsi
 * - Modified dialog_root_r() so that it waits for the dialog to be
 *  realised but does not enter the main loop.
 * - Added dialog_mainloop() to allow the calling app. to decide
 *   when the main loop should execute.  The intention here is to allow
 *   apps to get something on the display while the continue to
 *   initialize.
 *
 * Revision 1.3  1994/11/05  21:18:26  davidsi
 * - dialog_init() did not declare the types of its arguments.
 *
 * Revision 1.2  1993/01/31  18:45:57  davidsi
 * - When a font cannot be loaded write the message to stderr and state
 *   which font is giving the problems.
 *
 * Revision 1.1  1993/01/17  16:21:21  davidsi
 * Initial revision
 *
 */

/* 
 * Sections of code marked C001 are:
 *
 * Copyright 1993, David Koblas (koblas@netcom.com)  
 *
 * Permission to use, copy, modify, and to distribute this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.  There is no
 * representations about the suitability of this software for
 * any purpose.  this software is provided "as is" without express
 * or implied warranty.
 */



/*
 * Include Files
 */
#include <stdio.h>
#include <varargs.h>

#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#include <Xm/MainW.h>

#include <Xm/Xm.h>

#define IN_DIALOGS

#include "d_base.h"
#include "d_form.h"


d_root *base_node;		/* the start of the universe!! */

static void initCursors(Widget);
static Cursor watchCursor;
static Cursor normalCursor;


/*
 * dialog_destroy2 tree flag
 *
 * Destroys a dialog tree - see dialog_destroy
 *
 * INPUTS :	tree		The root of the tree to destroy.
 *		flag		TRUE if siblings of the root of the specified
 *				tree are to be destroyed, else FALSE
 * RETURNS:	None
 */
static void
dialog_destroy2(tree,flag)
dialog tree;
int flag;
{
	if (tree->child) 
		dialog_destroy2(tree->child,TRUE);

	/* 
	 * May need to break the parents link, but we cannot
	 * simply patch it to NULL because the parent would
	 * then lose all the siblings as well.  So point the
	 * parent at the next sibling
 	 */
	if (tree->parent)
		tree->parent->child=tree->sibling_f;	

	/*
	 * Remove all siblings
	 */
	if (flag) {
		while (tree->sibling_f!=tree) {
			dialog_destroy2(tree->sibling_f,FALSE);
		}
	}

	/*
	 * detach from the sibling list
	 */
	tree->sibling_f->sibling_b=tree->sibling_b;
	tree->sibling_b->sibling_f=tree->sibling_f;
	if (tree->destroy)
		tree->destroy(tree);
	free(tree);
}

/*
 * dialog_destroy tree
 *
 * Destroys the entire dialog tree starting at the root node given
 * by the tree argument.
 *
 * INPUTS :	tree		Root of tree to remove
 * RETURNS:	None
 */
void
dialog_destroy(tree)
dialog tree;
{
	dialog_destroy2(tree,FALSE);
	XtUnmanageChild(tree->widget);
	XtDestroyWidget(tree->widget);
}


/*
 * dialog_add parent child
 *
 * Adds a child dialog to a parent dialog maintaining all lists as
 * necessary. 
 *
 * INPUTS :	parent		The parent dialog
 *		child		The child dialog to add.
 * RETURNS:	None
 */
static void
dialog_add(parent,child)
dialog parent,child;
{
	if (parent==NULL)
		return;
	/*
	 * Check for no children
	 */
	if (parent->child==NULL) {
		child->sibling_f=child;
		child->sibling_b=child;
	} else {
		dialog curchild=parent->child;
		child->sibling_f=curchild;
		child->sibling_b=curchild->sibling_b;
		curchild->sibling_b->sibling_f=child;
		curchild->sibling_b=child;
	}
	parent->child=child;
	child->parent=parent;
}


/*
 * dialog_realise2 tree flag
 *
 * Realises a dialog tree - see dialog_realise
 *
 * INPUTS :	tree		The root of the tree to destroy.
 *		flag		TRUE if siblings of the root of the specified
 *				tree are to be realises, else FALSE
 * RETURNS:	None
 */
static void
dialog_realise2(tree,flag)
dialog tree;
int flag;
{
	dialog stree;

	stree=tree;
	do {
		if (stree->child) 
			dialog_realise2(stree->child,TRUE);
		if (stree->realise)
			stree->realise(stree);
		if (!flag)
			return;
		stree=stree->sibling_f;
	} while (stree!=tree);
}



/*
 * dialog_realise tree
 *
 * Realises the entire dialog tree starting at the root node given
 * by the tree argument.
 *
 * INPUTS :	tree		Root of tree to realise
 * RETURNS:	None
 */
void
dialog_realise(tree)
dialog tree;
{
	dialog_realise2(tree,FALSE);
}


/*
 * dialog_waitformap dialog
 *
 * This procedure is by David Brooks of the OSF and appears in
 * the Motif FAQ list.  It has been renamed.
 *
 * Wait for a dialog to be realized.  This is ideal if you need to
 * open a dialog and then write to it, without waiting for a callback.
 *
 * INPUTS :	dialog		The dialog to wait upon
 * RETURNS:	None
 */
void
dialog_waitformap(d)
dialog d;
{
	Widget w;
	Widget diashell, topshell;
	Window diawindow, topwindow;
	Display *dpy;
	XWindowAttributes xwa;
	XEvent event;
	XtAppContext cxt;
	XWindowAttributes wa;
    
	w=d->widget;

	for (diashell=w; !XtIsShell(diashell); diashell = XtParent(diashell)) ;
		for (topshell = diashell;
			!XtIsTopLevelShell(topshell);
			topshell = XtParent(topshell));
    
	if (XtIsRealized(diashell) && XtIsRealized(topshell)) {
		dpy = XtDisplay(topshell);
		diawindow = XtWindow(diashell);
		topwindow = XtWindow(topshell);
		cxt = XtWidgetToApplicationContext(diashell);
	
		/* Wait for the dialog to be mapped. */
		while (XGetWindowAttributes(dpy, diawindow, &xwa),
		       xwa.map_state != IsViewable) {
	    
			/*
			 * unless the primary is (or becomes) unviewable
			 * or unmapped, it's probably iconified, and nothing
			 * will happen.
			 */
	    
			if (XGetWindowAttributes(dpy, topwindow, &xwa),
			    xwa.map_state != IsViewable)
				break;
		       /*
			* At this stage, we are guaranteed there will be
			* an event of some kind.  Beware; we are presumably
			* in a callback, so this can recurse.
			*/
			XtAppNextEvent(cxt, &event);
			XtDispatchEvent(&event);
		}
	}
    
	/*
	 * The next XSync() will get an expose event if the dialog
	 * was unmapped. 
	 */
	XmUpdateDisplay(topshell);
}



/*
 * dialog_base dialog
 *
 * Locates the base dialog node of any dialog, the base is typically
 * the shell dialog.
 *
 * INPUTS :	dialog		The dialog to search back from
 * RETURNS:	The base node.
 */
dialog
dialog_base(d)
dialog d;
{
	while (d->parent!=NULL && !(d->privateflags&DIALOG_BASENODE))
		d=d->parent;
	return(d);
}



/*
 * dialog_root_r dialog
 *
 * A special realisation function for the root node.
 *
 * INPUTS :	dialog		The root dialog
 * RETURNS:	None
 */
static void
dialog_root_r(dr)
d_root *dr;
{
	XtManageChild(dr->widget);
	XtRealizeWidget(dr->main_app_shell);

	(void)dialog_waitformap((dialog)dr);

	return;
}


/*
 * dialog_mainloop
 *
 * Handles (forever) the input events.
 *
 * INPUTS :	dialog		The root dialog
 * RETURNS:	None
 */
void
dialog_mainloop(d)
dialog d;
{
	d_root *dr=(d_root *)d;

	XtAppMainLoop(dr->app_context);
}



/*
 * dialog_root prefs
 *
 * Creates a root dialog.  This is the dialog that all others will
 * be children of.
 *
 * INPUTS :	prefs		The preferences to be used by the tree
 * RETURNS:	None
 */
dialog
dialog_root(np)
new_prefs *np;
{
	char *option;              /* options from Xdefaults */
	d_root *dr;
	Arg al[10];                /* arg list */
	int ac;

	base_node=dr=(d_root *)malloc(sizeof(d_root));
	bzero(base_node,sizeof(d_root));

	/*
	 * When this dialog gets realised it starts managing!
	 */
	dr->realise=dialog_root_r;
	XtToolkitInitialize ();      /* Initialize toolkit and open display.  */
	dr->app_context=XtCreateApplicationContext();
	dr->display=XtOpenDisplay(dr->app_context,NULL,np->argv[0],"name",
					NULL, 0, &(np->argc), np->argv);
	if (!dr->display) {
		XtWarning ("This program requires X to be running");
		exit (0);
	}
	dr->main_app_shell = XtAppCreateShell (np->name, "name1",
			applicationShellWidgetClass, dr->display,NULL,0);


	ac = 0;
	XtSetArg (al[ac], XmNscrollingPolicy, XmAPPLICATION_DEFINED);  ac++;
	dr->widget=XmCreateMainWindow(dr->main_app_shell, "main", al, ac);
	initCursors(dr->widget);
	return((dialog)dr);
}


/*
 * dialog_setfonts fontspec
 *
 *
 * INPUTS :	fontspec		The fonts to be used
 * RETURNS:	None
 */
void
dialog_setfonts(np)
font_specifier *np;
{
	int iter;
	char buffer[1024];
	XrmDatabase database=NULL;

	if (np->npfont==0)  {
		fprintf(stderr,"Must be at least one defined font\n");
		exit(1);
	}

	base_node->nfonts=np->npfont;
	for(iter=0;iter<np->npfont;iter++) {
		base_node->fname[iter]=(char *)malloc(strlen(np->fname[iter])+1);
		strcpy(base_node->fname[iter],np->fname[iter]);
		base_node->pfontinfo[iter]=XLoadQueryFont(base_node->display,
							   np->pfont[iter]);
		if (base_node->pfontinfo[iter]!=NULL)
			base_node->pfontlist[iter]=XmFontListCreate(
					base_node->pfontinfo[iter],
					np->fname[iter]);
		else {
			fprintf(stderr,
			  "Cannot obtain desired font(%s)\n",np->pfont[iter]);
			exit(1);
		}
	}
}



#ifdef ALLVISIBLE
/*
 * fname_2_fontlist name
 *
 * Maps a font name back to the original font list structure
 * 
 * INPUTS :	name		The name of the font
 * RETURNS:	A pointer to the font list structure
 */
XmFontList
fname_2_fontlist(name)
char *name;
{
	int iter;

	for(iter=0;iter<base_node->nfonts;iter++) 
		if (!strcmp(base_node->fname[iter],name)) 
			return(base_node->pfontlist[iter]);
	printf("cannot find font %s\n",name);
	return(NULL);
}
#endif




#ifdef ALLVISIBLE
/*
 * fname_2_fontinfo name
 *
 * Maps a font name back to the original font info structure
 * 
 * INPUTS :	name		The name of the font
 * RETURNS:	A pointer to the font info structure
 */
XFontStruct *
fname_2_fontinfo(name)
char *name;
{
	int iter;

	for(iter=0;iter<base_node->nfonts;iter++) 
		if (!strcmp(base_node->fname[iter],name)) 
			return(base_node->pfontinfo[iter]);
	printf("cannot find font %s\n",name);
	return(base_node->pfontinfo[0]);
}
#endif



#ifdef ALLVISIBLE
/*
 * dialog_position position arglist argcount
 *
 * Adds arguments to an argument list to ensure a dialog is correctly
 * positioned.  Though not static this function is private to the
 * dialog library.
 *
 * It is the callers responsibility to ensure there is enough space
 * in the argument array. 
 *
 * INPUTS :	position	A position structure
 * 		arglist		The Argument list
 *		argcount	Pntr to a value giving the initial
 *				num of args in the arglist.
 * RETURNS:	argcount	Pntr to a value given the final num of
 *				args in the list.
 */
dialog_position(p,al,ac)
position *p;
Arg al[];
int *ac;
{
	switch(p->flags) {
	case POSITION_PERC:
		XtSetArg(al[*ac],XmNrightAttachment,XmATTACH_POSITION);
		XtSetArg(al[*ac+1],XmNrightPosition,p->right);
		XtSetArg(al[*ac+2],XmNleftAttachment,XmATTACH_POSITION);
		XtSetArg(al[*ac+3],XmNleftPosition,p->left);
		XtSetArg(al[*ac+4],XmNtopAttachment,XmATTACH_POSITION);
		XtSetArg(al[*ac+5],XmNtopPosition,p->top);
		XtSetArg(al[*ac+6],XmNbottomAttachment,XmATTACH_POSITION);
		XtSetArg(al[*ac+7],XmNbottomPosition,p->bottom);
		(*ac)+=8;
		break;
	default:
		printf("bad siwtch in dialog_position");
		exit(1);
	}
}
#endif



#ifdef ALLVISIBLE
/*
 * dialog_default_r dialog 
 *
 * Default realisation function.
 *
 * INPUTS :	dialog		The dialog to realise
 * RETURNS:	None
 */
void
dialog_default_r(d)
dialog d;
{
	XtManageChild(d->widget);
}
#endif



#ifdef ALLVISIBLE
/*
 * dialog_default_d dialog 
 *
 * Default destruction function.
 *
 * INPUTS :	dialog		The dialog to destroy
 * RETURNS:	None
 */
void
dialog_default_d(d)
dialog d;
{
	if (d->widget==0) {
		fprintf(stderr,
			"PANIC:Widget is not defined on dialog 0x%x\n",d);
		exit(1);
	}
}

#endif


/*
 * dialog_setmodal dialog flag 
 *
 * Sets whether a dialog is modal or not.
 *
 * INPUTS :	dialog		The dialog whose modality is to be changed.
 *		flag		TRUE/FALSE if dialog is to be modal or not
 * RETURNS:	None
 */
void
dialog_setmodal(d,flag)
dialog d;
int flag;
{
	if (d->widget==0) {
		fprintf(stderr,
			"PANIC:Widget is not defined on dialog 0x%x\n",d);
		exit(1);
	}

	if (flag)
		XtAddGrab(d->widget,TRUE,FALSE);
	else 
		XtRemoveGrab(d->widget);
}




#ifdef ALLVISIBLE
/*
 * dialog_init parent size
 *
 * Initialises a dialog structure.  The dialog will be cleared and the
 * default realisation and destruction functions set.  It will also
 * be linked to the parent and any siblings (as necessary).  This
 * function is private to the library.
 *
 * INPUTS :	parent		The parent of the dialog to set.
 *		size		The size of dialog structure needed
 * RETURNS:	A pointer to the newly created dialog structure
 */
dialog
dialog_init(parent,size)
dialog parent;
int size;
{
	dialog d;

	d=(dialog)malloc(size);
	bzero(d,size);
	d->child=NULL;
	d->parent=NULL;
	d->sibling_f=d;
	d->sibling_b=d;
	d->destroy=dialog_default_d;
	d->realise=dialog_default_r;
	dialog_add(parent,d);

	return(d);
}
#endif


/*
 * handle_event
 *
 * INPUTS :	None
 * RETURNS:	None
 */
void
handle_event()
{
	XEvent xe;

	XtAppNextEvent(base_node->app_context,&xe);
	XtDispatchEvent (&xe);
}



/*
 * wid_selbox_remove widget,num,va_alist
 *
 * Unmanges a number of children on a selection box widget.
 *
 * INPUTS:	widget		The widget where the data is to be stored.
 *		num		Number of values to store
 *		va_alist	The childrem (XmDIALOG_XXXX) to unmanage
 * RETURNS:	None
 */
#ifdef ALLVISIBLE
int
wid_selbox_remove(w,num,va_alist)
Widget w;
int num;
va_dcl
{
	va_list ap;
	long value;
	Widget kid[10];     /* arg list */
	int i;

	if (num>10)
		fprintf(stderr,
		 "SANITY:wid_selbox_remove can only remove upto 10 children\n");
	va_start(ap);
	for(i=0;i<num;i++) {
		value = va_arg(ap, long);
		kid[i] = (Widget)XmSelectionBoxGetChild(w, value);
	}
	va_end(ap);
	XtUnmanageChildren(kid, num);
}
#endif



/*
 * dialog_updatedisplay dialog
 *
 * Wait for a dialog to be realized.  This is ideal if you need to
 * open a dialog and then write to it, without waiting for a callback.
 *
 * INPUTS :	dialog		The dialog to wait upon
 * RETURNS:	None
 */
void
dialog_updatedisplay(d)
dialog d;
{
	Widget w;
	Widget shell;
	Widget topshell;
	XEvent event;
    
	w=d->widget;
	XFlush(base_node->display);

	for (shell = w;!XtIsShell(shell);shell=XtParent(shell))
		;
	for( topshell = shell; !XtIsTopLevelShell(topshell);
	     topshell = XtParent(topshell))
		;
    
    	/*
	 * The next XSync() will get an expose event if the
	 * dialog was unmapped. 
	 */
	XmUpdateDisplay(topshell);
	
	while(XCheckMaskEvent(base_node->display,Expose|GraphicsExpose|
				NoExpose|VisibilityNotify|ResizeRequest|
				CirculateRequest|CirculateNotify,&event)){
		XtDispatchEvent(&event);
	}
}


/*
 * C001v 
 */

#define wait1_width 24
#define wait1_height 24
static unsigned char wait1_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3e, 0x00,
  0x80, 0xcf, 0x00, 0xc0, 0x0f, 0x01, 0xe0, 0x0f, 0x02, 0xf0, 0x0f, 0x04,
  0xf0, 0x0f, 0x04, 0xf8, 0x0f, 0x08, 0xf8, 0x0f, 0x08, 0xf8, 0xff, 0x0f,
  0x08, 0xf8, 0x0f, 0x08, 0xf8, 0x0f, 0x10, 0xf8, 0x07, 0x10, 0xf8, 0x07,
  0x20, 0xf8, 0x03, 0x40, 0xf8, 0x01, 0x80, 0xf9, 0x00, 0x00, 0x3e, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };

#define wait2_width 24
#define wait2_height 24
static unsigned char wait2_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3e, 0x00,
  0x80, 0xff, 0x00, 0xc0, 0xff, 0x01, 0xe0, 0xff, 0x03, 0x90, 0xff, 0x04,
  0x10, 0x7f, 0x04, 0x08, 0x3e, 0x08, 0x08, 0x1c, 0x08, 0x08, 0x08, 0x08,
  0x08, 0x1c, 0x08, 0x08, 0x3e, 0x08, 0x10, 0x7f, 0x04, 0x90, 0xff, 0x04,
  0xe0, 0xff, 0x03, 0xc0, 0xff, 0x01, 0x80, 0xff, 0x00, 0x00, 0x3e, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };

#define wait3_width 24
#define wait3_height 24
static unsigned char wait3_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3e, 0x00,
  0x80, 0xf9, 0x00, 0x40, 0xf8, 0x01, 0x20, 0xf8, 0x03, 0x10, 0xf8, 0x07,
  0x10, 0xf8, 0x07, 0x08, 0xf8, 0x0f, 0x08, 0xf8, 0x0f, 0xf8, 0xff, 0x0f,
  0xf8, 0x0f, 0x08, 0xf8, 0x0f, 0x08, 0xf0, 0x0f, 0x04, 0xf0, 0x0f, 0x04,
  0xe0, 0x0f, 0x02, 0xc0, 0x0f, 0x01, 0x80, 0xcf, 0x00, 0x00, 0x3e, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };

#define wait4_width 24
#define wait4_height 24
static unsigned char wait4_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3e, 0x00,
  0x80, 0xc1, 0x00, 0x40, 0x00, 0x01, 0x60, 0x00, 0x03, 0xf0, 0x80, 0x07,
  0xf0, 0xc1, 0x07, 0xf8, 0xe3, 0x0f, 0xf8, 0xf7, 0x0f, 0xf8, 0xff, 0x0f,
  0xf8, 0xf7, 0x0f, 0xf8, 0xe3, 0x0f, 0xf0, 0xc1, 0x07, 0xf0, 0x80, 0x07,
  0x60, 0x00, 0x03, 0x40, 0x00, 0x01, 0x80, 0xc1, 0x00, 0x00, 0x3e, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };

static struct {
	Cursor		cursor;
	int		width, height;
	unsigned char	*bits;
} cursorInfo[] = {
	{ None, wait1_width, wait1_height, wait1_bits },
	{ None, wait2_width, wait2_height, wait2_bits },
	{ None, wait3_width, wait3_height, wait3_bits },
	{ None, wait4_width, wait4_height, wait4_bits },
};

/*
 * C001^ 
 */


/*
 * dialog_setcursor2 dialog cursor
 *
 * Set a cursor for a dialog and EVERYTHING that can be found in
 * the same hierarchy.  Best called via dialog_setcursor().
 *
 * INPUTS :	dialog		The dialog to which the cursor applies.
 *		cursortype	The cursor to set
 * RETURNS:	None
 */
static void
dialog_setcursor2(d,cursortype)
dialog d;
enum cursortype_e cursortype;
{
	dialog tree;
	Widget w;
	Window window;
	static int counter=0;
	Cursor c;

	switch (cursortype){
	case CURSOR_BUSY:
		counter++;
		if (counter>= XtNumber(cursorInfo))
			counter=0;
		c=cursorInfo[counter].cursor;
		break;
	case CURSOR_PROP:
		c=cursorInfo[counter].cursor;
		break;
	case CURSOR_NORMAL:
		c=normalCursor;
		break;
	case CURSOR_WAIT:
		c=watchCursor;
		break;
	}

	/*
	 * Process all siblings
	 */
	tree=d;
	do {
		if (tree->child) {
			if (cursortype==CURSOR_BUSY)
				dialog_setcursor2(tree->child,CURSOR_PROP);
			else
				dialog_setcursor2(tree->child,cursortype);
		}


		if ( w=(tree->widget) ) {
			window=XtWindow(w);
			if (window)  
				XDefineCursor(base_node->display,window, c);
		} 
    		tree=tree->sibling_f;
	} while (tree && tree!=d);

}


/*
 * dialog_setcursor2 dialog cursor
 *
 * Set a cursor for a dialog and EVERYTHING that can be found in
 * the same hierarchy.  Best called via dialog_setcursor().
 *
 * INPUTS :	dialog		The dialog to which the cursor applies.
 *		cursortype	The cursor to set
 * RETURNS:	None
 */
void
dialog_setcursor(d,cursortype) 
dialog d;
enum cursortype_e cursortype;
{
	dialog_setcursor2(d,cursortype);
	XFlush(base_node->display);
}

/*
 * C001v 
 */


/*
 *  Create a XImage
 */
static XImage *
NewXImage(dpy, visual, depth, width, height)
Display *dpy;
Visual *visual;
int depth;
int width;
int height;
{
	XImage	*xim;
	int	pad;

	if (depth > 16)
		pad = 32;
	else if (depth > 8)
		pad = 16;
	else
		pad = 8;

	xim = XCreateImage(dpy,visual,depth,ZPixmap,0,NULL,width,height,pad,0);	

	if (xim == NULL)
		return NULL;

	xim->data = (char*)XtMalloc(xim->bytes_per_line * height);

	if (xim->data == NULL) {
		XDestroyImage(xim);
		xim = NULL;
	}

	return xim;
}


static void 
initCursors(w)
Widget w;
{
	static int	inited = False;
	Display		*dpy = XtDisplay(w);
	Screen		*screen = XtScreen(w);
	GC		gc = None;
	Pixmap		pix, mask;
	int		i, width, height;
	XImage		*src, *msk;
	int		x, y;
	XColor		xcols[2];

	if (inited)
		return;
	inited = True;

	watchCursor = XCreateFontCursor(dpy, XC_watch);
	normalCursor= XCreateFontCursor(dpy, XC_left_ptr);

	xcols[0].pixel = BlackPixelOfScreen(screen);
	xcols[1].pixel = WhitePixelOfScreen(screen);

	XQueryColors(dpy, DefaultColormapOfScreen(screen), xcols, XtNumber(xcols));

	for (i = 0; i < XtNumber(cursorInfo); i++) {
		width = cursorInfo[i].width;
		height = cursorInfo[i].height;
		pix=XCreatePixmapFromBitmapData(dpy,RootWindowOfScreen(screen),
			(char*)cursorInfo[i].bits,width,height,True,False, 1);
		src=XGetImage(dpy,pix,0,0,width,height,AllPlanes,ZPixmap);
		msk=NewXImage(dpy, NULL, 1, width, height);
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				Boolean	flg = (Boolean)XGetPixel(src, x, y);

				if (!flg && x > 0)
					flg = XGetPixel(src, x - 1, y);
				if (!flg && x < width - 1)
					flg = XGetPixel(src, x + 1, y);
				if (!flg && y > 0)
					flg = XGetPixel(src, x, y - 1);
				if (!flg && y < height - 1)
					flg = XGetPixel(src, x, y + 1);
				XPutPixel(msk, x, y, flg);
			}
		}

		mask = XCreatePixmap(dpy, pix, width, height, 1);
		if (gc == None)
			gc = XCreateGC(dpy, mask, 0, 0);
		XPutImage(dpy, mask, gc, msk, 0, 0, 0, 0, width, height);

		XDestroyImage(src);
		XDestroyImage(msk);

		cursorInfo[i].cursor = XCreatePixmapCursor(dpy, pix, mask, 
				&xcols[0], &xcols[1], width / 2, height / 2);
		XFreePixmap(dpy, pix);
		XFreePixmap(dpy, mask);
	}
	if (gc != None)
		XFreeGC(dpy, gc);
}
/*
 * C001^ 
 */
