
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: slist.C,v 2.3 1995/06/15 20:09:40 davidsi Exp $
// 
// $Log: slist.C,v $
// Revision 2.3  1995/06/15  20:09:40  davidsi
// - When loading a list from a file, if the input file cannot be
//   opened don't issue an error/warning message. Let the caller decide
//   whether they are going to inform the user - in many cases (e.g.
//   when the program is first used) there won't be a file to load.
//
// Revision 2.2  1995/05/18  20:19:43  davidsi
// - Added SlistSetpath() so that paths can be specified for the files
//   that store the data on the disk.
// - SlistMakeconsistent() now builds the complete file path.
// - SlistCopycallbacks() allows callbacks to be copied from one slist
//   instance to another.
// - Call to add_callback() renamed to CallbackAdd() to reflect changes
//   to the callback class.
// - The make_callback() function is now CallbackMake().
//
// Revision 2.1  1995/04/25  21:08:59  davidsi
// - In slist::load() pass to the callbacks a pointer to the list instance so
//   that they can insert onto the list without needing to know the precise
//   list.
//
// Revision 2.0  1995/01/19  21:15:14  davidsi
// - Because callback::add_callback now returns the created node added
//   a cast to disgard the value.
// - To reflect changes to the callback class add_load_vect() now requires
//   a function that takes a ptr and vararg arguments.
// - BUG FIX. Fixed the spelling of DEFAULTVSERION.
//
// Revision 1.1  1994/06/27  21:10:39  davidsi
// Initial revision
//
//

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "localmacros.h"

#include "slist.h"
#include "record.h"
#include "dialog.h"

//
//////////////////////////////////////////////////////////////////////////////
//
// Functions for the  slist_node class
//
//////////////////////////////////////////////////////////////////////////////
//

//
// slist_node::save_record record
//
// The save_record function returns the string that should be written 
// to a file when the list is being stored pernamently. It is expected
// that derived classes will provide their own version of this function
// (it is declared as a virtual function). The default function below
// returns an empty string.
//
// INPUTS :	record	The record class member that is to be used for
//			storing the data.
// RETURNS:	None
//
void
slist_node::save_record(class record *rec)
{
	rec->create("");
}

//
// These are to save casts
//
class slist_node *
slist_node::next()
{
	return( LISTREMAP(slist_node,list_node,next,this) );
}


//
//////////////////////////////////////////////////////////////////////////////
//
// Functions for the slist class
//
//////////////////////////////////////////////////////////////////////////////
//



//
// slist::version
//
// The version function returns an ascii string specifying a version id.
// It is expected that derived classes will provide their own version of
// this function (it is declared as a virtual function). This default
// function below returns "DEFAULTVERSION".
//
// INPUTS :	None
// RETURNS:	String specifying a version identifier
//
char *
slist::version()
{
	return("DEFAULTVERSION");
}



//
// slist::save
//
// Writes a slist to a file.  The file is created using the mode
// specifed at open time. Version information is written prior
// to traversing the list and writing individual records.
//
// INPUTS :	None
// RETURNS:	None
//
void
slist::save()
{
	class slist_node *ptr;
	class record *rec=new class record;
	unsigned long writemode;

	switch(mode) {
		case WRITE_BY_LINE:
			writemode=OPEN_WRITE_LINE;
			break;
		case WRITE_BY_SIZE:
			writemode=OPEN_WRITE_SIZE;
			break;
		case WRITE_BY_PERC:
			writemode=OPEN_WRITE_PERC;
			break;
	}

	ptr=(class slist_node *)head();
	if ( rec->open(fullpath,writemode)==-1 ) {
		fprintf(stderr,"Cannot open output file\n");
		exit (1);
	}

	//
	// Write version identifier
	//
	rec->create( version() );
	rec->write();


	while (! eol(ptr) )  {
		ptr->save_record(rec);
		rec->write();
		ptr=ptr->next();
	}
	rec->close();
}



//
// slist::load
//
// Loads a slist from a file.
//
// INPUTS :	None
// RETURNS:	None
//
void
slist::load()
{
	class record *rec=new class record;

	if (!fullpath) {
		fprintf(stderr,"No file specified\n");
		goto done;
	}

	if ( rec->open(fullpath,OPEN_READ)==-1) {
		//
		// Leave callers to warn about failure to load an
		// input file
		//
		// fprintf(stderr,"Cannot open input file\n");
		//
		goto done;
	}

	//
	// Keep loading records from the file where there are records to load
	//
	while ( rec->read() ) {
		load_vect.CallbackMake(0,rec,this);
	}
	rec->close();

done:
	//
	// Free up rec 
	//
	delete rec;
}


//
// slist::add_load_vect function
//
// Allows a callback,that will be executed when the slist is loaded
// from a file, to be added.  All callbacks should be added prior
// to the slist being loaded.
//
// INPUTS :	function	The function to be called at load time
// RETURNS:	None
//
void 
slist::add_load_vect( void (*f)(char *, va_list ) )
{
	(void)load_vect.CallbackAdd(f);
}


//
// slist::SlistMakeconsistent 
//
// Ensures the internal (full) path is correct when either the file
// name component or directory component change.
//
// INPUTS :	None
// RETURNS:	None
//
void
slist::SlistMakeconsistent()
{
	char *path;

	if (fullpath)
		delete fullpath;

	path=pathname;

	//
	// If no path yet set assume current directory
	//
	if (path==NULL)
		path=".";

	//
	// +2 allows for both null terminator and the additional /
	//
	fullpath=new char[strlen(filename)+strlen(path)+2];
	sprintf(fullpath,"%s/%s",path,filename);
}



//
// slist::set_filename filename mode
//
// Sets the file name and file layout used by a stored list
//
// INPUTS :	filename	The file to be used by the stored list
//		imode		The file layout
// RETURNS:	None
//
void
slist::set_filename(char *name,
	unsigned long imode
	)
{
	switch (imode) {
		case WRITE_BY_LINE:
		case WRITE_BY_SIZE:
		case WRITE_BY_PERC:
			break;
		default:
			DialogPanic("slist::set_filename imode is bad");
	}
	if (filename)
		delete filename;
	filename=new char[strlen(name)+1];
	strcpy(filename,name);
	mode=imode;
	SlistMakeconsistent();
}


//
// SlistSetpath pathname
//
// INPUTS :	pathname	The new pathname
// RETURNS:	None
//
void
slist::SlistSetpath(char *path)
{
	int ret;
	struct stat statbuf;
	char buffer[2048];

	if (pathname)
		delete pathname;
	pathname=new char[strlen(path)+1];
	strcpy(pathname,path);
	
	SlistMakeconsistent();

	if ( stat(path,&statbuf) != -1)
		return;
	if (errno==ENOENT) {
		sprintf(buffer,"%s does not exist - Will be Created",path);
		DialogMessage(buffer);
		if (mkdir(path,S_IEXEC|S_IREAD|S_IWRITE)==-1) 
			DialogMessage("Cannot create directory");
	} else {
		sprintf(buffer,"Cannot locate %s - error %d ",path,errno);
		DialogMessage(buffer);
	}
}


//
// slist::SlistCopycallbacks base
//
// Copies the callbacks assigned to instance to another instance
//
// INPUTS : base	The instance to copy from
// RETURNS: None
//
void
slist::SlistCopycallbacks(class slist &base)
{
	load_vect.CallbackCopy(base.load_vect);
}

//
// slist
//
// Constructor for the slist (stored list) class
//
// INPUTS :	None
// RETURNS:	None
//
slist::slist()
{
	filename=NULL;
	pathname=NULL;
	fullpath=NULL;
}

//
// ~slist
//
// Destructor for the slist (stored list) class
//
// INPUTS :	None
// RETURNS:	None
//
slist::~slist()
{
	if (filename)
		delete filename;
	if (pathname)
		delete pathname;
	if (fullpath)
		delete fullpath;
}
