/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/****************************************************************
  Server callable routines

  This module contains routines to be called by the
  License server.

      license_init_db()
      freeze_tickets()
      thaw_tickets()
      free_ticket()
      reclaim_tickets()

 ****************************************************************/

/* LICENSE_INIT_DB
   Prtns_init will fill the database datastructures
   for a local database from the text database.
 */
int license_init_db(char *db)
{
  int rc;

  if ( db == NULL ) db = LICENSE_TEXT_DB;

  /* Release old database info */
  if ( License_State == 1 ) {
    struct license_str *lp;
    struct range_str *rp;

    for ( rc = 0; rc < License_List_lim; rc++ ){
      lp = License_List[rc];
      free(lp->service);
      free_list(lp->class);
      rp = lp->range;
      free(rp->lower);
      if ( rp->upper != NULL ) free (rp->upper);
      free(rp);
      free_list(lp->profile);
      free(lp);
    }
    License_List_lim = 0;
    License_State = 0;
  }

  license_open_db(db);

  rc = license_db_parse();

  license_close_db();

  if ( rc ) {
    return(LICENSE_FAILURE);
  }
  License_State = 1;  /* initialized */
  return(LICENSE_SUCCESS);
}

/* FREEZE_TICKETS
   Save all the active tickets to a file.
 */
int freeze_tickets(char *fname)
{
  FILE *f;
  int rc;
  int i;

  if ( fname == NULL ) fname = LICENSE_TEXT_TB;
  f = fopen(fname,"w");
  if ( f == NULL ) {
    return(LICENSE_FAILURE);
  }

  /* save the sequence number */
  fprintf(f,"%d\n",License_Seq);

  /* save the outstanding licenses */
  for ( i = 0; i < License_Tickets_lim; i++ ) {
    /* the routine thaw_tickets() depends upon this format! */
    rc = fprintf(f,"%i %i %i %i %s %s\n",License_Tickets[i].seq,
            License_Tickets[i].issued,
            License_Tickets[i].update,
            License_Tickets[i].reclaim,
            License_Tickets[i].license, 
            License_Tickets[i].assoc ? License_Tickets[i].assoc:":");
    if ( ferror(f) ) {
      fclose(f);
      return(LICENSE_FAILURE);
    }
  }
  rc = fclose(f);
  if ( rc == EOF ) {
    return(LICENSE_FAILURE);
  }
  return(LICENSE_SUCCESS);
}

/* THAW_TICKETS
   Retrieve tickets from file and add to list.
   Discard tickets currently in use.

   Used when starting the server to retreive state information.
 */
int thaw_tickets(char *fname)
{
  int rc,i;
  char line[80];
  char *p,*p1, *p2;
  struct tlist_str *tp,tlist;
  FILE *f;
  time_t now;

  if ( fname == NULL ) {
    fname = LICENSE_TEXT_TB;
  }

  f = fopen(fname,"r");
  if ( f == NULL ) {
    return(LICENSE_FAILURE);
  }

  p = fgets(line,sizeof(line),f);
  if ( p != NULL ) {
    License_Seq = atoi(p);
  }

  for (;;) {
    p = fgets(line,sizeof(line),f);
    if ( p == NULL ) break;

    p1 = strchr(p,' ');
    if ( p1 == NULL ) continue;
    tlist.seq = atoi(p);
    p = p1 +1;

    p1 = strchr(p,' ');
    if ( p1 == NULL ) continue;
    tlist.issued = atoi(p);
    p = p1 +1;

    p1 = strchr(p,' ');
    if ( p1 == NULL ) continue;
    /* tlist.update=atoi(p);  */
    time(&tlist.update); /* last update is now! */

    p = p1 +1;

    p1 = strchr(p,' ');
    if ( p1 == NULL ) continue;
    tlist.reclaim = atoi(p);
    p = p1 +1;

    p2 = strchr(p,' ');       /* assoc */
    if ( p2 ) {
      *p2 = '\0';
      p2++;
      p1 = p2 + (strlen(p2) -1);
    } else {                  /* no assoc */
      p1 = p + (strlen(p) -1);
    }
    *p1 = '\0';     /* remove newline */

    tp = find_ticket(p);
    if ( tp != NULL ) continue;    /* ticket already exists! */

    tlist.license = mkstr(p);
    if ( p2 ) {
      tlist.assoc = mkstr(p2);
    } else {
      tlist.assoc = NULL;
    }

    rc =  add_ticket(&tlist);
    if ( rc != LICENSE_SUCCESS ) {
      free(tlist.license);
      break;
    }
  }

  rc = fclose(f);
  return(LICENSE_SUCCESS);
}

/* RECLAIM_TICKETS
   Reclaim tickets which have not been updated as required.
   Returns number of tickets reclaimed.
 */
int reclaim_tickets(time_t now)
{
  int rc;
  int i;
  int j;

  /* Check each of the outstanding licenses */
  /* Delete_ticket will modify the value of License_Tickets_lim! */
  for (rc = 0, i = 0, j = License_Tickets_lim; i < j; i++) {
    if ( License_Tickets[i].reclaim != 0 &&
         License_Tickets[i].update + License_Tickets[i].reclaim < now ) {
      j = delete_ticket(License_Tickets[i].license);
      if ( j == LICENSE_SUCCESS )rc++;
      j = License_Tickets_lim;  /* this had better not be optimized out. */
    }
  }
  return(rc);
}

