/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/* interface.c
   Copyright 1991 University of Washington
   Author:  Steve Jones, UCS

   This module contains the applications interface routines
   to the license management service for the
   Univerity of Washington Information Network.

      get_ticket()
      update_ticket()
      return_ticket()
      validate_ticket()
      license_server()

      ticket2str()
      str2ticket()
      free_ticket()
      license_message()
 */

#include <license.h>
#include <license_servers.h>

static char *mknstr(char *s, int l);

/* GET_TICKET
   This routine gets a ticket from the ticket server.
 */
int get_ticket(struct ticket_str **ticket, char *service, char *class,
               char *assoc) 
{

#ifdef LICENSE_SERVER
   struct license_str *lp;
   int rc;
   int buflen;


   /* validate arguments */
   if ( ticket == NULL || service == NULL || class == NULL ) {
     return(LICENSE_FAILURE);
   }

   lp = findpool(service,class);
   if ( lp == NULL ) return(LICENSE_FAILURE);
   rc = grant_ticket(ticket,lp,assoc);
   return(rc);
#else
   char line[100];
   char *buf, *p;
   int rc;
   struct ticket_str *tp;
   int profile_index;
   int buflen;

   /* validate arguments */
   if ( ticket == NULL || service == NULL || class == NULL ) {
     return(LICENSE_FAILURE);
   }

   if ( assoc != NULL ) { 
     sprintf(line,"request=ticket class=%s service=%s assoc=%s",
             class,service,assoc);
   } else {
     sprintf(line,"request=ticket class=%s service=%s",class,service);
   }

   rc = license_open(0);
   if ( rc != LICENSE_SUCCESS) return(rc);

   rc = license_send(line);
   if ( rc == LICENSE_SUCCESS) {
     rc = license_recv(&buf,&buflen);
   }
   license_close();
   if ( rc != LICENSE_SUCCESS) {
     return(rc);
   }

   tp = str2ticket(buf,buflen);
   free(buf);

   *ticket = tp;
   if ( tp == NULL ) return(LICENSE_FAILURE);
   return(LICENSE_SUCCESS);
#endif
}

/*
   UPDATE_TICKET
   This routine updates a ticket received from the ticket server.
 */
int update_ticket(struct ticket_str *ticket)
{
#ifdef LICENSE_SERVER
   char *tick;
   struct tlist_str *tp;

   if ( ticket == NULL ) {
     return(LICENSE_FAILURE);
   }
   tick = ticket->license;
   tp = find_ticket(tick);
   if ( tp == NULL ) {
     return(LICENSE_FAILURE);
   }
   time(&tp->update);
   return(LICENSE_SUCCESS);
#else
   char line[80];
   char *buf;
   int rc,buflen;

   if ( ticket == NULL ) {
     return(LICENSE_FAILURE);
   }
   sprintf(line,"request=update ticket=%s",ticket->license);

   rc = license_open(0);
   if ( rc != LICENSE_SUCCESS) return(rc);

   rc = license_send(line);
   if ( rc == LICENSE_SUCCESS) {
     rc = license_recv(&buf,&buflen);
   }
   license_close();
   if ( rc != LICENSE_SUCCESS) {
     return(rc);
   }

   if ( strncmp(buf,"ack",3) ) {
     rc = LICENSE_FAILURE;
   } else {
     rc = LICENSE_SUCCESS;
   }
   free(buf);
   return(rc);
#endif
}

/*
   RETURN_TICKET
   This routine returns a ticket to the ticket server.
 */
int return_ticket(struct ticket_str *ticket)
{
#ifdef LICENSE_SERVER
  char *tick;
  int rc;

  if ( ticket == NULL ) return(LICENSE_SUCCESS);
  tick = ticket->license;
  if ( tick != NULL ) {
    rc = delete_ticket(tick);
  }
  free_ticket(ticket);
  return(LICENSE_SUCCESS);
#else
  char line[80];
  char *buf;
  int j,rc,buflen;

  if ( ticket == NULL ) {
    return(LICENSE_FAILURE);
  }
  sprintf(line,"request=return ticket=%s",ticket->license);

  free_ticket(ticket);

  rc = license_open(0);
  if ( rc != LICENSE_SUCCESS) return(rc);

  rc = license_send(line);
  if ( rc == LICENSE_SUCCESS) {
    rc = license_recv(&buf,&buflen);
  }
  license_close();
  if ( rc != LICENSE_SUCCESS) {
    return(rc);
  }

  free(buf);
  return(LICENSE_SUCCESS);
#endif
}
/*
   VALIDATE_TICKET
   This routine validates a ticket key with the license manager.
 */
int validate_ticket(char *key, char *assoc)
{
#ifdef LICENSE_SERVER
  int rc,RC;
  struct tlist_str *tp1, *tp2;
  char *association;
  
  RC = LICENSE_SUCCESS;
  if ( key == NULL ) return(RC);
  if ( *key != ':' ) return(LICENSE_FAILURE);  /* sanity check, make_key() */
  
  if ( assoc != NULL ){   /* association checking */
    /* Get ticket(key) */
    tp1 = find_ticket(key);
    if ( !tp1 ) RC = LICENSE_FAILURE;
  
    /* Get ticket(assoc) */
    tp2 = find_ticket(assoc);
    if ( ! tp2 ) RC = LICENSE_FAILURE;

    /* Verify both tickets have same association */
    if ( tp1 && tp2 ) {
      if ( tp1->assoc && tp2->assoc && ! strcmp(tp1->assoc,tp2->assoc)){
        /* Ok, Remove association, it is one time use for validation */
        if ( *tp1->assoc == ':' ) {  /* sanity check, make_key() */
          rc = delete_ticket(tp1->assoc);
          free(tp2->assoc);   
          tp2->assoc = NULL;   /* don't want multiple validations */
          if ( rc != LICENSE_SUCCESS ) RC = rc;
        }
      } else {
        if ( tp1->assoc || tp2->assoc ) RC = LICENSE_FAILURE;
      }
    }
  }
  /* Remove key, it is one time use for validation */
  rc = delete_ticket(key);
  if ( rc != LICENSE_SUCCESS ) RC = rc;

  /* Return status */
  return(RC);
#else
  char line[80];
  char *buf;
  int j,rc,buflen;

  if ( key  == NULL ) {
    return(LICENSE_FAILURE);
  }
  if ( assoc != NULL ) {
    sprintf(line,"request=validate key=%s assoc=%s",key,assoc);
  } else {
    sprintf(line,"request=validate key=%s",key);
  }

  rc = license_open(0);
  if ( rc != LICENSE_SUCCESS) return(rc);

  rc = license_send(line);
  if ( rc == LICENSE_SUCCESS) {
    rc = license_recv(&buf,&buflen);
  }
  license_close();
  if ( rc != LICENSE_SUCCESS) {
    return(rc);
  }

  rc = strcmp("ack",buf) ? LICENSE_FAILURE : LICENSE_SUCCESS;
  free(buf);
  return(rc);
#endif
}

/* LICENSE_SERVER 
   Use to set list of license servers to contact for service.
   The list must be NULL terminated.

      E.G.    char *list[]={"host1",NULL};

*/
char **license_server(char **list)
{
  char **p;
  p = license_servers;
  license_servers = list;
  return(p);
}

struct ticket_str *str2ticket(char *buffer, int buflen)
{
  struct ticket_str *tp;
  int i,j,profile_index;
  char *p,*p1,*p2,*l;

  /* validate buffer */
  if ( buffer == NULL ) return(NULL);
  if ( strncmp(buffer,"ticket",6) ) return(NULL);

  /* get ticket structure */
  tp = (struct ticket_str *)malloc(sizeof(struct ticket_str));
  if ( tp == NULL ) return(NULL);
  tp->license   = NULL;
  tp->assoc     = NULL;
  profile_index = 0;
  tp->profile[profile_index] = NULL;
  tp->connect   = 0;
  tp->idle      = 0;
  tp->update    = 0;
  tp->reclaim   = 0;

  /* fill fields */
  p = buffer;
  l = buffer + buflen;

  for ( ; p < l; ) {
    /* each line in buf is of form "key=value\n" */

    p1 = strchr(p,'=');
    if ( p1 == NULL ) break;
    p2 = strchr(p1,'\n');
    if ( p2 == NULL ) break;
    i = p1 - p;    /* key len */
    p1++;
    j =  p2 - p1;  /* value len */
    p2++;
    if ( ! strncmp(p,"ticket",i) ) {
      tp->license = mknstr(p1,j);
      if ( tp->license == NULL ) goto error;
    } else if ( ! strncmp(p,"assoc",i) ) {
      tp->assoc = mknstr(p1,j);
      if ( tp->assoc == NULL ) goto error;
    } else if ( !strncmp(p,"seq",i) ) {
      tp->seq = atoi(p1);
    } else if ( !strncmp(p,"connect",i) ) {
      tp->connect = atoi(p1);
    } else if ( !strncmp(p,"idle",i) ) {
      tp->idle = atoi(p1);
    } else if ( !strncmp(p,"update",i) ) {
      tp->update = atoi(p1);
    } else if ( !strncmp(p,"reclaim",i) ) {
      tp->reclaim = atoi(p1);
    } else if ( !strncmp(p,"profile",i) ) {
      tp->profile[profile_index] = mknstr(p1,j);
      if ( tp->profile[profile_index] == NULL ) goto error;
      /* translate RS to linefeed */
      for ( p1 = tp->profile[profile_index]; *p1; p1++) {
        if ( *p1 == '\036' ) {
          *p1 = '\n';
        }
      }
      profile_index++;
    }
    p = p2;

  }
  tp->profile[profile_index] = NULL;

  return(tp);

error:
  free_ticket(tp);
  return(NULL);
}

char *ticket2str(struct ticket_str *tp)
{
  char *p,*p1;
  int rc, i;
  long int slen;

  if ( tp == NULL ) return(NULL);  /* sanity check */

  /* Make a conservative calculation as to how large a buffer to allocate */
  slen = MIN_MALLOC + 6 * sizeof(struct ticket_str);
  for ( i = 0; tp->profile[i] != NULL; i++ ) {
     slen += strlen(tp->profile[i]) +20;
  }
  p = (char *) malloc(slen);
  if ( p == NULL ) return(NULL);  /* egads */

  /* Marshall the ticket structure into the buffer */
  p1 = p;
  sprintf(p1,"ticket=%s\nseq=%d\nassoc=%s\n",tp->license,tp->seq,tp->assoc);
  p1 += strlen(p1);
  sprintf(p1,"connect=%d\nidle=%d\nupdate=%d\nreclaim=%d\n",
          tp->connect, tp->idle, tp->update, tp->reclaim);
  p1 += strlen(p1);
  for ( i = 0; tp->profile[i] != NULL; i++ ) {
    sprintf(p1,"profile=%s\n",tp->profile[i]);
    p1 += strlen(p1);
  }
  return(p);
}
/* MKNSTR
   Make a copy of a C-string (first "l" characters)
 */
static char *mknstr(char *s, int l)
{
  char *p;
  p = (char *) malloc( l < MIN_MALLOC ? MIN_MALLOC : l+1 );
  if ( p == NULL ) return ( NULL );
  strncpy(p,s,l);
  *(p+l) = '\0';
  return(p);
}
/* FREE_TICKET
   Frees the space allocated for a ticket
 */
free_ticket(struct ticket_str *ticket)
{
  int j;

  if ( ticket == NULL ) return;

  if ( ticket->license != NULL ) free(ticket->license);
  if ( ticket->assoc != NULL ) free(ticket->assoc);
  for ( j = 0; ticket->profile[j] != NULL; j++){
    free(ticket->profile[j]);
  }
  free(ticket);
  return;
}
char *license_message(int rc)
{
  static char *messages[LICENSE_MESSAGES] = {
             "Success.",
             "Failure.",
             "Network error.",
             "Server failure.",
             "Error reading from network.",
             "Network timeout.",
             "Malloc failure."
            };
  if ( rc <= LICENSE_MESSAGES ) {
    return(messages[rc-1]);
  }
  return(NULL);
}
