/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/************************************************************************
 * driver.c
 * This is core of the table-driven database driver. Hopefully this
 * file can be used pretty much unchanged, along with a customized
 * host.c and table.c to implement new drivers.
 * 
 * See willow/doc/Customization for information about the flags
 * for this driver.
 *
 * Some routines have the suffix _wcb or _hcb on their names. These
 * are "Callback" routines, modeled after those in X. They are
 * triggered when certain events occur. _wcb routines are triggered by
 * incoming packets from Willow, while _hcb routines are triggered by
 * incoming strings from the database host.
 ************************************************************************/

#include <stdio.h>
/* The makedepend on Solaris is too stupid to
 handle a complex define like: 
 #if  defined(SOLARIS) || defined(_SEQUENT_)
 so we have to make it ugly. */
#ifdef SOLARIS
#include <string.h>
#include <libgen.h>
#else  /* SOLARIS */
#ifdef _SEQUENT_
#include <string.h>
#include <libgen.h>
#else  /* _SEQUENT_ */
#include <strings.h>
#endif /* _SEQUENT_ */
#endif /* SOLARIS */

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#if defined(_AIX) || defined(_SEQUENT_)
#include <sys/select.h>
#endif

#include <license.h>
#include <license_status.h>

static void updater(void);

#include "driver.h"
#include "protocol.h"
#include "host.h"

/* These two codes are only used for time-logging */
#define TELNETTED 99
#define DBED      100

/* External variables */
extern summ_tab_entry *summ_tab;
extern int summary_long_len;
extern char logging;

/* Extern routines */
extern int atoi (char *);
extern int browse_wcb(packet_rec *packet);
extern int list_close (void);
extern int close (int);
extern int disconnect_wcb (void);
extern int dup (int);
extern int execlp (char *, char *, ...);
extern int init_host (void);
extern int init_table (void);
extern int log_trans (int type, char *comment);
extern char *lameify(char *record, hl_rec *hl_list);
extern int make_search (search_desc_rec *text_fields, search_desc_rec *limit_fields);
extern int malarm_add (int interval, void (*handler) (/* ??? */));
extern int pipe (int *);
extern int rand (void);
extern int read (int, void *, unsigned int);
extern int record_req_wcb (packet_rec *packet);
extern int save_login_info (char *uid, char *pw, char *db_name, char *t_fields, char *s_fields, char *ext_fields);
extern int search_cancel_wcb (packet_rec *packet);
extern int title_req_wcb (packet_rec *packet);
extern int user_resp_wcb(packet_rec *packet);
extern int ext_req_wcb(packet_rec *packet);
extern int write (int, void *, unsigned int);
#ifndef vax
extern pid_t getpid (void);
extern pid_t vfork (void);
#endif
extern unsigned int sleep (unsigned int);
extern void *malloc (size_t);
extern void *realloc (void *, size_t);
extern void free (void *);
extern void malarm_cancel (int id);
extern void parse_fields (void);
extern void srand (unsigned int);
extern packet_rec *netify(packet_rec *packet);
extern packet_rec *denetify(packet_rec *packet);
extern void validate_class(int sock, char *class, char *class_file);

/* Forward Routines */
char *env_get(char *s);
char *strndup(char *s, int n);
int browse_wcb(packet_rec *packet);
int cheat_wcb (packet_rec *packet);
int cleanup (void);
int de_0_fy (char *buffer, int size);
int dump (char *str);
int dump_state (int state);
static void echo_packet (packet_rec *packet, char send);
int echo_wcb (packet_rec *packet);
int error (char *msg);
int grow_parse_buf (void);
int host_send (char *str);
int login_wcb (packet_rec *packet);
int main_loop (void);
static int misc_error_hcb (char *msg);
int read_from_host (void);
int read_from_willow (void);
int scan (char *string);
int search_wcb (packet_rec *packet);
int update_hcb (char *msg);
void dead_kid(void);
void force_quit_hcb (void);
void force_transition(int which);
void send_user_req(int type, char *query_text, char *query_tag,
		   char *yes_label, char *no_label);

/***** GLOBALS *****/

/* Hold text read between matches in this buffer */
char *parse_buf;

/* Level of debugging, passed in on command line */
int debug_level = 0;

/* Are we being accessed via a network socket rather than a local pipe? */
char socketed = FALSE;

/* file descriptors for communication with willow. Could be a pipe
   or a socket */
int to_willow, from_willow;

/* Current characters in parse_buf */
int    parse_buf_size;
int    parse_buf_used = 0;

char   fresh_input;  /* flag */

int    host_pid = NULL;     /* process id */

extern state_table states;           /* The state table */

/* The current state of the remote computer */
int    state;

/* The current state of our connection with willow */
char   connected_to_willow = TRUE;

/* Are we currently echoing interactions with host? */
char   echo = FALSE;

/* a couple of pipes for talking to telnet */
int  to_host[2],
     from_host[2];

/* File descriptor mask for select call */
fd_set read_fds;
int    num_fds;

/* Head and tail of linked list of hilite points */
hl_rec *hl_head = NULL, *hl_tail = NULL;
int     num_hilites = 0;

/* Environment list */
env_rec env[MAX_ENV];
int env_cnt = 0;

/* Maximum number of circ requests we will do */
int max_circ = DFLT_MAX_CIRC;

/* Timeout value for circ requests */
int circ_time = DFLT_CIRC_TIMEOUT;

/* Which display mode for circ queries */
char *circ_mode;

/* Name of the machine we query for circ status */
char *circ_host;

/* Which is key field for circ queries */
char *circ_key;

/* Name of database */
char *db_name;

/* Name of the file containing IP address class info */
char *class_file = NULL;

/* Flag for whether circulation is done for this database */
char do_circ;

/* Is this a manual log-in? */
char manual_login = FALSE;

/* ticket-updater alarm id */
int updater_id = NULL;

/* Tag value for the most recently sent user request */
char *last_sent_req;

/* structure for license manager, global cause we allocate in one routine,
   but free in another */

static  struct ticket_str *app_ticket=NULL; 
static  struct ticket_str *db_ticket=NULL; 

/* directory where willow config files live. Not used by uwbrs,
   but needed by z3950 driver */
char *config_dir;

#define USAGE "USAGE: uwbrs [-debug <number>] [-socketed] [-outfile <file>]\n\
             [-cfg_path <path>] [-class_file <file>]"

/* Takes one argument, an integer representing the debugging level */
main(int argc, char **argv)
{
  FILE *outfile;
  char *outfile_name = "/dev/null";
  char  redirect = FALSE;

  /* Process the command line args */
  argc--; argv++;
  while (argc) {
    if (strcmp("-debug", argv[0]) == 0) {
      if (argv[1]) {
	debug_level = atoi(argv[1]);
	argc -= 2;
	argv += 2;
      }
      else
	error(USAGE);
    }
    else if (strcmp("-socketed", argv[0]) == 0){
      socketed = TRUE;
      argc -= 1;
      argv += 1;
    }
    else if (strcmp("-outfile", argv[0]) == 0) {
      redirect = TRUE;
      if (argv[1]) {
	outfile_name = argv[1];
	argc -= 2;
	argv += 2;
      }
      else
	error(USAGE);
    }
    else if (strcmp("-cfg_path", argv[0]) == 0) {
      if (argv[1]) {
	config_dir = argv[1];
	argc -= 2;
	argv += 2;
      }
      else
	error(USAGE);
    }
    else if (strcmp("-class_file", argv[0]) == 0) {
      if (argv[1]) {
	class_file = argv[1];
	argc -= 2;
	argv += 2;
      }
      else
	error(USAGE);
    }
    else
      error(USAGE);
  }

  /* initialize willow file descriptors. If we are running socketed,
     then 0 is socket, passed from inetd. */
  if (socketed)
    to_willow = from_willow = 0;
  else {
    to_willow = FROM_DRIVER_FD;
    from_willow = TO_DRIVER_FD;
  }

  /* Redirect stdout and stderr. Must do this if socketed, otherwise 
     optional. */
  if (socketed || redirect) {
    if (!(outfile = fopen(outfile_name, "a"))) {
      char err_buf[256];
      
      sprintf(err_buf,"Could not open outfile '%s'", outfile_name);
      error(err_buf);
    }
    close(fileno(stdout));
    close(fileno(stderr));
    dup2(fileno(outfile),fileno(stdout));
    dup2(fileno(outfile),fileno(stderr));
  }

  parse_buf = malloc(PARSE_BUF_SIZE);
  parse_buf[0] = 0;
  parse_buf_size = PARSE_BUF_SIZE;
  init_table();
  init_host();
  state = NO_CON;

  num_fds = 0;	/* initialize */
  main_loop();
}

/* This is the main loop of the program. We block on select until
   we get input from one of the pipes we are listening on, then
   take appropriate action.
 */
main_loop(void)
{
  while (1) {
    FD_ZERO(&read_fds);
    if (connected_to_willow) {
      num_fds = MAX(num_fds, (from_willow) + 1);
      FD_SET(from_willow, &read_fds);
    }
    if (state != NO_CON) {
      num_fds = MAX(num_fds, (from_host[R]) + 1);
      FD_SET(from_host[R], &read_fds);
    }

    /* Block forever on select */
    if (select(num_fds, &read_fds, NULL, NULL, NULL) < 0) {
      if (errno == EINTR){  /* Interrupt due to timer */
	errno = 0;
	FD_ZERO(&read_fds);
	continue;
      }
      else 
	error("error on select");
    }
    /* Decide which pipe has input */
    if (FD_ISSET(from_willow, &read_fds)) {
      read_from_willow();
    }
    else if (FD_ISSET(from_host[R], &read_fds)) {
      read_from_host();
    }
    else
      error("select on unknown pipe");
  }
}

/* This is the heart of the finite state machine. It scans the 
   input coming in from the remote host, and follows the state table.
 */
read_from_host(void)
{
  int i;
  char reset_ready;  /* flag for resetting READY state */

  fresh_input = TRUE;
  for (i=0; i < MAX_TRANSITIONS; i++) {
    if (!states[state][i].in_string)
      return;
    if (scan(states[state][i].in_string)) {
      /* call action routine first, in case it needs to modify table
	 in order to branch. */
      if (states[state][i].action)
	states[state][i].action();
      
      if (states[state][i].out_string)
	host_send(states[state][i].out_string);
      
      /* Reset the parse_buf. Do this out here because the action
	 routine might have needed it. */
      parse_buf_used = 0;
      parse_buf[0] = 0;
      
      reset_ready = (state == READY);
      /* Do this last, since action routine might need to know state */
      state = states[state][i].next_state;

      if (debug_level)
	  dump_state(state);

      /* READY is special state, used to queue single commands, must
         reset after use */
      if (reset_ready) {
	states[READY][0].out_string = "";
	states[READY][0].next_state = READY;
      }

      /* But if we transitioned to READY, there may be an extra command
	 still queued there (particularly if a retrieve full was hit
	 immediately after a 30 more) */
      if ((state == READY) && (states[READY][0].out_string[0]))
	force_transition(0);
      return;
    }
  }      
}

/* This is called when we want to force the state machine to move.
   Pretend the in_string[which] has just been matched for whatever state
   we are currently in.
 */
void force_transition(int which)
{
  char reset_ready;  /* flag for resetting READY state */

  reset_ready = (state == READY);

  if (states[state][which].out_string[0])
    host_send(states[state][which].out_string);
  state = states[state][which].next_state;

  /* READY is special state, used to queue single commands, must
     reset after use */
  if (reset_ready) {
    states[READY][0].out_string = "";
    states[READY][0].next_state = READY;
  }
}

/* This routine takes a packet in off the pipe from willow, and
   takes an appropriate action.
 */
read_from_willow(void)
{
  packet_rec packet;   /* store packet we read here */
  int  got_count = 0;  /* how many bytes of packet have we gotten */
  int  n = 0;
  
  /* First read in just the header */
  while (got_count < PACKET_NET_SIZE) {
    n = read(from_willow, &packet + got_count, 
	     PACKET_NET_SIZE - got_count);
    if (n == 0) {
      error("Willow has unexpectedly died");
      return;
    }
    if (n < 0) {
      perror(NULL);
      error("error reading input pipe (header)");
      return;
    }
    got_count += n;
  }
  denetify(&packet);

  /* Grab enough storage, and read in rest of packet */
  packet.data = malloc(packet.data_len);

  /* Now read until we get the entire packet in */
  got_count = 0;
  while (got_count != (packet.data_len)) {
    n = read(from_willow, packet.data+got_count, packet.data_len - got_count);
    if (n == 0) {
      error("Willow has unexpectedly died");
      return;
    }
    if (n < 0) {
      perror(NULL);
      error("error reading input pipe (data)");
      return;
    }
    got_count += n;
  }

  /* Call a handler routine for this packet type */
  if (debug_level)
    echo_packet(&packet, FALSE);

  switch (packet.type) {
  case CONNECT:
    login_wcb(&packet);
    break;
  case SEARCH:
    search_wcb(&packet);
    break;
  case CANCEL:
    if (logging)
      log_trans(packet.type,"");
    search_cancel_wcb(&packet);
    break;
  case TITLE_REQ:
    title_req_wcb(&packet);
    break;
  case RECORD_REQ:
    record_req_wcb(&packet);
    break;
  case EXT_REQ:
    ext_req_wcb(&packet);
    break;
  case DISCONNECT:
    if (logging)
      log_trans(packet.type, "");
    disconnect_wcb();
    break;
  case TOGGLE_ECHO:
    echo_wcb(&packet);
    break;
  case CHEAT:
    cheat_wcb(&packet);
    break;
  case USER:
    user_resp_wcb(&packet);
    break;
  case BROWSE_REQ:
    browse_wcb(&packet);
    break;
  default:
    error("unknown packet type");
  }
  free(packet.data);
}

/* Scan input for requested string, return true if found, else false,
   We always re-read the old buffer, unless the fresh_input flag
   has been reset.

   Occasionally input stream will break in middle of a target string. 
   To guard against that, we search the parse buffer, which starts from 
   the end of our previous hit. This is slightly inefficient.

   The calling routine has responsibility for resetting the parse buffer.
 */
scan(char *string)
{
  int    n;
  char   read_buffer[IN_BUF_SIZE];
  
  if (fresh_input)
    {
      fresh_input = FALSE;
      
      /* The -1 is so there is always room to append a \0 */
      n = read(from_host[R], read_buffer, sizeof(read_buffer) - 1);
      de_0_fy(read_buffer, n);

      /* Save old buffer */
      if ((parse_buf_used += strlen(read_buffer)) > parse_buf_size)
	grow_parse_buf();
      strcat(parse_buf, read_buffer);
      
      if (echo) {
	dump(read_buffer);
      }
    }
  
  /* Use built in regular expression handler to check for string */
#if  defined(SOLARIS) || defined(_SEQUENT_) || defined(sco)
  {
      char *rexp;
      if ((rexp = regcmp(string, (char *)0)) == NULL)
	  error("could not regcmp string");
      if (regex(rexp, parse_buf) == NULL) {
	  free (rexp);
	  return 0;
      } else {
	  free (rexp);
	  return 1;
      }
  }
#else
  if (re_comp(string) != 0)
    error("could not re_comp string");
  return(re_exec(parse_buf));
#endif
}

/* Take out any NULL's from the buffer, cause we want to treat this
   as a string. While we are at it, take out \r's, because they are
   so darned annoying. Then stick a NULL at the end.
 */
de_0_fy(char *buffer, int size)
{
  int sdex = 0, tdex = 0; /* source and target indices */
  
  while (sdex < size) {
    if ((buffer[sdex] != 0) && (buffer[sdex] != '\r')) {
      buffer[tdex] = buffer[sdex];
      tdex++;
    }
    sdex++;
  }
  buffer[tdex] = 0;
}

/* This function sends a string out on the pipe to the remote host */
host_send(char *str)
{
  write(to_host[W], str, strlen(str));
}

/* Send some data down the pipe back to willow. If len is NULL, send 
   strlen() bytes, plus 1 for terminator.
 */
willow_send(void *data, int len)
{
  if (len)
    write(to_willow, data, len);
  else
    write(to_willow, data, strlen(data) + 1);
}

/* Dump the string to wherever we want all the interactions dumped.
   For now, just send it to standard err. 
 */
dump(char *str)
{
  fprintf(stdout, "%s", str);
  fflush(stdout);
}

/*
 * These are the non-host-specific willow-callback routines.
 */

/* This routine is called when we receive a connect request from
   willow and we are not currently logged in. We have
   to fork off the telnet process to log into the database host.
 */
login_wcb(packet_rec *packet)
{
  char    *uid;
  char    *pw;
  char    *title_fields;
  char    *summ_fields;
  char    *ext_fields;
  char    *ptr;
  char    *machine;
  char    *log_chance_str;
  int      log_chance;
  int      rc;
  char     err_msg[256];
  char    *dummy;
  static char *serve_list[2];

  /* set up pipes for communicating with telnet process */
  pipe(to_host);
  pipe(from_host);
  
  ptr = packet->data;

  db_name = strdup(ptr);
  ptr += strlen(ptr) + 1;
  
  title_fields = strdup(ptr);
  ptr += strlen(ptr) + 1;

  summ_fields = strdup(ptr);
  ptr += strlen(ptr) + 1;

  ext_fields = strdup(ptr);
  ptr += strlen(ptr) + 1;

  /* Build environ list */
  env_cnt = 0;
  while ((ptr - packet->data) < packet->data_len) {
    if (env[env_cnt].name)
      free(env[env_cnt].name);
    env[env_cnt].name = strdup(ptr);
    ptr += strlen(ptr) + 1;

    if (env[env_cnt].val)
      free(env[env_cnt].val);
    env[env_cnt].val = strdup(ptr);
    ptr += strlen(ptr) + 1;
    env_cnt++;
  }
  
  circ_host = env_get("CircHost");
  if (do_circ = (circ_host != NULL)) {
    if (dummy = env_get("MaxCirc"))
      max_circ =  atoi(dummy);
    if (dummy = env_get("CircTimeout"))
      circ_time = atoi(dummy);
    if (!(circ_mode = env_get("CircMode")))
      error("Missing CircMode in a Circulation-querying database");
    if (!(circ_key  = env_get("CircKey")))
      error("Missing CircKey in a Circulation-querying database");
  }
  
  if (!env_get("ManualLogin")) {
    /* Get account and password from server. */
    if (debug_level)
      update_hcb("Getting login-licenses from server...");
    
    /* Before asking for ticket, if this is a socketed driver we may
       want to make the sure the client is coming from a place that
       matches the class it wants to be */
    if (socketed && class_file)
      validate_class(from_willow, env_get("class"), class_file);
    
    /* Check if a license-manager host was specified */
    if (serve_list[0] = env_get("LicenseHost")) {
      serve_list[1] = NULL;
      license_server(serve_list);
    }

    if ((rc = get_ticket(&app_ticket, env_get("app_pool"), 
			 env_get("class"), NULL)) != LICENSE_SUCCESS) {
      if (rc == LICENSE_FAILURE)
	sprintf(err_msg, "Too many users already logged in. Please try again later.\nCould not get application ticket");
      else
	sprintf(err_msg, "Could not get application ticket\nError Code = %s",
		license_message(rc));
      error(err_msg);
    }
    if ((rc = get_ticket(&db_ticket, env_get("db_pool"), env_get("class"), 
			 app_ticket->assoc)) != LICENSE_SUCCESS) {
      if (rc == LICENSE_FAILURE)
	sprintf(err_msg, "Too many users already logged in. Please try again later.\nCould not get database ticket");
      else
	sprintf(err_msg, "Could not get database ticket\nError Code = %s",
		license_message(rc));
      error(err_msg);
    }
    uid = app_ticket->profile[0];
    machine = app_ticket->profile[1];
    pw = db_ticket->profile[0];
    if (db_ticket->profile[1])
      got_info(db_ticket->profile[1]);

    /* Set up an alarm so we can periodically update the ticket.*/
    if (app_ticket->update)
      updater_id = malarm_add(app_ticket->update, updater);
  }
  else { /* This is a manual login, no license. */
    /* We ask user to type in uid and pw */
    uid = pw = NULL; 
    /* HostName is supposed to be provided in db.conf */
    if (!(machine = env_get("HostName")))
      error("Missing name of database host");
    
    manual_login = TRUE;
  }

  /* Decide if we want to do logging of this session */
  log_chance_str = env_get("log_chance");
  if (log_chance_str) {
    log_chance = atoi(log_chance_str);
    if (log_chance == 0)
      logging = FALSE;
    else if (log_chance >= 100)
      logging = TRUE;
    else {
      srand(getpid());
      logging = ((rand() % 100) < log_chance);
    }
  }
  if (logging) {
    char comment_buf[64];
    
    sprintf(comment_buf, "%s/%s", db_name, uid);
    log_trans(packet->type, comment_buf);
  }
  
  save_login_info(uid, pw, db_name, title_fields, summ_fields, ext_fields);
  state = AWAIT_LOGIN;
  
  /* Set up signal in case child dies */
  signal(SIGCHLD, dead_kid);

  /* fork off a child, which will overlay telnet on itself */
  update_hcb("Establishing network connection to database server...");
  host_pid = vfork();

  if (host_pid == 0) {
    /* in the child, redirect stdin and out */
    close(fileno(stdin)); 
    dup(to_host[R]);
    close(fileno(stdout));
    dup(from_host[W]);
    close(to_host[R]);
    close(to_host[W]); 
    close(from_host[R]); 
    close(from_host[W]);

    /* Doing this somehow prevents weird runaway telnet processes */
    close(fileno(stderr));
    
    /* We're out of here */
    if (execlp(TELNET_PATH, TELNET_NAME, machine, NULL) < 0) {
      /* at this point we are in a child process, so we can't do much */
      perror(NULL);
      /* Must _exit after vfork */
      _exit(-1);
    }
  }
  else if (host_pid > 0) {
    /* in the parent, close extra pipes */
    close(to_host[R]);
    close(from_host[W]);
    
  }
  else {
    error("Could not fork");
  }
}

/* Called by SIGALRM when it is time to update the ticket */
static void updater(void)
{
  int      rc;
  char     err_msg[256];

  if ((rc = update_ticket(app_ticket)) != LICENSE_SUCCESS) {
    sprintf(err_msg, "License manager unable to update application ticket\nError Code = %s", license_message(rc));
    error(err_msg);
  }
  if ((rc = update_ticket(db_ticket)) != LICENSE_SUCCESS) {
    sprintf(err_msg, "License manager unable to update db ticket\nError Code = %s", license_message(rc));
    error(err_msg);
  }
  updater_id = malarm_add(app_ticket->update, updater);
}

/* Called when we get an echo toggle packet */
echo_wcb(packet_rec *packet)
{
  echo = !echo;
}

/* Called when we get a search packet from willow. Pull fields
   out, and put them in two linked lists, one for normal text
   field searches, and one for limit searches. Then pass lists
   to make_search(), a host specific routine. This routine
   frees the data structures created here.
 */
search_wcb(packet_rec *packet)
{
  search_desc_rec *text_fields = NULL,
                  *limit_fields = NULL,
                  *cur_field = NULL,
                  *last_tfield = NULL,
                  *last_lfield = NULL;
  char *data_ptr;
  int i;

  data_ptr = packet->data;
  
  for (i = 0; i < packet->var.search.num_text_fields; i++) {
    cur_field = (search_desc_rec *) malloc(sizeof(search_desc_rec));
    if (*data_ptr <= DAY)
      cur_field->range_type = (int) (*data_ptr++);
    else
      cur_field->range_type = NONE;
    cur_field->text = strdup(data_ptr);
    data_ptr += strlen(data_ptr) + 1;
    cur_field->command = strdup(data_ptr);
    data_ptr += strlen(data_ptr) + 1;
    cur_field->next = NULL;

    /* If it is a range add to limit list, otherwise text list */
    if (cur_field->range_type)
      if (!limit_fields) 
	limit_fields = last_lfield = cur_field;
      else {
	last_lfield->next = cur_field;
	last_lfield = cur_field;
      }
    else
      if (!text_fields) 
	text_fields = last_tfield = cur_field;
      else {
	last_tfield->next = cur_field;
	last_tfield = cur_field;
      }
  }
  for (i = 0; i < packet->var.search.num_limit_fields; i++) {
    cur_field = (search_desc_rec *) malloc(sizeof(search_desc_rec));
    cur_field->range_type = NONE;
    cur_field->command = strdup(data_ptr);
    data_ptr += strlen(data_ptr) + 1;
    cur_field->next = NULL;
    if (!limit_fields) 
      limit_fields = last_lfield = cur_field;
    else {
      last_lfield->next = cur_field;
      last_lfield = cur_field;
    }
  }
  make_search(text_fields, limit_fields);
}

cheat_wcb(packet_rec *packet)
{
  host_send(packet->data);
}

got_info(char *info)
{
  packet_rec packet;
  
  if (!info)
    info = NO_INFO_STR;
  else if (!info[0])
    info = NO_INFO_STR;
  packet.type = INFO;
  packet.var.info.type = COPYRIGHT_INFO;
  packet.data_len = strlen(info) + 1;
  packet.data = NULL;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(info, packet.data_len);
}

send_fields_info(int fields_cnt, char *info, int len)
{
  packet_rec packet;

  packet.type = INFO;
  packet.var.info.type = FIELDS_INFO;
  packet.var.info.fields_cnt = fields_cnt;
  packet.data_len = len;
  packet.data = NULL;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(info, packet.data_len);
}

/*
 * These are the non-host-specific host-callback routines.
 */

/* Called from the state table when we have finished connecting */
connected_hcb(void)
{
  packet_rec packet;

  packet.type = CONNECTED;
  packet.data_len = 0;
  packet.data = NULL;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);

  parse_fields();

  if (logging)
    log_trans(packet.type,"");

}


static int misc_error_hcb(char *msg)
{
  packet_rec packet;

  packet.type = ERROR;
  packet.data_len = strlen(msg) + 1;
  packet.data = NULL;
  
  packet.var.error.type = MISC;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(msg, packet.data_len);
  if (logging)
    log_trans(packet.type, msg);
}

timed_out_hcb(void)
{
  error("System Timeout Limit Exceeded");
}

misc_warning_hcb(char *msg)
{
  packet_rec packet;

  packet.type = ERROR;
  packet.data_len = strlen(msg) + 1;
  packet.data = NULL;
  
  packet.var.error.type = WARNING;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(msg, packet.data_len);
  if (logging)
    log_trans(packet.type, "Warning");
}

void bad_search(char *msg)
{
  packet_rec packet;

  packet.type = ERROR;
  packet.data_len = strlen(msg) + 1;
  packet.data = NULL;
  
  packet.var.error.type = BAD_SEARCH;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(msg, packet.data_len);
  if (logging)
    log_trans(packet.type,"Bad Search");
}

update_hcb(char *msg)
{
  packet_rec packet;

  packet.type = ERROR;
  packet.data_len = strlen(msg) + 1;
  packet.data = NULL;
  
  packet.var.error.type = UPDATE;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(msg, packet.data_len);
}

/* Send the request for password, if appropriate */
ask_pw_hcb(void)
{
  if (manual_login)
    send_user_req(MASKED_REQ, PW_REQ_TEXT, PW_TAG, "Ok", "Cancel");
}

login_msg_hcb(void)
{
  /* Send the request for the user-id */
  if (manual_login)
    send_user_req(TEXT_REQ, UID_REQ_TEXT, UID_TAG, "Ok", "Cancel");

  update_hcb("Logging-in to database server...");
  if (logging) {
    int sdex, len;
    char log_host[64];

    len = strlen(parse_buf);
    for (sdex = 0; sdex < len; sdex++)
      if (parse_buf[sdex] == '\n')
	if (strncmp(parse_buf + ++sdex, "AIX telnet", 10) == 0) {
	  strncpy(log_host, parse_buf + sdex + strlen("AIX telnet  "), 6);
	  break;
	}
    log_trans(TELNETTED, log_host);
  }
}

db_select_hcb(void)
{
  if (logging)
    log_trans(DBED, "");
  update_hcb("Opening selected database...");
}

starting_msg_hcb(void)
{
  update_hcb("Starting database program...");
}


bad_uid_hcb(void)
{
  error("Invalid User Id");
}

bad_login_hcb(void)
{
  error("Invalid Account or Password");
}

bad_db_hcb(void)
{
  error("Unable to access specified database");
}

db_gone_hcb(void)
{
  error("Specified database is currently unavailable.\nPlease try again later.");
}

void force_quit_hcb(void)
{
  signal(SIGCHLD, SIG_IGN);
  sleep(1);
  if (host_pid)
    kill(host_pid, SIGKILL);
  wait( NULL );
  close(to_host[W]);
  close(from_host[R]);
  cleanup();
}


/* Close things up, and quit */
cleanup(void)
{

  if (app_ticket) {
    malarm_cancel(updater_id);
    return_ticket(app_ticket);
  }
  if (db_ticket)
    return_ticket(db_ticket);

  list_close();

  /* We are out of here */
  exit(0);
}

/* close off our connection to Willow, we are on our own */
close_willow(void)
{
  connected_to_willow = FALSE;
  close(from_willow);
  if (!socketed)  /* don't want to close socket twice */
    close(to_willow);

  if (!echo) {
    close(fileno(stdout));
    close(fileno(stderr));
    close(fileno(stdin));
  }
}

/* Send the stats back to willow */
send_stats_cmd(char *stats, int num_results, int search_num)
{
  packet_rec packet;
  
  packet.type = STATS;
  packet.data = NULL;
  packet.var.stats.result_id = search_num; 
  packet.var.stats.num_results = num_results;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);

  if (stats == NULL) {
    packet.data_len = 0;
    willow_send(netify(&packet), PACKET_NET_SIZE);
    if (logging) {
      char comment[16];
      sprintf(comment, "%d hits", num_results);
      log_trans(packet.type, comment);
    }
  } 
  else {
    packet.data_len = strlen(stats) + 1;
    willow_send(netify(&packet), PACKET_NET_SIZE);
    willow_send(stats, packet.data_len);
    if (logging)
      log_trans(packet.type,stats);
  }
}

/* Send a summary back to willow */
send_title(int num, char *title)
{
  packet_rec packet;

  packet.type = TITLE;
  packet.var.title.title_num = num;
  packet.data = NULL;
  packet.data_len = strlen(title) + 1;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(title, NULL);
}

/* Send back an extension record */
void send_ext(int num, char *text)
{
  packet_rec packet;

  packet.type = EXTS;
  packet.data = NULL;
  packet.var.exts.ext_num = num;
  packet.data_len = strlen(text) + 1;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(text, NULL);
}

/* Send a user request packet */
void send_user_req(int type, char *query_text, char *query_tag,
		   char *yes_label, char *no_label)
{
  packet_rec packet;

  packet.type = USER_REQ;
  packet.var.user_req.type = type;
  packet.data = NULL;
  packet.data_len = strlen(query_text) + 1 +
                    strlen(query_tag) + 1 +
		    strlen(yes_label) + 1 +
		    strlen(no_label) + 1;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(query_text, NULL);
  willow_send(query_tag, NULL);
  willow_send(yes_label, NULL);
  willow_send(no_label, NULL);
  last_sent_req = query_tag;
}

/* Send back a short record */
void send_summ_record(int num, char *text)
{
  packet_rec packet;
  char *hacked_text;

  hacked_text = lameify(text, NULL);
  packet.type = RECORD;
  packet.var.record.summary = TRUE;
  packet.var.record.num_hilites = 0;
  packet.data = NULL;
  packet.var.record.ref_num = num;
  packet.data_len = strlen(hacked_text) + 1;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  willow_send(hacked_text, NULL);
  if (text != hacked_text)
    free(hacked_text);
}


/* Send back full text retrieval */
send_full_record(int num, char *text)
{
  packet_rec packet;
  hl_rec    *hl, *hl_prev;
  netint     hl_buf;
  char *hacked_text;

  hacked_text = lameify(text, hl_head);
  packet.type = RECORD;
  packet.var.record.summary = FALSE;
  packet.var.record.num_hilites = num_hilites;
  packet.data = NULL;
  packet.var.record.ref_num = num;
  packet.data_len = strlen(hacked_text) + 1 + 
    packet.var.record.num_hilites * 2 * sizeof(int);
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);

  /* Send out the hi-lite information, and destroy the list */
  hl = hl_head;
  while (hl) {
    hl_buf = htonl(hl->start);
    willow_send(&hl_buf, sizeof(netint));
    hl_buf = htonl(hl->end);
    willow_send(&hl_buf,   sizeof(netint));
    hl_prev = hl; 
    hl = hl->next;
    free(hl_prev);
  }
  num_hilites = 0;
  hl_head = hl_tail = NULL;
  willow_send(hacked_text, NULL);
  if (logging)
    log_trans(packet.type, "");
  if (text != hacked_text)
    free(hacked_text);
}

int send_browse(int ref_num, char *list_data, int line_pos)
{
  packet_rec packet;

  packet.type = BROWSE;
  packet.var.browse.ref_num = ref_num;
  packet.var.browse.line_pos = line_pos;
  packet.data = NULL;
  if ((int) list_data != -1)
    packet.data_len = strlen(list_data) + 1;
  else {
    packet.data_len = 0;
    packet.var.browse.line_pos = -1;
  }
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  willow_send(netify(&packet), PACKET_NET_SIZE);
  if ((int) list_data != -1)
    willow_send(list_data, NULL);
}

/* Miscellaneous */
error(char *msg)
{
  misc_error_hcb(msg);
  fprintf(stderr, "FATAL ERROR IN DRIVER: %s\n",msg);fflush(stderr);
  disconnect_wcb();
}

/* Set up an alarm handler */
set_alarm(int time_out)
{
  malarm_add(time_out, force_quit_hcb);
}

/* append a carriage return to given malloced string */
char *append_cr(char *s)
{
  char *new;

  new = malloc(strlen(s) + 2);
  strcpy(new, s);
  strcat(new, "\n");
  free(s);
  return(new);
}

/* Copy a string into a newly created string of length n*/
char *strndup(char *s, int n)
{
  char *new_str;

  new_str = malloc(n);
  return(strcpy(new_str, s));
}

/* Called if parse buf gets full */
grow_parse_buf(void)
{
  static int call_count = 0;
  call_count++;
  if (call_count > 100)
    error("Parse buf getting way too big");
  parse_buf_size = (call_count + 1) * PARSE_BUF_SIZE;
  parse_buf = realloc(parse_buf, parse_buf_size);
}

/* If child dies */
void dead_kid(void)
{
  if (state != NO_CON)
    /* Could we be inviting a race condition here? */
    error("Connection to database host closed");

  /* short pause so our packets don't over-run each other */
  sleep(1);
  cleanup();
}

/* Create a node in our linked list of hilite points, then record the
   starting offset for it. */
hl_start(int point)
{
  hl_rec *hl;

  num_hilites++;
  hl = (hl_rec *) malloc(sizeof(hl_rec));
  hl->start = point;
  hl->end = -1;
  hl->next = NULL;

  if (!hl_head)
    hl_head = hl_tail = hl;
  else {
    hl_tail->next = hl;
    hl_tail = hl;
  }
}

/* Record the end-offset of a hilite region, check for errors to avoid
   crashing because of bad data, but don't bother reporting them. */
hl_end(int point)
{
  if (hl_tail)
    if (hl_tail->end == -1)
      hl_tail->end = point;
}

static void echo_packet(packet_rec *packet, char send)
{
  if (send)
    fprintf(stderr,"DRIVER SENDING ");
  else
    fprintf(stderr,"DRIVER RECEIVED ");
  switch (packet->type) {
    case CONNECT   : fprintf(stderr, "CONNECT");     break;
    case CONNECTED : fprintf(stderr, "CONNECTED");   break;
    case ERROR:
      switch (packet->var.error.type) {	
        case MISC:      fprintf(stderr, "MISC-ERR");      break;
	case BAD_SEARCH:fprintf(stderr, "BAD_SEARCH-ERR");break;
	case DEAD:	fprintf(stderr, "DEAD-ERR");      break;
	case UPDATE:	fprintf(stderr, "UPDATE");        break;
	case WARNING:   fprintf(stderr, "WARNING");       break;
      }
      break;    
    case SEARCH     : fprintf(stderr, "SEARCH");     break;
    case STATS      : fprintf(stderr, "STATS");      break;
    case TITLE_REQ  : fprintf(stderr, "TITLE_REQ");  break;
    case TITLE      : fprintf(stderr, "TITLE");      break;
    case RECORD_REQ : fprintf(stderr, "RECORD_REQ"); break;
    case RECORD     : fprintf(stderr, "RECORD");     break;
    case DISCONNECT : fprintf(stderr, "DISCONNECT"); break;
    case TOGGLE_ECHO: fprintf(stderr, "TOGGLE_ECHO");break;
    case CHEAT      : fprintf(stderr, "CHEAT");      break;
    case CANCEL     : fprintf(stderr, "CANCEL");     break;
    case INFO       : 
      switch (packet->var.info.type) {
        case COPYRIGHT_INFO: fprintf(stderr, "INFO-COPYRIGHT");break;
        case FIELDS_INFO   : fprintf(stderr, "INFO-FIELDS");   break;
      }
      break;
    case EXT_REQ    : fprintf(stderr, "EXT_REQ");    break;
    case EXTS       : fprintf(stderr, "EXTS");       break;
    case USER_REQ   : 
      switch (packet->var.user_req.type) {
        case BOOL_REQ   : fprintf(stderr, "USER_REQ-BOOL");  break;
        case TEXT_REQ   : fprintf(stderr, "USER_REQ-TEXT");  break;
        case MASKED_REQ : fprintf(stderr, "USER_REQ-MASKED");break;
      }
      break;
    case USER       : fprintf(stderr, "USER");       break;
    case BROWSE_REQ : fprintf(stderr, "BROWSE_REQ"); break;
    case BROWSE     : fprintf(stderr, "BROWSE");     break;
    default: fprintf(stderr, "UNKNOWN???");
    }
  fprintf(stderr, "\n");
  fflush(stderr);
}

dump_state (int state)
{
  fprintf(stderr, "  NEW STATE: %s\n", 
	  state == NO_CON       ? "NO_CON" :
	  state == AWAIT_LOGIN  ? "AWAIT_LOGIN" :
	  state == AWAIT_PW     ? "AWAIT_PW" :
	  state == AWAIT_DB     ? "AWAIT_DB" :
	  state == AWAIT_SEARCH ? "AWAIT_SEARCH" :
	  state == READY        ? "READY" :
	  state == AWAIT_STATS  ? "AWAIT_STATS" :
	  state == DISP_SHRT    ? "DISP_SHRT" :
	  state == DISP_LONG    ? "DISP_LONG" :
	  state == AWAIT_INFO   ? "AWAIT_INFO" :
	  "UNKNOWN???");
  fflush(stderr);
}

char *env_get(char *s)
{
  int i;
  
  for (i = 0; i < env_cnt; i++)
    if (strcmp(s, env[i].name) == 0)
      return(env[i].val);
  return(NULL);
}
