/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/************************************************************************
 * getclass.c
 * This module contains code for figuring out where a given IP address
 * comes from. It is only used when the driver is run under a socket.
 * That way we can put our driver up on a public server, but when the
 * the driver contacts our license manager it will use the correct 
 * class.
 * 
 * Code written by Steve Jones, noyd@cac.wahington.edu
 ************************************************************************/

#include <stdio.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif

#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

#define TOKEN_LIM      5*sizeof(long int) 
#define STRLIM         256

#ifdef DEBUG
#  undef CLASSFILE
#  define CLASSFILE      "./class.data"
#endif

static char Class[TOKEN_LIM];

struct c_entry {
      struct in_addr network;
      struct in_addr netmask;
      char class[TOKEN_LIM];
      struct c_entry *next;
};

static char *getclass(char *file,struct in_addr *inp);

extern int error(char *msg);

/* Compare whether the passed in class string matches the actual class
   of our socket connection.
 */
void validate_class(int sock, char *class, char *class_file)
{
  struct sockaddr_in name;
  char *peer_class;
  int len;
  char err_buf[256];

  len = sizeof(struct sockaddr);
  if (getpeername(sock, &name, &len) < 0) {
    sprintf(err_buf, "Could not getpeername() ERR %d", errno);
    error(err_buf);
  }

  if (!(peer_class = getclass(class_file, &name.sin_addr))){
    error("Could not determine class of socket connection");
  }

  if (!class) {
    sprintf(err_buf, "Specification of connection class missing");
    error(err_buf);
  }
    

  if (strcmp(class, peer_class)) {
    sprintf(err_buf, "Class mismatch:\nRequesting license for class '%s',\nbut connection is from class '%s'", class, peer_class);
    error(err_buf);
  }
}


/*  Line parser for getclass

   Mangles input line with strtok and fills c_entry structure 
   pointed to by ep.

   returns 1 on success, 0 on failure 

   *SJ  Presumption is "s_addr" is an unsigned long
*/
static int parse2(char *line, struct c_entry *ep)
{
  char *np, *cp, *ap, *mp;
  long int len, count;
  int i;

  /* extract tokens from line */
  np = strtok(line," \t");
  if ( !np ) return(0);
  cp = strtok(NULL," \t");
  if ( !cp ) return(0);

  /* parse address  and set network and netmask */
  ap = strtok(np,"/");        /* network or IP address */
  if ( !ap ) return(0);
  mp = strtok(NULL,"/");      /* network address length */
  
  ep->network.s_addr = 0;     
  ep->netmask.s_addr = 0;     

  for (count = 0, ap = strtok(ap,".");
       ap ; 
       ap = strtok(NULL,"."), count++ ) {
    i = atoi(ap);
    ep->network.s_addr = (ep->network.s_addr << 8) | i;
  }
  for ( i = count; i < 4; i++ ) {
    ep->network.s_addr = ep->network.s_addr << 8;
  }

  ep->network.s_addr = htonl(ep->network.s_addr);

  if ( mp ) {  /* network mask length */
    len = atol(mp);
  } else {
    len = count * 8;
  }
  /*SJ Presumption is "s_addr" is an unsigned long */
  if ( len > 0 ) {
    /* left shift with zero fill */
    ep->netmask.s_addr = (~0 << (sizeof(unsigned long)*8 - len));
  } else { 
    ep->netmask.s_addr = 0;
  }
  ep->netmask.s_addr = htonl(ep->netmask.s_addr);
  
  /* copy class */
  strncpy(ep->class,cp,TOKEN_LIM);
  ep->class[TOKEN_LIM-1] = '\0';

  /* done */
  return(1);
}

/* repetitive checking for class using name and number. Not used
   by willow database drivers, since we only have to check once. */
static char *getclassr(char *file,struct in_addr *inp)
{
  int             match;
  char            line[STRLIM];
  register char   *cp;
  struct c_entry  *np,*nextp;
  struct c_entry  entry;
  int             i;

  static FILE   *class_fp;
  static struct c_entry *head = NULL;

  /* check for reset */
  if ( inp == NULL ){
    for ( np = head; np; np = nextp) {
      nextp = np->next;
      free(np);
    }
    head = NULL;
    return (NULL);
  }

  if ( ! head ) {
    /* open classes file */
    class_fp = fopen(file, "r");
    if (class_fp == NULL) {
      char err_buf[256];

      sprintf(err_buf, "Could not open class-file '%s'", file);
      error(err_buf);
    }

    /* read file and save parsed information on linked list */
    for ( ;; ) {
      cp  = fgets(line, sizeof(line), class_fp);
      if ( cp == NULL ) break;
      if ((i = strcspn (line, "#\n")) < strlen (line)) cp[i] = '\0';
      if (*line == '\0') continue;

      if ( !parse2(line,&entry) ) continue; 

      /* build list */
      np = (struct c_entry *) malloc(sizeof(struct c_entry));
      if ( !np) break;
      *np = entry;
      if ( head == NULL ) {
        head = np;
        nextp = head;
      } else {
        nextp->next = np;
        nextp = np;
      }
    }
    (void) fclose(class_fp);
  }

  /* search list for matches. */
  match = 1;
  for (nextp = head; nextp; nextp = nextp->next ) {
    if ( (inp->s_addr & nextp->netmask.s_addr) ==
         nextp->network.s_addr ) {
      match = 0;
      break;
    }
  }

  if ( match ) return ( NULL );
  strcpy(Class,nextp->class);
  return(Class);
}


/* Single use checking for class with name and number */
static char *getclass(char *file,struct in_addr *inp)
{
        int             match;
        char            line[STRLIM];
        register char   *cp;
        FILE   *class_fp;
        struct c_entry  entry; 
        int             i;
	
        /* open classes file */
        class_fp = fopen(file, "r");
        if (class_fp == NULL) {
	  char err_buf[256];

          sprintf(err_buf, "Could not open class-file '%s'", file);
	  error(err_buf);
        }

        match = 1;
        for ( ;; ) {
          cp  = fgets(line, sizeof(line), class_fp);
          if ( cp == NULL ) break;
          if ((i = strcspn (line, "#\n")) < strlen (line)) cp[i] = '\0';
          if (*line == '\0') continue;
 
          if (! parse2(line,&entry)) continue;

          /* perform match */
          if ( (inp->s_addr & entry.netmask.s_addr) ==
               entry.network.s_addr ) {
            match = 0;
            break;
          }
          
        }

        (void) fclose(class_fp);
        if ( match ) return ( NULL );
        strcpy(Class,entry.class);
        return(Class);
}
