/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/***************************************************************************
 * iii.c
 * This file contains the routines that are used to get circulation status
 * from the III box.
 ***************************************************************************/

#include  <stdio.h>
#include  <ctype.h>
#ifdef SOLARIS
#include <string.h>
#else /* SOLARIS */
#include  <strings.h>
#endif /* SOLARIS */

#include "isc/isc.h"
#include "hash.h"
#include "host.h"
#include "driver.h"

#define CNT_STR   "LH"
#define IT_STR    "IT"
#define OR_STR    "OR"
#define PR_STR    "PR"
#define SS_STR    "SS"
#define ERR       "<UNABLE TO RETRIEVE CIRCULATION INFORMATION>\n"
#define TOO_MANY  "<TOO MANY VOLUMES TO DISPLAY, PLEASE CHECK THE SHELVES>\n"

static int sock = 0;

extern int circ_time;
extern int max_circ;
extern char *circ_mode;
extern char *circ_key;
extern int  summary_long_len;
extern char *circ_host;
extern hash_tbl *summ_form_tab;     /* hash table of field format records */
extern summ_tab_entry *summ_tab;

/* External routines */
extern void *malloc (size_t);
extern int  hash_compare(char *key, format_rec *element);

/* Forward routines */
void open_circ(void);
void close_circ(void);
char *get_circ(int num);
static char *errify(int num, char *msg);
static char *gcat_errify(int num);

/* This routine initializes a circ connection. It returns nothing. If there 
   was a failure, it is reported in get_circ(). */
void open_circ(void)
{
  if (sock > 0)
    close_circ();
  sock = IscInit(circ_host);
}

/* Close a circ connection */
void close_circ(void)
{
  if (sock <= 0)
    return;

  IscTerm(sock);
  sock = 0;
}

/* Take out weird escape chars */
char *strip(char *buffer)
{
  int sdex = 0, tdex = 0; /* source and target indices */
  int len;
  
  len = strlen(buffer);
  
  while (sdex < len) {
    /* isprint() does not work for >128 */
    if (!(buffer[sdex] & 128))
      if (isprint(buffer[sdex]) || (buffer[sdex] == '\n'))
	buffer[tdex++] = buffer[sdex];
    sdex++;
  }
  buffer[tdex] = 0;
  return(buffer);
}

/* This routine takes an index into our summary table, and sends back
   the unexpanded circ message.
 */
char *get_circ(int num)
{
#define CIRC_AVAILABLE
#ifdef CIRC_AVAILABLE   
  int res_len;
  int sdex = 0, tdex = 0;
  char key[64];
  char *circ;
  static char broken = FALSE;
  format_rec *summ_form;
  char *data;
  int  circ_cnt;

  /* Once we get an error, never try again */
  if (broken)
    return(errify(num, ERR));

  IscSetParams(circ_time);
  if (sock == 0)
    open_circ();

  if (sock < 0) {
    broken = TRUE;
    return(errify(num, ERR));
  }

  /* Find the current value of LH, which indicates how many items, but it
     is ok if there is no value */
  if (summ_form = (format_rec *) hash_Lookup(summ_form_tab, 
		      hash_HashFunction(CNT_STR, 2), hash_compare, CNT_STR)){
    if (summ_tab[num].fields[summ_form->idex]) {
      circ_cnt = atoi(summ_tab[num].fields[summ_form->idex] + ABBREV_INDENT);
      if (circ_cnt> max_circ)
	return(errify(num, TOO_MANY));
      /* The -1 flag means we are in GCAT and need to get circ from OR/PR */
      else if (circ_cnt == -1)
	return(gcat_errify(num));
    }
  }

  /* Extract the key */
  summ_form = (format_rec *) hash_Lookup(summ_form_tab, 
	         hash_HashFunction(circ_key, 2), hash_compare, circ_key);
  if (summ_tab[num].fields[summ_form->idex])
    data = summ_tab[num].fields[summ_form->idex] + ABBREV_INDENT;
  else
    return(errify(num, ERR));
  tdex = sdex = 0;
  while (data[sdex] && (data[sdex] != '\n') && (data[sdex] != '.') &&
	 (data[sdex] != ' '))
    key[tdex++] = data[sdex++];
  key[tdex] = 0;

  if (!IscGetHoldingStatus(sock, circ_mode, key, FALSE, TRUE, 
			   78 - summary_long_len + ABBREV_INDENT,
			   TRUE, ABBREV_INDENT, &circ, &res_len)) {
    broken = TRUE;
    return(errify(num, ERR));
  }
  return(strip(circ));
#else
  return(strdup("No CIRC status."));
#endif
}

/* Return an error message, and contents of IT field if possible. */
static char *errify(int num, char *msg)
{
  int len = 0;
  char *err_buf;
  format_rec *summ_form;
  char *it;

  summ_form = (format_rec *) hash_Lookup(summ_form_tab, 
	         hash_HashFunction(IT_STR, 2), hash_compare, IT_STR);
  if (summ_form)
    it = summ_tab[num].fields[summ_form->idex];
  else 
    it = NULL;
  if (it) {
    it += ABBREV_INDENT;
    len = strlen(it);
  }
  err_buf = malloc(len + strlen(msg) + 16);

  if (!it)
    sprintf(err_buf, "%s   %s", SS_STR, msg);
  else
    sprintf(err_buf, "%s   %s\n     %s", SS_STR, it, msg);

  return(err_buf);
}

/* Return contents of OR or PR field for a GCAT circ request */
static char *gcat_errify(int num)
{
  format_rec *summ_form;
  char *alt_val;

  summ_form = (format_rec *) hash_Lookup(summ_form_tab, 
	         hash_HashFunction(OR_STR, 2), hash_compare, OR_STR);
  if (summ_form)
    alt_val = summ_tab[num].fields[summ_form->idex];
  else 
    return(NULL);

  if (!alt_val) {
    summ_form = (format_rec *) hash_Lookup(summ_form_tab, 
		       hash_HashFunction(PR_STR, 2), hash_compare, PR_STR);
    if (summ_form)
      alt_val = summ_tab[num].fields[summ_form->idex];
    else 
      return(NULL);
  }
  return(alt_val);
}
