/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/* list-server.c

   Connects to our network list server.
 */

/*
 * 
 * MODIFICATION HISTORY:
 *
 * xxx  Mathew Freedman Created.
 *
 * 8/93 Tom Unger       Added logic to keep connection to client open.
 *                      Added request number to requests and responces and
 *                      look for EOT character at end of responce.
 *
 * CONNECTED CLIENTS:
 * 
 *	Originally, the cliente created a connection to the server,
 *	sent a request, waited for the responce, and closed to
 *	connection.  To eliminate the overhead of creating
 *	connections for every requeset I've changed the code so that
 *	the client keeps a connection to the server open.
 *	
 *	In order that the client can match responces with it's
 *	requests I've added a request number that the client sends at
 *	the beginning of all requests.  The server replies with the
 *	same rquest number in the reply.
 *	
 *	Previously, after sending the reply the server closed the
 *	connection.  The client used this to know when it had read
 *	the complete reply.  Since the connection stays open
 *	something else is needed.  The server sends an EOT character
 *	at the end of the responce.
 *	
 *	If for some reason a requests fails the client closes the
 *	connection and creates a new connection.  This is easier than
 *	trying to re-sync the current connection.
 *	
 *
 * RECURSIVE CALLS
 *	
 *	Originally this list client code was written to with a mini X11
 *	event loop that it executed while waiting for the reply.  This
 *	could lead to a recursive call to the list_server () function.
 *	For now I've removed that code.  All that remains is the funkey
 *	way that read_it uses longjump to return (instead of simply 'return').
 */



#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/param.h>
#include <sys/file.h>
#include <netdb.h>
#include <setjmp.h>

#ifdef vax
char *malloc();
char *realloc();
#else
#include <malloc.h>
#endif
#include <errno.h>
#include <sys/ioctl.h>

#ifdef _AIX
#include <sys/select.h>
#endif /* AIX */


#ifdef TOM_DEBUG
#include "mc.h"
#else
#include "host.h"
#endif

#define SERV_HOST_NAME  "brs_server.u.washington.edu"
/*#define SERV_HOST_NAME  "haley4.u.washington.edu"*/
/*#define SERV_ADDR        0x8c8e3501 */


#define SERV_PORT_DEFAULT 3050 /* HARD_DEFAULT */

#define LINEMAX  1024
#define LINE_AVG 25
#define EOT	((char) 0x04)

/* Forward/extern declarations */
int list_close (void);
extern int malarm_add (int interval, void (*handler)());
extern void malarm_cancel (int id);
extern int close (int);
extern int write (int, void *, unsigned int);
extern long int atol (char *);
extern int read (int, void *, unsigned int);
static void read_it(int lines_wanted, int *fid);
static void cancel_read(void);

/* Our file descriptor. Only one across recursive calls */
static int   sockfd = 0;

/* Request count. */
static int   rq_count = 0;

/* Our data area */
static char *list_area = NULL;


/* Jump-buffer */
static jmp_buf env;

/* Statistics that we gather. */
#ifdef TOM_DEBUG
extern ChildMSG MSG;
extern BOOL	ChildDebug;
#endif

/* For debugging */
static int BytesRead;

/*
 * Establish a connection to the server.
 */
int
list_connect (void)
{
  struct sockaddr_in serv_addr;
  static int  port = 0;
  static int  addr = 0;
#if 0
  int		f;
  int		rc;
#endif
  int         alarm_id;

  if (sockfd != 0)
	  list_close ();

#ifdef TOM_DEBUG
  if (ChildDebug) 
	  fprintf (stderr, "list_server:  Looking up host name\n");
#endif
  if (!(port && addr)) {
    struct hostent    *host;

    port = htons(SERV_PORT_DEFAULT);

#ifdef SERV_ADDR
    addr = htonl (SERV_ADDR);
#else
    if ((host = gethostbyname(SERV_HOST_NAME)) == 0) {
      perror("list_lookup: can't get list-server address");
#ifdef TOM_DEBUG
      ++MSG.reqNetFail;
#endif
      sockfd = 0;
      return (-1);
    }
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memmove(&addr, host->h_addr, sizeof(addr));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
    bcopy(host->h_addr, &addr, sizeof(addr));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
#endif
  }
#ifdef TOM_DEBUG
  if (ChildDebug) 
      fprintf (stderr, "list_server:  Got address\n");
#endif
#if defined(POSIX) || defined (_POSIX_SOURCE)
  memset( (char *) &serv_addr, 0, sizeof(serv_addr));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
  bzero( (char *) &serv_addr, sizeof(serv_addr));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  serv_addr.sin_family              = AF_INET;
#if defined(POSIX) || defined (_POSIX_SOURCE)
  memmove(&serv_addr.sin_addr, &addr, sizeof(addr));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
  bcopy(&addr, &serv_addr.sin_addr, sizeof(addr));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  serv_addr.sin_port                = port;

  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("list_lookup: can't open stream socket");   
#ifdef TOM_DEBUG
    ++MSG.reqLFail;
#endif
    sockfd = 0;
    return (-1);
  } 

  /* Set a timeout to interrupt the connect if it takes too long */
  alarm_id = malarm_add(LIST_TIMEOUT, (void(*)()) NULL);
  if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0){
    perror("list_lookup: can't connect to server");
    list_close();
    switch (errno) {
    case ENETDOWN: case ENETUNREACH: 
#ifdef TOM_DEBUG
	++MSG.reqNetFail;
#endif
	break;

    default:
#ifdef TOM_DEBUG
	++MSG.rConnFail;
#endif
	break;
    }
    sockfd = 0;
    return (-1);
  } 
  malarm_cancel(alarm_id);
  rq_count = 0;

#if 0
  /* Set the socket to non-blocking I/O.*/
  f = TRUE;
  rc = ioctl (sockfd, FIONBIO, (caddr_t) &f);
  if (rc == -1) {
      perror ("ioctl to set pipe non blocking.");
      return (-1);
  }
#endif

  

  return (0);
}




/*
 * Close the connection to the server.
 */
int
list_close (void)
{
  if (sockfd)
    close (sockfd);
  sockfd = 0;
}



/*
 * Blocking routine to send request to the server.
 * Expects a connection to exist, but if there is none will create one.
 */
char *
list_lookup(char *db_name, 
	    char *list_name, 
	    char search_type, 
	    char *prefix_st, 
	    int lines_wanted, 
	    int *line_pos, 
	    int line_offset)
{
  int   n;
  struct sockaddr_in serv_addr;
  char str[LINEMAX];
  int  rc;
  char	*la;
  int	write_attempts;
  int	rtn_rq_count;


  /* If no socket, try to open one. */
  if (sockfd == 0) {
      rc = list_connect ();
      if (rc != 0)
	  return ((char *)-1);
  }


  sprintf(str,"%s\n%d\n%s\n%s\n%c\n%s\n%d\n%d\n", "willow", ++rq_count, db_name, 
	  list_name, search_type, prefix_st, lines_wanted, line_offset);
  
  n = strlen(str);

  /* 
   * Write message to server.  
   * If the write failed because the socket was closed then perhaps the
   * server timed us out and closed the connection on us.  try re-establishing
   * the connection before complaining to the user. 
   */
  write_attempts = 0;
 WriteAgain:
  if (write(sockfd, str, n) != n) {
      if (++write_attempts < 2 && (errno == EBADF ||
				   errno == ETIMEDOUT ||
				   errno == ECONNRESET)) {
	  list_close ();
	  if (list_connect () == 0) 
	      goto WriteAgain;
      }
      perror("list_lookup: write error on socket");
#ifdef TOM_DEBUG
      ++MSG.reqFail;
#endif
      return (char *) -1;
  } 


  /* Read the responce. */
 ReadAgain:
  BytesRead = 0;
  rc = setjmp(env);
  if (rc == 0) {
      read_it(lines_wanted, NULL);
    
  }
  else if (rc < 0) {
      /* rc -2 means we timed out.  Close the connection. */
      if (rc == -2) {
	  fprintf (stderr, "  Bytes read before timeout:  %d\n", BytesRead);
	  list_close ();
      }

      /* Error return from read_it, read_timeout */
      return (char *) -1;
  }

  /* Compare the returned request count with the expected request count
   * if it does not match then this may be the responce to a previous
   * request. 
   *
   * The responce is in the format:
   * 	NNNNNNNN\n		// request number.
   *	NNNNNNNN\n		// line_pos.
   *	(text data)
   *
   * line_pos of 0 seems to indicate that there is an error.
   *
   * Unless it is a header request in which case the request number
   * is followed by different information.
   */
  
  la = list_area;
  rtn_rq_count = atol(la);
  if (rtn_rq_count != rq_count)
	  goto ReadAgain;
  la += 9;
  
  /* If we make it here, we are returning healthily from read_it() */
  *line_pos = atol(la);
  if (search_type == 'H') {
      if (*line_pos == 0) {
#ifdef TOM_DEBUG
	  ++MSG.reqLFail;
#endif
	  return (char *) -1;
      }
      else
	  return(la);
  }
  else
	  return (&la[9]);
}  /* end of list_lookup */


static void read_it(int lines_wanted, int *fid)
{
  char *str_ptr;
  int   n, room;
  static int size;
  fd_set read_fds;
  int    num_fds;
  struct timeval timeout;
  int    status;
  int    alarm_id;


  if (!list_area) {
      size = lines_wanted * LINE_AVG + LINEMAX;
      if ((list_area = malloc(size)) == NULL){
	  perror("list_lookup: no memory available to store the list");
#ifdef TOM_DEBUG
	  ++MSG.reqLFail;
#endif
	  longjmp(env, -1);
      } 
  }


  BytesRead = 0;
  room = size - 1;
  *list_area = '\0'; 
  str_ptr = list_area; 

  

  /* Establish timeout.  cancel_read() does a long jump out of this
   * routine so we don't need any timeout logic here. */
  alarm_id = malarm_add(LIST_TIMEOUT, cancel_read);

  /* Read data until we receive an EOT character */
  do {
      n = read(sockfd, str_ptr, room);
      if (n < 0) {
	  perror ("list_lookup:  error reading data");
#ifdef TOM_DEBUG
	  ++MSG.reqLFail;
#endif
	  malarm_cancel(alarm_id);
	  longjmp (env, -1);
      }
      str_ptr += n;
      BytesRead += n;
      room -= n;
      if (room < 100){
	  int diff;
	  
	  diff = str_ptr - list_area;
	  size += LINEMAX;
	  room += LINEMAX;
	  list_area = realloc(list_area, size);
	  str_ptr = list_area + diff;
      }
  } while (*(str_ptr-1) != EOT);

  malarm_cancel(alarm_id);

  *(--str_ptr) = '\0';
  longjmp(env, 1);
}


/* Called if read_it not called first by AddInput */
static void cancel_read(void)
{
  perror("list_lookup: timed out on read");
#ifdef TOM_DEBUG
  ++MSG.reqTimeOut;
#endif
  longjmp(env, -2);
}

