/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/***************************************************************************
 * config_db.c
 * This file contains code for reading in the database configuration
 * file for Willow and for Wilco.
 * 
 * config_dbs(db_config_file_name, config_path, db_hash_table, db_menu_list)
 *       is called to read the file. It returns a count of the number of dbs.
 *
 * The following application-specific routines must be declared externally.
 *   error(msg):  fatal error handling routine
 *
 * The following utility functions are needed, and are defined in utils.c
 *   get_line(str, fp) 
 ***************************************************************************/

/*
 * 05/17/94  slh  allow more than one possible comment char
 */

#include <stdio.h>
#include <ctype.h>

#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef WILLOW
#include <Mrm/MrmAppl.h>
#else
#define XtMalloc  malloc
#define XtRealloc realloc
#define XtFree    free

extern char *malloc(size_t);
extern char *realloc(void *, size_t);
#endif

#include "db.h"
#include "hash.h"

/* Global exported variables */
/* Name of current config file, used for error messages */
char *config_file_name = "";

/* Number of lines read from config file, for error messages */
int  config_line_num = 0;

/* Private Globals */
/* A table for holding the possible fields for the database config file */
static char *db_config_fields[NUM_CONFIG_FIELDS];

/* hash table for quick lookup of config fields */
hash_tbl *conf_hash_tab;     /* hash table of field format records */

/* Forward declarations */
static db_desc_rec *new_db(void);
static db_field_desc_rec *new_t_field(void);
static db_ext_rec *new_ext(void);
static void config_err(char *msg);
extern int error (char *);
extern int atoi ( char *);
extern char *get_line (char *s, FILE *f);
int scan_config (FILE *infile, char *value);
int conf_compare(char *key, conf_hash_elmt *element);
int db_compare(char *key, db_desc_rec *element);

/* Read the database configuration file, and build the database menus
   and data-structures based on that info. Return a count of the number
   of databases.
 */ 
int config_dbs(char *db_config_file_name,
	       char *path,
	       hash_tbl **db_hash_tab,
               db_menu_rec **db_menu_list)
{
  char  *file_name;
  FILE *infile;
  int    field;              /* Which field we are at in config file */
  char   value[MAX_LINE];    /* value of current field */
  db_desc_rec *db_desc;      /* description of a single database */
  db_field_desc_rec *db_field, /* description of a single text/xlimit field */
                    *prev_db_field; /* previous field, for building list */
  db_ext_rec  *ext, *prev_ext;  /* multimedia extensions ptr */
  int    num, i;
  conf_hash_elmt *element;     /* element for hash table insertion */
  db_menu_rec *db_menu_ptr = NULL;  /* ptr for building menu list */
  int db_count = 0;
  char    default_menu = TRUE; /* should make up our own MENU directives? */

  *db_menu_list = NULL;
  /* Initialize the db_config_fields table */
  db_config_fields[DB_TAG]       = DB_TAG_S;
  db_config_fields[DB_NAME]      = DB_NAME_S;
  db_config_fields[DB_LONG]      = DB_LONG_S;
  db_config_fields[DRIVER]       = DRIVER_S;
  db_config_fields[DB_INT_NAME]  = DB_INT_NAME_S;
  db_config_fields[SUMM_FIELDS]  = SUMM_FIELDS_S;
  db_config_fields[TITLE_FIELDS] = TITLE_FIELDS_S;
  db_config_fields[T_FIELD_NAME] = T_FIELD_NAME_S;
  db_config_fields[T_FIELD_CMD]  = T_FIELD_CMD_S;
  db_config_fields[LIST_NAME]    = LIST_NAME_S;
  db_config_fields[LIST_HIER]    = LIST_HIER_S;
  db_config_fields[LIST_MNEMONIC]= LIST_MNEMONIC_S;
  db_config_fields[DFLT_AT]      = DFLT_AT_S;
  db_config_fields[L_FIELD_NAME] = L_FIELD_NAME_S;
  db_config_fields[L_FIELD_CMD]  = L_FIELD_CMD_S;
  db_config_fields[ECHO_HOST]    = ECHO_HOST_S;
  db_config_fields[USE_LOCAL]    = USE_LOCAL_S;
  db_config_fields[USE_FTP]      = USE_FTP_S;
  db_config_fields[USE_EMAIL]    = USE_EMAIL_S;
  db_config_fields[FTP_HOST]     = FTP_HOST_S ;
  db_config_fields[FTP_UID]      = FTP_UID_S;
  db_config_fields[FTP_APPEND]   = FTP_APPEND_S;
  db_config_fields[PRINT_CMD]    = PRINT_CMD_S;
  db_config_fields[MAIL_ADDR]    = MAIL_ADDR_S;
  db_config_fields[MAIL_SUBJ]    = MAIL_SUBJ_S;
  db_config_fields[OPAC_MODE]    = OPAC_MODE_S;
  db_config_fields[OPAC_ONLY]    = OPAC_ONLY_S;
  db_config_fields[OPAC_FIELD]   = OPAC_FIELD_S;
  db_config_fields[RANGE_TYPE]   = RANGE_TYPE_S;
  db_config_fields[YEAR_RANGE]   = YEAR_RANGE_S;
  db_config_fields[USE_PN]       = USE_PN_S;
  db_config_fields[ENV_NAME]     = ENV_NAME_S;
  db_config_fields[ENV_VAL]      = ENV_VAL_S;
  db_config_fields[ABBR_FIELDS]  = ABBR_FIELDS_S;
  db_config_fields[NO_INFO]      = NO_INFO_S;
  db_config_fields[EXT_LABEL]    = EXT_LABEL_S;
  db_config_fields[EXT_PROG]     = EXT_PROG_S;
  db_config_fields[EXT_FIELDS]   = EXT_FIELDS_S;
  db_config_fields[BEG_MENU]     = BEG_MENU_S;
  db_config_fields[END_MENU]     = END_MENU_S;
  db_config_fields[MENU_ITEM]    = MENU_ITEM_S;
  db_config_fields[MENU_SEP]     = MENU_SEP_S;
  db_config_fields[USER_INFO]    = USER_INFO_S;
  db_config_fields[CUSTOM_FIELDS]= CUSTOM_FIELDS_S;
  db_config_fields[IS_LIMIT]     = IS_LIMIT_S;
  db_config_fields[INCLUDE_QUERY]= INCLUDE_QUERY_S;

  /* Initialize hash table for quick lookup of config field strings */
  conf_hash_tab = hash_Init(HASH_TAB_SIZE);

  /* Insert each item into the hash table */
  for (i = 0; i < NUM_CONFIG_FIELDS; i++) {
    element = (conf_hash_elmt *) XtMalloc(sizeof(conf_hash_elmt));
    element->idex = i;
    element->str = db_config_fields[i];
    hash_Insert(conf_hash_tab, 
		hash_HashFunction(element->str, strlen(element->str)),
		conf_compare, element->str, element);
  }
  
  /* if db_config_file_name is a full path, use it. 
     Else append to path */
  file_name = XtMalloc((path?strlen(path):0) + strlen(db_config_file_name) +1);
  if ((db_config_file_name[0] == '/') || !path)
    strcpy(file_name, db_config_file_name);
  else {
    strcpy(file_name, path);
    strcat(file_name, db_config_file_name);
  }

  if (strcmp(file_name, "-") == 0)
    infile = stdin;
  else if (!(infile = fopen(file_name,"r"))) {
    char buf[1024];
    
    sprintf(buf, "Can't open db config file:\n\"%s\"", file_name);
    error(buf);
  }
  config_line_num = 0;
  config_file_name = file_name;

  /* Initialize hash table for quick lookup of the databases */
  *db_hash_tab = hash_Init(HASH_TAB_SIZE);

  /* Loop through the config file, reading in database descriptions.
     First parse the menu building instructions, if any. They are simply
     read and stored for the time being, because it does not make sense
     to build the menu until we have read in all the database entries. */
  field = scan_config(infile, value);

  while (field != DB_TAG) {
    if (field == EOF)
      config_err("Could not find an initial 'db_tag'");
    if (db_menu_ptr) {
      db_menu_ptr->next = (db_menu_rec *) XtMalloc(sizeof(db_menu_rec));
      db_menu_ptr = db_menu_ptr->next;
    }
    else {
      *db_menu_list = (db_menu_rec *) XtMalloc(sizeof(db_menu_rec));
      db_menu_ptr = *db_menu_list;
      default_menu = FALSE;
    }
    db_menu_ptr->type = field;
    db_menu_ptr->value = strdup(value);
    db_menu_ptr->next = NULL;
    field = scan_config(infile, value);
  }

  while (field == DB_TAG) {
    db_desc = new_db();
    db_desc->tag = strdup(value);

    /* If we did not find any menu directives in the db.conf file, then
       just create a simple flat menu ourselves. */
    if (default_menu) {
      if (db_menu_ptr) {
	db_menu_ptr->next = (db_menu_rec *) XtMalloc(sizeof(db_menu_rec));
	db_menu_ptr = db_menu_ptr->next;
      }
      else {
	*db_menu_list = (db_menu_rec *) XtMalloc(sizeof(db_menu_rec));
	db_menu_ptr = *db_menu_list;
      }
      db_menu_ptr->type = MENU_ITEM;
      db_menu_ptr->value = strdup(value);
      db_menu_ptr->next = NULL;
    }

    field = scan_config(infile, value);
    while ((field != EOF) && (field != DB_TAG)) {
      switch (field) {
      case DB_NAME:
	db_desc->short_name = strdup(value);
	break;
      case OPAC_MODE:
	db_desc->opac_mode = TRUE;
	break;
      case NO_INFO:
	db_desc->no_info = TRUE;
	break;
      case USE_PN:
	db_desc->use_prev_next = TRUE;
	break;
      case DB_LONG: 
	db_desc->long_name = strdup(value);
	break;
      case DRIVER:
	db_desc->driver = strdup(value);
	break;
      case DB_INT_NAME:
	db_desc->internal_name = strdup(value);
	break;
      case SUMM_FIELDS:
	db_desc->summary_fields = strdup(value);
	break;
      case TITLE_FIELDS:
	db_desc->title_fields = strdup(value);
	break;
      case ENV_NAME:
	if (!db_desc->environ) {
	  db_desc->environ = XtMalloc(MAX_ENV_SIZE);
	  db_desc->env_len = 0;
	}
	strcpy(db_desc->environ + db_desc->env_len, value);
	db_desc->env_len += strlen(value) + 1;
	if (db_desc->env_len > MAX_ENV_SIZE)
	  error("Environment list too long in db.conf");
	break;
      case ENV_VAL:
	if (!db_desc->environ)
	  config_err("env_val without a previous env_name?");
	strcpy(db_desc->environ + db_desc->env_len, value);
	db_desc->env_len += strlen(value) + 1;
	if (db_desc->env_len > MAX_ENV_SIZE)
	  error("Environment list too long in db.conf");
	break;
      case EXT_FIELDS:
	db_desc->ext_fields = strdup(value);
	break;
      case EXT_LABEL:
	ext = new_ext();
	ext->label = strdup(value);
	ext->next = NULL;
	if (!db_desc->extensions)
	  db_desc->extensions = ext;
	else
	  prev_ext->next = ext;
	prev_ext = ext;
	break;
      case EXT_PROG:
	if (!ext)
	  config_err("extension_program without previous extension_name?");
	ext->program = strdup(value);
	break;
      case T_FIELD_NAME:
	db_desc->num_text_fields++;
	db_field = new_t_field();
	db_field->name = strdup(value);
	db_field->next = NULL;
	if (!db_desc->text_fields) {
	  db_desc->text_fields = db_field;
	  prev_db_field = db_field;
	}
	else {
	  prev_db_field->next = db_field;
	  prev_db_field = db_field;
	}
	break;
      case L_FIELD_NAME:
	db_desc->num_limit_fields++;
	db_field = (db_field_desc_rec *) XtMalloc(sizeof(db_field_desc_rec));
	db_field->name = strdup(value);
	db_field->next = NULL;
	if (!db_desc->limit_fields) {
	  db_desc->limit_fields = db_field;
	  prev_db_field = db_field;
	}
	else {
	  prev_db_field->next = db_field;
	  prev_db_field = db_field;
	}
	break;
      case T_FIELD_CMD:
      case L_FIELD_CMD:
	db_field->command = strdup(value);
	break;
      case DFLT_AT:
	num = atoi(value);
	if (!((num > 0) && (num <= NUM_TEXT_FIELDS)))
	  config_err("dflt_at value out of valid range");
	db_desc->defaults[num-1] = db_field;
	break;
      case DB_TAG:
      case EOF:
	break;
      case LIST_NAME:
	db_field->list_names[db_field->list_cnt] = strdup(value);
	if (db_field->list_cnt++ > MAX_LISTS)
	  config_err("too many lists for this field");
	break;
      case LIST_MNEMONIC:
	db_field->list_mnemonic = strdup(value);
	break;
      case LIST_HIER:
	db_field->list_hier = TRUE;
	break;
      case IS_LIMIT:
	db_field->is_limit = TRUE;
	break;
      case OPAC_FIELD:
	db_field->opac_field = TRUE;
	break;
      case OPAC_ONLY:
	db_field->opac_field = TRUE;
	db_field->opac_only = TRUE;
	break;
      case RANGE_TYPE:
	if (strcasecmp(value, DAY_S) == 0)
	  db_field->range_type = DAY;
	else if (strcasecmp(value, MONTH_S) == 0)
	  db_field->range_type = MONTH;
	else if (strcasecmp(value, YEAR_S) == 0)
	  db_field->range_type = YEAR;
	else if (strcasecmp(value, NUMERIC_S) == 0)
	  db_field->range_type = NUMERIC;
	else
	  config_err("invalid range_type");
	break;
      case YEAR_RANGE: {
	char *p;
	int   diff;

	db_field->first_year = atoi(value);
	p = value;
	while (*p && isdigit(*p))
	  p++;
	while (*p && !isdigit(*p))
	  p++;
	db_field->last_year = atoi(p);
    
	/* Check validity of first/last year entries */
	diff = db_field->last_year - db_field->first_year;

	if (!((diff > 0)  && (diff <= MAX_YEAR_RANGE)))
	  db_field->last_year = db_field->first_year = 0;
	break;
      }
      default:
	config_err("unknown db.conf field");
      }
      field = scan_config(infile, value);
    }
    /* Make sure required fields got read in */
    if (!(db_desc->long_name && db_desc->short_name && db_desc->driver &&
	  db_desc->internal_name && db_desc->summary_fields &&
	  db_desc->num_text_fields))
      config_err("missing a required db.conf field");

    /* Add this db to the hash table */
    if (0 > hash_Insert(*db_hash_tab, 
			hash_HashFunction(db_desc->tag, strlen(db_desc->tag)),
			db_compare, db_desc->tag, db_desc)) {
      sprintf(value, "Duplicate db_tag '%s'", db_desc->tag);
      config_err(value);
    }
    db_count++;
  }
  return db_count;
}

/* Scan a line from the config file. File can be the db.conf file or
   the .willowrc. Get passed a file pointer, and a string in which we
   write the user-supplied value. Return value is the numerical constant
   indicating which field this is. -1 for end of file.
 */
int 
scan_config(FILE *infile,
	    char *value)
{
  char  line_buf[MAX_LINE];
  char  field_buf[MAX_LINE]; 
  int   sdex, tdex = 0, field;
  int   len;
  conf_hash_elmt *element;

  /* Skip empty lines and comment lines */
  while (TRUE) {
    config_line_num++;
    if (!get_line(line_buf, infile))
      return(EOF);
    if ((sdex = first_good(line_buf)) > -1)
      break;
  }

  /* Extract the string representing the field name */
  len = strlen(line_buf);
  while (!(isspace(line_buf[sdex]) || (line_buf[sdex] == CONFIG_SEP_CHAR))) {
    field_buf[tdex++] = line_buf[sdex++];
    if (sdex == len)
      break;
  }
  field_buf[tdex] = 0;

  /* Look up the field name in the hash table */
  if (!(element = (conf_hash_elmt *) hash_Lookup(conf_hash_tab,
	               hash_HashFunction(field_buf, strlen(field_buf)),
		       conf_compare, field_buf)))
    config_err("unknown field name");
  else
    field = element->idex;

  /* Extract the field value */
  while (line_buf[sdex] != CONFIG_SEP_CHAR) {
    sdex++;
    if (sdex >= len) {
      value[0] = 0;
      return(field);
    }
  }
  sdex++;
  while (isspace(line_buf[sdex]))
    sdex++;
  strcpy(value, &line_buf[sdex]);
  return(field);
}

/* Return the index of first "good" character on a line. -1 if 
   there are none, or line is a comment .
 */
int first_good(char *line)
{
  int pos = 0;

  while (line[pos]) {
    if (line[pos] == '\n')
      return(-1);
    else if (strchr(COMMENT_CHARS,line[pos]))
      return(-1);
    else if (isspace(line[pos]))
      pos++;
    else
      return(pos);
  }
  return(-1);
}

/* Allocate and initialize a new database description record */
static db_desc_rec *
new_db(void)
{
  db_desc_rec *db;

  db = (db_desc_rec *) XtMalloc(sizeof(db_desc_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
  memset(db, 0, sizeof(db_desc_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
  bzero(db, sizeof(db_desc_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  return(db);
}

/* Allocate and initialize a new field record */
static db_field_desc_rec *
new_t_field(void)
{
  db_field_desc_rec *db;

  db = (db_field_desc_rec *) XtMalloc(sizeof(db_field_desc_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
  memset(db, 0, sizeof(db_field_desc_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
  bzero(db, sizeof(db_field_desc_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  return(db);
}

static db_ext_rec *
new_ext(void)
{
  db_ext_rec *ext;
  
  ext = (db_ext_rec *) XtMalloc(sizeof(db_ext_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
  memset(ext, 0, sizeof(db_ext_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
  bzero(ext, sizeof(db_ext_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  return(ext);
}

static void 
config_err(char *msg)
{
  char error_buf[MAX_LINE];

  sprintf(error_buf,"Error in %s at line %d\n%s",
	  config_file_name, config_line_num, msg);
  error(error_buf);
}

/* Comparison function to be passed into hash routines */
int conf_compare(char *key, conf_hash_elmt *element)
{
  return(!strcasecmp(key, element->str));
}

/* Comparison function for db's to be passed into hash routines */
int db_compare(char *key, db_desc_rec *element)
{
  return(!strcasecmp(key, element->tag));
}
