/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/***************************************************************************
 * dbio.c
 *
 * This file contains all the code for dealing with the 
 * database driver program. It can be called by both Willow and Wilco.
 * See the file protocol.h for details on the driver communication protocol.
 *
 * The following routines are available:
 *    send_cheat_cmd(command) : pass the command string on to the driver,
 *        so it can be sent directly on to the database host.
 *    Boolean send_connect_cmd(db, config_path) : takes a database record
 *        pointer and a path for where config files can be found. Returns
 *        FALSE if db is manual, but there is no user-id set.
 *    send_disconnect_cmd(): Tells the driver to drop a connection. Will
 *        take care of closing file descriptors etc.
 *    send_echo_cmd(): Toggle whether or not driver should echo interactions
 *        with the host system.
 *    send_record_req_cmd(item, summary, abbreviated): ask the driver to
 *         send us the text of a record. Indicate whether it should be
 *         summary or full, and abbreviated or expanded fields.
 *    send_ext_req_cmd(item): ask the driver to send us the extension record
 *         for a given title.
 *    send_search_cmd(num_text_fields, num_limit_fields, data, data_len) : 
 *        sends the driver information needed to launch a search. See comment
 *        for that routine for information on expected data format.
 *    send_search_cancel_cmd(): Tell the driver to try to cancel the
 *        currently executing search.
 *    send_title_req_cmd(summ_num): request the driver to send us the  
 *        specified title display string.
 *    send_user_resp(answer, text): send a user's response back to the driver.
 *
 * The following application-specific items must be declared externally
 *    debug_level: What level of debugging is currently set. This is
 *        passed on to the driver as well. Default is 0. At 1, received
 *        packets are echoed. At 2, sent packets are echoed as well. At 3,
 *        xcodecenter is launched instead of the driver. 
 *    dispatch_packet() : takes a packet that has come in from the driver
 *        and passes it on to an appropriate handler. Will need defs from
 *        protocol.h
 *    echo_on(): routine to tell whether echo mode is currently enabled.
 *    error(msg): fatal error handling routine
 *    full_record_fields: a pointer to a block containing any full-record
 *        field specification currently set.
 *    full_record_fields_len: how many bytes of data in that block.
 *
 ***************************************************************************/

/*
 * 05/17/94  slh  wait for (two) specific pid(s)
 *		    rm close() of one fds after fork()
 */

#include <stdio.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#ifdef WILLOW
#include <Mrm/MrmAppl.h>

#include "willow.h"
#else
#define XtMalloc  malloc
#define XtFree    free

extern char *malloc(size_t);
extern char *realloc(void *, size_t);

#define DRIVER_DIR "drivers/"
#endif

#include "db.h"
#include "protocol.h"

/* 
 * Private Global Data 
 */

/* a couple of file descriptors for talking to db-drivers. May be
   pipes or a socket */
static int  to_db, from_db;

/* This is the callback id for reading from the pipe */
#ifdef WILLOW
static XtInputId    input_id;
#else
static int          input_id;
#endif

#ifndef   WILLOW
static  int driver_pid = -1;
static  int last_driver_pid = -1;
#endif /*!WILLOW*/

/*
 * External Data
 */

extern char  full_record_fields[];
extern int   full_record_fields_len;

#ifdef WILLOW
extern XtAppContext app_context;
#else
typedef void (*PFV)();
extern int add_input (int fd, PFV func, void *data);
#endif
extern int debug_level;

extern int pipe (int *);
#ifdef vax
extern int vfork ();
#else
extern pid_t vfork (void);
#endif
extern int dup2 (int, int);
extern int close (int);
extern int execlp ( char *,  char *, ...);
extern int error (char *);
extern void remove_input (int fd);
extern int read (int, void *, unsigned int);
extern void free (void *);
extern int write (int,  void *, unsigned int);
#ifdef vax
extern int wait3 ();
#else
extern pid_t wait3 (int*, int, void *);
#endif
extern packet_rec *netify(packet_rec *packet);
extern packet_rec *denetify(packet_rec *packet);

/*
 * Forward Routines
 */
static char *db_driver_launch(char *driver, char *config_path);
static char *local_launch(char *driver, char *config_path);
static char *socket_launch(char *driver);
static void db_driver_close(void);
#ifdef WILLOW
static void db_read(caddr_t data, int *fd, XtInputId *input_id);
#else
static void db_read(caddr_t data, int *fd, int *input_id);
#endif
static void db_send(char *data, int   len);
static void echo_packet(packet_rec *packet, char send);

void send_echo_cmd (void);

/* Start up a driver. Can be a forked/exec'ed local program with a pipe,
   or via a remote socket. */
static char *
db_driver_launch(char  *driver,      /* name of the driver program */
		 char  *config_path) /* Where do willow/wilco files live? */
{
  char *msg;

  if (strchr(driver, SUB_SEP_CHAR)) {
    if (msg = socket_launch(driver))
      return(msg);
  }
  else {
    if (msg = local_launch(driver, config_path))
      return(msg);
  }

  /* Now add a callback to read from pipe/socket whenever something
     appears */
#ifdef WILLOW
  input_id = XtAppAddInput(app_context, from_db, 
			   XtInputReadMask, db_read, NULL);
#else
  input_id = add_input(from_db, db_read, NULL);
#endif
  return(NULL);
}

/* make a socket connection to a driver-server */
static char *socket_launch(char *driver)
{
  int sock;
  struct sockaddr_in server;
  struct hostent *hp;
  char *host;
  char *port_s;
  static char err_buf[256];

  /* driver is of form "program:host:port" */
  driver = strchr(driver, SUB_SEP_CHAR);
  driver++;
  host = strdup(driver);
  port_s = strchr(host, SUB_SEP_CHAR);
  *port_s = 0;
  port_s++;

  /* Create socket */
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0)
    return("Could not open stream socket for driver");

  server.sin_family = AF_INET;
  if ((hp = gethostbyname(host)) == 0) {
    sprintf(err_buf, "%s: unknown host", host);
    return(err_buf);
  }
#if defined(POSIX) || defined (_POSIX_SOURCE)
  memmove(&server.sin_addr, hp->h_addr, hp->h_length);
#else
  bcopy(hp->h_addr, &server.sin_addr, hp->h_length);
#endif
  server.sin_port = htons(atoi(port_s));

  if (connect(sock, &server, sizeof(server)) < 0) {
    sprintf(err_buf, "Could not connect to %s port %s", host, port_s);
    return(err_buf);
  }
  to_db = from_db = sock;

#ifdef WILLOW
  XtFree(host);
#else
  free(host);
#endif
  return(NULL);
}

/* Set up pipes, and start up the database driver */
static char *
local_launch(char  *driver,      /* name of the driver program */
		 char  *config_path) /* Where do willow/wilco files live? */
{
  char     full_path[MAX_LINE]; /* full path name of driver */
  char     debug_level_str[MAX_LINE]; /* String version of debug_level */
  int      to_db_pipe[2];
  int      from_db_pipe[2];
#ifdef   WILLOW
  int      driver_pid;
#endif /*WILLOW*/

  /* set up pipes for communicating with driver */
  pipe(to_db_pipe);
  pipe(from_db_pipe);
   
  /* fork off a child, which will overlay driver on itself */
#ifndef   WILLOW
  last_driver_pid = driver_pid;
#endif /*!WILLOW*/
  driver_pid = vfork();
  if (driver_pid == 0) {
    /* in the child, redirect stdin and out */
    dup2(to_db_pipe[R], TO_DRIVER_FD);
    dup2(from_db_pipe[W], FROM_DRIVER_FD);
#ifdef   WILLOW
    close(to_db_pipe[R]);
#endif /*WILLOW*/
    close(to_db_pipe[W]);
    close(from_db_pipe[R]);
    close(from_db_pipe[W]);

    strcpy(full_path, config_path);
    strcat(full_path, DRIVER_DIR);
    strcat(full_path, driver);
    sprintf(debug_level_str,"%d",debug_level);

    /* Exec driver, pass single argument which is level of debugging */
    if (debug_level > 2)
      execlp("xcodecenter", "xcodecenter", "-p", 0);
    else {
      if (execlp(full_path, driver, "-debug", debug_level_str, "-cfg_path", 
		 config_path, NULL) < 0) {
	/* at this point we are in a child process, so we can't do much */
	perror(NULL);
	/* Must _exit after vfork */
	_exit(-1);
      }
    }
  }
  else if (driver_pid > 0) {
    /* in the parent, close extra pipes */
    to_db = to_db_pipe[W];
    close(to_db_pipe[R]);

    from_db = from_db_pipe[R];
    close(from_db_pipe[W]);
  }
  else {
    return("Could not fork driver");
  }
  return(NULL);
}

/* Close down the pipes to the db_driver */
static void
db_driver_close(void)
{
  if (input_id) {
#ifdef WILLOW
    XtRemoveInput(input_id);
#else
    remove_input(input_id);
#endif
    input_id = NULL;
    close(to_db);
    close(from_db);
  }
}

/* This function is called when there is data waiting on the pipe.
   Read it and take action depending on program state 
   */
static void 
db_read(caddr_t    data,
	int       *fd,
#ifdef WILLOW
	XtInputId *input_id
#else
        int       *input_id
#endif
	) /* This ) is out here for the benefit of Bill Cattey's editor */
{
  packet_rec packet;   /* store packet we read here */
  int  got_count = 0;  /* how many bytes of packet have we gotten */
  int  n = 0;
  
  /* First read in just the header */
  while (got_count < PACKET_NET_SIZE) {
    n = read(from_db, &packet + got_count, PACKET_NET_SIZE - got_count);
    if (n < 0) {
      perror(NULL);
      error("error reading input pipe (header)");
    }
    if (n == 0)   /* Means driver is gone */
      break;
    got_count += n;
  }
  denetify(&packet);
  if (n) { 
    /* Grab enough storage, and read in rest of packet */
    packet.data = XtMalloc(packet.data_len);
    
    /* Now read until we get the entire packet in */
    got_count = 0;
    while (got_count != (packet.data_len)) {
      n = read(from_db, packet.data+got_count, packet.data_len - got_count);
      if (n < 1) {
	perror(NULL);
	error("error reading input pipe (data)");
      }
      got_count += n;
    }
  }
  else {  /* driver died, or at closed pipe */
    packet.type = ERROR;
    packet.var.error.type = DEAD;
    packet.data = NULL;
    /* Close pipes, remove handler */
    db_driver_close();
  }
  
  if (debug_level) 
    echo_packet(&packet, FALSE);

  /* Call a handler routine for this packet type */
  dispatch_packet(&packet);

  if (packet.data)
    XtFree(packet.data);
}

/* Send some data out on the pipe to the database. Take a char* and
   a length. If length is NULL, then take strlen of the data, and add
   1 for null byte. If data is NULL, send a NULL byte.
 */
static void
db_send(char *data, int len)
{
  /* don't send if the driver is gone */
  if (!input_id)
    return;
  if (!data)
    write(to_db, "", 1);
  else if (len)
    write(to_db, data, len);
  else
    write(to_db, data, strlen(data) + 1);
}

/* Put together a packet for a connect command. We also need to 
   get a user-id and password.
 */
char *send_connect_cmd(db_desc_rec *db,    /* record of db we are to load */
		       char        *config_path) /* Where config files live */
{
  packet_rec  packet;
  char *msg;

  packet.type = CONNECT;
  packet.data = NULL;
  if (msg = db_driver_launch(db->driver, config_path))
    return(msg);
  
  /* If echo is selected, send off an echo-cmd before anything else */
  if (echo_on())
    send_echo_cmd();

  /* Need to add 1s for NULL chars. */
  packet.data_len = strlen(db->internal_name) + 1 +
                    strlen(db->title_fields) + 1 +
                    strlen(db->summary_fields) + 1 +
	            (db->ext_fields ? strlen(db->ext_fields) : 0) + 1 +
                    db->env_len;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);

  /* Send out the header, then each data field. Some might be empty */
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
  db_send(db->internal_name, NULL);
  db_send(db->title_fields, NULL);
  db_send(db->summary_fields, NULL);
  db_send(db->ext_fields, NULL);
  db_send(db->environ, db->env_len);

  return(NULL);
}

/* Send out the search command. All the hard work has already been done. 
   In the data block we expect to see a number of adjacent null-terminated
   strings. First will be the regular text-search fields. The user typed 
   text will be immediately followed by the command string for that field. 
   There will be num_text_fields of those pairs. Then there will be 
   num_limit_fields limit commands. Datalen is the entire size of the data
   block, including nulls.
*/
void 
send_search_cmd(int num_text_fields,
		int num_limit_fields,
		char *data,
		int data_len)
{
  packet_rec packet;

  packet.type = SEARCH;
  packet.data_len = data_len; 
  packet.var.search.num_text_fields = num_text_fields;
  packet.var.search.num_limit_fields = num_limit_fields;
  packet.data = NULL;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
  db_send(data, data_len);
}

/* Send out the user's response to a query */
void send_user_resp(int answer, char *text)
{
  packet_rec packet;

  packet.type = USER;
  packet.var.user_resp.answer = answer;
  packet.data = NULL;
  packet.data_len = strlen(text) + 1;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
  db_send(text, NULL);
}

/* Send out a simple disconnect packet */
void
send_disconnect_cmd(void)
{
  packet_rec packet;
  
  packet.type = DISCONNECT;
  packet.data_len = 0;
  packet.data = NULL;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
  db_driver_close();
}

/* Send a simple echo packet */
void
send_echo_cmd(void)
{
  packet_rec packet;
  
  packet.type = TOGGLE_ECHO;
  packet.data_len = 0;
  packet.data = NULL;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
}

/* Ask for the next summary */
void
send_title_req_cmd(int title_num)
{
  packet_rec packet;
  
  packet.type = TITLE_REQ;
  packet.data_len = 0;
  packet.data = NULL;
  packet.var.title_req.title_num = title_num;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
}

/* Ask for an extension record */
void
send_ext_req_cmd(int ext_num)
{
  packet_rec packet;
  
  packet.type = EXT_REQ;
  packet.data_len = 0;
  packet.data = NULL;
  packet.var.exts.ext_num = ext_num;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
}

/* Send command to cancel the search */
void
send_search_cancel_cmd(void)
{
  packet_rec packet;
  
  packet.type = CANCEL;
  packet.data_len = 0;
  packet.data = NULL;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
}

/* send command to retrieve full text */
void
send_record_req_cmd(int item, char summary, char abbreviated)
{
  packet_rec packet;

  packet.type = RECORD_REQ;
  if (!summary)
    packet.data_len = full_record_fields_len;
  else
    packet.data_len = 0;
  packet.data = NULL;
  packet.var.record_req.ref_num = item;
  packet.var.record_req.summary = summary;
  packet.var.record_req.abbreviated = abbreviated;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
  if (!summary && full_record_fields_len)
    db_send(full_record_fields, full_record_fields_len);
}

/* send command to be executed directly */
void
send_cheat_cmd(char *command)
{
  packet_rec packet;

  packet.type = CHEAT;
  packet.data_len = strlen(command) + 1;
  packet.data = NULL;
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
  db_send(command, strlen(command) + 1);
}

/* Called on SIGCHLD, wait on dead process. Could be a driver, mail, 
   ftp, etc. 
 */
void
wait_child(int sig, int code, caddr_t  data)
{
#ifdef   WILLOW
#if defined(POSIX) || defined (_POSIX_SOURCE)
    waitpid(-1, NULL, WNOHANG);
#else
    wait3(NULL, WNOHANG, NULL);
#endif
#else  /*WILLOW*/
  if (driver_pid != -1)
    if (waitpid( driver_pid , NULL , WNOHANG ) == driver_pid)
      driver_pid = -1;
  /* since new one may have started and reset driver_pid */
  if (last_driver_pid != -1)
    if (waitpid( last_driver_pid , NULL , WNOHANG ) == last_driver_pid)
      last_driver_pid = -1;
#endif /*WILLOW*/
}

static void echo_packet(packet_rec *packet, char send)
{
  if (send)
    fprintf(stderr,"WILLOW SENDING ");
  else
    fprintf(stderr,"WILLOW RECEIVED ");
  switch (packet->type) {
    case CONNECT   : fprintf(stderr, "CONNECT");     break;
    case CONNECTED : fprintf(stderr, "CONNECTED");   break;
    case ERROR:
      switch (packet->var.error.type) {	
        case MISC:      fprintf(stderr, "MISC-ERR");      break;
	case BAD_SEARCH:fprintf(stderr, "BAD_SEARCH-ERR");break;
	case DEAD:	fprintf(stderr, "DEAD-ERR");      break;
	case UPDATE:	fprintf(stderr, "UPDATE");        break;
	case WARNING:   fprintf(stderr, "WARNING");       break;
      }
      break;    
    case SEARCH     : fprintf(stderr, "SEARCH");     break;
    case STATS      : fprintf(stderr, "STATS");      break;
    case TITLE_REQ  : fprintf(stderr, "TITLE_REQ");  break;
    case TITLE      : fprintf(stderr, "TITLE");      break;
    case RECORD_REQ : fprintf(stderr, "RECORD_REQ"); break;
    case RECORD     : fprintf(stderr, "RECORD");     break;
    case DISCONNECT : fprintf(stderr, "DISCONNECT"); break;
    case TOGGLE_ECHO: fprintf(stderr, "TOGGLE_ECHO");break;
    case CHEAT      : fprintf(stderr, "CHEAT");      break;
    case CANCEL     : fprintf(stderr, "CANCEL");     break;
    case INFO       : 
      switch (packet->var.info.type) {
        case COPYRIGHT_INFO: fprintf(stderr, "INFO-COPYRIGHT");break;
        case FIELDS_INFO   : fprintf(stderr, "INFO-FIELDS");   break;
      }
      break;
    case EXT_REQ    : fprintf(stderr, "EXT_REQ");    break;
    case EXTS       : fprintf(stderr, "EXTS");       break;
    case USER_REQ   : 
      switch (packet->var.user_req.type) {
        case BOOL_REQ   : fprintf(stderr, "USER_REQ-BOOL");  break;
        case TEXT_REQ   : fprintf(stderr, "USER_REQ-TEXT");  break;
        case MASKED_REQ : fprintf(stderr, "USER_REQ-MASKED");break;
      }
      break;
    case USER       : fprintf(stderr, "USER");       break;
    case BROWSE_REQ : fprintf(stderr, "BROWSE_REQ"); break;
    case BROWSE     : fprintf(stderr, "BROWSE");     break;
    default: fprintf(stderr, "UNKNOWN???");
    }
  fprintf(stderr, "\n");
  fflush(stderr);
}


/* Send a browse-list lookup to the driver */
int list_lookup(char *db_name, 
		  char *list_name, 
		  char search_type, 
		  char *prefix_st, 
		  int lines_wanted, 
		  int line_offset)
{
  static int browse_ref_num = 0;
  packet_rec packet;

  if (!list_name)
    return(-1);
  if (!list_name[0])
    return(-1);

  packet.type = BROWSE_REQ;
  packet.var.browse_req.ref_num      = ++browse_ref_num;
  packet.var.browse_req.search_type  = search_type;
  packet.var.browse_req.lines_wanted = lines_wanted;
  packet.var.browse_req.line_offset  = line_offset;

  /* Need to add 1s for NULL chars. */
  packet.data_len = (list_name ? strlen(list_name) : 0) + 1 +
		    (prefix_st ? strlen(prefix_st) : 0) + 1;

  /* Send out the header, then each data field. Some might be empty */
  if (debug_level > 1)
    echo_packet(&packet, TRUE);
  db_send((char *) netify(&packet), PACKET_NET_SIZE);
  db_send(list_name, NULL);
  db_send(prefix_st, NULL);
  return(browse_ref_num);
}

