/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/*******************************************************
 * extensions.c
 * routines for implementing the Willow multi-media 
 * extensions mechanism.
 ******************************************************/

#include <Mrm/MrmAppl.h>
#include <Xm/PushB.h>
#include <stdio.h>

#include "willow.h"
#include "db.h"

/* Externs */
extern Widget widgets[], search_shell;

extern Boolean list_get_selected (Widget w, int **pos_list, int *num);
extern void center_summ(Widget w, XtPointer data, XEvent *event, Boolean *cont);
extern void force_extensions(void);
extern char *extensions[];
extern XtAppContext app_context;

/* Pipes into extension, and error back */
typedef struct {
  int to_ext[2], from_ext[2];
  int ext_pid;
} ext_pid_rec;

#define IN  0
#define OUT 1

/* Forwards */
ext_pid_rec *open_ext_pipe(char *program);
void ext_activate_cb(Widget w, db_ext_rec *ext, u_long *reason);
void ext_read(ext_pid_rec *pid_p, int *fid, XtInputId *iid);

/* Initialize the extensions. */
void init_extensions(db_desc_rec *db)
{
  WidgetList  children;
  Widget      button;
  int         num_children, i, n;
  db_ext_rec *ext;
  Arg         args[10];

  if (!db->extensions) {
    XtUnmanageChild(widgets[k_ext_menu]);
    XtManageChild(widgets[k_summ_disp]);
    return;
  }
  XtUnmanageChild(widgets[k_summ_disp]);
  XtManageChild(widgets[k_ext_menu]);
  center_summ(widgets[k_sum_dum], NULL, NULL, NULL);

  /* Just destroy the old children (except first one */
  XtVaGetValues(widgets[k_ext_pane], XmNnumChildren, &num_children,
		XmNchildren, &children, NULL);
  for (i = 1; i < num_children; i++)
    XtDestroyWidget(children[i]);

  /* Build a menu item per extension */
  for (ext = db->extensions; ext != NULL; ext = ext->next) {
    n = 0;
    XtSetArg(args[n], XmNlabelString,
	     XmStringCreateSimple(ext->label));n++;
    button = XmCreatePushButton(widgets[k_ext_pane], "button", args, n);
    XtAddCallback(button, XmNactivateCallback, ext_activate_cb, ext);
    XtManageChild(button);
  }
}

/* Extension button pressed. Get extension records for current
   selections, and pipe them through the designated extension
   program.
 */
void ext_activate_cb(Widget w, db_ext_rec *ext, u_long *reason)
{
  char *result;
  int  *position_list;
  int   position_count, i;
  ext_pid_rec *pid_p;

  if (list_get_selected(widgets[k_summary_titles], &position_list, 
			   &position_count) && ext->program) {
    pid_p = open_ext_pipe(ext->program);
    force_extensions();
    for (i = 0; i < position_count; i++) {
      result = extensions[position_list[i]];
      write(pid_p->to_ext[OUT], result, strlen(result));
      write(pid_p->to_ext[OUT], "\n\n", 2);
    }
    close(pid_p->to_ext[OUT]);
    XtFree(position_list);
  }
}

/* Open a pipe to the extension command */
ext_pid_rec *open_ext_pipe(char *program)
{
  ext_pid_rec *pid_p;

  /* set up pipes for communicating with extension program */
  pid_p = (ext_pid_rec *) XtMalloc(sizeof(ext_pid_rec));
  pipe(pid_p->to_ext);
  pipe(pid_p->from_ext);
   
  /* fork off a child, which will overlay extension prog on itself */
  pid_p->ext_pid = vfork();

  if (pid_p->ext_pid == 0) {
    /* in the child, redirect stdin, stdout, and stderr. */
    dup2(pid_p->to_ext[IN], fileno(stdin));
    dup2(pid_p->from_ext[OUT],fileno(stdout));
    dup2(pid_p->from_ext[OUT],fileno(stderr));
    close(pid_p->to_ext[IN]);
    close(pid_p->to_ext[OUT]); 
    close(pid_p->from_ext[IN]); 
    close(pid_p->from_ext[OUT]);

    /* Exec sh, pass in extension command */
    if (execlp("/bin/sh", "sh", "-c", program, NULL) < 0) {
      /* at this point we are in a child process, so we can't do much */
      perror(NULL);
      /* Must _exit after vfork */
      _exit(-1);
    }
  }
  else if (pid_p->ext_pid > 0) {
    /* in the parent, close extra pipes */
    close(pid_p->to_ext[IN]);
    close(pid_p->from_ext[OUT]);
    
    /* Now add a callback to read from pipe whenever something
       appears */
    XtAppAddInput(app_context, pid_p->from_ext[IN], 
		  XtInputReadMask, ext_read, pid_p);
  }
  else {
    perror(NULL);
    error("Could not fork sh/extension");
  }
  return(pid_p);
}

/* Read from standard error from extension process. 
 */
void ext_read(ext_pid_rec *pid_p, int *fid, XtInputId *iid)
{
  static char ext_msg_buf[CHUNK_SIZE];
  int  n = 0;
  static int err_len = 0;
  
  /* The -1 is so there is always room to append a \0 */
  n = read(pid_p->from_ext[IN], ext_msg_buf + err_len, 
	   sizeof(ext_msg_buf) - err_len - 1);  

  if (n)
    err_len += n;
    
  else {
    XtRemoveInput(*iid);
    close(pid_p->from_ext[IN]);
    XtFree(pid_p);

    if (err_len) {
      /* Got an error message, ext failed */
      ext_msg_buf[err_len] = 0;
      pop_warning("Error from External Program:\n%s", ext_msg_buf, search_shell);
      err_len = 0;
    }
  }
}
