/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/*******************************************************
 * field_pick.c
 * This file contains the routines for handling user-choice of
 * of display fields.
 ******************************************************/

#include <stdio.h>

#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif

#include <Mrm/MrmAppl.h>
#include <Xm/ToggleB.h>
#include "willow.h"
#include "field_pick.h"
#include "db.h"
#include "hash.h"

extern int error (char *msg);
void pop_warning(char *template, char *arg, Widget parent);

void field_pick_hide(void);
void field_pick_show(void);
void field_pick_ok(void);
static void full_record_fields_build(void);
static void field_pick_set_prev_picked(void);
void field_pick_all_cb(Widget w, int *tag, u_long *reason);
void field_pick_clear_cb(Widget w, int *tag, u_long *reason);
void field_pick_defaults_cb(Widget w, int *tag, u_long *reason);
int field_pick_compare(char *key, field_pick_hash_elmt *element);
void field_pick_hash_free(char *data);
int db_compare(char *key, db_desc_rec *element);

extern Widget     widgets[],
                  search_shell,
                  summary_shell,
                  dict_shell,
                  results_shell,
                  toplevel;
extern MrmHierarchy s_MrmHierarchy;  /* MRM database hierarchy ID */
extern MrmType dummy_class;         /* and class variable. */
extern db_desc_rec *cur_db;
extern hash_tbl *db_hash_tab;

/* hash table for quick lookup of display fields */
static hash_tbl *field_pick_hash_tab = NULL;

/* Block of text showing the currently selected fields, if any, for
   transmission back to the driver */
char  full_record_fields[FULL_FIELDS_LEN];
int   full_record_fields_len = 0;

/* List of custom fields settings from X-resource */
char *custom_fields_list;

/* When we receive a list of fields from the database driver,
   set up the dialog box. 
 */
void got_field_list(int count, char *fields)
{
  WidgetList old_btns;
  Widget button;
  int num_children;
  int fptr = 0;
  int total_btns, i;
  char *sname, lname[MAX_LINE];
  XmString label;
  field_pick_hash_elmt *element;
  Boolean is_default;

  if (!field_pick_hash_tab)
    field_pick_hash_tab = hash_Init(MAX_PICK_FIELDS);
  else
    hash_Reset(field_pick_hash_tab, field_pick_hash_free);

  XtVaGetValues(widgets[k_field_pick_rc], XmNnumChildren, &num_children,
		XmNchildren, &old_btns, NULL);
  
  /* Need to get old_btns into our own storage, else Xt will re-use it */
  if (num_children)
    old_btns 
      = (WidgetList) bdup((char *) old_btns, num_children * sizeof(old_btns[0]));
  
  total_btns = MAX(count, num_children);
  for (i = 0; i < total_btns; i++) {
    if (i < count) {
      /* Extract default-flag, short and long field names. 
	 Short needs to be held on to */
      is_default = fields[fptr++];
      sname = strdup(fields + fptr);
      fptr += strlen(sname) + 1;
      strcpy(lname, fields + fptr);
      fptr += strlen(lname) + 1;
      label = XmStringCreateSimple(lname);
      
      if (i < num_children)
	button = old_btns[i];
      else
	button = XmCreateToggleButton(widgets[k_field_pick_rc], "field_pick",
				      NULL, NULL);
      XtVaSetValues(button, XmNlabelString, label, XmNuserData, sname, 
		    XmNhighlightThickness, 0, XmNtraversalOn, 0, NULL);
      XtManageChild(button);

      /* Throw this button into our hash table */
      element = (field_pick_hash_elmt *)XtMalloc(sizeof(field_pick_hash_elmt));
      element->str = sname;
      element->button = button;
      element->is_default = is_default;
      hash_Insert(field_pick_hash_tab, 
		  hash_HashFunction(element->str, strlen(element->str)),
		  field_pick_compare, element->str, element);
    }
    else  {
      XtVaGetValues(old_btns[i], XmNuserData, &sname, NULL);
      XtFree(sname);
      XtUnmanageChild(old_btns[i]);
    }

  }
  field_pick_show();
  if (cur_db->display_fields) {
    field_pick_set_prev_picked();
    full_record_fields_build();
  }
  else {
    field_pick_defaults_cb(NULL, NULL, NULL);
    field_pick_ok();
  }
  XtFree(old_btns);
}

/* Hide the field-picking frame and readjust necessary attachments */
void field_pick_hide(void)
{
  XtUnmanageChild(widgets[k_field_pick_form]);  
  XtVaSetValues(widgets[k_field_pick_sep], XmNtopWidget,
		widgets[k_show_search_sep], NULL);
}

/* Show the field-picking frame and readjust necessary attachments.
   For some reason we need to reset the size of the RowColumn after
   the manage */
void field_pick_show(void)
{
  Dimension  form_w, form_h;
  char label[MAX_LINE];
  
  /* Set the label to match current database */
  sprintf(label, FIELD_PICK_FMT, cur_db ? cur_db->short_name : "");
  XtVaSetValues(widgets[k_field_pick_label], XmNlabelString, 
		XmStringCreateSimple(label), NULL);

  XtVaGetValues(widgets[k_field_pick_form], XmNheight, &form_h,
		XmNwidth, &form_w, NULL);
  
  XtManageChild(widgets[k_field_pick_form]);  
  XtVaSetValues(widgets[k_field_pick_sep], XmNtopWidget,
		widgets[k_field_pick_form], NULL);
  
  XtVaSetValues(widgets[k_field_pick_form], XmNheight, form_h,
		XmNwidth, form_w, NULL); 

}


/* Callbacks for setting/clearing all toggles */
void field_pick_all_cb(Widget w, int *tag, u_long *reason)
{
  WidgetList buttons;
  int num_children, i;
  
  XtVaGetValues(widgets[k_field_pick_rc], XmNnumChildren, &num_children,
		XmNchildren, &buttons, NULL);
  
  for (i = 0; i < num_children; i++) {
    if (XtIsManaged(buttons[i]))
      XmToggleButtonSetState(buttons[i], TRUE, TRUE);
    else
      break;
  }
}

void field_pick_clear_cb(Widget w, int *tag, u_long *reason)
{
  WidgetList buttons;
  int num_children, i;
  
  XtVaGetValues(widgets[k_field_pick_rc], XmNnumChildren, &num_children,
		XmNchildren, &buttons, NULL);
  
  for (i = 0; i < num_children; i++) {
    if (XtIsManaged(buttons[i]))
      XmToggleButtonSetState(buttons[i], FALSE, TRUE);
    else
      break;
  }
}

void field_pick_defaults_cb(Widget w, int *tag, u_long *reason)
{
  WidgetList buttons;
  int num_children, i;
  char *name;
  field_pick_hash_elmt *element;

  XtVaGetValues(widgets[k_field_pick_rc], XmNnumChildren, &num_children,
		XmNchildren, &buttons, NULL);
  
  for (i = 0; i < num_children; i++) {
    if (XtIsManaged(buttons[i])) {
      XtVaGetValues(buttons[i], XmNuserData, &name, NULL);
      element = (field_pick_hash_elmt *) 
	hash_Lookup(field_pick_hash_tab, 
		    hash_HashFunction(name, strlen(name)),
		    field_pick_compare, name);
      XmToggleButtonSetState(buttons[i], element->is_default, TRUE);
    }
    else
      break;
  }
}

/* Set whichever fields were previously selected for this database */
static void field_pick_set_prev_picked(void)
{
  int   i, j;
  field_pick_hash_elmt *element;
  
  if (!cur_db)
    return;
  if (!cur_db->display_fields) {
    field_pick_hide();
    return;
  }
  field_pick_show();
  field_pick_clear_cb(NULL, NULL, NULL);

  for (i = 0; cur_db->display_fields[i]; i++) {
    if (element = (field_pick_hash_elmt *) 
	hash_Lookup(field_pick_hash_tab, 
		    hash_HashFunction(cur_db->display_fields[i], 
				      strlen(cur_db->display_fields[i])),
		    field_pick_compare, cur_db->display_fields[i]))
      XmToggleButtonSetState(element->button, TRUE, TRUE);
    else {
      /* If the field was not found in the hash table, then we have an
	 invalid entry in display_fields, so we must get rid of it */
      XtFree(cur_db->display_fields[i]);
      for (j = i + 1; cur_db->display_fields[j]; j++)
	cur_db->display_fields[j - 1] = cur_db->display_fields[j];
      cur_db->display_fields[j - 1] = NULL;
      i--;
    }
  }
}

/* called after record-display options have been ok'ed */
void field_pick_ok(void)
{
  int        i, j, num_children;
  WidgetList buttons;
  char *sname;
  Boolean    at_least_one = FALSE;
  Boolean    all_default = TRUE;
  field_pick_hash_elmt *element;

  if (!cur_db)
    return;
  
  /* Handle the standard fields case */
  if (!XtIsManaged(widgets[k_field_pick_form])) {
    at_least_one = TRUE;
    if (cur_db->display_fields) {
      for (i = 0; cur_db->display_fields[i]; i++)
	XtFree(cur_db->display_fields[i]);
      XtFree(cur_db->display_fields);
      cur_db->display_fields = NULL;
    }
  }

  /* Otherwise do custom fields */
  else {
    if (!cur_db->display_fields)
      cur_db->display_fields = (char **) XtMalloc(sizeof(char *) * MAX_PICK_FIELDS);
  
    XtVaGetValues(widgets[k_field_pick_rc], XmNnumChildren, &num_children,
		  XmNchildren, &buttons, NULL);
    
    if (num_children > MAX_PICK_FIELDS)
      error("More fields in database than MAX_PICK_FIELDS");
  
    for (i = 0, j = 0; i < num_children; i++) {
      if (XtIsManaged(buttons[i])) {
	XtVaGetValues(buttons[i], XmNuserData, &sname, NULL);
	if (XmToggleButtonGetState(buttons[i])) {
	  cur_db->display_fields[j++] = strdup(sname);
	  at_least_one = TRUE;
	}

	/* We need to check check current selections against defaults
	   for this database. */
	if (all_default) {
	  element = (field_pick_hash_elmt *) 
	    hash_Lookup(field_pick_hash_tab, 
			hash_HashFunction(sname, strlen(sname)),
			field_pick_compare, sname);
	  if (element->is_default != XmToggleButtonGetState(buttons[i]))
	    all_default = FALSE;
	}
      }
      else
	break;
    }
    cur_db->display_fields[j] = NULL;

    /* Record whether we are using defaults for this database */
    cur_db->all_default_fields = all_default;
  }
  full_record_fields_build();
  if (!at_least_one)
    pop_warning("You have selected 0 fields for display. Your Full Record\nretrievals will be empty.", NULL, search_shell);

}

void field_pick_cancel(void)
{
  field_pick_set_prev_picked();
}

/* Comparison function to be passed into hash routines */
int field_pick_compare(char *key, field_pick_hash_elmt *element)
{
  return(!strcasecmp(key, element->str));
}

/* Place-holder function for hash_Reset */
void field_pick_hash_free(char *data)
{
}

/* Copy all the selected fields from the current database into
   a block suitable for transmission to the driver. */
static void full_record_fields_build(void)
{
  int i;

  full_record_fields_len = 0;
  if (!cur_db)
    return;
  if (!cur_db->display_fields)
    return;
  
  for (i = 0; cur_db->display_fields[i]; i++) {
    if ((full_record_fields_len + strlen(cur_db->display_fields[i]) + 1) >
	FULL_FIELDS_LEN)
      error("Total length of selected field identifiers over-ran buffer");
    strcpy(full_record_fields + full_record_fields_len, 
	   cur_db->display_fields[i]);
    full_record_fields_len += strlen(cur_db->display_fields[i]) + 1;
  }
}

/* Parse a string containing a database name and fields list, install
   into the database record. Strings comes either from .willowrc or 
   X-resources.
 */
int parse_field_pick_str(char *str)
{
  char *db_tag;
  char *fields;
  db_desc_rec *db;
  int  i, sdex;
  Boolean done = FALSE;

  db_tag = str;
  if (!(fields = strchr(db_tag, SUB_SEP_CHAR)))
    return -1;
  *fields = 0;
  fields++;

  db = (db_desc_rec *) hash_Lookup(db_hash_tab,
				   hash_HashFunction(db_tag, strlen(db_tag)),
				   db_compare, db_tag);
  /* If the referenced database does not exist now, no big deal */
  if (!db)
    return (0);

  if (!db->display_fields)
    db->display_fields = (char **) XtMalloc(sizeof(char *) * MAX_PICK_FIELDS);

  i = 0;
  while (!done) {
    sdex = 0;
    while ((fields[sdex] != FIELD_PICK_SEP) && (fields[sdex]))
      sdex++;
    if (!fields[sdex])
      done = TRUE;
    fields[sdex] = 0;
    if (fields[0])
      db->display_fields[i++] = strdup(fields);
    fields += sdex + 1;
  }
  db->display_fields[i] = NULL;

  return 0;
}

/* Parse this list of custom-field settings, as created by 
   customFields X resource.
 */
void parse_field_pick_list()
{
  char *p, *idex;
  int len, total_len;

  p = custom_fields_list;
  total_len = strlen(p);
  while (p < (custom_fields_list + total_len)) {
    if (idex = strchr(p, '\n'))
      *idex = 0;
    len = strlen(p);
    parse_field_pick_str(p);
    p += len + 1;
  }
}
