/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/*******************************************************
 * print.c
 * This file contains routines for printing full record
 * results and summaries
 ******************************************************/

#include <stdio.h>
#include <sys/types.h>

#include <Mrm/MrmAppl.h>
#include <Xm/Text.h>
#include "willow.h"

#define IN  0
#define OUT 1

extern Widget widgets[], results_shell, summary_shell, toplevel, search_shell;
extern XtAppContext app_context;
extern Boolean printing;

extern char *print_cmd;
extern char *summaries[];

void print_read(caddr_t data, int *fid, XtInputId *iid);
int open_print_pipe (int which);
extern int write (int, void *, unsigned int);
extern int close (int);
extern Boolean list_get_selected (Widget w, int **pos_list, int *num);
extern void force_summaries (void);
extern Widget XmSelectionBoxGetChild (Widget, unsigned int);
extern int desensitize_selection_buttons (void);
extern int pop_working (char *, Widget);
extern char *get_query_display_str(void);

extern int pipe (int *);
#ifdef vax
extern int vfork ();
#else
extern pid_t vfork (void);
#endif
extern int dup2 (int, int);
extern int execlp (char *, char *, ...);
extern int error (char *);
extern int read (int, void *, unsigned int);
extern int sensitize_selection_buttons (void);
extern void pop_warning (char *, char *, Widget);
extern int not_ready_buttons (void);
extern int clear_working (void);
extern int ready_buttons (void);


/* Pipes into print, and error back */
int to_print[2],
    from_print[2];

/* length of error message from sh/print */
int err_len = 0;

/* Should the "Printing..." dialog box be popped up? Initialized in
   resources.c */
Boolean show_print_dialog;

/* Send a full-record result through a one way pipe to whatever the
   print command is. 
 */

void print_full_cb(Widget w, XtPointer data, u_long *reason)
{
  char *text;
  
  open_print_pipe(FULL_FLAG);
  text = XmTextGetString(widgets[k_text]);
  
  write(to_print[OUT], get_query_display_str(), 
	strlen(get_query_display_str()));
  write(to_print[OUT], text, strlen(text));
  close(to_print[OUT]);
  XtFree(text);
}

/* Send the summaries through one way pipe to print command. */
void print_summ_cb(Widget w, XtPointer data, u_long *reason)
{
  char *result;
  int  *position_list;
  int   position_count, i;

  if (list_get_selected(widgets[k_summary_titles], &position_list, 
			   &position_count)) {
    open_print_pipe(NULL);
    force_summaries();
    write(to_print[OUT], get_query_display_str(), 
	  strlen(get_query_display_str()));
    for (i = 0; i < position_count; i++) {
      result = summaries[position_list[i]];
      write(to_print[OUT], result, strlen(result));
      write(to_print[OUT], "\n", 1);
    }
    close(to_print[OUT]);
    XtFree(position_list);
  }
}

void cancel_print_opt_cb(Widget w, int *tag, u_long *reason)
{
  XmTextSetString(XmSelectionBoxGetChild(widgets[k_print_opt_box], 
					 XmDIALOG_TEXT), print_cmd);
}

void ok_print_opt_cb(Widget w, int *tag, u_long *reason)
{
  print_cmd = 
    XmTextGetString(XmSelectionBoxGetChild(widgets[k_print_opt_box],
					   XmDIALOG_TEXT));

}

/* Open a pipe to the print command */
open_print_pipe(int which)
{
  int print_pid = 0;

  XtSetSensitive(widgets[k_full_print], FALSE);
  desensitize_selection_buttons();
  not_ready_buttons();

  printing = TRUE;
  if (show_print_dialog)
    pop_working("Printing...", which == FULL_FLAG ? results_shell : summary_shell);

  /* set up pipes for communicating with print */
  pipe(to_print);
  pipe(from_print);
   
  /* fork off a child, which will overlay print on itself */
  print_pid = vfork();

  if (print_pid == 0) {
    /* in the child, redirect stdin, stdout, and stderr. */
    dup2(to_print[IN], fileno(stdin));
    dup2(from_print[OUT],fileno(stdout));
    dup2(from_print[OUT],fileno(stderr));
    close(to_print[IN]);
    close(to_print[OUT]); 
    close(from_print[IN]); 
    close(from_print[OUT]);

    /* Exec sh, pass in print command */
    if (execlp("/bin/sh", "sh", "-c", print_cmd, NULL) < 0) {
      /* at this point we are in a child process, so we can't do much */
      perror(NULL);
      /* Must _exit after vfork */
      _exit(-1);
    }
  }
  else if (print_pid > 0) {
    /* in the parent, close extra pipes */
    close(to_print[IN]);
    close(from_print[OUT]);
    
    /* Now add a callback to read from pipe whenever something
       appears */
    XtAppAddInput(app_context, from_print[IN], 
		  XtInputReadMask, print_read, NULL);
  }
  else {
    perror(NULL);
    error("Could not fork sh/print");
  }
}

/* Read from standard error from print process. 
 */
void print_read(caddr_t data, int *fid, XtInputId *iid)
{
  static char print_msg_buf[CHUNK_SIZE];
  int  n = 0;
  
  printing = FALSE;
  if (show_print_dialog)
    clear_working();

  /* The -1 is so there is always room to append a \0 */
  n = read(from_print[IN], print_msg_buf + err_len, 
	   sizeof(print_msg_buf) - err_len - 1);  

  if (n)
    err_len += n;
    
  else {
    sensitize_selection_buttons();
    ready_buttons();

    if (print_cmd[0])
      XtSetSensitive(widgets[k_full_print], TRUE);
    XtRemoveInput(*iid);
    close(from_print[IN]);

    if (err_len) {
      /* Got an error message, print failed */
      print_msg_buf[err_len] = 0;
      pop_warning("Printing Error:\n%s", print_msg_buf, search_shell);
      err_len = 0;
    }
  }
}
