/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/******************************************************
 protocol.h
 This file contains definitions used for the willow
 protocol. It is included by willow, and by the driver
 programs.
 
Willow Sends                     Driver Can Respond
============                     ==================
Connect                          Connected     			           
  database name                  Misc. Error 
  title fields			   error text
  summary fields		 Info Display
  extension fields
  environment list

 [extension fields, and environment list may be null.]

Search				 Stats
  num text fields		   result id
  type val cmd		           num results
  type val cmd		           stats string
  type val cmd		         Error
  num limit fields		   error text
  limit
  limit

Search Cancel                    Stats, no results

Get Title			 Title
  ref #				   ref #
  		                   title text   
				 Error
				   error text

Get Record			 Record
  ref #				   ref #
  short/long                       short/long
  abbrev/full names                number of match words       
				   list of match word offsets  
				   result-text                 
				 Error
				   error-text

Get Extension Fields             Extension
  ref #                            ref #
                                   extensions text
				 Error
				   error text

Get Browse List Segment          Browse List Segment
  ref #                            ref #
  search type                      line position
  lines wanted                     list data
  offset wanted
  list name
  prefix string

Disconnect			 no response

echo_toggle                      no response

cheat
  text to passed to dbms         no response


Driver Sends                     Willow Responds
============                     ===============
User Request                     User Response
  Type (boolean, text, masked)     Yes/No
  Name                             Text
  Text                             
  Yes-label
  No-label


 ******************************************************/

#define PROTOCOL

/* Define constants for i/o redirection */
#define R 0
#define W 1

/* File descriptors for willow<==>driver pipe */
#define TO_DRIVER_FD   3
#define FROM_DRIVER_FD 4

/* These are the codes for packet types */
#define CONNECT      0
#define CONNECTED    1
#define ERROR        2
#define SEARCH       3
#define STATS        4
#define TITLE_REQ    5
#define TITLE        6
#define RECORD_REQ   7
#define RECORD       8
#define DISCONNECT   9
#define TOGGLE_ECHO  10
#define CHEAT        11
#define CANCEL       12
#define INFO         13
#define EXT_REQ      14
#define EXTS         15
#define USER_REQ     16
#define USER         17
#define BROWSE_REQ   18
#define BROWSE       19

/* Codes for error types */
#define MISC       0
#define BAD_SEARCH 1
#define DEAD       2
#define UPDATE     3
#define WARNING    4

/* Codes for user-request types */
#define BOOL_REQ   0
#define TEXT_REQ   1
#define MASKED_REQ 2

/* Codes for Info types */
#define COPYRIGHT_INFO 0
#define FIELDS_INFO    1

/* Pre-define tags for two user-request types */
#define UID_TAG    "UserID"
#define PW_TAG     "Password"

#ifdef __alpha
typedef int netint; /* 32 bit network integer */
#else
typedef long int netint; /* 32 bit network integer */
#endif

/* Structures for each possible packet type. If a packet has string data in
   it, there will a list of lengths for each each string field, then a
   data area at the end.

   The final type, conversion_rec, must be as long as the longest of these
   variants. This is used in protocol.c for converting byte order.
 */

/* error structure */
typedef struct {
  netint   type;
} err_rec;

/* search structure. This one is much more complex than the others.
   The data will contain a series of null terminated strings. First 
   the text from each used line in the search composition window, followed
   by the proper search command for that field. Num_text_fields tells how
   many such pairs to expect. Then a string for each limit field that is
   set. Num_limit_fields tells how many of those to expect.
 */
typedef struct { 
  netint num_text_fields;
  netint num_limit_fields;
} search_rec;

/* stats structure */
typedef struct {
  netint  result_id;
  netint  num_results;
} stats_rec;

/* Get-title structure */
typedef struct {
  netint title_num;
} title_req_rec;

/* returned-title structure */
typedef struct {
  netint title_num;
} title_rec;

/* Get-record-request structure. There can be an optional field specifier
   in the database, looked at only for full-records, triggered by 
   the Willow user setting custom field-display. */
typedef struct {
  netint ref_num;
  netint summary;
  netint abbreviated;
} record_req_rec;

/* Returned record structure */
typedef struct {
  netint ref_num;
  netint summary;
  netint num_hilites;
} record_rec;

/* Get-extension-request structure */
typedef struct {
  netint ext_num;
} ext_req_rec;

/* Returned extension structure */
typedef struct {
  netint ext_num;
} ext_rec;

/* A database information record sent spontaneously (and optionally) 
   from the driver to Willow. The data portion contains either the
   copyright information string or a list of all the available fields
   for this database. If it is a fields list, the first character
   of each field is either 0 or 1 to indicate if this field is retrieved
   by default in the full-record. Followed by short-name/long-name pairs.
 */
typedef struct {
  netint type;        /* Either COPYRIGHT_INFO or FIELDS_INFO */
  netint fields_cnt;  /* this only needed for FIELD_INFO sub-type */
} info_rec;

/* A request from the driver to the user. Can be for a boolean yes/no
   response, typed text, or typed text which gets masked by asterisks.

   The driver tags the question with a label, i.e. "UserId", so that if 
   during the same Willow session Willow sees the the same question from
   same database, it can retype the user's previous response.
 */
typedef struct {
  netint type;
/* In the data portion, the driver should provide four strings, any of
   which can be empty...
    char *question_text;
    char *question_tag; 
    char *yes_button_label;
    char *no_button_label;
 */
} user_req_rec;

/* Response from user back to the driver. */
typedef struct {
  netint answer; /* 1 for yes, 0 for no */
/* In the data portion is whatever the user typed in, possibly empty string
    char *answer_text;
 */
} user_resp_rec;

/* Browse list request */
typedef struct {
  netint  ref_num;
  netint  search_type;  /* 'H'eader 'C'haracter or 'N'umeric */
  netint  lines_wanted; /* How many lines of data wanted */
  netint  line_offset;  /* Where in returned block to start match */
/* In the data portion, are two strings:
    char *list_name;
    char *prefix_string; */
} browse_req_rec;

/* Browse list result */
typedef struct {
  netint  ref_num;
  netint  line_pos;   /* offset in complete list, or -1 if error */
  /* plus string of actual result in the data portion */
} browse_rec;

/* Dummy record type for converting byte-orders. This must be as
   long as the longest of the above records (currently browse_rec) */
typedef struct {
  netint int1;
  netint int2;
  netint int3;
  netint int4;
} conversion_rec;

/* This is the structure for an entire packet */
typedef struct {
  netint  type;
  netint  data_len;
  union {
    err_rec         error;
    search_rec      search;
    stats_rec       stats;
    title_req_rec   title_req;
    title_rec       title;
    record_req_rec  record_req;
    record_rec      record;
    ext_req_rec     ext_req;
    ext_rec         exts;
    browse_req_rec  browse_req;
    browse_rec      browse;
    info_rec        info;
    user_req_rec    user_req;
    user_resp_rec   user_resp;
    conversion_rec  conversion;
  } var; 
  /* This data pointer is not actually transferred between willow and
     driver, nor counted in total size of the packet rec for transfer
     purposes. */
  char *data;
} packet_rec;

/* Use this instead of actual size of packet for network read/write purposes */
#define PACKET_NET_SIZE sizeof(packet_rec) - sizeof(char *)

/* Define a structure for building linked list of hilite points, used
   by both driver and willow */
typedef struct _hl_rec {
  netint   start, end;
  struct _hl_rec *next, *prev;
} hl_rec;
 
 

