/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/****************************************************************
 * save_search.c
 * This file contains the code necessary for adding and deleting
 * searches to and from the search menu, as well as code for 
 * reading and writing the search sets from the menu to and from
 * files.
 ****************************************************************/
#include <stdio.h>
#include <sys/time.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif


#include <Mrm/MrmAppl.h>
#include <Xm/Text.h>
#include <Xm/SelectioB.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>

#include "willow.h"
#include "db.h"
#include "save_query.h"
#include "dict.h"

/* A linked list of named searches */
Search_list_ptr  search_head = NULL,
                  search_tail = NULL;

/* External data structures */
extern Widget     widgets[],
                  toplevel,
                  search_shell;

extern MrmHierarchy s_MrmHierarchy; /* MRM database hierarchy ID */
extern MrmType dummy_class;         /* and class variable. */
extern db_desc_rec *cur_db;         /* current database pointer */
extern int opac_mode;               /* Are we in basic mode? */
extern sel_rec_def *sel_list;       /* need this for basic mode displays */

/* Textual display of current search strategy, for pre-pending onto
   saved search results */
static char *query_display_str;

/* Should we do the pre-pend? */
Boolean show_search;

/* External functions */
extern void clear_dialog_cb (Widget, Widget, u_long *);
extern char *xmstring_to_string (XmString string);
extern void pop_warning (char *template, char *arg, Widget parent);
extern char *get_line (char *s, FILE *f);
extern int atoi (char *);
extern Widget range_get_widget (int type, int val);
extern int error (char *msg);
extern void XmListDeleteItemsPos (Widget, int, int);
extern void XmListAddItemUnselected (Widget, XmString, int);
extern void setup_range_cb (Widget w, int which, u_long *reason);
extern void range_ok_cb (Widget w, int *tag, u_long *reason);


/* Forward function declarations */
void unpack_search_cb(Widget w, Search_ptr search, caddr_t reason);
void free_search(Search_ptr search);
void add_search_cb(Widget w, int *tag, u_long *reason);
void really_add_search_cb(Widget w, int *tag, XmSelectionBoxCallbackStruct *reason);
void del_search_cb(Widget w, int *tag, u_long *reason);
void really_del_search_cb(Widget w, int *tag, XmListCallbackStruct *reason);
void really_open_search_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason);
void really_save_search_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason);
Search_ptr package_search(void);

/* pop up a file selection box in order to choose a file to save
   a search set into.
 */

int really_really_del_search (Search_list_ptr p);
int really_really_add_search (Search_ptr search, XmString name);
int zap_list_items (Widget w);
int add_list_items (Widget w, Search_list_ptr p);


/* Called after selecting a file to save */
void really_save_search_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason)
{
  FILE  *outfile;
  char  *filename,
        *menu_name;
  int   i;
  Search_list_ptr  p;

  /* Sometimes get called bogusly */
  if (w && !XmToggleButtonGetState(widgets[k_q_local_save_btn]))
    return;

  clear_dialog_cb(w, widgets[k_q_save_box], NULL);

  filename = xmstring_to_string(reason->value);
  outfile = fopen(filename,"w"); 
  if (!outfile){
    pop_warning("Unable to open %s", filename, search_shell);
    XtFree(filename);
    return;
  }
  XtFree(filename);

  /* First write a 'magic cookie' into the file, so we can identify it
     as valid later. */
  fprintf(outfile, "%d\n", SEARCH_FILE_COOKIE);
  fprintf(outfile, "%s\n", DONT_EDIT_WARNING);

  /* Go through the list, and record each search item. */
  p = search_head;
  while (p) 
    {
      menu_name= xmstring_to_string(p->name);
      fprintf(outfile, "%s\n", menu_name);
      XtFree(menu_name);
      for (i = 0; i < NUM_TEXT_FIELDS; i++){
	fprintf(outfile,"%s\n%s\n",
		p->search->field_names[i], 
		p->search->text_vals[i]);
	fprintf(outfile,"%d%d\n%d\n%d\n%d\n",
		p->search->froms[i]->limited,
		p->search->froms[i]->low,
		p->search->froms[i]->month_val,
		p->search->froms[i]->day_val,
		p->search->froms[i]->numeric);
	fprintf(outfile,"%d%d\n%d\n%d\n%d\n",
		p->search->tos[i]->limited,
		p->search->tos[i]->low,
		p->search->tos[i]->month_val,
		p->search->tos[i]->day_val,
		p->search->tos[i]->numeric);
      }
      for (i = 0; i < NUM_BOOL_FIELDS; i++)
	fprintf(outfile, "%s\n%d\n", 
		p->search->bool_names[i],
		p->search->bool_vals[i]);
      p = p->next;
    }
  fclose(outfile);
}

/* Delete all saved searches */
del_all_searches(void)
{
  Search_list_ptr p;
  
  p = search_head;
  while (p) {
    p = search_head->next;
    really_really_del_search(search_head);
    search_head = p;
  }
  search_head = search_tail = NULL;
}

/* Called after selecting a file to open */
void really_open_search_cb(Widget w, int *tag, XmFileSelectionBoxCallbackStruct *reason)
{
  FILE  *infile;
  char  *filename;
  int    i, cookie;
  char   line_buf[MAX_LINE];
  Search_ptr        s;
  XmString search_name;

  /* Sometimes get called bogusly */
  if (w && !XmToggleButtonGetState(widgets[k_q_local_save_btn]))
    return;

  if (w)
    clear_dialog_cb(w, widgets[k_q_save_box], (u_long *) reason);

  filename = xmstring_to_string(reason->value);
  infile = fopen(filename,"r"); 
  if (!infile){
    pop_warning("Unable to open file %s", filename, search_shell);
    XtFree(filename);
    return;
  }

  /* Check that first line of file contains 'magic cookie' */
  get_line(line_buf, infile);
  sscanf(line_buf, "%d", &cookie);
  if (cookie != SEARCH_FILE_COOKIE) {
    pop_warning("%s appears to be invalid!", filename, search_shell);
    return;
  }
  XtFree(filename);

  /* Skip over warning line */
  get_line(line_buf, infile);

  /* delete all the items currently on the search list */
  del_all_searches();

  /* read and package each search out of the file, then insert it */
  while (get_line(line_buf, infile))  {
    s = (Search_ptr) XtMalloc(sizeof(Search_rec));
    search_name = XmStringCreateSimple(line_buf);
    
    /* read each option menu setting and text value */
    for (i = 0; i < NUM_TEXT_FIELDS; i++) {
      get_line(line_buf, infile);
      s->field_names[i] = strdup(line_buf);
      get_line(line_buf, infile);
      s->text_vals[i] = strdup(line_buf);
      get_line(line_buf, infile);

      s->froms[i] = (range_stat_rec *) XtMalloc(sizeof(range_stat_rec));
      s->froms[i]->limited = (line_buf[0] == '1');
      s->froms[i]->low     = (line_buf[1] == '1');

      get_line(line_buf, infile);
      s->froms[i]->month_val = atoi(line_buf);
      s->froms[i]->month = range_get_widget(MONTH, s->froms[i]->month_val);

      get_line(line_buf, infile);
      s->froms[i]->day_val = atoi(line_buf);
      s->froms[i]->day = range_get_widget(DAY, s->froms[i]->day_val);

      get_line(line_buf, infile);
      s->froms[i]->numeric = atoi(line_buf);

      s->tos[i] = (range_stat_rec *) XtMalloc(sizeof(range_stat_rec));
      get_line(line_buf, infile);
      s->tos[i]->limited = (line_buf[0] == '1');
      s->tos[i]->low     = (line_buf[1] == '1');

      get_line(line_buf, infile);
      s->tos[i]->month_val = atoi(line_buf);
      s->tos[i]->month = range_get_widget(MONTH, s->tos[i]->month_val);

      get_line(line_buf, infile);
      s->tos[i]->day_val = atoi(line_buf);
      s->tos[i]->day = range_get_widget(DAY, s->tos[i]->day_val);

      get_line(line_buf, infile);
      s->tos[i]->numeric = atoi(line_buf);
    }
    
    /* read the boolean fields */
    for (i = 0; i < NUM_BOOL_FIELDS; i++) {
      get_line(line_buf, infile);
      s->bool_names[i] = strdup(line_buf);
      fscanf(infile,"%d\n", &s->bool_vals[i]);
    }      
    really_really_add_search(s, search_name);
  }
}

/* Really add a new search name to the searches menu */
void really_add_search_cb(Widget w, int *tag, XmSelectionBoxCallbackStruct *reason)
{
  Search_ptr  search;

  search = package_search();
  really_really_add_search(search, XmStringCopy(reason->value));
}

/* Insert the necessary information about this search into our linked
   list of searches, and create a menu item for it */
really_really_add_search(Search_ptr search, XmString name)
{
  Search_list_ptr  p;
  Arg   args[10];
  int   n;
  Widget button;

  /* first add the menu button to the searches menu */
  n = 0;
  XtSetArg(args[n], XmNlabelString, name);n++;
  button = XmCreatePushButton(widgets[k_searches_pdm], "button", args,n);
  XtManageChild(button);
    
  XtAddCallback(button, XmNactivateCallback, unpack_search_cb, search);

  /* now insert into search list */
  p = (Search_list_ptr) XtMalloc(sizeof(Search_list_rec));
  p->button = button;
  p->search = search;
  p->name = name;
  p->next = NULL;

  if (!search_head) {
    /* Empty list case */
    search_head = search_tail = p;
  }
  else {
    search_tail->next = p;
    search_tail = p;
  }
}

/* delete a named search from the current searches menu */
void del_search_cb(Widget w, int *tag, u_long *reason)
{
  Widget  button;
  
  if (!search_head)
    return;

  if (widgets[k_del_search] == NULL) {
    if (MrmFetchWidget(s_MrmHierarchy, 
		       "del_search_box", 
		       search_shell,
		       &widgets[k_del_search], 
		       &dummy_class) != MrmSUCCESS)
      error("can't fetch delete search window widget");
    
    /* We don't want help button, so get it and remove it */
    button = XmSelectionBoxGetChild(widgets[k_del_search], XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);
  }
  
  /* put together the list of choices */
  zap_list_items(XmSelectionBoxGetChild(widgets[k_del_search], XmDIALOG_LIST));
  add_list_items(XmSelectionBoxGetChild(widgets[k_del_search], XmDIALOG_LIST),
		 search_head);
  XmTextSetString(XmSelectionBoxGetChild(widgets[k_del_search], XmDIALOG_TEXT),
		  "");

  map_dialog(widgets[k_del_search]);  
}

/* Clear out any existing items on the XmList */
zap_list_items(Widget w)
{
  /* Not using XmListDeleteAllItems here, cause it seems flakey */
  Arg  args[10];
  int  n;
  int  count;

  n = 0;
  XtSetArg(args[n], XmNitemCount, &count);n++;
  XtGetValues(w, args, n);

  XmListDeleteItemsPos(w, count, 1);
}

/* Add add an item to the XmList, from the search list. Do it recursively,
   for fun.
 */
add_list_items(Widget w, Search_list_ptr p)
{
  if (!p)
    return;
  else
    {
      XmListAddItemUnselected(w, p->name, 0);
      add_list_items(w, p->next);
    }
}

/* really delete a named search from the current searches menu */
void really_del_search_cb(Widget w, int *tag, XmListCallbackStruct *reason)
{
  Search_list_ptr  p,prev;

  /* When this callback is activated via <return>, we get called twice for
     some reason. Therefore we could get in trouble unless we do the following
     check.
   */
  if (!search_head)
    return;

  /* first search through the linked list, then find and unlink item */
  p = prev = search_head;
  while (!XmStringCompare(p->name, reason->item)) {
    prev = p;
    p = p->next;
    if (!p)
      return;
  }
  
  if (search_head == search_tail)
    search_head = search_tail = NULL;
  else if (p == search_head)
    search_head = p->next;
  else if (p == search_tail) {
    search_tail = prev;
    prev->next = NULL;
  }
  else
    prev->next = p->next;

  /* De-allocate all the storage, and unmanage menu button */
  really_really_del_search(p);
}

/* Really really delete a search from the menu, and the search
   list.
 */
void free_search(Search_ptr search)
{
  int    i;

  for (i = 0; i < NUM_TEXT_FIELDS; i++) {
    XtFree(search->field_names[i]);
    XtFree(search->text_vals[i]);
    XtFree(search->froms[i]);
    XtFree(search->tos[i]);
  }
  for (i = 0; i < NUM_BOOL_FIELDS; i++) {
    XtFree(search->bool_names[i]);
  }
  XtFree(search);
}

really_really_del_search(Search_list_ptr p)
{
  XtUnmanageChild(p->button);
  free_search(p->search);
  XmStringFree(p->name);
  XtFree(p);
}

/* Bundle up all the information which describes the current search
   into a neat package.
   Return a pointer to a structure which contains everything.
 */
Search_ptr package_search(void)
{
  int        i;
  Search_ptr result;
  Arg        args[10];
  int        n;
  Widget     button;
  XmString   label;
  range_stat_rec *range_stats;
  
  result = (Search_ptr) XtMalloc(sizeof(Search_rec));

  for (i = 0; i < NUM_TEXT_FIELDS; i++) {
    /* retrieve user-entered text. del_search() should free strings */
    result->text_vals[i] = XmTextGetString(widgets[k_field1 + i]);
    
    /* figure out which option is chosen on the menu: first, retrieve
       the id of the currently selected button */
    n = 0;
    XtSetArg(args[n], XmNmenuHistory, &button);n++;
    XtGetValues(widgets[k_label1 + i], args, n);
    n = 0;
    XtSetArg(args[n], XmNlabelString, &label);n++;
    XtGetValues(button, args, n);
    result->field_names[i] = xmstring_to_string(label);

    /* Record the state of the range buttons */
    XtVaGetValues(widgets[k_from1 + i], XmNuserData, &range_stats, NULL);
    result->froms[i] = (range_stat_rec *) XtMalloc(sizeof(range_stat_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memmove(result->froms[i], range_stats, sizeof(range_stat_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
    bcopy(range_stats, result->froms[i], sizeof(range_stat_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */

    XtVaGetValues(widgets[k_to1 + i], XmNuserData, &range_stats, NULL);
    result->tos[i] = (range_stat_rec *) XtMalloc(sizeof(range_stat_rec));
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memmove(result->tos[i], range_stats, sizeof(range_stat_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
    bcopy(range_stats, result->tos[i], sizeof(range_stat_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  }

  /* Retrieve and store state of boolean fields */
  for (i = 0; i < NUM_BOOL_FIELDS; i++) {
    n = 0;
    XtSetArg(args[n], XmNlabelString, &label);n++;
    XtGetValues(widgets[k_bool1 + i], args, n);
    result->bool_names[i] = xmstring_to_string(label);
    result->bool_vals[i] = (int) XmToggleButtonGetState(widgets[k_bool1 + i]);
  }
  return(result);
}

/* Take a search out of a Search_rec, and install it as the current
   search.
 */
void unpack_search_cb(Widget w, Search_ptr search, caddr_t reason)
{
  int    i,j;
  int    n;
  Arg    args[10];
  Widget menu_pane, button;
  int    num_children;
  WidgetList buttons;
  XmString label;
  char  *button_name;
  range_stat_rec *range_stats;
  db_field_desc_rec *field;
  
  if (!cur_db)
    return;

  /* Set up the text fields */
  for (i = 0; i < NUM_TEXT_FIELDS; i++) {
    if (search->text_vals[i][0] || search->froms[i]->numeric ||
	search->tos[i]->numeric){
      /* First get the menu pane, so we can look at buttons */
      n = 0;
      XtSetArg(args[n], XmNsubMenuId, &menu_pane);n++;
      XtGetValues(widgets[k_label1 + i], args, n);
      /* Now get buttons */
      n = 0;
      XtSetArg(args[n], XmNnumChildren, &num_children);n++;
      XtSetArg(args[n], XmNchildren, &buttons);n++;
      XtGetValues(menu_pane, args, n);
      
      /* Loop through buttons, searching for one with the right name */
      button = NULL;
      n = 0;
      XtSetArg(args[n], XmNlabelString, &label);n++;
      for (j = 0; j < num_children; j++) {
	if (!XtIsManaged(buttons[j]))
	  break;
	XtGetValues(buttons[j], args, n);
	button_name = xmstring_to_string(label);
	if (strcmp(button_name, search->field_names[i]) == 0) {
	  button = buttons[j];
	  XtFree(button_name);
	  break;
	}
	XtFree(button_name);
      }
      
      /* If we found a matching button, install it */
      if (button) {
	n = 0;
	XtSetArg(args[n],XmNmenuHistory, button);n++;
	XtSetValues(widgets[k_label1 + i], args, n);
	setup_range_cb(button, i, NULL);
	
	/* If this is a range field, adjust buttons with saved data */
	if (!XtIsManaged(widgets[k_field1 + i])) {
	  /* Track down field_record for range_stats */
	  for (field = cur_db->text_fields; 
	       strcmp(search->field_names[i], field->name) != 0;
	       field = field->next)
	    ;
	  search->froms[i]->field = field;
	  search->tos[i]->field = field;
	  
	  XtVaGetValues(widgets[k_from1 + i], XmNuserData, &range_stats, NULL);
#if defined(POSIX) || defined (_POSIX_SOURCE)
	  memmove(range_stats, search->froms[i], sizeof(range_stat_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
	  bcopy(search->froms[i], range_stats, sizeof(range_stat_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
	  range_ok_cb(widgets[k_from1 + i], NULL, NULL);
	  
	  XtVaGetValues(widgets[k_to1 + i], XmNuserData, &range_stats, NULL);
#if defined(POSIX) || defined (_POSIX_SOURCE)
	  memmove(range_stats, search->tos[i], sizeof(range_stat_rec));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
	  bcopy(search->tos[i], range_stats, sizeof(range_stat_rec));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
	  range_ok_cb(widgets[k_to1 + i], NULL, NULL);
	}
      }
    }
    XmTextSetString(widgets[k_field1 + i], search->text_vals[i]);
  }

  /* Set up the boolean fields */
  for (i = 0; i < NUM_BOOL_FIELDS; i++) {
    n = 0;
    XtSetArg(args[n], XmNlabelString, &label);n++;
    for (j = 0; j < (cur_db ? 
		     cur_db->num_limit_fields: NUM_BOOL_FIELDS); 
	 j++) {
      XtGetValues(widgets[k_bool1 + j], args, n);
      button_name = xmstring_to_string(label);
      if (strcmp(button_name, search->bool_names[i]) == 0) {
	XmToggleButtonSetState(widgets[k_bool1 + i], 
			       (Boolean)search->bool_vals[i], FALSE); 
	XtFree(button_name);
	break;
      }
      XtFree(button_name);
    }
  }
}

/* Prepare a user-readable representation of the current query
   for inclusion with the saved results.
 */
void make_query_display(void)
{
  time_t secs;
  struct tm *real_time;
  char   date_str[MAX_LINE];
  int        i, j;
  Widget     button;
  XmString   label, range_label1, range_label2;
  char    *field_text, *field_label, *range_text1, *range_text2;
  int      len;
  sel_rec_def *selected;
  Boolean  first = TRUE;

  if (!cur_db)
    return;
  if (!query_display_str)
    query_display_str = XtMalloc(NUM_TEXT_FIELDS * MAX_LINE * 2);
  
  /* put time/date stamp and database name in */
  secs = time(NULL);
  real_time = localtime(&secs);
  strftime(date_str, MAX_LINE, "%A, %B %d, %Y. %r", real_time);
  
  sprintf(query_display_str, QUERY_DISPLAY_HEADER, date_str, cur_db->long_name);

  /* Check if we are doing a basic-mode search */
  if (opac_mode > 0) {
    XtVaGetValues(widgets[k_lookup_label], XmNlabelString, &label, NULL);
    field_label = xmstring_to_string(label);

    strcat(query_display_str, QUERY_DISPLAY_PREFIX);
    len = strlen(field_label);
    for (j = len; j < FIELD_DISP_WID; j++)
      strcat(query_display_str, " ");
    strcat(query_display_str, field_label);
    strcat(query_display_str, ": ");
    
    if (!sel_list) {
      /* this must be keywords-type search */
      field_text = XmTextGetString(widgets[k_dict_search]);
      strcat(query_display_str, field_text);
      XtFree(field_text);
      XtFree(field_label);
    }
    else {
      /* Must take the display-items out of the list */
      for (selected = sel_list; selected != NULL; selected = selected->next) {
	if (first)
	  first = FALSE;
	else {
	  strcat(query_display_str, "\n");
	  strcat(query_display_str, QUERY_DISPLAY_PREFIX);
	  for (j = 0; j < FIELD_DISP_WID + 2; j++)
	    strcat(query_display_str, " ");
	}
	strcat(query_display_str, selected->disp_value);
      }
    }
  }
  else { /* regular search */
    /* Go through search configuration and add labels/values */
    for (i = 0; i < NUM_TEXT_FIELDS; i++) {
      XtVaGetValues(widgets[k_label1 + i], XmNmenuHistory, &button, NULL);
      XtVaGetValues(button, XmNlabelString, &label, NULL);
      field_label = xmstring_to_string(label);
    
      if (XtIsManaged(widgets[k_field1 + i])) 
	field_text = XmTextGetString(widgets[k_field1 + i]);
      else {
	/* Using a range */
	field_text = XtMalloc(MAX_LINE);
	XtVaGetValues(widgets[k_from1 + i], XmNlabelString, &range_label1, NULL);
	range_text1 = xmstring_to_string(range_label1);
	XtVaGetValues(widgets[k_to1 + i], XmNlabelString, &range_label2, NULL);
	range_text2 = xmstring_to_string(range_label2);
	sprintf(field_text, QUERY_DISPLAY_RANGE, range_text1, range_text2);
      }
    
      strcat(query_display_str, QUERY_DISPLAY_PREFIX);
      len = strlen(field_label);
      for (j = len; j < FIELD_DISP_WID; j++)
	strcat(query_display_str, " ");
      strcat(query_display_str, field_label);
      strcat(query_display_str, ": ");
      strcat(query_display_str, field_text);
      strcat(query_display_str, "\n");
      XtFree(field_text);
      XtFree(field_label);
    }
    strcat(query_display_str, QUERY_DISPLAY_PREFIX);
    strcat(query_display_str, "\n");
    strcat(query_display_str, QUERY_DISPLAY_PREFIX);
    
    /* Retrieve and store states and labels of boolean fields */
    for (i = 0; i < NUM_BOOL_FIELDS; i++) {
      if (XtIsManaged(widgets[k_bool1 + i])) {
	XtVaGetValues(widgets[k_bool1 + i], XmNlabelString, &label, NULL);
	field_label = xmstring_to_string(label);
	if (XmToggleButtonGetState(widgets[k_bool1 + i]))
	  field_text = "  [X]";
	else
	  field_text = "  [ ]";
	strcat(query_display_str, field_text);
	strcat(query_display_str, field_label);
	XtFree(field_label);
      }
    }
  }
  strcat(query_display_str, QUERY_DISPLAY_FOOTER);
}

/* Return the query display string if the user has asked for it,
   otherwise return an empty string */
char *get_query_display_str(void)
{
  if (XmToggleButtonGetState(widgets[k_show_search_btn]))
    return(query_display_str);
  else
    return("");
}
