/***********************************************************************
 *                                                                     *
 *              WILLOW(tm) Information Retrieval Software              *
 *     (Washington Information Looker-upper Layered Over Windows)      *
 *             Copyright 1992-1994 University of Washington            *
 *                                                                     *
 *   Willow is a trademark of the University of Washington.  No        *
 *   commercial use of this trademark or software may be made without  *
 *   prior written permission of the University of Washington.         *
 *                                                                     *
 * Permission to use, copy, modify, and distribute Willow(tm) software *
 * and its documentation for any noncommercial purpose is hereby       *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the above copyright notice and this permission *
 * notice appear in supporting documentation, and that the name of the *
 * University of Washington not be used in advertising or publicity    *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  This software is made available "as is", *
 * and THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS  *
 * OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT         *
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS    *
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE UNIVERSITY OF   *
 * WASHINGTON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL     *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA  *
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT (INCLUDING       *
 * NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION    *
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                       *
 *                                                                     *
 * For further information please contact:                             *
 * Willow Project                                                      *
 * Information Systems                                                 *
 * University of Washington                                            *
 * 4545 15th NE, 1st floor                                             *
 * Seattle, Washington, 98105, USA                                     *
 * Internet: willow@cac.washington.edu                                 *
 ***********************************************************************/

/****************************************************************
 * utils.c
 *   Contains various handy utilities used by Willow and Wilco.
 * These routines are used by Wilco to mimic the Xt event model.
 *    add_input(fd, func, data) : Just like XtAppAddInput, this allows
 *        you to register func() to be a handler when there is data
 *        waiting to be read on fd. data is passed as an argument to
 *        func, along with the addresses of the file descriptor and 
 *        the input id (which is the same fd. The function returns an 
 *        input_id, which is identical to fd.
 *     remove_input(fd) : unregister fd.
 *     main_loop() : 
 * The following miscellaneous routines are also available.
 *    bdup(s, size) : return a duplicate of byte string s. Non
 *        necesarrily null-terminated.
 *    get_line(str, f) : read a null-terminated, \n-free line of data
 *        into str, from FILE *f.
 *    splice() : like the perl function.
 *    str_clean(str) : strip strings of undesirable characters.
 *    strndup(s) : return a duplicate of s, with malloc-ed memory, n long.
 *    usleep(us) : sleep for a number of microseconds 
 *    xmstring_to_string : convert annoying motif compound strings to
 *        real strings.
 *    xmtext_at_bottom(w): is the bottom of a text widget currently 
 *        visible?
 *
 * The following application-specific items must be declared externally
 *    error(msg): fatal error handling routine
 ****************************************************************/

#include <stdio.h>
#include <sys/types.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif

#include <ctype.h>
#include <errno.h>
#include <sys/time.h>

#if defined(_AIX) || defined(_SEQUENT_)
#include  <sys/select.h>
#endif

#ifdef WILLOW
#include <Mrm/MrmAppl.h>
#include <Xm/Text.h>
#else
#define XtMalloc  malloc
#define XtRealloc realloc
#define XtFree    free
#endif
#include "db.h"

extern void *malloc (size_t);
extern int error (char *);
extern void *realloc (void *, size_t);
extern void free (void *);
extern int getdtablesize (void);

/* Read a line from file f into null-terminated string s */
char *
get_line(char  *s,
	 FILE  *f)
{
  char *val;

  val = fgets(s, MAX_LINE, f);
  if (val)
    s[strlen(s) - 1] = 0;
  else
    s[0] = 0;

  return(val);
}

/* Sleep for a number of microseconds */
void
usleep(unsigned int nusecs)
{
  struct timeval tval;

  tval.tv_sec =  nusecs / 1000000;
  tval.tv_usec = nusecs % 1000000;
  select(0, NULL, NULL, NULL, &tval);
}

/* Replace non-printables and \n's with blanks. Necessary because
   Motif allows you to paste new-lines into a 1-line text widgets.
 */
char *
str_clean(char *s)
{
  int len,i;

  len = strlen(s);
  for (i = 0; i < len; i++)
    if (!isprint(s[i]) || (s[i] == '\n'))
      s[i] = ' ';

  return(s);
}

/* Copy a string into a newly created string, exactly n bytes. Make
   sure it is terminated */
char *
strndup(char *s, int n)
{
  char *new_str;

  new_str = XtMalloc(n + 1);

#if defined(POSIX) || defined (_POSIX_SOURCE)
  memmove(new_str, s, n);
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
  bcopy(s, new_str, n);
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  new_str[n] = 0;
  return(new_str);
}

/* Copy a string of bytes into new storage */
char *
bdup(char *s,
     int   size)
{
  char *new_str;
  
  new_str = XtMalloc(size);
#if defined(POSIX) || defined (_POSIX_SOURCE)
  memmove(new_str, s, size);
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
  bcopy(s, new_str, size);
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */
  return(new_str);
}

/* Splice one string into another, modeled after Perl function. Not 
   the most efficient way, but it is simple.
 */
void 
splice(char **dest,    /* addr of string getting changed */
       int    start,   /* start removed segment */
       int    end,     /* end of removed segment */
       char  *source,  /* string being inserted */
       int    source_len)
{
  int  dest_len,   
       new_len;     /* Length of string when we are finished */
  int  d,s;         /* dest and source indices */
  char *temp;

  if (!*dest)
    error("Attempt to splice into NULL string");

  dest_len = strlen(*dest);
  new_len = dest_len - (end - start) + source_len;

  if (new_len > dest_len)
    *dest = XtRealloc(*dest, new_len + 1);

  temp = strdup(*dest);
  
  d = 0;
  while (d < start) {
    (*dest)[d] = temp[d];
    d++;
  }

  s = 0;
  while (s < source_len) 
    (*dest)[d++] = source[s++];

  s = end;
  while (s < dest_len)
    (*dest)[d++] = temp[s++];

  (*dest)[d] = '\0';
  XtFree(temp);
}

#ifdef WILLOW
/* Convert annoying compound strings to real strings */
char *
xmstring_to_string(XmString string)
{ 
  XmStringContext   context; 
  char             *text; 
  XmStringCharSet   charset; 
  XmStringDirection dir; 
  Boolean           separator; 
  char             *buf = NULL; 
  
  int               done = FALSE;
  XmStringInitContext (&context, string);
  while (!done)
    if(XmStringGetNextSegment (context, &text, &charset,
			       &dir, &separator)){ 
      if(separator) /* Stop when next segment is a separator */
        done = TRUE;
      
      if(buf){
        buf = XtRealloc(buf, strlen(buf) + strlen(text) + 2);
        strcat(buf, text);			
      }
      else {
        buf = (char *) XtMalloc(strlen(text) +1);
        strcpy(buf, text);
      }
      XtFree(text);
    }
    else
      done = TRUE;
  
  XmStringFreeContext (context);
  return buf;
}

/* Figure out if the end of a text widget is already visible */
Boolean 
xmtext_at_bottom(Widget w)
{
  short rows;
  XmTextPosition top;
  char *text;
  int   i, lines = 0;
  Boolean  result;

  XtVaGetValues(w, XmNrows, &rows, XmNtopCharacter, &top, NULL);
  text = XmTextGetString(w);
  for (i = top; text[i]; i++) {
    if (text[i] == '\n') {
      lines++;
      if (lines > rows) {
	result = FALSE;
	break;
      }
    }
  }
  if (lines <= rows)
    result = TRUE;

  XtFree(text);
  return(result);
}
#endif

#ifndef WILLOW

/* Mimic of the functionality of a XtAppAddInput in a non-X program.
   Of course we don't use it in Willow, only Wilco.

   First, define type for Pointer to Function returning Void, then an
   array of them to use to dispatch input handlers. */

typedef void (*PFV)();
static PFV   *func_vec = NULL;
static void **data_vec = NULL;
static int    num_fds;

/* file descriptors for select call */
static fd_set master_set;

int
add_input(int fd,
	  PFV func,
	  void *data)
{
  int    i;

  if (!func_vec) {
    num_fds = getdtablesize();
    func_vec = (PFV *) malloc(sizeof(PFV) * num_fds);
    data_vec = (void *) malloc(sizeof(void *) * num_fds);
    for (i = 0; i < num_fds; i++)
      func_vec[i] = NULL;
    FD_ZERO(&master_set);
  }
  FD_SET(fd, &master_set);
  func_vec[fd] = func;
  data_vec[fd] = data;
  return(fd);
}

void
remove_input(int fd)
{
  FD_CLR(fd, &master_set);
  func_vec[fd] = NULL;
  data_vec[fd] = NULL;
}

/* Mimic the function of XtAppMainLoop(), call previously registered
   handlers as file descriptors become readable.
 */
void
main_loop(void)
{
  int  fd;
  fd_set  read_fds;

  while (TRUE) {
#if defined(POSIX) || defined (_POSIX_SOURCE)
    memmove(&read_fds, &master_set, sizeof(fd_set));
#else /* defined(POSIX) || defined (_POSIX_SOURCE) */
    bcopy(&master_set, &read_fds, sizeof(fd_set));
#endif /* defined(POSIX) || defined (_POSIX_SOURCE) */    
    if (select(num_fds, &read_fds, NULL, NULL, NULL) < 0) {
      if (errno == EINTR){  /* don't worry about it */
	errno = 0;
	continue;
      }
      else 
	error("error on select");
    }
    /* Decide which pipe has input */
    for (fd = 0; fd < num_fds; fd++)
      if (FD_ISSET(fd, &read_fds)) {
	if (func_vec[fd])
	  (*func_vec[fd])(data_vec[fd], &fd, &fd);
	else 
	  error("data available on handler-less fd");
      }
  }
}

#endif



