/* browse.c- browse translation routines */

#include <series.h>
#include <strlst.h>
#include <constant.h>
#include <session.h>
#include <msg.h>
#include <protocol.h>
#include <scaninfo.h>
#include <browse.h>

extern int debug_level; /* debugging flag from driver.c */

extern char *malloc();
char *makelower(char *s);

/* Handle a willow request for list service using z39.50-browse: take packet with willow protocol information, make into valid z39.50 scan request info, send using api, take scan response info, parse into proper willow protocol format, and send back to willow. */

void browse_wcb(SESSION **p_session, packet_rec *packet)
{

    char *ptr;
    int	refnum;
    /* Since the ref_id is optional in Z39.50, but since it's required for willow we use the Willow ref_id, but we wish we could compare the ref_id returned to be sure what we got was what we expected. */
    char searchtype;
    char *list_data;
    int	num_terms;
    int	line_pos;
    char *index_name;
    char *prefix_st;
    char *tmp;
    int collate;
    SCAN_INFO *scan;
    STRLST *sl;
    int stepsize = 0; /* default to 0(continuous) */
    int  code;
    static int prev_numeric = -1; /* use -1 if the prev scan is not numeric */
    static char *prev_numeric_text = NULL;
    int numeric;


    if (debug_level) {
	fprintf(stderr,"Entering browse_wcb\n");
	fflush(stderr);
    }

    if (prev_numeric_text == NULL) {
	prev_numeric_text = malloc(1);
	*prev_numeric_text = NULL;
    }

    /* extract request info from willow packet */
    refnum = packet->var.browse_req.ref_num;
    searchtype = packet->var.browse_req.search_type;
    num_terms = packet->var.browse_req.lines_wanted;
    line_pos = packet->var.browse_req.line_offset + 1;
    ptr = packet->data;

    index_name = strdup(ptr);

    ptr +=strlen(ptr) + 1;

    if (debug_level) {
	fprintf(stderr,"index = %s\n",index_name);
	fflush(stderr);
    }

    /* determine ASCII(default) or EBCDIC ordering */
    if ((tmp = strstr(index_name,":EBCDIC")) != NULL) {
	*tmp = NULL; /* remove suffix */
	collate = EBCDIC_COLLATING;

    } else if ((tmp = strstr(index_name,":ebcdic")) != NULL) {
	*tmp = NULL; /* remove suffix */
	collate = EBCDIC_COLLATING;

    } else if ((tmp = strstr(index_name,":ASCII")) != NULL) {
	
	*tmp = NULL; /* remove suffix */
	collate = ASCII_COLLATING;

    } else if ((tmp = strstr(index_name,":ascii")) != NULL) {
	
	*tmp = NULL; /* remove suffix */
	collate = ASCII_COLLATING;

    } else {
	/* neither specified, use ASCII */
	collate = ASCII_COLLATING;
    }


    if (searchtype == 'H') {
	/* header request */

	if (debug_level) {
	    fprintf(stderr,"searchtype = header");
	    fflush(stderr);
	}
	prev_numeric = -1;
	send_browse_header_resp(refnum, index_name);


    } else if (searchtype == 'C') {
	/* character based request */

	if (debug_level) {
	    fprintf(stderr,"searchtype = character based\n");
	    fflush(stderr);
	}
	prefix_st = makelower(ptr);
	if (prev_numeric_text != NULL)
	    free(prev_numeric_text);
	prev_numeric_text = prefix_st;
	prev_numeric = string2number(collate,prefix_st);

	/* send z39.50 scan request */
	code = smart_scan(*p_session, index_name, prefix_st, num_terms, line_pos, stepsize, &sl, collate);

	if (code == NOTOK) {
	    misc_warning_hcb("z39 browse returned NOTOK status.");
	} else if (code == OK) {
	    return_browse(refnum,prev_numeric,sl);
	} else if (code == TRUE) {
	    misc_warning_hcb("z39.50 diagnostics returned");
	} else { /* code == FAILED_BROWSE */
	    misc_warning_hcb("Browse translation failed");
	}
	strlst_dispose(&sl);

    } else { /* searchtype == 'N' */
	/* numeric request */

	if (debug_level) {
	    fprintf(stderr,"searchtype = numeric based\n");
	    fflush(stderr);
	}
	numeric = atoi(ptr) % 60466176; /* ignore bits higher than 36^5 */

	if (prev_numeric != -1 && abs(prev_numeric - numeric) < BROWSE_SCREEN_SIZE) {

	    /* scan using character based offset */

	    prefix_st = prev_numeric_text;

	    /* send z39.50 scan request */

	    if (debug_level) {
		fprintf(stderr,"before smartscan in numeric, line_pos = %d, num = %d, prev = %d\n",line_pos,numeric,prev_numeric);
		fflush(stderr);
	    }

	    code = smart_scan(*p_session, index_name, prefix_st, num_terms, line_pos - numeric + prev_numeric, stepsize, &sl, collate);

	    if (code == NOTOK) {
		misc_warning_hcb("z39 browse returned NOTOK status.");
	    } else if (code == FAILED_BROWSE) {
		misc_warning_hcb("Browse translation failed");
	    } else if (code == TRUE) {
		misc_warning_hcb("z39.50 diagnostics returned");
	    } else { /* code == OK */
		if (strlst_get_numitems(sl) < numeric + 1 && strlst_get_numitems(sl) < line_pos) {
		    misc_warning_hcb("Browse translation failed");
		} else {

		    if (numeric < line_pos)
		    /* item is near head of list, so line_pos won't match */
			prev_numeric_text = strdup(strlst_get_item(sl,numeric + 1));
		    else
			prev_numeric_text = strdup(strlst_get_item(sl,line_pos));

		    prev_numeric = numeric;
		    return_browse(refnum,numeric,sl);
		}
	    }
	    strlst_dispose(&sl);

	} else { /* scan using mapping to 5 character word */

	    prev_numeric = numeric;

	    if (numeric == 0) { /* beginning of list */
		if (collate == EBCDIC_COLLATING) /* use "a" */
		    prefix_st = strdup("a");
		else /* ASCII, use "0" */
		    prefix_st = strdup("0");
		line_pos = 1;
	    } else /* normal, map to 5 characters */
		prefix_st = number2string(collate,numeric);

	    /* send z39.50 scan request */

	    if (debug_level) {
		fprintf(stderr,"before smart scan in mapping\n");
		fflush(stderr);
	    }
	    code = smart_scan(*p_session, index_name, prefix_st, num_terms, line_pos, stepsize, &sl, collate);

	    if (code == NOTOK) {
		misc_warning_hcb("z39 browse returned NOTOK status.");
	    } else if (code == FAILED_BROWSE) {
		misc_warning_hcb("Browse translation failed");
	    } else if (code == TRUE) {
		misc_warning_hcb("z39.50 diagnostics returned");
	    } else { /* code == OK */
		if (strlst_get_numitems(sl) < numeric + 1 && strlst_get_numitems(sl) < line_pos) {
		    misc_warning_hcb("Browse translation failed");
		} else {

		    if (strlst_get_numitems(sl) < numeric + 1 && strlst_get_numitems(sl) < line_pos) {
			misc_warning_hcb("Browse translation failed");
		    } else {

			if (numeric < line_pos)
		        /* item is near head of list, so line_pos won't match */
			    prev_numeric_text = strdup(strlst_get_item(sl,numeric + 1));
			else
			    prev_numeric_text = strdup(strlst_get_item(sl,line_pos));

			return_browse(refnum,numeric,sl);
		    }
		}
	    }
	    strlst_dispose(&sl);
	}
    }
    if (debug_level) {
	fprintf(stderr,"leaving browse_wcb\n");
	fflush(stderr);
    }
}


void send_browse_header_resp(int refnum, char *index_name)
/* sends a response to willow describing the scaninfo-willow mapping */
{

    packet_rec packet;
    char *resp;
    char *rest_of_index_name;

    if (debug_level) {
	fprintf(stderr,"Sending header response");
	fflush(stderr);
    }

    if (strlen(index_name) < 2) {
	rest_of_index_name = malloc(1);
	*rest_of_index_name = NULL;
    } else
	rest_of_index_name = index_name + 1;

    resp = malloc(HEADER_RESP_SIZE + strlen(index_name));
    sprintf(resp,"%d%c%d%c%d%c%s%c%d%c%d%c%c%c%c%c%c%c%c%s%c%d%c%d%c%c%c%c%c%c",
	    NUM_RECORDS,WILLOW_SEPARATOR,
	    NUM_FIELDS,WILLOW_SEPARATOR,
	    NUM_LINES,WILLOW_SEPARATOR,

	    "Entries",WILLOW_SEPARATOR,
	    OCCUR_FIELD_WIDTH,WILLOW_SEPARATOR,
	    OCCUR_POSITION,WILLOW_SEPARATOR,
	    OCCUR_BROWSABLE,
	    OCCUR_SEARCHABLE,
	    OCCUR_VISIBLE_VALUE,
	    OCCUR_VISIBLE_LABEL,
	    OCCUR_RIGHT_JUSTIFY,

	    WILLOW_SEPARATOR,

	    toupper(*index_name),rest_of_index_name,WILLOW_SEPARATOR,
	    0,WILLOW_SEPARATOR,
	    TERM_POSITION,WILLOW_SEPARATOR,
	    TERM_BROWSABLE,
	    TERM_SEARCHABLE,
	    TERM_VISIBLE_VALUE,
	    TERM_VISIBLE_LABEL,
	    TERM_RIGHT_JUSTIFY);

    packet.type = BROWSE;
    packet.var.browse.line_pos = 0;
    packet.var.browse.ref_num = refnum;
    packet.data_len = strlen(resp) + 1;

    willow_send(netify(&packet),PACKET_NET_SIZE);
    willow_send(resp,NULL);

    free(resp);

    if (debug_level) {
	fprintf(stderr,"Finished sending header response");
	fflush(stderr);
    }

}


int smart_scan(SESSION *session, char *indexname, char *termname, int num_terms, int position, int stepsize, STRLST **p_sl, int collate )
/*makes repeated calls to z39_scan until the actual requested information can be returned.  There are 4 possibilities for unwanted outcomes; the loop will modify the z39_scan parameters for a better attempt, and try again, up to a certain number of tries(MAX_NUM_TRIES). returns OK if the scan was eventually successful, NOTOK if the z39.50 scan had errors, TRUE if the z39.50 scan had diagnostics, FAILED_BROWSE if the requested scan could not be completed.*/

{
    int code;
    int num_tries = 0;
    int gotten_position;
    int gotten_num_terms;
    int difference;
    int i;
    int offset;
    SERIES *new_series;
    char *firststring = NULL;
    char *laststring = NULL;
    SCAN_INFO *tmpscan;
    STRLST *sl, *sl2;

    if (debug_level) {
	fprintf(stderr,"\nentering smart_scan\n");
	fprintf(stderr,"termname = %s, num = %d, pos = %d\n",termname,num_terms,position);
	fflush(stderr);
    }

    /* check if position is out of normal range */

    if (position <= 0) {
	/* desired range is far after termname, shift forward */

	if (debug_level) {
	    fprintf(stderr,"after smart scan in out of range(low), recursing into smart_scan\n");
	    fflush(stderr);
	}
	code = smart_scan(session,indexname,termname,-position+2,1,stepsize,p_sl,collate);
	if (code != OK)
	    return code;

	if (strlst_get_numitems(*p_sl) == 0) {
	    msg_make(DEFAULT_MSG,"The driver cannot convert the server's scan responses(no items returned)");
	    return FAILED_BROWSE;
	} else if (-position+2 > strlst_get_numitems(*p_sl)) {
	    /* beyond end of list, use last item */
	    termname = strdup(strlst_get_item(*p_sl,strlst_get_numitems(*p_sl)));
	} else
	    termname = strdup(strlst_get_item(*p_sl,-position+2));

	position = 1;
	strlst_dispose(p_sl);

    } else if (position > num_terms) {
	/* desired range is far before termname, shift backward */

	 if (debug_level) {
	     fprintf(stderr,"after smart scan in out of range(high), recursing into smart_scan\n");
	     fflush(stderr);
	 }
	 code = smart_scan(session,indexname,termname,position,position,stepsize,p_sl,collate);
	 if (code != OK)
	     return code;

	 position = num_terms;
	 if (strlst_get_numitems(*p_sl) == 0) {
	     msg_make(DEFAULT_MSG,"The driver cannot convert the server's scan responses(no items returned)");
	     return FAILED_BROWSE;
	 } else if (position > strlst_get_numitems(*p_sl)) {
	     /* beyond end of list, use last item */
	     position = 1;
	     termname = strdup(strlst_get_item(*p_sl,strlst_get_numitems(*p_sl)));
	 } else
	     termname = strdup(strlst_get_item(*p_sl,position));
	 strlst_dispose(p_sl);
    }


    if (debug_level) {
	fprintf(stderr,"After out of range check:\n");
	fprintf(stderr,"termname = %s, num = %d, pos = %d\n",termname,num_terms,position);
	fflush(stderr);
    }

    tmpscan = scaninfo_new();
    *p_sl = (STRLST *) NULL;

    /* call z39.50 scan for an initial try */
    code = z39_scan(session,indexname,termname,num_terms,position,stepsize,tmpscan);

    if (code != OK)
	return code;

    if (debug_level) {
	fprintf(stderr,"After initial z39_scan, scaninfo is:\n");
	scaninfo_output(session,tmpscan);
	fprintf(stderr,"\n");
	fflush(stderr);
    }

    if (scaninfo_get_stepsize(tmpscan) != stepsize && (scaninfo_get_stepsize(tmpscan) != 1 || stepsize != 0)) {
	msg_make(DEFAULT_MSG,"The server cannot fulfill the stepsize request.");
	return FAILED_BROWSE;
    }

    sl = series2strlst(tmpscan->records);
    *p_sl = sl;

    scaninfo_dispose(&tmpscan);

    do { /* repeatedly try to refine results to match request */

	if (debug_level) {
	    fprintf(stderr,"\nbeginning of smartscan while\n");
	    strlst_display(sl);
	    fflush(stderr);
	}

	num_tries++;

	gotten_position = strlst_get_position(sl,termname,collate);
	if (gotten_position == 0) {
	    msg_make(DEFAULT_MSG,"The driver cannot convert the server's scan responses(no items returned)");		    
	    return FAILED_BROWSE;
	}
	if (gotten_position < 0) {
	    /* didn't match exactly, use the closest term */
	    gotten_position = -gotten_position;
	}

	gotten_num_terms = strlst_get_numitems(sl);


	if ((gotten_position == position) && (gotten_num_terms == num_terms))
	    break; /* matched request, break out of loop */


	else if ((num_terms - position) < (gotten_num_terms - gotten_position)) {

	    /* dump end */

	    if (debug_level) {
		fprintf(stderr,"\n***dump end***\n");
		fflush(stderr);
	    }
	    difference = (gotten_num_terms - gotten_position) - (num_terms - position);

	    strlst_prune_list(sl,strlst_get_numitems(sl) - difference + 1);


	} else if (position < gotten_position) {

	    /* dump start */

	    if (debug_level) {
		fprintf(stderr,"\n***dump start***\n");
		fflush(stderr);
	    }
	    difference = gotten_position - position;

	    if (difference >= strlst_get_numitems(sl))
		/* can't dump more than we have, so dump all */
		strlst_prune_list(sl,1);
	    else
		for (i=1;i<=difference;i++)
		    strlst_remove(sl,1);


	} else if (position > gotten_position) {

	    /* scan beginning */

	    if (strlst_get_numitems(sl) == 0) {
		msg_make(DEFAULT_MSG,"The driver cannot convert the server's scan responses(no items returned)");
		return FAILED_BROWSE;
	    }
	    firststring = strdup(strlst_get_item(sl,1));
	    if (debug_level) {
		fprintf(stderr,"\n***scan beginning***, new term = %s\n",firststring);
		fflush(stderr);
	    }

	    tmpscan = scaninfo_new();
	    code = z39_scan(session,indexname,firststring,num_terms,num_terms,stepsize,tmpscan);
	    if (code != OK)
		return code;

	    sl2 = series2strlst(tmpscan->records);
	    scaninfo_dispose(&tmpscan);

	    offset = strlst_get_position(sl2,firststring,collate);
	    if (offset == 1) {
		
		/* first try to scan a earlier term in case of server weirdness */

		if (debug_level) {
		    fprintf(stderr,"term repeatedly matches in first position\n");
		    fflush(stderr);
		}

                /* pick a slightly earlier term */
		if (collate == EBCDIC_COLLATING) {
		    if (firststring[strlen(firststring)-1] == 'a')
			firststring[strlen(firststring)-1] = '.';
		    else if (firststring[strlen(firststring)-1] == '0')
			firststring[strlen(firststring)-1] = 'z';
		    else
			firststring[strlen(firststring)-1]--;
		} else { /*ASCII*/
		    if (firststring[strlen(firststring)-1] == 'a')
			firststring[strlen(firststring)-1] = '9';
		    else if (firststring[strlen(firststring)-1] == '0')
			firststring[strlen(firststring)-1] = '!';
		    else
			firststring[strlen(firststring)-1]--;
		}

		    
		if (debug_level) {
		    fprintf(stderr,"Trying a new scan with term = %s\n",firststring);
		    fflush(stderr);
		}

		/* try scanning with this new term */
		strlst_dispose(&sl2);
		tmpscan = scaninfo_new();
		code = z39_scan(session,indexname,firststring,num_terms,1,stepsize,tmpscan);
		if (code != OK)
		    return code;

		sl2 = series2strlst(tmpscan->records);
		if (debug_level)
		    strlst_display(sl2);
		scaninfo_dispose(&tmpscan);

		if (strcmp(strlst_get_item(sl,1),strlst_get_item(sl2,1)) == 0) {
          	    /* try chopping letters */
		    while ((strlen(firststring) != 1) && (strcmp(strlst_get_item(sl,1),strlst_get_item(sl2,1)) == 0)) {
			firststring[strlen(firststring)-1] = NULL;

			if (debug_level) {
			    fprintf(stderr,"Trying scan with term = %s\n",firststring);
			    fflush(stderr);
			}

		       /* try scanning with this new term */
			strlst_dispose(&sl2);
			tmpscan = scaninfo_new();
			code = z39_scan(session,indexname,firststring,num_terms,1,stepsize,tmpscan);
			if (code != OK)
			    return code;

			sl2 = series2strlst(tmpscan->records);
			if (debug_level)
			    strlst_display(sl2);
			scaninfo_dispose(&tmpscan);
		    }

		    if (strlen(firststring) > 1) {
			/* got something, combine lists */
			strlst_combine(sl2,&sl);
			*p_sl = sl2;
			sl = sl2;
			continue;

		    } else {
			/* seems that term is near head of list; just return what we can */

			if (debug_level) {
			    fprintf(stderr,"search term is before head of list\n");
			    fflush(stderr);
			}
			strlst_dispose(&sl2);
			if (strlst_get_numitems(sl) < num_terms) {
			    /* need to scan end; force position to be correct */
			     position = gotten_position;
			     continue;
			} else if (strlst_get_numitems(sl) > num_terms) {
		             /* just return the right number of items */
			      strlst_prune_list(sl,num_terms + 1);
			}
			*p_sl = sl;
			return OK;

		    }
		}
	    }

	    if (offset <= 0) {
		if (offset == 0) { /* z39_scan didn't return any terms */
		    msg_make(DEFAULT_MSG,"The driver cannot convert the server's scan responses(no items returned)");
		    return FAILED_BROWSE;
		}
		/* didn't match exactly, use closest term */
		offset = -offset;
	    }

	    /* paste lists together */
	    strlst_prune_list(sl2,offset);
	    strlst_combine(sl2,&sl);
	    sl = sl2;
	    *p_sl = sl;


	} else { /* ((num_terms - position) > (gotten_num_terms - gotten_position)) */ 

	    /* scan end */

	    if (strlst_get_numitems(sl) == 0) {
		msg_make(DEFAULT_MSG,"The driver cannot convert the server's scan responses(no items returned)");
		return FAILED_BROWSE;
	    }
	    laststring = strdup(strlst_get_item(sl,strlst_get_numitems(sl)));
	    if (debug_level) {
		fprintf(stderr,"\n***scan end***, laststring = %s\n",laststring);
		fflush(stderr);
	    }

	    tmpscan = scaninfo_new();
	    code = z39_scan(session,indexname,laststring,num_terms,1,stepsize,tmpscan);
	    if (code != OK)
		return code;

	    sl2 = series2strlst(tmpscan->records);
	    scaninfo_dispose(&tmpscan);

	    offset = strlst_get_position(sl2,laststring,collate);

	    if (offset == strlst_get_numitems(sl2)) {

		/* first check if we can get any more terms */
		strlst_dispose(&sl2);
		tmpscan = scaninfo_new();
		code = z39_scan(session,indexname,laststring,2,1,stepsize,tmpscan);
		if (code != OK)
		    return code;

		sl2 = series2strlst(tmpscan->records);
		scaninfo_dispose(&tmpscan);

		if (strlst_get_position(sl2,laststring,collate) == strlst_get_numitems(sl2)) {

		    /* first try to scan ahead, in case of server weirdness */
		    if (debug_level) {
			fprintf(stderr,"term matches in last position\n");
			fflush(stderr);
		    }

		    /* pick a slightly forward term */
		    firststring = malloc(strlen(laststring) + 2);
		    strcpy(firststring,laststring);
		    if (collate == EBCDIC_COLLATING) 
			firststring[strlen(laststring)] = '.';
		    else /*ASCII*/
			firststring[strlen(laststring)] = '!';
		    firststring[strlen(laststring) + 1] = NULL;

		    if (debug_level) {
			fprintf(stderr,"Trying a new scan with term = %s\n",firststring);
			fflush(stderr);
		    }
		    
		    /* try scanning with this new term */
		    strlst_dispose(&sl2);
		    tmpscan = scaninfo_new();
		    code = z39_scan(session,indexname,firststring,num_terms,num_terms,stepsize,tmpscan);
		    if (code != OK)
			return code;

		    sl2 = series2strlst(tmpscan->records);
		    if (debug_level)
			strlst_display(sl2);
		    scaninfo_dispose(&tmpscan);

		    if (strcmp(strlst_get_item(sl,strlst_get_numitems(sl)),strlst_get_item(sl2,strlst_get_numitems(sl2))) == 0) {

		        /* seems that term is near end of list; just return as
		           many items as possible, with term in the correct
		           position. */

			if (debug_level) {
			    fprintf(stderr,"reached end of list\n");
			    fflush(stderr);
			}

			strlst_dispose(&sl2);
			*p_sl = sl;
			return OK;

		    } else { /* we got something different, just combine */
			strlst_combine(sl,&sl2);
			*p_sl = sl;
			return OK;
		    }


		} else { /* found more terms on retry */
		    if (strlst_get_numitems(sl) != num_terms) {
			/* combine lists and rescan */
			if (strcmp(strlst_get_item(sl,1),strlst_get_item(sl2,1)) == 0) { /* sl2 is just a bigger list */
			    strlst_dispose(&sl);
			    *p_sl = sl2;
			    sl = sl2;
			    continue;
			} else { /* combine */
			    offset = strlst_get_position(sl,strlst_get_item(sl2,1),collate);
			    strlst_prune_list(sl,offset);
			    strlst_combine(sl,&sl2);
			    *p_sl = sl;
			    continue;
			}
		    } else {
			do { /* try scanning less items until we get the end of list */
			 strlst_dispose(&sl2);
			 num_terms--;
			 tmpscan = scaninfo_new();
			 code = z39_scan(session,indexname,laststring,num_terms,1,stepsize,tmpscan);
			 if (code != OK)
			     return code;

			 sl2 = series2strlst(tmpscan->records);
			 scaninfo_dispose(&tmpscan);

			 offset = strlst_get_position(sl2,laststring,collate);

			} while (num_terms > 1 && offset == strlst_get_numitems(sl2));

			if (num_terms <= 1) { /* we've hit the end(or unfriendly server) */
			    strlst_dispose(&sl2);
			    *p_sl = sl;
			    return OK;

			} else { /* combine and return */
			    strlst_prune_list(sl,strlst_get_numitems(sl) - offset + 1);
			    strlst_combine(sl,&sl2);
			    *p_sl = sl;
			    return OK;
			}
		    }
		}
	    }
	    
	    if (offset <= 0) {
		if (debug_level) {
		    fprintf(stderr,"laststring not matched, but matched before!\n");
		    fflush(stderr);
		}
		msg_make(DEFAULT_MSG,"The driver cannot convert the server's scan responses(inconsistent matching)");
		return FAILED_BROWSE;

	    } else { /* paste lists together */
		strlst_prune_list(sl,strlst_get_numitems(sl) - offset + 1);
		strlst_combine(sl,&sl2);
		*p_sl = sl;
	    }
	}

    } while (num_tries <= MAX_NUM_TRIES);


    if (debug_level) {
	fprintf(stderr,"AFTER WHILE, numtries = %d\n",num_tries);
	fflush(stderr);
    }

    if (num_tries >= MAX_NUM_TRIES) {
	if (debug_level) {
	    fprintf(stderr,"exceeded max number of tries\n");
	    fflush(stderr);
	}
	msg_make(DEFAULT_MSG,"The driver cannot convert the server's scan responses(too many tries)");
	return FAILED_BROWSE;

    } else { /* we're ok */
	return OK;
    }
}


void return_browse(int refnum, unsigned int position, STRLST *sl)
/* parse information from z39.50, return in willow packet. */

{
#define MAX_OCCUR_LEN 15

    packet_rec packet;
    int i,listsize;
    char *list_data;
    char occur_str[MAX_OCCUR_LEN];

    if (debug_level) {
	fprintf(stderr,"\nentering return_browse, called with numeric= %d\n",position);
	strlst_display(sl);
	fflush(stderr);
    }

    /* convert from string list to willow list data */

    listsize = 0;
    for (i=1;i<=strlst_get_numitems(sl);i++)
	listsize += MAX_OCCUR_LEN + strlen(strlst_get_item(sl,i)) + 1;
	
    list_data = malloc(listsize + 1);
    *list_data = NULL;
    for (i=1;i<=strlst_get_numitems(sl);i++) {
	sprintf(occur_str,"%d%c",strlst_get_occurrences(sl,i),WILLOW_SEPARATOR);
	strcat(list_data,occur_str);
	strcat(list_data,strlst_get_item(sl,i));
	strcat(list_data,"\n");
    }

    packet.type = BROWSE;
    packet.var.browse.ref_num = refnum;
    packet.var.browse.line_pos = position;
    packet.data = NULL;
    if ((int) list_data != -1)
	packet.data_len = strlen(list_data) + 1;
    else {
	packet.data_len = 0;
	packet.var.browse.line_pos = -1;
    }
    willow_send(netify(&packet), PACKET_NET_SIZE);

    if ((int) list_data != -1)
	willow_send(list_data, NULL);
    free(list_data);

    if (debug_level) {
	fprintf(stderr,"leaving return_browse\n");
	fflush(stderr);
    }
}


char browse_i2c(int collate, int index)
/* ASCII: maps 0-9 to '0'-'9', 10-35 to 'a'-'z'
   EBCDIC: maps 0-25 to 'a'-'z', 26-35 to '0'-'9' */
{
    if (collate == EBCDIC_COLLATING) {
	if (index >= 0 && index <= 25)
	    return (char) (index + 'a');
	else if (index <= 35)
	    return (char) (index - 26 + '0');
	else return ' ';

    } else { /* collate == ASCII_COLLATING */
	if (index >= 0 && index <= 9)
	    return (char) (index + '0');
	else if (index <= 35)
	    return (char) (index + 'a' - 10);
	else return ' ';
    }
}


int browse_c2i(int collate, char c)
/* ASCII: maps '0'-'9' to 0-9, 'a'-'z' to 10-35
   EBCDIC: maps 'a'-'z' to 0-25, '0'-'9' to 26-35 */
{
    if (collate == EBCDIC_COLLATING) {
	if (c >= 'a' && c <= 'z')
	    return(c - 'a');
	else if (c >= '0' && c <= '9')
	    return(26 + c - '0');
	else return 0;

    } else { /* collate == ASCII_COLLATING */
	if (c >= '0' && c <= '9')
	    return(c - '0');
	else if (c >= 'a' && c <= 'z')
	    return(10 + c - 'a');
	else return 0;
    }
}


char *number2string(int collate, int num)
/* returns a 5-letter string corresponding to the number num using the following scheme:  the base 36 number abcde (higher order bits are ignored) is converted to string fghij, where f=browse_i2c(a), etc. */
{
    int i;
    int bit = 1679616; /* 36 ^ 4 */
    char *outstr;

    outstr = malloc(6);
    for (i=0;i<5;i++) {
	outstr[i] = browse_i2c(collate, num / bit);
	num %= bit;
    }
    outstr[5]=NULL;
    return outstr;
}


int string2number(int collate, char *str)
/* returns the number corresponding to the string str using the following scheme:  the string abcde(further characters are ignored) is converted to the base 36 number fghij, where f=browse_c2i(a), etc. */
{
    int i;
    int value = 0;
    int bit = 1679616; /* 36 ^ 4 */

    value = browse_c2i(collate, str[0]);

    for (i=0;i<=5;i++) {
	if (str[i] == NULL)
	    break;
	else
	    value += browse_c2i(collate,str[i]) * bit;
	bit /= 36;
    }

    return value;

}

char *makelower(char *s)
/* copies a string to a new string, making all uppercase chars lowercase */
{
    char *new_str;
    int i;

    new_str = malloc(strlen(s) + 1);
    for(i=0;i<strlen(s);i++)
	new_str[i] = tolower(s[i]);
    new_str[strlen(s)] = NULL;
    return new_str;
}
