/*                            HOST.C                             */

/*
 *  Copyright (c) 1992 Stanford Leland Junior University
 *
 *  Commercial use prohibited
 *
 *  Author: Harold Finkbeiner (Harold@Forsythe.Stanford.Edu)
 */

#include <memory.h>
#include <stddef.h>

#include "constant.h"
#include "host.h"
#include "utility.h"

/*--------------------------------------------------------------*
 |                    SEARCH_TABLE                              |
 | Function:                                                    |
 |    Search the table for a tag.                               |
 | Parameters:                                                  |
 |    string      : The string to search.                       |
 |    entry       : Array of tags to search for a match.        |
 |    table_count : Number of entries in the array.             |
 | Returns:                                                     |
 |    0 if match, otherwise non-zero.                           |
 | Description:                                                 |
 |                                                              |
 | Note: Just sees if the start of the string matches a tag.    |
 |                                                              |
 *--------------------------------------------------------------*/

FORMAT_RECORD *tag_from_table(char *string,
                              FORMAT_RECORD *entry,
                              int table_count)
{
    int i;

    for (i = table_count; i > 0; i--, entry++)
        {
            if (memcmp(string, entry->lname, entry->width) == 0)
                return entry;
        }
    return (FORMAT_RECORD *) NULL;
}

/*--------------------------------------------------------------*
 |                    TAGS_INTO_TABLE                           |
 | Function:                                                    |
 |    Save the tags into a table.                               |
 | Parameters:                                                  |
 |    tags        : The string to which contains the tags.      |
 |    table       : Array where to store entries.               |
 |    table_count : Pointer to store number of entries.         |
 | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |                                                              |
 | Note:  The tags are seperated by commas.  A tag may be       |
 |        followed by options which are seperated from the tag  |
 |        with a semicolon.  For example                        |
 |             TITLE,AUTHOR,PUBLISHER                           |
 |             LANGUAGE:E,LOC SUBJECT,Summary_Field_Count:6     |
 |        If it contains a comma or semicolon                   |
 |             " Subject:":E," Total,"6                         |
 |        There are basically two options, a number and a       |
 |        character.                                            |
 *--------------------------------------------------------------*/

void tags_into_table(char *tags,
                     FORMAT_RECORD *table,
                     int *table_count)
{
    char *field;
    char *options;
    int fld_length;
    int opt_length;

    do
        {
            fld_length = tokenize(&tags, &field, ",:");

            if (fld_length > 0)
                {
                    FORMAT_RECORD *entry = &table[(*table_count)++];
                    int flags = 0;

                    memset(entry, 0, sizeof(FORMAT_RECORD));

                    UNQUOTE(field, fld_length);
                    memcpy(entry->lname, field, fld_length);
                    entry->lname[fld_length] = EOS;
                    entry->width = fld_length;

                    if (*tags == ':')
                        {
                            tags++;       /* Skip the colon */

                            opt_length = tokenize(&tags, &options, ",");

                            if (opt_length > 0)     /* Options found */
                                {
                                    char *end;

                                    if (*tags) tags++;

                                    options[opt_length] = EOS;

                                    entry->count = strtol(options, &end, 10);

                                    while (*end == ' ' || *end == ',') end++;

                                    entry->flag = *end;
                                }
                        }
                    else
                        if (*tags) tags++;
                }
        }
    while (fld_length > 0);
}
