/*
 *  Copyright (C) 1994 by the Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/* Author: Tom Owens MIT Library Systems Office */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <memory.h>
#include "marc.h"

/*
_______________________________________________________________________________
mrInit (struct MARC_RECORD *marc)
 this routine initializes the elements of a marc structure. data arrays are
 malloc'd for the smallest size possible to be freed and increased later
 
   returns:
	MEMORY_ERROR if any malloc fails, otherwise GOOD

   revision history:
   who		what	    when	why
   tjo		creation    10/92	needed marc routine library
________________________________________________________________________________
*/
int     mrInit (marc)
MARC_RECORD *marc;
{

    /* malloc the variable arrays to one unit */

    marc -> tagMap = (TAG_MAP_ENTRY *) malloc (sizeof (TAG_MAP_ENTRY));
    if (marc -> tagMap == NULL)
	return MEMORY_ERROR;

    marc -> data = (char *) malloc (sizeof (char));
    if (marc -> data == NULL)
	return MEMORY_ERROR;

    /* if we got here, things are fine */
    return GOOD;
}

void mrFree (marc)
MARC_RECORD *marc;
/* free memory created by the marc routines */
{
    free (marc -> tagMap);
    free (marc -> data);
}
/*_______________________________________________________________________________
 mrMapMarc (rawRec,marc)
 this loads a true marc record into a MARC_RECORD structure which
 has previously been initialized by mrInit

the directory is mapped into an array of structures called tagMapEntries.  Each entry contains
  .tag = the tag number described (i.e. 003)
  .tagLength   the length of the tag
  .tagOffset   the offset where the tag begins within the data structure
 
  this means that [x].tag is a string of [x].tagLength characters beginning at data offset [x].tagOffset
  
tagOccurrences is an array of the number of occurrences of each tag number.  the tag number is the offset of the array which contains the data, e.g. tagOccurrences[100] would tell how many occurrences of the 100 tag are in the
structure.tagOccurrences[0] is the total number of tags in the structure

  returns:
	    GOOD if the structure could be mapped
	    MEMORY_ERROR if mallocs fail
	    TOO_MANY_TAGS if number of tags exceeded MAXTAGS

________________________________________________________________________________
*/
int     mrMapMarc (rawRec, marc)
char   *rawRec;
MARC_RECORD *marc;
{
    int     baseAddress,position,tagCounter,directoryEnd;

    int     GetInt ();		/* returns the int expressed by the ascii   chars */

    /* get the position of the tag data (base address of data) */
    baseAddress = GetInt (&rawRec[BASEADR_OFFSET], BASEADR_LENGTH);

    if ( (baseAddress < MIN_MARC_SIZE) || (baseAddress > MAX_MARC_SIZE) )
	return BAD_MARC_SIZE;

    /* length of the incoming record */
    marc -> length = GetInt (&rawRec[RECLEN_OFFSET], RECORD_LENGTH_LENGTH);
    if ( (marc -> length > MAX_MARC_SIZE) || (marc -> length < baseAddress) )
	return BAD_MARC_SIZE;

    /* length of the data portion of the record */
    marc -> length = marc -> length - baseAddress;

    /* move the data into the data array, first free the memory and re-malloc */
    free (marc -> data);
    marc -> data = (char *) malloc (marc -> length * sizeof (char));
    if (marc -> data == NULL)
	return MEMORY_ERROR;

    strncpy (marc -> data, &rawRec[baseAddress], marc -> length);

    /* map the tags into the local arrays */
    free(marc -> tagMap);
    marc -> tagMap = (TAG_MAP_ENTRY *) malloc(sizeof(TAG_MAP_ENTRY)); 
    if (marc -> tagMap == NULL)
	return MEMORY_ERROR;
    /* clear the array of the number of tags to zeroes */
    memset(marc -> tagOccurrences, 0,sizeof (marc -> tagOccurrences));

    /*start the mapping */
    directoryEnd = baseAddress - DIRECTORY_LENGTH;
    for (position = DIRECTORY_OFFSET, tagCounter = 0; position < directoryEnd; ) {

	if ( (marc -> tagMap[tagCounter].tag = GetInt(&rawRec[position],DIRECTORY_TAG_LENGTH) ) <= 0 )
	    return BAD_TAG_DIRECTORY;
	position+=DIRECTORY_TAG_LENGTH;
	marc -> tagOccurrences[marc ->tagMap[tagCounter].tag]++;

	if ((marc -> tagMap[tagCounter].tagLength = GetInt(&rawRec[position],DIRECTORY_LENGTH_LENGTH) ) <= 0 )
	    return BAD_TAG_DIRECTORY;
	position+=DIRECTORY_LENGTH_LENGTH;

	if ( (marc -> tagMap[tagCounter].tagOffset = GetInt(&rawRec[position],DIRECTORY_OFFSET_LENGTH) ) < 0 )
	    return BAD_TAG_DIRECTORY;
	position+=DIRECTORY_OFFSET_LENGTH;

	if (marc -> tagMap[tagCounter].tagOffset > marc -> length)
	    return BAD_TAG_DIRECTORY;

	tagCounter++;
	marc -> tagMap = (TAG_MAP_ENTRY *) realloc(marc -> tagMap,(tagCounter+1) * sizeof(TAG_MAP_ENTRY) );
	if (marc -> tagMap ==NULL)
	    return MEMORY_ERROR; 
    }

    marc -> tagOccurrences[0] = tagCounter;

    marc -> status = rawRec[STATUS_OFFSET];
    marc -> type = rawRec[TYPE_OFFSET];
    marc -> level = rawRec[LEVEL_OFFSET];
    marc -> encod = rawRec[ENCOD_OFFSET];
    marc -> dcf = rawRec[DCF_OFFSET];
    marc -> linking = rawRec[LINKING_OFFSET];

/* and the internal structure information gets re-initialized */
    marc -> lastTag = marc -> lastTagOccurrence = marc -> lastTagPosition = 0;
    
/*  debug */
/*

    for(position=0;position<marc->tagOccurrences[0];position++) {
	printf("tag: %d\tlen:%d\toff:%d\n",marc->tagMap[position].tag,marc->tagMap[position].tagLength,marc->tagMap[position].tagOffset);
    }
  
    getchar(); */
/* looks good, let's tell them */
    return GOOD;

}

/* 
______________________________________________________________________________
|GetInt (array of chars,length of array we are interested in)
|
|   returns an int of the numeric value of the numerals in a non-NULL terminated
|   character array; the length of the array can't be less than one or more than
|   ten
|  
|   returns:
|	-1 if the length is out of bounds, else the value of the numerals
|
_______________________________________________________________________________
*/

int     GetInt (string, len)
char   *string;
register int     len;
{

    register int counter,value;

#define NO_INT_FOUND -1

    if ((len > 10) || (len < 1))
	return NO_INT_FOUND;

    for (counter=0,value=0; counter<len; counter++) {
	if ( !( isdigit(string[counter]) ) )
	    return NO_INT_FOUND;
	value=(value*10)+( (int) string[counter] - 48);
    }
    return value;

}

/* clear an array of integers to a common value */
void ClearIntArray (intArray,arrayLength,clearValue) 
register int *intArray, arrayLength, clearValue;
{
    
    while (arrayLength>-1) {
	*(intArray+arrayLength)=clearValue;
	arrayLength--;
    }
    return;
}

/*__________________________________________________________________________________________
 int mrGetField(marc structure,field number to be searched, the tag returned, the offset of the data found, the length of the tag)
  finds a tag by the number of its occurrence in the full MARC record, 

   returns:   
	GOOD found the tag
	TAG_NOT_FOUND tag does not exist in record
              BAD_OCCURRENCE_REQUESTED illegal occurrence number
   revision history:
	owens	  10/92	    creation
*/
int     mrGetField (marc, fieldRequested, tagFound, tagDataOffset, tagDataLength)
MARC_RECORD *marc;
int     fieldRequested,
        *tagDataLength,
        *tagFound;
char    **tagDataOffset;
{

    /* either an illegal tag asked for or the tag does not exist */
    if (fieldRequested < 1)
	return BAD_OCCURRENCE_REQUESTED;
    if (fieldRequested > marc->tagOccurrences[0])
	return TAG_NOT_FOUND;
    
    /*ask for tags at tag 1, but tags begin at offset 0 */
    fieldRequested--;

    *tagDataOffset = &marc->data[marc->tagMap[fieldRequested].tagOffset];
    *tagDataLength = marc->tagMap[fieldRequested].tagLength;
    *tagFound = marc->tagMap[fieldRequested].tag;
    return GOOD;
}

/* given a pointer to tag and the length of the tag, return a specified subfield occurrence 
  returns: BAD_OCCURRENCE_REQUESTED : illegal occurrence number (less than 1)
                  BAD_TAG_DATA: something is wrong with the data in the tag sent down, probably a bad length
                  OCCURRENCE_NOT_FOUND: total number of subfield occurrences is less than the occurrence requested
                  other : a bad return from FindSubFieldDelimiter
*/
int mrGetSubField  (requestedOccurrence,tagData,tagLength,subFieldName,subFieldData,subFieldLength)

int requestedOccurrence, tagLength, *subFieldLength;
char *tagData,*subFieldName,**subFieldData;
{

    int     position,foundOccurrence,returnValue;

    if (requestedOccurrence < 1)
	return BAD_OCCURRENCE_REQUESTED;

    if (tagLength <1) 
	return BAD_TAG_DATA;

    /*start looping through the subfields */
    position=0;
    foundOccurrence=0;
    while ( (returnValue= FindSubFieldDelimiter(tagData,&position,tagLength)) == GOOD) {
	/* subfield name is character after delimiter */
	position++;
	if (position>tagLength)
	    return BAD_TAG_DATA;

	foundOccurrence++;

	if (foundOccurrence == requestedOccurrence){
	    *subFieldName = tagData[position];

	    /*data begins after the subfield name */
	    position++;
	    if (position>tagLength)
		return BAD_TAG_DATA;
	    *subFieldData = &tagData[position];
 
	    *subFieldLength=CalculateSubFieldLength(*subFieldData,tagLength-position);

	    return GOOD;
	}
    }
    
    if (foundOccurrence !=0) 
	return OCCURRENCE_NOT_FOUND;
    
    return returnValue;
}

/* find the subfield delimiter within string of known size
  returns:
        GOOD
        SUBFIELD_NOT_FOUND: there is no delimiter in the string 
*/
int FindSubFieldDelimiter (tagString,offsetToStart,offsetToStop)
register char *tagString;
register int *offsetToStart,offsetToStop;
{


    for( ;*offsetToStart<offsetToStop;(*offsetToStart)++ ) {
	if (tagString[*offsetToStart]==DELIMITER)
	    return GOOD;
	if (tagString[*offsetToStart]==END_OF_FIELD)
	    break;
    }
    return SUBFIELD_NOT_FOUND;
}

/* given a string of known length, determine the length of the current subfield data */
int CalculateSubFieldLength (subFieldString,lengthToExamine)
register char *subFieldString;
register int lengthToExamine;
{
    register int position;
    register char thisChar;
    
    /* loop through the string until we find a character which would indicate the end of the subfield -- or exhaust our search */
    for(position=0;position<lengthToExamine;position++) {
	thisChar=subFieldString[position];

	if ( (thisChar == DELIMITER) || (thisChar == END_OF_FIELD) || (thisChar == END_OF_RECORD) )
	    return position;
    }
    return position;
}
/* find a specific occurrence of a specific tag within a marc structure 
  returns: BAD_OCCURRENCE_REQUESTED (i.e.<1)
                  BAD_TAG_REQUESTED (i.e.<1 or >999)
                 TAG_NOT_FOUND the requested tag does not exist
                 OCCURRENCE_NOT_FOUND the tag exists but the occurrence doesn't
                 UNKNOWN_ERROR  probably a bad marc structure
                  GOOD
*/

int     mrFindTagName (marc, tagRequested, requestedOccurrence, tagDataOffset, tagDataLength)
MARC_RECORD *marc;
int     tagRequested, requestedOccurrence, *tagDataLength;
char	**tagDataOffset;
{
    int     position,
            foundOccurrence;

    if (requestedOccurrence < 1)
	return BAD_OCCURRENCE_REQUESTED;

    if ( (tagRequested < 1) || (tagRequested > MAX_TAG_NUMBER) )
	return BAD_TAG_REQUESTED;

    if (marc -> tagOccurrences[tagRequested] < 1)
	return TAG_NOT_FOUND;

    if (marc -> tagOccurrences[tagRequested] < requestedOccurrence)
	return OCCURRENCE_NOT_FOUND;
    /* we found this guy the last time we were in here */
    if ( (tagRequested == marc -> lastTag)  && (requestedOccurrence >= marc -> lastTagOccurrence) ) {
	position = marc -> lastTagPosition;
	foundOccurrence = marc -> lastTagOccurrence-1;
    }
    else {
	position = 0;
	foundOccurrence = 0;
    }


    /* leaf through the tag array looking for the tag we want */
    for ( ; position < marc -> tagOccurrences[0]; position++) {

	if (marc -> tagMap[position].tag == tagRequested) {
	    foundOccurrence++;
	    if (foundOccurrence == requestedOccurrence) {
		/* we found what we were looking for, move it in */
		*tagDataOffset = &marc -> data[marc -> tagMap[position].tagOffset];
		*tagDataLength = marc -> tagMap[position].tagLength;

		/* some structure housekeeping */
		marc -> lastTag = tagRequested;
		marc -> lastTagOccurrence = requestedOccurrence;
		marc -> lastTagPosition = position;

		/* we got what we want; let's go back */
		return GOOD;

	    }			/* if the right occurrence */

	}			/* if the right tag */

    }				/* looking through the tags */

    return UNKNOWN_ERROR;
}
/* given the offset and length of a tag, find a specific occurrence of a specific subfield
  returns:
        BAD_OCCURRENCE_REQUESTED i.e.<1
        BAD_TAG_DATA : something is wrong with the tag as it is defined, probably the length
        OCCURRENCE_NOT_FOUND: at least one instance of the subfield exists, but not the one you asked for
        SUBFIELD_NOT_FOUND: no occurrence of the subfield within the tag
        other returned by FindSubFieldDelimiter
        GOOD
 */
int     mrFindSubFieldName (subFieldRequested, requestedOccurrence, tagData, tagDataLength, subFieldOffset, subFieldLength)

char    subFieldRequested,
        **subFieldOffset,
        *tagData;
int     requestedOccurrence,
        tagDataLength,
        *subFieldLength;

{
    int     returnValue,
            foundOccurrence,
            position;


    if (requestedOccurrence < 1)
	return BAD_OCCURRENCE_REQUESTED;	/* must be at least one occurrence */
               

/* look from the beginning of the tag */

    position = 0;
    foundOccurrence = 0;


    while ( (returnValue = FindSubFieldDelimiter(tagData,&position,tagDataLength) ) == GOOD) {
	position++;			/* subfield name right after field marker*/
	if (position>tagDataLength)
	    return BAD_TAG_DATA;
	if (tagData[position] == subFieldRequested) {
	    foundOccurrence++;
	    if (foundOccurrence == requestedOccurrence) {
		position++;		/* data begins after name */
		if (position>tagDataLength)
		    return BAD_TAG_DATA;
		*subFieldOffset = &tagData[position];
		*subFieldLength = CalculateSubFieldLength(*subFieldOffset,tagDataLength-position);
		return GOOD;
	    }
	}
    }

    if (foundOccurrence != 0)
	return OCCURRENCE_NOT_FOUND;
    else
	return returnValue;
}

char *mrStatString(int status)
{
    switch (status) {
	case GOOD:
	    return "OK";
	case TAG_NOT_FOUND:
	    return "TAG_NOT_FOUND";
	case OCCURRENCE_NOT_FOUND:
	    return "OCCURRENCE_NOT_FOUND";
	case SUBFIELD_NOT_FOUND:
	    return "SUBFIELD_NOT_FOUND";
	case MEMORY_ERROR:
	    return "MEMORY_ERROR";
	case BAD_MARC_SIZE:
	    return "BAD_MARC_SIZE";
	case TOO_MANY_TAGS:
	    return "TOO_MANY_TAGS";
	case BAD_OCCURRENCE_REQUESTED:
	    return "BAD_OCCURRENCE_REQUESTED";
	case BAD_TAG_DIRECTORY:
	    return "BAD_TAG_DIRECTORY";
	case BAD_TAG_REQUESTED:
	    return "BAD_TAG_REQUESTED";
	case BAD_TAG_DATA:
	    return "BAD_TAG_DATA";
	case UNKNOWN_ERROR:
	default:
	    return "UNKNOWN_ERROR";
    }
}
