
/*                           W_SUTR.C                              */

/*
 *  Copyright (c) 1992 Stanford Leland Junior University
 *
 *  Commercial use prohibited
 *
 *  Author: Harold Finkbeiner (Harold@Forsythe.Stanford.Edu)
 *
 *  This will attempt to display a SUTR record.  If there are
 *  no fields, just use the full or brief display returned by
 *  the server.
 *
 *  Unfortunately not all servers have brief displays so
 *  the routine extract_tags_from_display tries to extract
 *  fields specified in db.conf to create a summary record.
 */

#include <stdio.h>

#include "berutil.h"
#include "constant.h"
#include "host.h"
#include "pdutags.h"
#include "putget.h"
#include "record.h"
#include "strings.h"

/* Tag display */

typedef enum extract_tag_handler
{
    Tag_Records,
    Tag_Titles,
} EXTRACT_TAG_HANDLER;

/* Default number of fields to display for summary record */

#define DEFAULT_FIELD_COUNT 6

/*--------------------------------------------------------------*
 |                     GET_FIELD                                |
 | Function:                                                    |
 |    Get the next field from the text.                         |
 | Parameters:                                                  |
 |    field       : Pointer to a string to hold the field.      |
 |    text        : Address of pointer to text.                 |
 |    text_length : Address of text length.                     |
 | Returns:                                                     |
 |    Start of non-blank text in the field.                     |
 | Description:                                                 |
 |    Fill in the field information and update the variable     |    |
 |    "text" to point to the end of the field.  Also update     |
 |    the variable "text_length" to the new length.             |
 | Note:                                                        |
 |    A field is the text until a newline. A field may span     |
 |    lines if the lines are indented past the first word of    |
 |   the first line.  For example                               |
 |         INFO: This would span multiple lines because         |
 |               some lines which follow are indented past      |
 |               the first word "INFO:" on the first line.      |
 |         STOP: It would stop here.                            |
 |   and the following would not work.                          |
 |         SORRY: This would be two fields, one for each line   |
 |           because the second line is not indented enough.    |
 |                                                              |
 |   You could probably create an environment variable to       |
 |   control what is meant by a field (what indentation is      |
 |   used).                                                     |
 *--------------------------------------------------------------*/

static char *get_field(STRING *field,
                       char **text,
                       int *text_length)
{
    char *pt = *text;
    char *et = *text + *text_length;
    char *first_word;
    int indent = 0;

    /*
     * The indent is everything up to and including the
     * first word in the text. So the text
     *    "   TITLE: Indent"
     * would have an indent of 9 and the text
     *    "AUTHOR: Harold"
     * would have an indent of 7.
     */

    first_word = (*pt == ' ') ? NULL : pt;

    for (; pt < et; pt++)
        {
            if (indent == 0)
                {
                    if (*pt == ' ')
                        {
                            if (first_word)
                                indent = pt - *text;
                        }
                    else if (first_word == NULL)
                        first_word = pt;
                }
            if (*pt == '\n') break;
        }

    /*
     * If the next indent is greater or equal then append
     * the line to the field.
     */

    while (pt < et && indent > 0)
        {
            char *bt = pt + 1;    /* Skip newline */

            while (*bt == ' ' && bt < et) bt++;

            if ((bt - pt - 1) < indent)
                break;

            for (pt = bt; pt < et && *pt != '\n'; pt++);
        }

    if (first_word == NULL)
        first_word = *text;

    field->location = *text;
    field->length = pt - *text;

    *text = (pt < et) ? pt + 1 : pt;
    *text_length = et - pt;

    return first_word;
}


/*--------------------------------------------------------------*
 |                EXTRACT_TAGS_FROM_TEXT                        |
 | Function:                                                    |
 |    Extract the tags from the text.                           |
 | Parameters:                                                  |
 |    text        : Address of pointer to text.                 |
 |    text_length : Address of text length.                     |
 |    table       : Table of tags to extract.                   |
 |    table_count : Number of tags in the table.                |
 |    tag_handler : How to handle the tags.                     |
 | Returns:                                                     |
 |    1 if text points to allocated memory. 0 if not.           |
 | Description:                                                 |
 |    Extract all fields which match a tag.                     |
 | Note:                                                        |
 |    If a field starts with a valid tag then the field is      |
 |    kept unless an options says otherwise.                    |
 *--------------------------------------------------------------*/

static int extract_tags_from_text(char **text,
                                  int *text_length,
                                  FORMAT_RECORD *table,
                                  int table_count,
                                  EXTRACT_TAG_HANDLER tag_handler)
{
    FORMAT_RECORD *entry;
    STRING new;
    STRING field;
    int field_count;
    int allocated = 0;
    int tags_found = 0;

    /*
     * The special tag "SUTR_Summary_Field_Count" will contain the
     * number of fields to output.  It must be the last tag in the
     * table.
     */

    entry = tag_from_table("SUTR_Summary_Field_Count", table, table_count);

    if (entry)
        {
            field_count = entry->count;
            table_count--;
        }
    else
        field_count = DEFAULT_FIELD_COUNT;

    new.length = 0;

    /* If there are no entries in the table then skip search */

    if (table_count <= 0)
        new.location = NULL;
    else
        new.location = (char *) malloc(*text_length);

    /* Try and extract fields which match the tags */

    if (new.location)
        {
            char *current = *text;
            int  length = *text_length;

            do
                {
                    char *start;

                    start = get_field(&field, &current, &length);

                    entry = tag_from_table(start, table, table_count);

                    if (entry)
                        {
                            /*
                             * A special flag of 'E' means that the field
                             * marks the end of any summary display fields.
                             * If no fields have been extracted then use
                             * everything before as the summary display.
                             */

                            if (entry->flag == 'E')
                                {
                                    if (new.length > 0)
                                        break;

                                    free(new.location);
                                    *text_length = field.location - *text;
                                    return 0;
                                }

                            /* For titles strip tag and do not add a NL */

                            if (tag_handler == Tag_Titles)
                                {
                                    char *end = field.location + field.length;

                                    start += entry->width;

                                    /* Skip punctuation */

                                    while (start < end && ispunct(*start)) start++;

                                    /* Skip blanks */

                                    while (start < end && isspace(*start)) start++;

                                    field.location = start;
                                    field.length   = end - start;
                                }
                            else
                                string_cconcatenate(&new, "\n");

                            string_concatenate(&new, &field);

                            /*
                             * Right now assume that a tag only occurs once in a
                             * record.  This is probably a bad assumption but it
                             * cuts down on the processing time.
                             */

                            if (++tags_found >= table_count) break;
                        }
                }
            while (length > 0);

            if (new.length == 0)
                free(new.location);
            else
                allocated = 1;
        }

    /*
     * If no fields were extracted then default to simply use
     * the first field_count fields.
     */

    if (new.length == 0)
        {
            new.location = *text;

            for (; field_count > 0; field_count--)
                 (void) get_field(&field, text, text_length);

            new.length = *text - new.location;
        }

    *text = new.location;
    *text_length = new.length;

    return allocated;
}


/*--------------------------------------------------------------*
 |                  OUTPUT_RECORD                               |
 | Function:                                                    |
 |    Output a record for display                               |
 | Parameters:                                                  |
 |    summary_flag : True if a summary record.                  |
 |    position     : Position in result set.                    |
 |    text         : Pointer to the text of the record.         |
 |    text_length  : The length of text.                        |
 | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |    Output a summary or full record accordingly.              |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/


static void output_record(int summary_flag,
                          int position,
                          char *text,
                          int text_length)
{
    char saved;
    int  allocated = 0;

    if (summary_flag)
        {
            extern FORMAT_RECORD summary_formats[];
            extern int summary_field_cnt;

            /*
             * If there are no summary fields assume a brief
             * record has been retrieved.
             */

            if (summary_field_cnt > 0)
                {
                    allocated = extract_tags_from_text(&text,
                                                       &text_length,
                                                       summary_formats,
                                                       summary_field_cnt,
                                                       Tag_Records);
                }
        }

    saved = text[text_length];

    text[text_length] = EOS;

    send_record(position, text, summary_flag);

    text[text_length] = saved;

    if (allocated) free(text);
}

/*--------------------------------------------------------------*
 |                  OUTPUT_TITLE                                |
 | Function:                                                    |
 |    Output a title of a record.                               |
 | Parameters:                                                  |
 |    position    : The ordinal position in the result.         |
 |    text        : The pointer to the SUTR record.             |
 |    text_length : Length of the record.                       |
 | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |    Try to get a one line display out the record.             |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/

static void output_title(int position,
                         char *text,
                         int text_length)
{
    extern FORMAT_RECORD title_formats[];
    extern int title_field_cnt;
    STRING title;
    char saved;
    int  i;
    int  allocated = 0;

    if (title_field_cnt > 0)
        {
            allocated = extract_tags_from_text(&text,
                                               &text_length,
                                               title_formats,
                                               title_field_cnt,
                                               Tag_Titles);

            title.length = text_length;
            title.location = text;
        }
    else
        (void) get_field(&title, &text, &text_length);

    for (i = 0; i < title.length; i++)
        if (title.location[i] == '\n')
            title.location[i] = ' ';

    string_squeeze(&title,&title);

    saved = title.location[title.length];

    title.location[title.length] = EOS;

    send_title(position, title.location);

    title.location[title.length] = saved;

    if (allocated) free(text);
}

/*--------------------------------------------------------------*
 |                  SUTR_WILLOW_RECORD                          |
 | Function:                                                    |
 |    Get a Simple-Unstructured-Text record for Willow.         |
 | Parameters:                                                  |
 |    record   : The pointer to the SUTR record.                |
 |    position : The ordinal position in the result.            |
 |    summary_flag : To output a summary form of the record.    |
 |    abbrev_flag  : To output abbreviated fields (ignored).    |
 | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |    Get a SUTR record.  Some may be stored in ASN.1 form.     |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/

void sutr_willow_record(RECORD *record,
                        int position,
                        int summary_flag,
                        int abbrev_flag)
{
    DATA_DIR *rec;

    if ((rec = dalloc(10)) == (DATA_DIR *) NULL)
        misc_error_hcb("unable to allocate memory for SUTR BER records.");
    else
        {
            if (   bld_dir((CHAR *) record->location, rec) == 0
                || rec->fldid != GENERALSTR)
                {
                    /*
                     * Some servers seem  to return the SUTR wrapped
                     * in BER while others just return the text.  This
                     * takes care of the case where there is no BER
                     * wrapper.
                     */

                    output_record(summary_flag,
                                  position,
                                  record->location,
                                  record->length);
                }
            else
                {
                    STRING text;

                    /* The text was wrapped in BER */

                    get_string(rec, GENERALSTR, &text, REQUIRED);

                    output_record(summary_flag,
                                  position,
                                  text.location,
                                  text.length);
                }
            dfree(rec);
        }
}

/*--------------------------------------------------------------*
 |                  SUTR_WILLOW_TITLE                           |
 | Function:                                                    |
 |    Get a Simple-Unstructured-Text title for Willow.          |
 | Parameters:                                                  |
 |    record   : The pointer to the SUTR record.                |
 |    position : The ordinal position in the result.            |
 | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |    Get a SUTR title.   Some may be stored in ASN.1 form.     |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/

int sutr_willow_title(RECORD *record,
                      int position)
{
    DATA_DIR *rec;

    if ((rec = dalloc(10)) == (DATA_DIR *) NULL)
        misc_error_hcb("unable to allocate memory for SUTR BER records.");
    else
        {
            if (   bld_dir((CHAR *) record->location, rec) == 0
                || rec->fldid != GENERALSTR)
                {
                    /*
                     * Some servers seem  to return the SUTR wrapped
                     * in BER while others just return the text.  This
                     * takes care of the case where there is no BER
                     * wrapper.
                     */

                    output_title(position, record->location, record->length);
                }
            else
                {
                    STRING text;

                    /* The text was wrapped in BER */

                    get_string(rec, GENERALSTR, &text, REQUIRED);

                    output_title(position, text.location, text.length);
                }
            dfree(rec);
        }
    return 0;
}
