/*
 *  Copyright (C) 1994 by the Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/*
 * Willow MARC formatting interface.
 * Interfaces Finkbeiner Z39.50 API to Owens MARC routines.
 *
 * Bill Cattey, MIT Information Systems,  wdc@mit.edu
 * Tom Owens MIT Libraries
 *
 */

# include <stdio.h>                     /* get standard i/o library defns */

#define OK 0
#define NOTOK -1

#include "marc.h"			/*marclib header file*/
#include "bib.h"			/*brief routines header file */

extern char *malloc(), *realloc();

extern int mrInit();
extern int mrMapMarc();	/* marclib */
extern void mrFree();
extern int GetBrief();		/* brief  */
extern int GetFull();		/* full  */
extern int GetExtract();
void send_record(int num, char *text, int summ_flag);

int wcb_marc(position, record, llen, summ_flag, abbr_flag)
int position;
char *record;
int llen, summ_flag, abbr_flag;
{
    int returnValue;
    char *summary = NULL, *marcOut = NULL;
    MARC_RECORD marcRec;
    char err_buf[BUFSIZ];

    if( llen < MIN_MARC_SIZE || llen > MAX_MARC_SIZE ) {
	sprintf (err_buf, "wcb_marc: Corrupt Marc record of bogus size: %d.\n", llen);
	misc_warning_hcb(err_buf);
	send_record(position, err_buf, summ_flag);
	return NOTOK;
    }

    if ((returnValue= mrInit(&marcRec))!=GOOD) {
	sprintf (err_buf, "wcb_marc: Corrupt Marc record with problem '%s' reported by mrInit.\n",  mrStatString(returnValue));
	misc_warning_hcb(err_buf);
	send_record(position, err_buf, summ_flag);
	return NOTOK;
    }

    if ((returnValue=mrMapMarc(record,&marcRec))!=GOOD) {
	sprintf (err_buf, "wcb_marc: Corrupt Marc record with problem '%s' reported by mrMapMarc.\n",  mrStatString(returnValue));
	misc_warning_hcb(err_buf);
	send_record(position, err_buf, summ_flag);
	return NOTOK;
    }

    if (summ_flag) {  /* summary */
	if ((returnValue=GetBrief(&marcRec, &marcOut, &summary))!=GOOD)
	{
	    if (summary) free(summary);
	    if (marcOut) free(marcOut);
	    mrFree(&marcRec);
	    sprintf (err_buf, "wcb_marc: No Marc record generated.\nProblem reported by GetBrief.\n");
	    misc_warning_hcb(err_buf);
	    send_record(position, err_buf, summ_flag);
	    return NOTOK;
	}
	send_record(position, marcOut, summ_flag);
	free(summary);
	free(marcOut);
	mrFree(&marcRec);
    } else {		/* full */
	if ((returnValue=GetFull(&marcRec,&marcOut))!=GOOD) {
	    if (marcOut) free(marcOut);
	    mrFree(&marcRec);
	    sprintf (err_buf, "wcb_marc: No Marc record generated.\nProblem reported by GetBrief.\n");
	    misc_warning_hcb(err_buf);
	    send_record(position, err_buf, summ_flag);
	    return NOTOK;
	}
	send_record(position, marcOut, summ_flag);
	free(marcOut);
	mrFree(&marcRec);
    }
    return OK;
}

/* This is a little inefficient right now.
 We call GetBrief, and throw the brief away and
 use the summary.
 That's a little like "Throw the stew away and eat
 the broom." */

int wcb_marc_title(position, record, llen)
int position;
char *record;
int llen;
{
    int returnValue;
    char *summary = NULL, *marcOut = NULL;
    MARC_RECORD marcRec;
    char err_buf[BUFSIZ];

    if( llen < MIN_MARC_SIZE || llen > MAX_MARC_SIZE ) {
	sprintf (err_buf, "wcb_marc_title: Corrupt Marc record of bogus size: %d.", llen);
	misc_warning_hcb(err_buf);
	send_title(position, err_buf);
	return NOTOK;
    }

    if ((returnValue= mrInit(&marcRec))!=GOOD) {
	sprintf (err_buf, "wcb_marc_title: Corrupt Marc record with problem '%s' reported by mrInit.",  mrStatString(returnValue));
	misc_warning_hcb(err_buf);
	send_title(position, err_buf);
	return NOTOK;
    }

    if ((returnValue=mrMapMarc(record,&marcRec))!=GOOD) {
	sprintf (err_buf, "wcb_marc_title: Corrupt Marc record with problem '%s' reported by mrMapMarc.",  mrStatString(returnValue));
	misc_warning_hcb(err_buf);
	send_title(position, err_buf);
	return NOTOK;
    }

    if ((returnValue=GetBrief(&marcRec, &marcOut, &summary))!=GOOD)
    {
	if (summary) free(summary);
	if (marcOut) free(marcOut);
	mrFree(&marcRec);
	sprintf (err_buf, "wcb_marc_title: Problem reported by GetBrief.");
	misc_warning_hcb(err_buf);
	send_title(position, err_buf);
	return NOTOK;
    }
    send_title(position, summary);
    free(summary);
    free(marcOut);
    mrFree(&marcRec);
    return OK;
}

int wcb_extract_marc(position, record, llen, key, extract)
int position;
char *record;
int llen;
char *key;
char **extract;
{
    int returnValue;
    char *marcOut = NULL;
    MARC_RECORD marcRec;
    char err_buf[BUFSIZ];

    *extract = NULL;	/* prepare for the worst */

    if( llen < MIN_MARC_SIZE || llen > MAX_MARC_SIZE ) {
	sprintf (err_buf, "wcb_extract_marc: Corrupt Marc record of bogus size: %d.\n", llen);
	misc_warning_hcb(err_buf);
	return NOTOK;
    }

    if ((returnValue= mrInit(&marcRec))!=GOOD) {
	sprintf (err_buf, "wcb_extract_marc: Corrupt Marc record with problem '%s' reported by mrInit.\n",  mrStatString(returnValue));
	misc_warning_hcb(err_buf);
	return NOTOK;
    }


    if ((returnValue=mrMapMarc(record,&marcRec)) != GOOD) {
	sprintf (err_buf, "wcb_extract_marc: Corrupt Marc record with problem '%s' reported by mrMapMarc.\n",  mrStatString(returnValue));
	misc_warning_hcb(err_buf);
	return NOTOK;
    }

    if ((returnValue=GetExtract(&marcRec, &marcOut, key)) != GOOD) {
	    if (marcOut) free(marcOut);
	    mrFree(&marcRec);
	    sprintf (err_buf, "wcb_extract_marc: No Marc record generated.\nProblem reported by GetExtract.\n");
	    misc_warning_hcb(err_buf);
	    return NOTOK;
    }
    *extract = marcOut;
    mrFree(&marcRec);
    return OK;
}

