
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: text.C,v 2.3 1995/06/24 15:01:04 davidsi Exp $
//
// $Log: text.C,v $
// Revision 2.3  1995/06/24  15:01:04  davidsi
// - Added GNU copyright message.
//
// Revision 2.2  1995/02/28  21:49:51  davidsi
// - Added locking for text instances, this allows a Text instance to
//   be modified once its being displayed: TextrealClear() sets the
//   lock, TextrealComplete() clears the lock, and TextrealIslocked()
//   returns the state of the lock.
//
// Revision 2.1  1995/02/02  21:18:14  davidsi
// - Many changes.  Factored out much of the functionality into 2 new classes.
//   The attr (attribute) class looks after text attributes. The textline
//   class looks after lines of text and their attributes.  This has resulted
//   in a much cleaner text class, without all the weird array manipulations.
//
// Revision 2.0  1995/01/19  21:25:04  davidsi
// - Removed time.h include.
//
// Revision 1.1  1994/12/11  21:38:35  davidsi
// Initial revision
//
//
//


//
// Include Files
//
#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "text.h"

#include "localmacros.h"

//
///////////////////////////////////////////////////////////////////////////////
//
// The attribute class
//
// This class stores information about text.
//
///////////////////////////////////////////////////////////////////////////////
//

//
// extract_attr
//
// Extracts attribute flag information
//
// INPUTS :	None
// RETURNS:	Flag attribute information
//
attr_t
attr::extract_attr()
{
	return(flags);
}

//
// extract_colour
//
// Extracts colour information
//
// INPUTS :	None
// RETURNS:	Colour information
//
colour_e
attr::extract_colour()
{
	return(colour);
}

//
// Constructors for the attribute class. 
//
// Attributes can be built on the basis of an existing attribute,
// colour/flag information, or solely flag information.
//
attr::attr(class attr *a)
{
	flags=a->flags;
	colour=a->colour;
}

attr::attr(attr_t iflags,colour_e icolour)
{
	flags=iflags;
	colour=icolour;
}

attr::attr(attr_t iflags)
{
	flags=iflags;
	colour=COLOUR_NOTSET;
}

//
// Destructor for attributes
//
attr::~attr()
{
}

//
///////////////////////////////////////////////////////////////////////////////
//
// The textline class
//
// This class acts as a container for text and its attributes.
//
///////////////////////////////////////////////////////////////////////////////
//

char *
textline::readtext(int fieldno)
{
int chrs;			// length of string of current field
int field;			// current field number
char *strptr;

	strptr=linetext;
	for(field=0;field<fieldno;field++) {
		chrs=fieldlens[field];
		strptr += chrs+1;
	}
	return(strptr);
}

class attr *
textline::readattribute(int fieldno)
{
	return (attrs[fieldno] );
}

void
textline::setattr(int fieldno,class attr *aptr)
{
	if (attrs[fieldno])
		delete attrs[fieldno];
	attrs[fieldno]=aptr;
}



//
// add line
//
// Adds a line of text.
//
// INPUTS :	line		The text to add (null terminated)
// RETURNS:	None
//
textline::textline(int fields,
		char *txt,
		int *fmax,
		class attr **defattrs,
		char fieldend)
{
	char *text;
	char *string;
	int i;
	int newlen;
	int field;
	int sublen;

	newlen=strlen(txt);
	//
	// NB. We add fields because the string passed in may not
	// have defined all (or any) fields in which case we will
	// need to add further null terminating bytes - 1 for each
	// ommitted field.
	//
	linetext=(char *)malloc(newlen+10+fields);
	strcpy(linetext,txt);


	//
	// Set up attributes - null terminated!
	//
	attrs=(class attr **)malloc( (fields+1) *sizeof(attr *));
	for(field=0;field<fields;field++) 
		attrs[field]=new attr(defattrs[field]);
	attrs[fields]=NULL;

	//
	// Compute field lengths
	//
	fieldlens=(int *)malloc(fields*sizeof(int));
	string=linetext;
	for(field=0;field<fields;field++) {
		//
		// If not on null terminator, search forward, else
		// the field has zero width
		// 
		if (*string!='\0') {
			text=string;
			while (*string!='\0' && *string!=fieldend)
				string++;
			sublen=(int)(string-text);
			if ( *string==fieldend ) {
				*string='\0';
				string++;
			}
		} else {
			//
			// Start to add null terminators for the missing fields
			//
			string++;
			*string='\0';
			sublen=0;
		}

		if (sublen>fmax[field])
			fmax[field]=sublen;
	
		fieldlens[field]=sublen;
	}
}

textline::~textline()
{
	int field;
	class attr **aptr;

	aptr=attrs;
	while (*aptr) {
		delete *aptr;
		aptr++;
	}

	free(linetext);
	free(fieldlens);
}

//
///////////////////////////////////////////////////////////////////////////////
//
// The text class
//
///////////////////////////////////////////////////////////////////////////////
//


//
// TextrealAdd line
//
// Adds a line of text.
//
// INPUTS :	line		The text to add (null terminated)
// RETURNS:	None
//
void
Textreal::TextrealAdd(char *text)
{
	if (n_lines>=max_n_lines) {
		max_n_lines+=100;
		if (lines==NULL)
			lines=(class textline **)
			    malloc(sizeof(class textline *)*max_n_lines);
		else
			lines=(class textline **)
			    realloc(lines,sizeof(class textline *)*max_n_lines);
	}
	lines[n_lines]=new textline(fields,text,fieldmax,df_attr,fieldend);
	n_lines++;
}


//
// TextrealCline 
//
// Return the current line number. If text has just been added then
// this function returns the line number that references that text.
//
// INPUTS :	None
// RETURNS:	The current line number
//
int
Textreal::TextrealCline()
{
	return(n_lines-1);
}


//
// TextrealClear
//
// Removes all previously added lines of text
//
// 
// INPUTS :	None
// RETURNS:	None
//
void
Textreal::TextrealClear()
{
	int iter;
	int field;

	if (lines) {
		for(iter=0;iter<n_lines;iter++)
			delete lines[iter];
	}

	n_lines = 0;
	textflags |= TEXT_FLAGS_LOCKED;
}

//
//  TextrealRead lineno
//
// Returns an ascii string for a given line.  An attempt is made at
// formatting the string according to the field attributes.
// 
// INPUTS :	lineno		The line for which the text is req'd
// RETURNS:	The text string for the specified line	
//
char *
Textreal::TextrealRead(int lineno)
{
register char *strptr;		// current string for field
int field;			// current field number
static char buffer[1024];
int iter;
int pos;

	for(strptr=buffer;strptr<buffer+1024;strptr++)
		*strptr=' ';

	pos=0;
	for(field=0;field<fields;field++) {
		strptr=lines[lineno]->readtext(field);
	/*	if (field_attr[field] & FIELDATTR_LEFT) */
		if (strptr)
			strncpy(&buffer[pos],strptr,strlen(strptr));
/*		else  FIELD RIGHT 
			strcpy(&buffer[pos+fieldmax[field]-chrs],strptr);
*/
		pos += fieldmax[field]+1;
	}
	buffer[pos]='\0';
	return(buffer);
}

//
// TextrealRead line,field
//
// Reads a field on a specific line.  No formatting is performed.
// 
// INPUTS :	lineno		The line upon which the field lies.
//				The field number that is req'd
// RETURNS:	A string containing the required field.
//
char *
Textreal::TextrealRead(int lineno, int fieldno)
{
	if (fieldno>fields)
		return( NULL );
	return( lines[lineno]->readtext(fieldno) );
}

//
// TextrealReadattribute line,field
//
// Reads the attribute that applies to a field on a specific line.
// 
// INPUTS :	lineno		The line upon which the field lies.
//		file		The field number that is req'd
// RETURNS:	The relevant attribute
//
class attr *
Textreal::TextrealReadattribute(int lineno, int fieldno)
{
	if (fieldno>fields)
		return( NULL );
	return( lines[lineno]->readattribute(fieldno) );
}


//
// TextrealOutput lineno
//
// This is a virtual function that will normally be overriden
// in derived classes, so as to provide a more useful output form!
//
// INPUTS : 	lineno		The line to be output
// RETURNS:	None
//
void
Textreal::TextrealOutput(int lineno)
{
	char *lineout;

	lineout=TextrealRead(lineno);
	printf("%s\n",lineout);
}

//
// TextrealNlines
//
// Returns the number of added lines
// 
// INPUTS :	None
// RETURNS:	The number of added lines
//
int
Textreal::TextrealNlines()
{
	return(n_lines);
}

//
// TextrealNfields
//
// Returns the number of supported fields.
// 
// INPUTS :	None
// RETURNS:	The number of supported fields.
//
int
Textreal::TextrealNfields()
{
	return(fields);
}



//
// TextrealSetdefaultattr field attr
//
// Sets the default attributes for a given field.
// 
// INPUTS :	field		The field whose attributes are to be set
//		attr		Attributes
// RETURNS:	None
//
void
Textreal::TextrealSetdefaultattr(int fieldno,class attr *aptr)
{
	if (fieldno<1 || fieldno>fields) {
		fprintf(stderr,
		         "setfieldattr field %d should be in range 1-%d\n",
			 fieldno,fields);
		exit(1);
	}
	fieldno--;

	delete df_attr[fieldno];

	df_attr[fieldno]=aptr;
}


//
// TextrealSetattr line field attr
//
// Sets the attributes for a given field on a specific line
// 
// INPUTS :	line		The line on which the field exists
//		field		The field whose attributes are to be set
//		attr		Attributes
// RETURNS:	None
//
void
Textreal::TextrealSetattr(int lineno,int fieldno,class attr *aptr)
{

	if (fieldno<1 || fieldno>fields) {
		fprintf(stderr,
		         "setfieldattr field %d should be in range 1-%d\n",
			 fieldno,fields);
		exit(1);
	}
	fieldno--;

	lines[lineno]->setattr(fieldno,aptr);
}

//
// TextrealSetattr line attr
//
// Sets the attributes for all fields on a specific line
// 
// INPUTS :	line		The line on which the field exists
//		attr		Attributes
// RETURNS:	None
//
void
Textreal::TextrealSetattr(int lineno,class attr *aptr)
{
	int fieldno;
	class attr *raptr;

	for(fieldno=0;fieldno<fields;fieldno++) {
		if (fieldno<fields-1)
			raptr=new attr(aptr);
		else
			raptr=aptr;
		lines[lineno]->setattr(fieldno,raptr);
	}
}

//
// TextrealUnref
//
// Marks the text instance as having one less reference. When there
// are no references to the instance it is deleted.
// 
// INPUTS :	None
// RETURNS:	None
//
void
Textreal::TextrealUnref()
{
	ref_cnt--;
	if (ref_cnt==0)
		delete this;
}

//
// TextrealRef
//
// Marks the text instance as having one more reference. An instance
// cannot be deleted while there is 1 or more references to it.
// 
// INPUTS :	None
// RETURNS:	None
//
void
Textreal::TextrealRef()
{
	ref_cnt++;
}

//
// TextrealComplete
//
// Marks a text instance as unlocked. A text that is locked won't
// be accessed for display updates. A text becomes locked when
// TextrealClear() is used to clear ready for text additions.
//
// INPUTS : 	None
// RETURNS:	None
//
void
Textreal::TextrealComplete()
{
	textflags &= ~TEXT_FLAGS_LOCKED;
}


//
// TextrealIslocked
//
// Returns the state of a text instance.
//
// INPUTS : 	None
// RETURNS:	True if the instance is locked, else false.
int
Textreal::TextrealIslocked()
{
	return(  (textflags & TEXT_FLAGS_LOCKED)==TEXT_FLAGS_LOCKED );
}


//
// Constructor function
//
Textreal::Textreal(int maxfields)
{
	int field;

	textflags=0;
	ref_cnt=1;

	fields=maxfields;

	max_n_lines=0;
	n_lines=0;
	lines=NULL;
	fieldend='\t';

	fieldmax=(int *)malloc(sizeof(int)*(1+maxfields));
	for(field=0;field<maxfields;field++)
		fieldmax[field]=0;

	df_attr=(class attr **)malloc(sizeof(class attr *) *(1+maxfields));
	for(field=0;field<maxfields;field++)
		df_attr[field]=new attr(FIELDATTR_LEFT,FOREGROUND);
	df_attr[maxfields]=NULL;
}


//
// Destructor function
//
Textreal::~Textreal()
{
	int iter;
	class attr **aptr;

	//
	// Give a warning if an attempt is made to delete an instance
	// that has refences to it. 
	//
	if (ref_cnt) {
		printf("PANIC:Text instance deleted while referenced\n");
		exit(1);
	}

	if (lines) {
		for(iter=0;iter<n_lines;iter++)
			delete lines[iter];
		free(lines);
	}

	//
	// Free off attribute information
	//
	aptr=df_attr;
	while (*aptr) {
		delete *aptr;
		aptr++;
	}
	free(df_attr);
	free(fieldmax);
}

//
///////////////////////////////////////////////////////////////////////////////
//
// The Text class inherits most of its functionality from the Textreal
// class and infact acts as a redirector to that the Textreal class.  
// The key different between Text and Textreal is that multiple
// Text classes can use the same Textreal class. This is particularly
// useful for text dialogs because printing the text requires
// access to the same lines as those being displayed within the dialog.
//
///////////////////////////////////////////////////////////////////////////////
//


//
// TextAdd  text
//
// Adds a line of text.
//
// INPUTS :	line		The text to add (null terminated)
// RETURNS:	None
//
void
Text::TextAdd(char *itext)
{
	text->TextrealAdd(itext);
}

//
// TextCline 
//
// Return the current line number. If text has just been added then
// this function returns the line number that references that text.
//
// INPUTS :	None
// RETURNS:	The current line number
//
int
Text::TextCline()
{
	return(text->TextrealCline());
}


//
// TextClear
//
// Removes all previously added lines of text
// 
// INPUTS :	None
// RETURNS:	None
//
void
Text::TextClear()
{
	text->TextrealClear();
}

//
// TextRead lineno
//
// Returns an ascii string for a given line.  An attempt is made at
// basic formatting.
// 
// INPUTS :	lineno		The line for which the text is req'd
// RETURNS:	The text string for the specified line	
//
char *
Text::TextRead(int lineno)
{
	return (text->TextrealRead(lineno));
}


//
// TextRead line,field
//
// Reads a field on a specific line.  No formatting is performed.
// 
// INPUTS :	lineno		The line upon which the field lies.
//		field		The field number that is req'd
// RETURNS:	A string containing the required field.
//
char *
Text::TextRead(int lineno, int fieldno)
{
	return(text->TextrealRead(lineno,fieldno) );
}


//
// TextReadattribute line,field
//
// Reads the attributes for a field on a specific line.
// 
// INPUTS :	lineno		The line 
//		field		The field number that is req'd
// RETURNS:	The attribute information
//
class attr *
Text::TextReadattribute(int lineno, int fieldno)
{
	return(text->TextrealReadattribute(lineno,fieldno));
}


//
// TextOutput lineno
//
// This is a virtual function that will normally be overriden
// in derived classes, so as to provide a more useful output form!
//
// INPUTS : 	lineno		The line to be output
// RETURNS:	None
//
void
Text::TextOutput(int lineno)
{
	text->TextrealOutput(lineno);
}

//
// TextNlines
//
// Returns the number of added lines
// 
// INPUTS :	None
// RETURNS:	The number of added lines
//
int
Text::TextNlines()
{
	return(text->TextrealNlines());
}

//
// TextNfields
//
// Returns the number of supported fields.
// 
// INPUTS :	None
// RETURNS:	The number of supported fields
//
int
Text::TextNfields()
{
	return(text->TextrealNfields());
}




//
// TextSetdefaultattr field attr
//
// Sets the default attributes for a given field.
// 
// INPUTS :	field		The field whose attributes are to be set
//		attr		Attributes
// RETURNS:	None
//
void
Text::TextSetdefaultattr(int fieldno,class attr *aptr)
{
	text->TextrealSetdefaultattr(fieldno,aptr);
}


//
// TextSetattr line field attr
//
// Sets the attributes for a given field on a given line
// 
// INPUTS :	field		The field whose attributes are to be set
//		attr		Attributes
// RETURNS:	None
//
void
Text::TextSetattr(int lineno,int fieldno,class attr *aptr)
{
	text->TextrealSetattr(lineno,fieldno,aptr);
}



//
// TextSetattr line attr
//
// Sets the attributes for all fields on a given line
// 
// INPUTS :	line		The line on which all fields are to have
//				their attributes set
//		attr		Attributes information
// RETURNS:	None
void
Text::TextSetattr(int lineno,class attr *aptr)
{
	text->TextrealSetattr(lineno,aptr);
}


//
// TextAssign othertext
//
// TextAssign allows one text instance to use the information stored
// in another without more memory being allocated.
//
void
Text::TextAssign(class Text *othertext)
{
	text->TextrealUnref();
	text=othertext->text;
	text->TextrealRef();
}

//
// TextIslocked
//
// Returns the state of a text instance.
//
// INPUTS : 	None
// RETURNS:	True if the instance is locked, else false.
//
int
Text::TextIslocked()
{
	return( text->TextrealIslocked() );
}

//
// TextComplete
//
// Marks a text instance as unlocked. A text that is locked won't
// be accessed for display updates. A text becomes locked when
// TextrealClear() is used to clear ready for text additions.
//
// INPUTS : 	None
// RETURNS:	None
//
void
Text::TextComplete()
{
	text->TextrealComplete();
}

//
// Constructor function
//
Text::Text()
{
	text=new class Textreal(MAXFIELDS);
}

//
// Destructor function
//
// When a Text instance is deleted, the underlying Textreal class is
// unreferenced.  Whether or not the text goes away will depend on
// whether there are other references to that Textreal class.
//
Text::~Text()
{
	text->TextrealUnref();
}
