
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: val.C,v 2.1 1995/05/19 18:50:45 davidsi Exp $
//
// $Log: val.C,v $
// Revision 2.1  1995/05/19  18:50:45  davidsi
// - The val_init() function now requires an argument specifying which
//   portfolio the valuation belongs to.
// - Added ValuationMakeconsistant() function member to allow a valuation
//   to be reset when the portfolio changes. Including specifying a different
//   pathname for the price information.
//
// Revision 2.0  1995/03/08  21:26:52  davidsi
// - Initial version.
//
//
//
//

#include <stdio.h>
#include <varargs.h>
#define _STDARG_H
#include <stdlib.h>
#include <string.h>
#include <dial.h>
#include <math.h>

#include "localmacros.h"

#include "val.h"
#include "security.h"
#include "menu.h"
#include "callback.h"


static class Valuation *vptr=NULL;

class Valuation  
{
	class security secvalue;
	class security seccost;
	class security secprofit;
	Callbackid callback_price;	// callback for price changes
        static void recompute(char *realthis,char *);
public:
	void ValuationPaths();
	void ValuationCompute(class date &);
	void ValuationMakeconsistant(char *pname);

	Valuation();
	~Valuation();
};

void
Valuation::ValuationPaths()
{
	secprofit.prices->PriceChecklocation();
	seccost.prices->PriceChecklocation();
	secvalue.prices->PriceChecklocation();
}

void
Valuation::ValuationCompute(class date &end_date)
{
	class security *ptr;
	class Holding *holding;
	double number;			// number of securities owned 
	double cost;			// amount spent 
	double received;		// amount received 
	double t_cost=0.0;		// total amount spent/received 
	double t_value=0.0;		// total value of securities at
					// period end 
	double price;			// price of security on(or about)

					// end_date 
	class date start_date;


	holding=portfolio->head();
	//
	// For every known security do..
	//
	for ( ; !portfolio->eol(holding) ; holding=holding->next() ) {
		ptr=holding->HoldingSecurity();
		holding->HoldingTransactionhead()->extract(&start_date,
				&end_date,&number,&cost,&received,NULL);
		if ( (fabs(number)<SMALL_NUMBER) &&
		     (fabs(cost)<SMALL_NUMBER) && 
		     (fabs(received)<SMALL_NUMBER) )
			continue;	// security never in persons portfolio
		price=ptr->prices->PriceFetch(&end_date,TRUE);
		if (price>=0.0 && (number>SMALL_NUMBER) )
			t_value += price*number;
		else if (number>SMALL_NUMBER)
			t_value += cost-received;
		t_cost += cost-received;
	}
	secprofit.prices->PriceStore(&end_date,t_value-t_cost);
	seccost.prices->PriceStore(&end_date,t_cost);
	secvalue.prices->PriceStore(&end_date,t_value);
}


void
Valuation::recompute(char *tmpthis,char *)
{
        class Valuation *rthis=(class Valuation *)tmpthis;
	class date end_date;

	end_date.setdate();             // Get today's date
	rthis->ValuationCompute(end_date);
}

//
// ValuationMakeconsistant
//
// Makes a valuation instance consistent with the security by removing
// all the price information.  This can then be reloaded on a as needed
// basis.
//
// INPUTS : None
// RETURNS: None
//
void
Valuation::ValuationMakeconsistant(char *pname)
{
	secvalue.prices->PriceClear();
	seccost.prices->PriceClear();
	secprofit.prices->PriceClear();

	secvalue.prices->PriceAssignpath(pname);
	seccost.prices->PriceAssignpath(pname);
	secprofit.prices->PriceAssignpath(pname);
}

Valuation::Valuation()
{
	secvalue.SecurityCreate("Portfolio total",
			SEC_FLAGS_DUMMY,NULL,0,SEC_TYPE_GROUP);
	secvalue.add();
	seccost.SecurityCreate("Portfolio cost",
			SEC_FLAGS_DUMMY,NULL,0,SEC_TYPE_GROUP);
	seccost.add();
	secprofit.SecurityCreate("Portfolio profit",
			SEC_FLAGS_DUMMY,NULL,0,SEC_TYPE_GROUP);
	secprofit.add();
        callback_price=main_price_list->change_vect.CallbackAdd(
                        Valuation::recompute,
                        (char *)this);
}

Valuation::~Valuation()
{
	//
	// Remove callbacks
	//
	main_price_list->change_vect.CallbackRemove(callback_price);
}



void
ValuationValidate(class Busy &busy)
{
	class date curdate;
	class date enddate;
	int totaldays;

	enddate.setdate();				// Get today's date


	//
	// Check paths on files are correct
	//
	vptr->ValuationPaths();

	//
	// Now recompute all the data...this sure will take a long time..
	//
	busy.BusyTotal( enddate.daysince1970()-curdate.daysince1970() );

	totaldays=0;
	for(;;) {
		if ( curdate.compare(&enddate)>0 )
			break;
		totaldays++;
		if ( totaldays%31 == 0 )
        		busy.BusyCompleted(totaldays,curdate.getdate());

		vptr->ValuationCompute(curdate);
		curdate.add(1);
	}
}


void
val_init(char *pname)
{
	//
	// If no valuation yet created, create one, otherwise
	// "patch up" the existing one.
	//
	if (vptr==NULL)
		vptr=new class Valuation;

	vptr->ValuationMakeconsistant(pname);
}
