
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: value.C,v 2.5 1995/06/17 20:29:11 davidsi Exp $
//
// $Log: value.C,v $
// Revision 2.5  1995/06/17  20:29:11  davidsi
// - Added operation() function member to the value_string class, since we
//   need to be able to compare strings.
// - Added operation() function member to the value_struct class, since it
//   is helpful to be able to perform equality/inequality checks on
//   structures.
// - Added operation() function member to the value_array class, since it
//   is helpful to be able to perform equality/inequality checks on
//   arrays.
// - Removed unwrap() in both value and value_struct classes.
// - Added a new type, the value_char to allow for characters rather
//   than integers.
// - Now that a char type exists, deref() has been added to strings, so
//   that characters can be extracted from them.
// - Teach integer and character class about the op_not operation.
//
// Revision 2.4  1995/06/07  20:48:20  davidsi
// - Now that the error stream understands value class members don't
//   do the decoding here - simply pass the value into the error
//   stream.
// - Set the value_flag member to 0 for the generic value class.
//
// Revision 2.3  1995/05/29  19:03:56  davidsi
// - The value::fetch_storage() function member now returns NULL so the
//   condition can be detected by the caller.
// - BUG FIX. Added value::clone() so bad calls to clone() can be spotted.
// - Added assign_storage() and clone() to the function type.
//
// Revision 2.2  1995/05/19  18:53:01  davidsi
// - Added type conversions for both integer and real operations, so
//   operands can be freely mixed.
//
// Revision 2.1  1995/04/25  21:06:35  davidsi
// - Add name() function member to value and all value derived classes. This
//   identifies the type using an ascii string - great for debugging.
// - Remove release_storage() function member - it is not needed.
// - Added function2index() to the Colour class to allow function names
//   to be mapped to their index (enumerated) values.
//
// Revision 2.0  1995/01/31  22:06:38  davidsi
// - First version.
//
//
//

#include "stdio.h"
#include "stdlib.h"

#include "value.h"
#include "error.h"
#include "context.h"
#include "dialog.h"
#include "builtin.h"

void
value::setstatus(const valuetype_e c_status)
{
	status=c_status;
}

void
value::dispose()
{
	if (locked_context)
		locked_context->unlock_heap(locked_heap);
	locked_context=NULL; 
	locked_heap=NULL; 
	if (status==temporary) {
		delete this;
	}
}


void
value::lock_2_heap(class context *cont)
{
	locked_heap=cont->lock_heap();
	locked_context=cont;
}


int
value::id()
{
	return(value_flag);
}

int
value::store(int num)
{
	cerror << "store(int) failed on typeid "<<this<<"\n";
	return(-1);
}

int
value::store(char *)
{
	cerror << "store(char *) failed on typeid "<<this<<"\n";
	return(-1);
}

int
value::store(double)
{
	cerror << "store(double) failed on typeid "<<this<<"\n";
	return(-1);
}

int
value::load(int &)
{
	cerror << "load failed on typeid "<<this<<"\n";
        return(-1);
}

int
value::load(char *&)
{
	cerror << "load failed on typeid "<<this<<"\n";
	return(-1);
}

int
value::load(double &)
{
	cerror << "load failed on typeid "<<this<<"\n";
	return(-1);
}


int
value::size()
{
	return(0);
}

void
value::to_mem(char *mem)
{
	cerror << "to_mem failed on typeid "<<this<<"\n";
}

void
value::from_mem(char *mem)
{
	cerror << "from_mem failed on typeid "<<this<<"\n";
}

void
value::assign_storage(char *)
{
	cerror << "assign_storage failed on typeid "<<this<<"\n";
}

char *
value::fetch_storage()
{
	cerror << "fetch_storage failed on typeid "<<this<<"\n";
	return(NULL);
}

class value *
value::clone()
{
	cerror << "clone failed on typeid "<<this<<"\n";
}

char *
value::name()
{
	return( "void" );
}

//
// By default types are equal if there value flags are the same. This
// is not TRUE for all types, eg. arrays etc.
//
int
value::type_equal(class value *v2)
{
	if (value_flag == v2->value_flag)
		return(1);
	return(0);
}

class value *
value::operation(operators_e,class value *)
{
	return(NULL);
}

class value *
value::deref(int index)
{
	cerror << "Derefence failed on typeid "<<this<<"\n";
	return(NULL);
}

class value *
value::deref(int index,char *&addr)
{
	cerror << "Derefence failed on typeid "<<this<<"\n";
	return(NULL);
}

class value *
value::lookup(char *name)
{
	cerror << field_lookup_failed <<name<<"\n";
	return(NULL);
}

class value *
value::lookup(char *name,char *&)
{
	cerror << field_lookup_failed <<name<<"\n";
	return(NULL);
}

virtual class node_funcdecl *
value::funcaddr()
{
	return(NULL);
}


//
// A fundamental value cannot be disposed of.  Values start as
// temporary , use setstatus() to upgrade to a fundamental value.
//
value::value()
{
	status=temporary;
	locked_heap=NULL; 
	value_flag=0;
}

value::~value()
{
}



//
//////////////////////////////////////////////////////////////////////////////
//
// Integer value operations
//
//////////////////////////////////////////////////////////////////////////////
//
int
value_int::store(int num)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_int)\n";
		exit(1);
	}
#endif
	*number=num;
	return(0);
}

int
value_int::store(char *str)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_int)\n";
		exit(1);
	}
#endif
	*number=atoi(str);
	return(0);
}

int
value_int::store(double num)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_int)\n";
		exit(1);
	}
#endif
	*number=(int)num;
	return(0);
}

int
value_int::load(int &reti)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_int)\n";
		exit(1);
	}
#endif
	reti=*number;
	return(0);
}

int
value_int::load(char *&retstr)
{
	static char buffer[64];

#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_int)\n";
		exit(1);
	}
#endif

	sprintf(buffer,"%d",*number);

	retstr=buffer;

	return(0);
}

int
value_int::load(double &retdoub)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_int)\n";
		exit(1);
	}
#endif
	retdoub=(double)*number;
	return(0);
}


int
value_int::size()
{
	return( sizeof(*number) );
}

void 
value_int::to_mem(char *mem)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_int)\n";
		exit(1);
	}
#endif
	*((int *)mem)=*number;
	INFO(cinfo << "Stored int "<< *number <<" at address" << (address)number << "\n";)
}

void
value_int::from_mem(char *mem)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_int)\n";
		exit(1);
	}
#endif
	*number=*((int *)mem);
	INFO(cinfo << "Read int" << *number << " from address" << (address)mem << "\n";)
}

void
value_int::assign_storage(char *mem)
{
	if (mem)
		number=(int *)mem;
	else {
		//
		// Only copy the current value if the pointer is non null
		//
		if (number)
			tmp_number=*number;
		number=&tmp_number;
	}
}

char *
value_int::fetch_storage()
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_int)\n";
		exit(1);
	}
#endif
	return((char *)number);
}

char *
value_int::name()
{
	return( "int" );
}

class value *
value_int::operation(operators_e op,class value *v2)
{
	class value_int *va,*vb;
	char va_alloc=0;
	int second;

	//
	// Handle unary operators first - they have no type checking
	//
	if (op==op_not) {
		if (status==temporary)
			va=this;
		else
			va=(class value_int *)clone();
		*(va->number) = !(*(va->number));
		return(va);
	}

	//
	// Do we have type mismatch problems?
	//
	if (value_flag != v2->value_flag) {
		//
		// If v2 is a real number, then upgrade the current
		// value to a real and then do the operation
		//
		if (v2->value_flag==VALUE_REAL) {
			class value_real *vc=new class value_real;

			vc->assign_storage(NULL);
			vc->store(*number);
			dispose();
			return(vc->operation(op,v2));
		} else if (v2->value_flag==VALUE_CHAR) {
			v2->load(second);
		} else {
			//
			// Otherwise type mismatch error
			//
			cerror << type_mismatch;
			return(NULL);
		}
	} else {
		//
		// Types match so simply load the value straight in
		//
		vb=(class value_int *)v2;
		second=*(vb->number);
	}
	//
	// If the value is temporary, then use it.  Otherwise
	// a temporary node must be created. 
	// 
	if (status==temporary) {
		va=this;
	} else {
		va=(class value_int *)clone();
		va_alloc=1;
	}

	//
	// As we are about to modify the value, point to a temporary
	// workarea.
	//
	va->assign_storage(NULL);
	

	switch(op) {
        	case op_add:*(va->number) += second;
			return(va);
        	case op_sub:*(va->number) -= second;
			return(va);
        	case op_mul:*(va->number) *= second;
			return(va);
	        case op_div:*(va->number) /= second;
			return(va);
	        case op_or:*(va->number) |= second;
			return(va);
	        case op_xor:*(va->number) ^= second;
			return(va);
	        case op_and:*(va->number) |= second;
			return(va);
	        case op_lor:*(va->number) = *(va->number) || second;
			return(va);
	        case op_land:*(va->number) = *(va->number) && second;
			return(va);
	        case comp_eq:
			*(va->number) = (*(va->number)==second);
			return(va);
		case comp_ne:
			*(va->number) = (*(va->number)!=second);
			return(va);
		case comp_lt:
			*(va->number) = (*(va->number)<second);
			return(va);
		case comp_le:
			*(va->number) = (*(va->number)<=second);
			return(va);
		case comp_gt:
			*(va->number) = (*(va->number)>second);
			return(va);
		case comp_ge:
			*(va->number) = (*(va->number)>=second);
			return(va);
	}
	cerror << bad_operation ;
	if (va_alloc) {
		va->dispose();
		va=NULL;
	}
	return(NULL);
}

class value *
value_int::clone()
{
        class value_int *newval=new value_int;

        newval->number=number;
        newval->tmp_number=tmp_number;

	return ((class value *)newval);
}


value_int::value_int()
{
	value_flag=VALUE_INT;
	number=NULL;
}

value_int::~value_int()
{
}

//
//////////////////////////////////////////////////////////////////////////////
//
// Real value operations
//
//////////////////////////////////////////////////////////////////////////////
//
int
value_real::store(int num)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_real)\n";
		exit(1);
	}
#endif
	*number=(double)num;
	return(0);
}

int
value_real::store(char *str)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_real)\n";
		exit(1);
	}
#endif
	*number=atof(str);
	return(0);
}

int
value_real::store(double num)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_real)\n";
		exit(1);
	}
#endif
	*number=num;
	return(0);
}

int
value_real::load(int &reti)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_real)\n";
		exit(1);
	}
#endif
	reti=(int)*number;
	return(0);
}

int
value_real::load(char *&retstr)
{
	static char buffer[64];

#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_real)\n";
		exit(1);
	}
#endif
	sprintf(buffer,"%8.4f",*number);

	retstr=buffer;
	//retstr=(char *)malloc(strlen(buffer)+1);
	////strcpy(retstr,buffer);

	return(0);
}

int
value_real::load(double &retdoub)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_real)\n";
		exit(1);
	}
#endif
	retdoub=*number;
	return(0);
}


int
value_real::size()
{
	return( sizeof(*number) );
}

void 
value_real::to_mem(char *mem)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_real)\n";
		exit(1);
	}
#endif
	*((double *)mem)=*number;
	INFO(cinfo << "Stored real "<< *number <<" at address" << (address)number << "\n";)
}

void
value_real::from_mem(char *mem)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_real)\n";
		exit(1);
	}
#endif
	*number=*((double *)mem);
	INFO(cinfo << "Read real" << *number << " from address" << (address)mem << "\n";)
}

void
value_real::assign_storage(char *mem)
{
	if (mem)
		number=(double *)mem;
	else {
		//
		// If a number pointer is given then we clone the value,
		//
		if (number)
			tmp_number=*number;
		
		number=&tmp_number;
	}
}

char *
value_real::fetch_storage()
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_real)\n";
		exit(1);
	}
#endif
	return((char *)number);
}

char *
value_real::name()
{
	return( "real" );
}

class value *
value_real::operation(operators_e op,class value *v2)
{
	class value_real *va,*vb;
	class value_int *val_bool;
	char va_alloc=0;
	int result;

	
	//
	// If 2nd argument is a int upgrade it to a real, before
	// doing any type checking.
	//
	if (v2->value_flag==VALUE_INT) {
		class value_real *vc=new class value_real;

		vc->assign_storage(NULL);
		v2->load(result);
		vc->store(result);
		v2->dispose();
		v2=vc;
	}

	//
	// Do we have type mismatch problems?
	//
	if (value_flag != v2->value_flag) {
		//
		// Otherwise type mismatch error
		//
		cerror << type_mismatch;
		return(NULL);
	}

	if (op==op_add || op==op_sub || op==op_mul || op==op_div ) {
		//
		// If the value is temporary, then use it.  Otherwise
		// a temporary node must be created. 
		// 
		if (status==temporary) {
			va=this;
		} else {
			va=(class value_real *)clone();
			va_alloc=1;
		}

		//
		// As we are about to modify the value, point to a temporary
		// workarea.
		//
		va->assign_storage(NULL);
		vb=(class value_real *)v2;

		switch(op) {
		case op_add:*(va->number) += *(vb->number);
			return(va);
        	case op_sub:*(va->number) -= *(vb->number);
			return(va);
        	case op_mul:*(va->number) *= *(vb->number);
			return(va);
	        case op_div:*(va->number) /= *(vb->number);
			return(va);
		}
		if (va_alloc) 
			delete va;
	}

	//
	// Not a type real operation, so assume it will be a 
	// comparison, so allocate a value_int.
	// 
	val_bool=new class value_int;
	val_bool->assign_storage(NULL);
	vb=(class value_real *)v2;
	
	switch(op) {
	        case comp_eq:
			val_bool->store(*(number)==*(vb->number));
			return(val_bool);
		case comp_ne:
			val_bool->store(*(number)!=*(vb->number));
			return(val_bool);
		case comp_lt:
			val_bool->store(*(number)<*(vb->number));
			return(val_bool);
		case comp_le:
			val_bool->store(*(number)<=*(vb->number));
			return(val_bool);
		case comp_gt:
			val_bool->store(*(number)>*(vb->number));
			return(val_bool);
		case comp_ge:
			val_bool->store(*(number)>=*(vb->number));
			return(val_bool);
	}
	cerror << bad_operation ;
	delete val_bool;

	return(NULL);
}

class value *
value_real::clone()
{
        class value_real *newval=new value_real;

        newval->number=number;
        newval->tmp_number=tmp_number;

	return ((class value *)newval);
}


value_real::value_real()
{
	number=NULL;
	value_flag=VALUE_REAL;
}

value_real::~value_real()
{
}



//
//////////////////////////////////////////////////////////////////////////////
//
// String value operations
//
//////////////////////////////////////////////////////////////////////////////
//
int
value_string::store(int num)
{
#ifdef ASSERTS
	if (str==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	sprintf(str,"%d",num);
	return(0);
}

int
value_string::store(char *s)
{
#ifdef ASSERTS
	if (str==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	strcpy(str,s);
	return(0);
}

int
value_string::store(double num)
{
#ifdef ASSERTS
	if (str==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	sprintf(str,"%8.4f",num);
	return(0);
}

int
value_string::load(int &retint)
{
#ifdef ASSERTS
	if (str==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	retint=atoi(str);
	return(0);
}

int
value_string::load(char *&retstr)
{
#ifdef ASSERTS
	if (str==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	retstr=str;
	return(0);
}

int
value_string::load(double &retdou)
{
#ifdef ASSERTS
	if (str==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	retdou=atof(str);
	return(0);
}


int
value_string::size()
{
	return(STRING_LENGTH);
}

void 
value_string::to_mem(char *mem)
{
#ifdef ASSERTS
	if (str==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	strcpy(mem,str);
}

void
value_string::from_mem(char *mem)
{
#ifdef ASSERTS
	if (str==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	strcpy(str,mem);
}

void
value_string::assign_storage(char *mem)
{
	str=mem;
}

char *
value_string::fetch_storage()
{
#ifdef ASSERTS
	if (str==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	return((char *)str);
}

char *
value_string::name()
{
	return( "string" );
}

class value *
value_string::operation(operators_e op,class value *v2)
{
	class value_string *vb;
	class value_int *val_bool;
	int result;

	
	//
	// Do we have type mismatch problems?
	//
	if (value_flag != v2->value_flag) {
		//
		// Otherwise type mismatch error
		//
		cerror << type_mismatch;
		return(NULL);
	}

	//
	// Not a type string operation - assume it will be a 
	// comparison, so allocate a value_int.
	// 
	val_bool=new class value_int;
	val_bool->assign_storage(NULL);
	vb=(class value_string *)v2;

	result=strcmp(str,vb->str);
	
	switch(op) {
	        case comp_eq:
			val_bool->store(result==0);
			return(val_bool);
		case comp_ne:
			val_bool->store(result!=0);
			return(val_bool);
		case comp_lt:
			val_bool->store(result<0);
			return(val_bool);
		case comp_le:
			val_bool->store(result<=0);
			return(val_bool);
		case comp_gt:
			val_bool->store(result>0);
			return(val_bool);
		case comp_ge:
			val_bool->store(result>=0);
			return(val_bool);
	}
	cerror << bad_operation ;
	delete val_bool;

	return(NULL);
}

class value *
value_string::deref(int index)
{
	class value_char *val;
	char *ret_addr=str+index;

#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_string)\n";
		exit(1);
	}
#endif
	if (index>=STRING_LENGTH) {
		cerror << index_too_large << "\n";
		return(NULL);
	}
	
	val=new class value_char;
	val->assign_storage( ret_addr );

	INFO(cinfo << "Dereference copy from address 0x"<<(address)ret_addr;)
	INFO(cinfo <<" on index="<<index<<"\n";)

	return(val);
}

class value *
value_string::deref(int index,char *&ret_addr)
{

#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_array)\n";
		exit(1);
	}
#endif
	if (index>=STRING_LENGTH) {
		cerror << index_too_large << "\n";
		return(NULL);
	}

	ret_addr=str+index;

	INFO(cinfo << "Dereference gives address 0x"<<(address)ret_addr;)
	INFO(cinfo <<" on index="<<index<<"\n";)
	return(CHAR_VALUE);
}


class value *
value_string::clone()
{
        class value_string *newval=new value_string;

	if (str) {
	        newval->str=(char *)malloc(strlen(str)+1);
		strcpy(newval->str,str);
	}

	return ((class value *)newval);
}


value_string::value_string()
{
	str=NULL;
	value_flag=VALUE_STRING;
}

value_string::~value_string()
{
	if (str)
		free(str);
}



//
//////////////////////////////////////////////////////////////////////////////
//
// Character value operations
//
//////////////////////////////////////////////////////////////////////////////
//
int
value_char::store(int num)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_char)\n";
		exit(1);
	}
#endif
	*number=num;
	return(0);
}

int
value_char::store(char *str)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_char)\n";
		exit(1);
	}
#endif
	*number=atoi(str);
	return(0);
}

int
value_char::store(double num)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_char)\n";
		exit(1);
	}
#endif
	*number=(char)num;
	return(0);
}

int
value_char::load(int &reti)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_char)\n";
		exit(1);
	}
#endif
	reti=*number;
	return(0);
}

int
value_char::load(char *&retstr)
{
	static char buffer[64];

#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_char)\n";
		exit(1);
	}
#endif

	sprintf(buffer,"%d",*number);

	retstr=buffer;

	return(0);
}

int
value_char::load(double &retdoub)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_char)\n";
		exit(1);
	}
#endif
	retdoub=(double)*number;
	return(0);
}


int
value_char::size()
{
	return( sizeof(*number) );
}

void 
value_char::to_mem(char *mem)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_char)\n";
		exit(1);
	}
#endif
	*mem=*number;
	INFO(cinfo << "Stored char "<< *number <<" at address" << (address)number << "\n";)
}

void
value_char::from_mem(char *mem)
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_char)\n";
		exit(1);
	}
#endif
	*number=*mem;
	INFO(cinfo << "Read char" << *number << " from address" << (address)mem << "\n";)
}

void
value_char::assign_storage(char *mem)
{
	if (mem)
		number=mem;
	else {
		//
		// Only copy the current value if the pointer is non null
		//
		if (number)
			tmp_number=*number;
		number=&tmp_number;
	}
}

char *
value_char::fetch_storage()
{
#ifdef ASSERTS
	if (number==NULL) {
		cerror << internal_address_not_set << " (value_char)\n";
		exit(1);
	}
#endif
	return((char *)number);
}

char *
value_char::name()
{
	return( "char" );
}

class value *
value_char::operation(operators_e op,class value *v2)
{
	class value_char *va,*vb;
	char va_alloc=0;

	//
	// Handle unary operators first - they have no type checking
	//
	if (op==op_not) {
		if (status==temporary)
			va=this;
		else
			va=(class value_char *)clone();
		*(va->number) = !(*(va->number));
		return(va);
	}


	//
	// Do we have type mismatch problems?
	//
	if (value_flag != v2->value_flag) {
		//
		// If v2 is an integer then upgrade the current
		// value to an integer and then do the operation
		//
		if (v2->value_flag==VALUE_INT) {
			class value_int *vc=new class value_int;

			vc->assign_storage(NULL);
			vc->store((int)(*number));
			dispose();
			return(vc->operation(op,v2));
		}
		//
		// Otherwise type mismatch error
		//
		cerror << type_mismatch;
		return(NULL);
	}
	//
	// If the value is temporary, then use it.  Otherwise
	// a temporary node must be created. 
	// 
	if (status==temporary) {
		va=this;
	} else {
		va=(class value_char *)clone();
		va_alloc=1;
	}

	//
	// As we are about to modify the value, point to a temporary
	// workarea.
	//
	va->assign_storage(NULL);
	vb=(class value_char *)v2;
	

	switch(op) {
        	case op_add:*(va->number) += *(vb->number);
			return(va);
        	case op_sub:*(va->number) -= *(vb->number);
			return(va);
        	case op_mul:*(va->number) *= *(vb->number);
			return(va);
	        case op_div:*(va->number) /= *(vb->number);
			return(va);
	        case op_or:*(va->number) |= *(vb->number);
			return(va);
	        case op_xor:*(va->number) ^= *(vb->number);
			return(va);
	        case op_and:*(va->number) |= *(vb->number);
			return(va);
	        case op_lor:*(va->number) = *(va->number) || *(vb->number);
			return(va);
	        case op_land:*(va->number) = *(va->number) && *(vb->number);
			return(va);
	        case comp_eq:
			*(va->number) = (*(va->number)==*(vb->number));
			return(va);
		case comp_ne:
			*(va->number) = (*(va->number)!=*(vb->number));
			return(va);
		case comp_lt:
			*(va->number) = (*(va->number)<*(vb->number));
			return(va);
		case comp_le:
			*(va->number) = (*(va->number)<=*(vb->number));
			return(va);
		case comp_gt:
			*(va->number) = (*(va->number)>*(vb->number));
			return(va);
		case comp_ge:
			*(va->number) = (*(va->number)>=*(vb->number));
			return(va);
	}
	cerror << bad_operation ;
	if (va_alloc) {
		va->dispose();
		va=NULL;
	}
	return(NULL);
}

class value *
value_char::clone()
{
        class value_char *newval=new value_char;

        newval->number=number;
        newval->tmp_number=tmp_number;

	return ((class value *)newval);
}


value_char::value_char()
{
	value_flag=VALUE_CHAR;
	number=NULL;
}

value_char::~value_char()
{
}



//
//////////////////////////////////////////////////////////////////////////////
//
// Array value operations
//
//////////////////////////////////////////////////////////////////////////////
//

int
value_array::size()
{
	return( memsize );
}

void 
value_array::to_mem(char *mem)
{
#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_array)\n";
		exit(1);
	}
#endif
	memcpy(mem,addr,memsize);
}

void
value_array::from_mem(char *mem)
{
#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_array)\n";
		exit(1);
	}
#endif
	memcpy(addr,mem,memsize);
}

void
value_array::assign_storage(char *mem)
{
	addr=mem;
}

char *
value_array::fetch_storage()
{
#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_array)\n";
		exit(1);
	}
#endif
	return((char *)addr);
}

char *
value_array::name()
{
	return( "array" );
}

int
value_array::type_equal(class value *v2)
{
	class value_array *va;

	if (value_flag!=v2->value_flag)
		return(0);

	//
	// The check above means this cast is safe to make!
	//
	va=(class value_array *)v2;

	//
	// Both types must be arrays - BUT the types are only
	// equal if the base types are also equal.
	//
	return( base_type->type_equal(va->base_type) );
}


class value *
value_array::operation(operators_e op,class value *v2)
{
	class value_array *vb;
	class value_int *val_bool;
	int result;

	
	//
	// Do we have type mismatch problems?
	//
	if (value_flag != v2->value_flag) {
		//
		// Otherwise type mismatch error
		//
		cerror << type_mismatch;
		return(NULL);
	}

	val_bool=new class value_int;
	val_bool->assign_storage(NULL);
	vb=(class value_array *)v2;

	//
	// If the arrays are not even the same basic type then they
	// cannot be  equal.
	//
	if (type_equal(vb))
		result=memcmp(addr,vb->addr,memsize);
	else
		result=1;
	
	//
	// Only support equality and inequality on structures
	//
	switch(op) {
	        case comp_eq:
			val_bool->store(result==0);
			return(val_bool);
		case comp_ne:
			val_bool->store(result!=0);
			return(val_bool);
	}
	cerror << bad_operation ;
	delete val_bool;

	return(NULL);
}


class value *
value_array::deref(int index)
{
	class value *val;
	char *ret_addr=addr+index*elesize;
#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_array)\n";
		exit(1);
	}
#endif
	
	val=base_type->clone();

	val->assign_storage( ret_addr );

	INFO(cinfo << "Dereference copy from address 0x"<<(address)ret_addr;)
	INFO(cinfo <<" on index="<<index<<"\n";)

	return(val);
}

class value *
value_array::deref(int index,char *&ret_addr)
{

#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_array)\n";
		exit(1);
	}
#endif

	ret_addr=addr+index*elesize;

	INFO(cinfo << "Dereference gives address 0x"<<(address)ret_addr;)
	INFO(cinfo <<" on index="<<index<<"\n";)
	return(base_type);
}


class value *
value_array::clone()
{
        class value_array *newval=new value_array(base_type,elements);

	INFO(cinfo << "Clone of value_array\n";)
	newval->addr=addr;

	return ((class value *)newval);
}


value_array::value_array( class value *cbase,int csize)
{
	base_type=cbase;
	elements=csize;

	elesize=base_type->size();
	memsize=elements*elesize;
	addr=NULL;
	value_flag=VALUE_ARRAY;
}

value_array::~value_array()
{
}


//
//////////////////////////////////////////////////////////////////////////////
//
// Function handling 
//
//////////////////////////////////////////////////////////////////////////////
//

void 
value_func::to_mem(char *mem)
{
#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_func)\n";
		exit(1);
	}
#endif

	*((class node_funcdecl **)mem)=*addr;

	INFO(cinfo << "this="<<(address)this<<"Function node read at address " << (address)addr <<"\n";)
}

void
value_func::from_mem(char *mem)
{
	*addr=*((class node_funcdecl **)mem);

	INFO(cinfo << "this="<<(address)this<<":Function node stored at address " << (address)addr <<"\n";)
}

void
value_func::assign_storage(char *mem)
{
	if (mem)
		addr=(class node_funcdecl **)mem;
	else {
		if (addr)
			tmp_addr=*addr;
		
		addr=&tmp_addr;
	}
}


class value  *
value_func::clone()
{
        class value_func *newfunc=new value_func;

	INFO(cinfo << "Clone of value_func\n";)
	newfunc->addr=addr;

	return ((class value *)newfunc);
}

int
value_func::size()
{
	return(sizeof(addr));
}


class node_funcdecl *
value_func::funcaddr()
{
	if (!addr) {
		cerror << "Function address is null\n";
		exit(1);
	}
		
	return(*addr);
}


char *
value_func::name()
{
	return( "function" );
}



value_func::value_func()
{
	addr=NULL;
	value_flag=VALUE_FUNC;
}

value_func::~value_func()
{
}


//
//////////////////////////////////////////////////////////////////////////////
//
// Structure value operations
//
//////////////////////////////////////////////////////////////////////////////
//
int
value_struct::size()
{
	return(current_offset);
}

void
value_struct::to_mem(char *mem)
{
#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_struct)\n";
		exit(1);
	}
#endif
	memcpy(mem,addr,current_offset);
}

void
value_struct::from_mem(char *mem)
{
#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_struct)\n";
		exit(1);
	}
#endif
	memcpy(addr,mem,current_offset);
}

void
value_struct::assign_storage(char *mem)
{
	addr=mem;
}

char *
value_struct::fetch_storage()
{
#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_struct)\n";
		exit(1);
	}
#endif
	return((char *)addr);
}

char *
value_struct::name()
{
	return( "structure" );
}

int
value_struct::type_equal(class value *v2)
{
	class value_struct *va;

	if (value_flag!=v2->value_flag)
		return(0);
	//
	// Structures are only equal if they use the same symbol
	// table.
	//
	va=(class value_struct *)v2;
	if (symt==va->symt)
		return(1);

	return(0);
}


class value *
value_struct::operation(operators_e op,class value *v2)
{
	class value_struct *vb;
	class value_int *val_bool;
	int result;

	
	//
	// Do we have type mismatch problems?
	//
	if (value_flag != v2->value_flag) {
		//
		// Otherwise type mismatch error
		//
		cerror << type_mismatch;
		return(NULL);
	}

	val_bool=new class value_int;
	val_bool->assign_storage(NULL);
	vb=(class value_struct *)v2;

	//
	// If the structures are not the same type then they cannot be
	// equal.
	//
	if (type_equal(vb))
		result=memcmp(addr,vb->addr,current_offset);
	else
		result=1;
	
	//
	// Only support equality and inequality on structures
	//
	switch(op) {
	        case comp_eq:
			val_bool->store(result==0);
			return(val_bool);
		case comp_ne:
			val_bool->store(result!=0);
			return(val_bool);
	}
	cerror << bad_operation ;
	delete val_bool;

	return(NULL);
}

class value *
value_struct::clone()
{
        class value_struct *newval=new value_struct();

	newval->addr=addr;
	newval->symt=symt;
	newval->current_offset=current_offset;
	
	//
	// Since newval uses the same symbol table (no point having
	// unnecessary copies) mark it as having a further reference.
	// It can't be deleted until the reference count drops
	// to 0.
	// 
	symt->SymboltableReferenced();

	return ((class value *)newval);
}



class value *
value_struct::lookup(char *name)
{
	class value *val;
	char *ret_addr;
        class typesymbol *tsym;

#ifdef ASSERTS
	INFO(cinfo << "Looking up "<<name<<"\n";)
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_struct)\n";
		exit(1);
	}
#endif
	if (symt==NULL) 
		DialogPanic("Lookup on empty symbol table for structure");

        tsym=(class typesymbol *)symt->findsymbol(name);
	if (tsym==NULL) {
		cerror << bad_type <<". Cannot decode field ";
		cerror << name << "\n";
		INFO(cinfo << "structure address = "<<(address)this<<"\n";)
		return(NULL);
	}
	
	ret_addr=addr+tsym->offset;
	val=tsym->type->clone();

	val->assign_storage( ret_addr );

	INFO(cinfo << "Dereference copy from address 0x"<<(address)ret_addr;)
	INFO(cinfo <<" on name="<<name<<"\n";)

	return(val);
}

class value *
value_struct::lookup(char *name,char *&ret_addr)
{
        class typesymbol *tsym;
#ifdef ASSERTS
	if (addr==NULL) {
		cerror << internal_address_not_set << " (value_struct)\n";
		exit(1);
	}
#endif

	if (symt==NULL) 
		DialogPanic("Lookup on empty symbol table for structure");

        tsym=(class typesymbol *)symt->findsymbol(name);
	if (tsym==NULL) {
		cerror << bad_type <<". Cannot decode field for address ";
		cerror << name << "\n";
		INFO(cinfo << "structure address = "<<(address)this<<"\n";)
		return(NULL);
	}

	ret_addr=addr+tsym->offset;
	INFO(cinfo << "Dereference gives address 0x"<<(address)ret_addr;)
	INFO(cinfo <<" on field="<<name<<"\n";)
	return(tsym->type);
}

void
value_struct::addfield(char *c_fieldname,class value *c_type)
{
	class typesymbol *tsym=new typesymbol;

	//
	// If not symbol table yet allocated, allocate one now
	//
	if (!symt)
		symt=new class symboltable;

	//
	// Must make sure the type for this field doesn't go away
	//
	c_type->setstatus(fundamental);


	tsym->type=c_type;
	tsym->offset=current_offset;
	current_offset += c_type->size();

	symt.addsymbol(c_fieldname,tsym);

	INFO(cinfo << "Adding field to structure ("<<(address)this;)
	INFO(cinfo << ") name "<< c_fieldname;)
	INFO(cinfo << " size="<<c_type->size()<<" offset="<<tsym->offset<<"\n";)
	INFO(cinfo << "Address of type is "<<(address)c_type<<"\n";)
}

value_struct::value_struct()
{
	//
	// symt is a ptr not a structure so that when value_structs
	// are cloned, the clone can use the original symbol table.
	// The sybol table is allocated on as needed basis!
	//
	symt=NULL;
	value_flag=VALUE_STRUCT;
	current_offset=0;
	addr=NULL;
}

value_struct::~value_struct()
{
	if ( symt )
		symt->SymboltableUnreferenced();
}

//
//////////////////////////////////////////////////////////////////////////////
//
// The type symbol class is only used by value_struct class to map
// field names in to ffset and type pairs.
//
//////////////////////////////////////////////////////////////////////////////
//
typesymbol::typesymbol()
{
}

typesymbol::~typesymbol()
{
}



//
//
//
int allocate_type_id()
{
	static int next_free_id=2000;

	return(next_free_id++);
}

