
//
//    X SHARE MANAGER (XSM)
//
//    Copyright (C) 1995  David Simons
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
// $Id: value.h,v 2.3 1995/06/17 20:29:33 davidsi Exp $
//
// $Log: value.h,v $
// Revision 2.3  1995/06/17  20:29:33  davidsi
// - Added operation() function member to the value_string class, since we
//   need to be able to compare strings.
// - Added operation() function member to the value_struct class, since it
//   is helpful to be able to perform equality/inequality checks on
//   structures.
// - Added operation() function member to the value_array class, since it
//   is helpful to be able to perform equality/inequality checks on
//   arrays.
// - Removed unwrap() in both value and value_struct classes.
// - Added a new type, the value_char to allow for characters rather
//   than integers.
// - Now that a char type exists, deref() has been added to strings, so
//   that characters can be extracted from them.
// - Teach integer and character class about the op_not operation.
//
// Revision 2.2  1995/05/29  19:01:46  davidsi
// - Added assign_storage() and clone() to the function type.
//
// Revision 2.1  1995/04/25  21:06:22  davidsi
// - Add name() function member to value and all value derived classes. This
//   identifies the type using an ascii string - great for debugging.
// - Remove release_storage() function member - it is not needed.
//
// Revision 2.0  1995/01/19  21:33:47  davidsi
// - Initial version.
//
//
//
#ifndef __VALUE_H
#define __VALUE_H
#include "symbol.h"

#define VALUE_INT	0x1
#define VALUE_STRING	0x2
#define VALUE_REAL	0x4
#define VALUE_ARRAY	0x8
#define VALUE_FUNC	0x10
#define VALUE_STRUCT	0x20
#define VALUE_CHAR	0x40

//
// A fundamental value cannot be disposed of.  Values start as
// temporary , use setstatus() to upgrade to a fundamental value.
//
enum valuetype_e {
	fundamental,		// came from a symbol definition - !delete able
	temporary		// temporary staging value
	};

enum operators_e {
	op_add,
	op_sub,
	op_mul,
	op_div,
	op_and,
	op_land,
	op_or,
	op_lor,
	op_xor,
	op_not,
	comp_eq,
	comp_ne,
	comp_lt,
	comp_le,
	comp_gt,
	comp_ge,
	};

class value  {
	char *locked_heap;	// address of locked heap
	class context *locked_context;

protected:
	valuetype_e status;
	int value_flag;		// see VALUE_XXX

public:
	int id();
	void setstatus(valuetype_e);
	void dispose();
	void lock_2_heap(class context *cont);

	virtual int store(int num);	
	virtual int store(char *str);
	virtual int store(double num);

	virtual int load(int &);
	virtual int load(char *&);
	virtual int load(double &);

	//
	// Every value should have these functions set
	// 
	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual class value *clone();
	virtual char *name();

	//
	// Generic functions can be used for these. 
	//
	virtual int type_equal(class value *);

	//
	// These provide specific behaviours
	//
	virtual class value *operation(operators_e,class value *);
	virtual class value *deref(int);
	virtual class value *deref(int,char *&addr);
	virtual class value *lookup(char *name);
	virtual class value *lookup(char *name,char *&);
	virtual class node_funcdecl *funcaddr();



	value();
	virtual ~value();
};


class value_int : public value  {
	int *number;
	int tmp_number;		// since ints are so small when workspace
				// is needed number points to this member
public:
	virtual int store(int num);	
	virtual int store(char *str);
	virtual int store(double num);

	virtual int load(int &);
	virtual int load(char *&);
	virtual int load(double &);

	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual char *name();


	virtual class value *operation(operators_e,class value *);

	virtual class value *clone();
	value_int();
	~value_int();
};

class value_real : public value  {
	double *number;
	double tmp_number;	// since doubles are so small when workspace
				// is needed number points to this member
public:
	virtual int store(int num);	
	virtual int store(char *str);
	virtual int store(double num);

	virtual int load(int &);
	virtual int load(char *&);
	virtual int load(double &);

	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual char *name();

	virtual class value *operation(operators_e,class value *);

	virtual class value *clone();
	value_real();
	~value_real();
};

#define STRING_LENGTH	200

class value_string : public value  {
	char *str;
public:
	virtual int store(int num);	
	virtual int store(char *str);
	virtual int store(double num);

	virtual int load(int &);
	virtual int load(char *&);
	virtual int load(double &);

	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual char *name();

	//
	// derefs on a string yield a char
	//
	virtual class value *deref(int);
	virtual class value *deref(int,char *&addr);

	virtual class value *operation(operators_e,class value *);
	virtual class value *clone();

	value_string();
	~value_string();
};

class value_char : public value  {
	char *number;
	char tmp_number;	// since ints are so small when workspace
				// is needed number points to this member
public:
	virtual int store(int num);	
	virtual int store(char *str);
	virtual int store(double num);

	virtual int load(int &);
	virtual int load(char *&);
	virtual int load(double &);

	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual char *name();


	virtual class value *operation(operators_e,class value *);

	virtual class value *clone();
	value_char();
	~value_char();
};


class value_array : public value  {
	int elements;
	class value *base_type;
	char *addr;			// storage
	int memsize;			// number of bytes in addr
	int elesize;			// size of each element
public:
	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual char *name();

	virtual int type_equal(class value *);
	virtual class value *operation(operators_e,class value *);

	virtual class value *deref(int);
	virtual class value *deref(int,char *&addr);

	virtual class value *clone();
	value_array(class value *,int);
	~value_array();
};

class value_func : public value  {
	class node_funcdecl **addr;
	class node_funcdecl *tmp_addr;
public:
	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual class value *clone();
	virtual class node_funcdecl *funcaddr();
	virtual char *name();
	value_func();
	~value_func();
};


class typesymbol : public symbol_value {
public:
	class value *type;
	int offset;

	typesymbol();
	~typesymbol();
};




class value_struct : public value  {
	symboltable *symt;	// symbol table for this structure
	char *addr;		// storage
	int current_offset;	// size of structure while struct being built
public:
	virtual int size();
	virtual void to_mem(char *mem);
	virtual void from_mem(char *mem);
	virtual void assign_storage(char *mem);
	virtual char *fetch_storage();
	virtual char *name();

	virtual int type_equal(class value *);
	virtual class value *lookup(char *name);
	virtual class value *lookup(char *name,char *&addr);
	virtual class value *clone();
	virtual class value *operation(operators_e,class value *);

	void addfield(char *fieldname,class value *type);


	value_struct();
	~value_struct();
};

#endif
